@@ -4,7 +4,12 @@
  */
 
 #include "hostenv.h"
+
 #include <stdio.h>
+#ifndef FILE /* Some systems don't have this as a MACRO.. */
+# define FILE FILE
+#endif
+#include <sfio.h>
 
 #ifndef __
 # ifdef __STDC__
@@ -21,7 +26,7 @@
 #include <sys/stat.h>
 #endif	/* S_IFMT */
 
-#include "malloc.h"
+/* #include "zmalloc.h" -- included also in begin of "listutils.h" */
 #include "listutils.h"
 #include "token.h"
 #include "splay.h"
