@@ -3,15 +3,25 @@
  *	This will be free software, but only when it is finished.
  */
 /*
- *	A plenty of changes, copyright Matti Aarnio 1990-1995
+ *	A plenty of changes, copyright Matti Aarnio 1990-1999
  */
 
+#ifndef _Z_TA_H_
+#define _Z_TA_H_
+
 #ifdef HAVE_CONFIG_H
 #include "hostenv.h"
 #endif
 
-struct address {
-	struct address	*link;		/* next sender / sender for this rcpt */
+#include <stdio.h>
+#ifndef FILE /* Some systems don't have this as a MACRO.. */
+# define FILE FILE
+#endif
+#include <sfio.h>
+
+
+struct taddress {
+	struct taddress	*link;		/* next sender / sender for this rcpt */
 	const char	*channel;
 	const char	*host;
 	const char	*user;
@@ -24,21 +34,24 @@ struct address {
 #define _DSN_NOTIFY_DELAY	0x004
 #define _DSN_NOTIFY_NEVER	0x008
 
+#define _DSN__DIAGDELAYMODE	0x800 /* Internal magic */
+
 /* `convertmode' controls the behaviour of the message conversion:
      _CONVERT_NONE (0): send as is
      _CONVERT_QP   (1): Convert 8-bit chars to QUOTED-PRINTABLE
      _CONVERT_8BIT (2): Convert QP-encoded chars to 8-bit
      _CONVERT_UNKNOWN (3): Turn message to charset=UNKNOWN-8BIT, Q-P..
 */
-#define _CONVERT_NONE	 0
-#define _CONVERT_QP	 1
-#define _CONVERT_8BIT	 2
-#define _CONVERT_UNKNOWN 3
-
+typedef enum {
+  _CONVERT_NONE,
+  _CONVERT_QP,
+  _CONVERT_8BIT,
+  _CONVERT_UNKNOWN
+} CONVERTMODE;
 
 struct rcpt {
 	struct rcpt	*next;
-	struct address	*addr;		/* addr.link is the sender address */
+	struct taddress	*addr;		/* addr.link is the sender address */
 	const char	*orcpt;		/*  DSN  ORCPT=  string */
 	const char	*inrcpt;	/* "DSN" INRCPT= string */
 	const char	*notify;	/*  DSN  NOTIFY= flags  */
@@ -58,6 +71,14 @@ struct rcpt {
 	int		status;		/* current delivery sysexit code */
 	struct ctldesc	*desc;		/* backpointer to descriptor */
 	/* XX: something needed for XOR address lists */
+
+	char		*delayslot;
+#if 0 /* not yet ?? */
+	/* Delayed diagnostics */
+	char		*diagdelaybuf;
+	int		diagdelaysize;
+	int		diagdelayspace;
+#endif
 };
 
 struct ctldesc {
@@ -69,6 +90,7 @@ struct ctldesc {
 	time_t		msgmtime;	/* Message file arrival time */
 	long		msgbodyoffset;	/* offset of message body in msgfile */
 	long		msgsizeestimate; /* Estimate of the msg size */
+	long		msginonumber;	/* message file inode number */
 	int		msgfd;		/* message file I/O descriptor */
 	int		ctlfd;		/* control file I/O descriptor */
 	int		ctlid;		/* control file id (inode number) */
@@ -76,11 +98,12 @@ struct ctldesc {
 	const char	*contents;	/* message file data */
 	long		contentsize;	/* message file size */
 	long		*offset;	/* array of indices into contents */
-	struct address	*senders;	/* list of sender addresses */
+	struct taddress	*senders;	/* list of sender addresses */
 	struct rcpt	*recipients;	/* list of selected recipients */
 	int		rcpnts_total;	/* number of recipients, total */
 	int		rcpnts_remaining;/* .. how many yet to deliver */
 	int		rcpnts_failed  ;/* .. how many failed ones */
+	long		format;		/* _CF_FORMAT_xxx */
 	char		***msgheaders;	/* pointer to all msg headers */
 	char		***msgheaderscvt; /* converted headers */
 #ifdef	HAVE_MMAP
@@ -119,7 +142,7 @@ struct mimestate {
 /* ctlopen.c: */
 extern void            ctlfree __((struct ctldesc *dp, void *anyp));
 extern void           *ctlrealloc __((struct ctldesc *dp, void *anyp, size_t size));
-extern struct ctldesc *ctlopen __((const char *file, const char *channel, const char *host, int *exitflag, int (*selectaddr)(const char *, const char *, void *), void *saparam, int (*matchrouter)(const char *, struct address *, void *), void *mrparam));
+extern struct ctldesc *ctlopen __((const char *file, const char *channel, const char *host, int *exitflag, int (*selectaddr)(const char *, const char *, void *), void *saparam, int (*matchrouter)(const char *, struct taddress *, void *), void *mrparam));
 extern void            ctlclose __((struct ctldesc *dp));
 extern int	       ctlsticky __((const char *spec_host, const char *addr_host, void *cbparam));
 
@@ -144,6 +167,8 @@ extern void	       diagnostic __((/* struct rcpt *, int, int, char *,... */));
 /* emptyline.c: */
 extern int	       emptyline __(( char *line, int size ));
 
+extern int zmalloc_failure;
+
 #if 0 /* actually better to include "libz.h" for this */
 /* esyslib.c: */
 extern int cistrcmp  __((const char *s1, const char *s2));
@@ -174,7 +199,7 @@ extern int erename __((char *from, char *to));
 #endif
 
 /* lockaddr.c: */
-extern int lockaddr __((int fd, char *map, long offset, int was, int new, const char *file, const char *host, const int mypid));
+extern int lockaddr __((int fd, char *map, int offset, int was, int new, const char *file, const char *host, const int mypid));
 
 /* markoff.c: */
 extern int markoff __((char *filecontents, int bytesleft, long offsets[], const char *filename));
@@ -193,7 +218,7 @@ extern int cte_check __((struct rcpt *rp));
 extern char **has_header __((struct rcpt *rp, const char *keystr));
 extern void delete_header __((struct rcpt *rp, char **hdrp));
 extern int  downgrade_charset __((struct rcpt *rp, FILE *verboselog));
-extern void downgrade_headers __((struct rcpt *rp, int downgrade, FILE *verboselog));
+extern int  downgrade_headers __((struct rcpt *rp, CONVERTMODE downgrade, FILE *verboselog));
 extern int qp_to_8bit __((struct rcpt *rp));
 
 /* mime2headers.c */
@@ -201,8 +226,10 @@ extern int headers_to_mime2 __((struct rcpt *rp, const char *defcharset, FILE *v
 extern int headers_need_mime2 __(( struct rcpt *rp ));
  
 
-/* writeheaders.c: */
-extern int writeheaders __((struct rcpt *rp, FILE *fp, const char *newline, int use_cvt, int maxwidth));
+/* fwriteheaders.c: */
+extern int fwriteheaders __((struct rcpt *rp, FILE *fp, const char *newline, CONVERTMODE use_cvt, int maxwidth, char **chunkbufp));
+/* swriteheaders.c: */
+extern int swriteheaders __((struct rcpt *rp, Sfio_t *fp, const char *newline, CONVERTMODE use_cvt, int maxwidth, char **chunkbufp));
 
 /* buildbndry.c: */
 extern char *mydomain __((void));
@@ -221,9 +248,6 @@ extern void warning __(());
 /* lib/skip821address.c */
 extern char *skip821address __((const char *s));
 
-/* lib/taspoolid.c */
-extern void taspoolid __((char *buf, int buflen, time_t mtime, const char *path));
-
 /* tasyslog.c */
 extern void tatimestr __((char *buf, int dt));
 extern void tasyslog __((struct rcpt *rp, int xdelay, const char *wtthost, const char *wttip, const char *stats, const char *msg));
@@ -234,3 +258,10 @@ extern void tasyslog __((struct rcpt *rp, int xdelay, const char *wtthost, const
 #endif
 
 extern int getmyuucename __((char *, int));
+
+/* nonblocking.c */
+extern int  fd_nonblockingmode __((int fd));
+extern int  fd_blockingmode __((int fd));
+extern void fd_restoremode __((int fd, int mode));
+
+#endif
