@@ -20,7 +20,7 @@
 extern char *progname;
 extern char *logfn;
 
-int
+/* RETSIGTYPE */ int
 loginit(sig)
      int sig;
 {
@@ -44,7 +44,8 @@ loginit(sig)
 	fcntl(FILENO(stdout), F_SETFL, flags);
 #endif	/* F_SETFL */
 	setvbuf(stdout, (char *)NULL, _IOLBF, 0);
+	freopen(logfn, "a", stderr);
 	setvbuf(stderr, (char *)NULL, _IOLBF, 0);
-	SIGNAL_HANDLE(SIGHUP, (RETSIGTYPE (*) __((int))) loginit);
+	SIGNAL_HANDLE(sig, (RETSIGTYPE(*)())loginit);
 	return 0;
 }
