@@ -39,7 +39,9 @@
 #ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
+#ifdef HAVE_NETDB_H
 #include <netdb.h>
+#endif
 #ifndef EAI_AGAIN
 # include "netdb6.h" /* IPv6 API stuff */
 #endif
@@ -55,6 +57,15 @@
 #endif
 #include <arpa/inet.h>
 #include <string.h>
+#include <errno.h>
+
+union sockaddr_uni {
+    struct sockaddr     sa;
+    struct sockaddr_in  v4;
+#if defined(AF_INET6) && defined(INET6)
+    struct sockaddr_in6 v6;
+#endif
+};
 
 #include "libz.h"
 
@@ -92,13 +103,13 @@ struct sockaddr ***sockaddrp;
 	int s;
 	int i;
 	int ifcount = 0;
-	int af = AF_INET;
+	int pf = PF_INET;
         struct ifconf ifc;
 	int ifbufsize = 4 * sizeof(struct ifreq) + 4;
 	char *interfacebuf = (void*)malloc(ifbufsize);
-	struct sockaddr **sap;
+	union sockaddr_uni **sap;
 
-	sap = (void*)malloc(sizeof(struct sockaddr*) * (ifcount + 2));
+	sap = (void*)malloc(sizeof(union sockaddr_uni*) * (ifcount + 2));
 	if (! sap || !interfacebuf) {
 	  if (interfacebuf)
 	    free(interfacebuf);
@@ -111,7 +122,7 @@ struct sockaddr ***sockaddrp;
 other_socktype:
 #endif
 
-	s = socket(af, SOCK_DGRAM, 0);
+	s = socket(pf, SOCK_DGRAM, 0);
 	if (s < 0) {
 	  free(interfacebuf);
 	  free(sap);
@@ -136,14 +147,12 @@ other_socktype:
 	  /* get the list of known IP address from the kernel */
 	  ifbufsize <<= 1;
 	  interfacebuf = (void*)realloc(interfacebuf,ifbufsize);
+	  memset(&ifc, 0, sizeof(ifc));
 	  ifc.ifc_buf = interfacebuf;
 	  ifc.ifc_len = ifbufsize;
-	  if (ioctl(s, SIOCGIFCONF, (char *)&ifc) < 0) {
-		close(s);
-		free(interfacebuf);
-		free(sap);
-		return -2;
-	  }
+	  if (ioctl(s, SIOCGIFCONF, (char *)&ifc) < 0)
+	    if (errno == EINVAL)
+	      continue;
 
 	  if (ifc.ifc_len < (ifbufsize - 2*sizeof(struct ifreq)))
 	    break;
@@ -157,14 +166,14 @@ other_socktype:
 	for (i = 0; i < ifc.ifc_len; ) {
 
 	  struct ifreq *ifr = (struct ifreq *) &ifc.ifc_buf[i];
-	  struct sockaddr *sa = &ifr->ifr_addr;
+	  union sockaddr_uni *sa = (union sockaddr_uni *) &ifr->ifr_addr;
 #ifdef SIOCGIFFLAGS
 	  struct ifreq ifrf;
 #endif
 
 #ifdef HAVE_SA_LEN
-	  if (sa->sa_len > sizeof ifr->ifr_addr)
-	    i += sizeof ifr->ifr_name + sa->sa_len;
+	  if (sa->sa.sa_len > sizeof ifr->ifr_addr)
+	    i += sizeof ifr->ifr_name + sa->sa.sa_len;
 	  else
 #endif
 	    i += sizeof *ifr;
@@ -189,10 +198,10 @@ other_socktype:
 
 	  if (ioctl(s, SIOCGIFFLAGS, (char *) &ifrf) < 0)
 	    continue; /* Failed.. */
-
-	  /* printf("name='%s'  ifrf_flags=0x%x\n",
-	     ifr->ifr_name,ifrf.ifr_flags); */
-
+#if 0
+	   printf("name='%s'  ifrf_flags=0x%x\n",
+	     ifr->ifr_name,ifrf.ifr_flags);
+#endif
 	  if (!(IFF_UP & ifrf.ifr_flags))
 	    continue;
 #else
@@ -203,7 +212,7 @@ other_socktype:
 	    continue;
 #endif
 
-	  sap = (void*)realloc(sap, sizeof(struct sockaddr*) * (ifcount + 2));
+	  sap = (void*)realloc(sap, sizeof(union sockaddr_uni*) * (ifcount + 2));
 	  if (! sap) {
 	    close(s);
 	    free(interfacebuf);
@@ -211,16 +220,16 @@ other_socktype:
 	  }
 
 	  /* XX: Use sa_len on BSD44 ??? */
-	  if (sa->sa_family == AF_INET) {
+	  if (sa->sa.sa_family == AF_INET) {
 	    struct sockaddr_in *si4 = (void*)malloc(sizeof(*si4));
 	    if (si4 == NULL)
 	      break;
 	    /* pick the whole sockaddr package! */
 	    memcpy(si4, &ifr->ifr_addr, sizeof(struct sockaddr_in));
-	    sap[ifcount] = (struct sockaddr*)si4;
+	    sap[ifcount] = (union sockaddr_uni *)si4;
 	  }
 #if defined(AF_INET6) && defined(INET6)
-	  if (sa->sa_family == AF_INET6) {
+	  if (sa->sa.sa_family == AF_INET6) {
 	    struct sockaddr_in6 *si6 = (void*)malloc(sizeof(*si6));
 	    if (si6 == NULL)
 	      break;
@@ -229,7 +238,7 @@ other_socktype:
 
 	    /* pick the whole sockaddr package! */
 	    memcpy(si6, &ifr->ifr_addr, sizeof(struct sockaddr_in6));
-	    sap[ifcount] = (struct sockaddr*)si6;
+	    sap[ifcount] = (union sockaddr_uni *)si6;
 	  }
 #endif
 	  ++ifcount;
@@ -238,14 +247,14 @@ other_socktype:
 	close(s);
 
 #if defined(AF_INET6) && defined(INET6)
-	if (af != AF_INET6) {
-	  af = AF_INET6;
+	if (pf != PF_INET6) {
+	  pf = PF_INET6;
 	  goto other_socktype;
 	}
 #endif
 
 	free(interfacebuf);
-	*sockaddrp = sap;
+	*sockaddrp = (struct sockaddr **)sap;
 
 	return ifcount;
 #else
@@ -256,7 +265,7 @@ other_socktype:
 #ifndef TESTMODE /* We test ONLY of  loadifaddresses() routine! */
 
 static             int    nmyaddrs = 0;
-static struct sockaddr  ** myaddrs = NULL;
+static union sockaddr_uni  ** myaddrs = NULL;
 
 static void stashmyaddress __((const char *));
 static void
@@ -332,21 +341,21 @@ stashmyaddress(host)
 	}
 	if (myaddrs == NULL) {
 	  nmyaddrs = 0;
-	  myaddrs = malloc((nmyaddrs + naddrs +1) * sizeof(struct sockaddr*));
+	  myaddrs = (void*)malloc((nmyaddrs + naddrs +1) * sizeof(union sockaddr_uni*));
 	} else
-	  myaddrs = realloc(myaddrs,
-			    (nmyaddrs + naddrs +1) * sizeof(struct sockaddr*));
+	  myaddrs = (void*)realloc((void*)myaddrs,
+				   (nmyaddrs + naddrs +1) * sizeof(union sockaddr_uni*));
 
 	if (!myaddrs) return; /* Uurgh.... */
 
 	for (hp_init(hp); *hp_getaddr() != NULL; hp_nextaddr()) {
 	  if (hp->h_addrtype == AF_INET) {
 	    struct sockaddr_in *si;
-	    si = malloc(sizeof(*si));
+	    si = (void*)malloc(sizeof(*si));
 	    if (!si) {
 	      return;
 	    }
-	    myaddrs[nmyaddrs++] = (struct sockaddr *) si;
+	    myaddrs[nmyaddrs++] = (union sockaddr_uni *) si;
 	    memset(si,0,sizeof(*si));
 	    si->sin_family = AF_INET;
 	    memcpy(&si->sin_addr.s_addr, *hp_getaddr(), hp->h_length);
@@ -354,11 +363,11 @@ stashmyaddress(host)
 #if defined(AF_INET6) && defined(INET6)
 	  if (hp->h_addrtype == AF_INET6) {
 	    struct sockaddr_in6 *si6;
-	    si6 = malloc(sizeof(*si6));
+	    si6 = (void*)malloc(sizeof(*si6));
 	    if (!si6) {
 	      return;
 	    }
-	    myaddrs[nmyaddrs++] = (struct sockaddr *) si6;
+	    myaddrs[nmyaddrs++] = (union sockaddr_uni *) si6;
 	    memset(si6,0,sizeof(*si6));
 	    si6->sin6_family = AF_INET;
 	    memcpy(&si6->sin6_addr.s6_addr, *hp_getaddr(), hp->h_length);
@@ -373,7 +382,7 @@ stashmyaddresses(host)
 const char *host;
 {
 	char *s1, *s2, *zenv;
-	struct sockaddr **sa;
+	union sockaddr_uni **sa;
 	int sacnt;
 
 	/* Clear them all away */
@@ -389,7 +398,7 @@ const char *host;
 
 	/* Now start fillig -- interface addresses, if you can get them */
 
-	sacnt = loadifaddresses(&sa);
+	sacnt = loadifaddresses((struct sockaddr ***)&sa);
 
 	if (sacnt > 0) {
 	  /* Okay, we GOT some addresses, I bet we got them all!
@@ -421,45 +430,42 @@ const char *host;
 }
 
 
+/* Here we compare only the address portion, not ports, nor anything else! */
+
+
 int
-matchmyaddress(sa)
-	struct sockaddr *sa;
+matchmyaddress(_sa)
+	struct sockaddr *_sa;
 {
 	int i;
+	union sockaddr_uni *sau = (union sockaddr_uni *) _sa;
 
 	if (!myaddrs)
 		stashmyaddresses(NULL);
 	if (!myaddrs) return 0; /* Don't know my addresses ! */
 	
 	/* Match loopback net.. */
-	if (sa->sa_family == AF_INET) {
-	  struct sockaddr_in *si;
-	  si = (struct sockaddr_in *)sa;
+	if (sau->sa.sa_family == AF_INET) {
+	  int net;
 
-#ifndef INADDR_LOOPBACK /* FreeBSD does not have this in  <netinet/in.h> ! */
-# define INADDR_LOOPBACK 0x7f000001UL
-#endif
-#define INADDR_LOOPBACKNET (IN_CLASSA_NET & INADDR_LOOPBACK)
-
-	  if ((ntohl(si->sin_addr.s_addr) & IN_CLASSA_NET)
-	      == INADDR_LOOPBACKNET) {
-	    return 2;
-	  }
+	  net = (ntohl(sau->v4.sin_addr.s_addr) >> 24) & 0xFF;
+	  if (net == 127)
+	    return 2; /* Loopback network */
+	  if (net == 0 || net == 127 || net > 223)
+	    return 3;
 	}
 
 	/* ... and then the normal thing -- listed interfaces */
 
 	for (i = 0; i < nmyaddrs; ++i) {
 	  /* if this is myself, skip to next MX host */
-	  if (sa->sa_family == myaddrs[i]->sa_family) {
-	    if (sa->sa_family == AF_INET &&
-		memcmp(sa, myaddrs[i],
-		       sizeof(struct sockaddr_in)) == 0)
+	  if (sau->sa.sa_family == myaddrs[i]->sa.sa_family) {
+	    if (sau->sa.sa_family == AF_INET &&
+		memcmp(&sau->v4.sin_addr, &myaddrs[i]->v4.sin_addr, 4) == 0)
 	      return 1;
 #if defined(AF_INET6) && defined(INET6)
-	    if (sa->sa_family == AF_INET6 &&
-		memcmp(sa, myaddrs[i],
-		       sizeof(struct sockaddr_in6)) == 0)
+	    if (sau->sa.sa_family == AF_INET6 &&
+		memcmp(&sau->v6.sin6_addr, &myaddrs[i]->v6.sin6_addr, 16) == 0)
 	      return 1;
 #endif
 	  }
@@ -505,7 +511,7 @@ char *argv[];
 #if defined(AF_INET6) && defined(INET6)
     case AF_INET6:
       inet_ntop(AF_INET6, &((struct sockaddr_in6 **) sa)[i]->sin6_addr, buf, sizeof(buf));
-      printf("IPv6: [IPv6:%s]\n");
+      printf("IPv6: [IPv6:%s]\n",buf);
       break;
 #endif
     default:
