@@ -8,6 +8,11 @@
 #include <stdio.h>
 #include <ctype.h>
 
+#ifndef __STDC__
+#define const
+#endif
+
+
 extern char *getenv();
 
 
