@@ -64,7 +64,7 @@ const char *postoffice;	/* may be extern or local */
   Define sending mail priority.
 */
 
-int mail_priority = 0;
+int mail_priority;
 
 /*
  * Makes a temporary file under the postoffice, based on a file name template.
@@ -106,8 +106,11 @@ _mail_fopen(filenamep)
 			strcpy(cp, post);
 		if ((fd = open(path, O_CREAT|O_EXCL|O_RDWR, 0600)) >= 0) {
 			fp = fdopen(fd, "w+");
-			mail_free(*filenamep);
-			*filenamep = path;
+			if (fp) {
+			  setvbuf(fp, NULL, _IOFBF, 8192);
+			  mail_free(*filenamep);
+			  *filenamep = path;
+			}
 			return fp;
 		}
 		eno = errno;
