@@ -1,7 +1,7 @@
 /* portability.h - include or define things that aren't present on all systems
  * vixie@decwrl 26dec92 [new]
  *
- * $Id: portability.h,v 8.14 1996/06/06 20:19:09 vixie Exp $
+ * $Id: portability.h,v 1.3 1999/11/16 16:40:08 mea Exp $
  */
 
 /*
@@ -259,14 +259,16 @@ struct timezoneBSD {
 # ifndef NeXT
 extern char *getenv __P((char *));
 # else
+#  ifndef sun /* SunOS 4.1.x */
 extern char *getenv __P((const char *));
+#  endif
 # endif
 extern int errno;
 
 # if !defined(DMALLOC) && !defined(NeXT)
 extern char *malloc(), *realloc(), *calloc();
 #  if defined(sun)
-extern int free();
+/* extern int free(); */ /* Not for SunOS 4.1.x */
 #  else
 extern void free();
 #  endif
@@ -590,7 +592,7 @@ extern int gettimeofday __P((struct timeval *, struct _TIMEZONE *));
 #endif
 
 #if defined(SVR4) && defined(sun)
-extern int gethostname __P((char *, size_t));
+extern int gethostname __P((char *, int));
 #endif
 
 #ifdef NEED_STRDUP
