@@ -43,13 +43,6 @@ int	real_stability = 0;
 /* maximum number of Received headers, for primitive loop detection */
 int	maxReceived = 50;
 
-/* we want those nice tabs between the header field name and value */
-#ifdef RFC822TABS
-int	wanttabs = RFC822TABS;
-#else
-int	wanttabs = 1;
-#endif
-
 /* depending on what you postmaster wants from   $(filepriv ...) to see:
    022 for usual        644 or more strict,
    002 for more relaxed 664 view of things
@@ -62,10 +55,9 @@ struct sptree *spt_databases, *spt_files, *spt_modcheck;
 struct sptree *spt_goodguys, *spt_uidmap, *spt_loginmap;
 struct sptree *spt_fullnamemap, *spt_incoredbs, *spt_headers;
 struct sptree *spt_eheaders;
-extern struct sptree *spt_builtins;
 
 struct sptree_init splaytrees[] = {
-{ &spt_incoredbs,	0 },		/* incore database name -> splay tree */
+  /* { &spt_incoredbs,	0 }, */		/* incore database name -> splay tree */
 { &spt_databases,	0 },		/* database name -> descriptor */
 { &spt_files,		0 },		/* file name -> FILE * or DB * */
 { &spt_modcheck,	0 },		/* modtime and nlinks */
