@@ -22,21 +22,6 @@
 #include <grp.h>			/* for run_grpmems() */
 #include <errno.h>
 
-#ifdef HAVE_DIRENT_H
-# include <dirent.h>
-#else /* not HAVE_DIRENT_H */
-# define dirent direct
-# ifdef HAVE_SYS_NDIR_H
-#  include <sys/ndir.h>
-# endif /* HAVE_SYS_NDIR_H */
-# ifdef HAVE_SYS_DIR_H
-#  include <sys/dir.h>
-# endif /* HAVE_SYS_DIR_H */
-# ifdef HAVE_NDIR_H
-#  include <ndir.h>
-# endif /* HAVE_NDIR_H */
-#endif /* HAVE_DIRENT_H */
-
 #include "zmsignal.h"
 #include "zsyslog.h"
 #include "mail.h"
@@ -71,9 +56,10 @@ static conscell *run_listexpand   __((conscell *avl, conscell *il));
 static conscell *run_newattribute __((conscell *avl, conscell *il));
 #endif
 
-static int run_doit      ARGCV;
-static int run_stability ARGCV;
-static int run_process   ARGCV;
+extern int run_doit      ARGCV;
+extern int run_stability ARGCV;
+extern int run_process   ARGCV;
+
 static int run_grpmems   ARGCV;
 static int run_praliases ARGCV;
 static int run_listaddrs ARGCV;
@@ -81,8 +67,8 @@ static int run_homedir   ARGCV;
 static int run_822date   ARGCV;
 static int run_filepriv  ARGCV;
 static int run_runas     ARGCV;
+static int run_cat       ARGCV;
 static int run_gensym    ARGCV;
-static void free_gensym __((void));
 static int run_uid2login ARGCV;
 static int run_login2uid ARGCV;
 static int run_basename  ARGCV;
@@ -139,6 +125,7 @@ struct shCmd fnctns[] = {
 {	"rfc822date",	run_822date,	NULL,	NULL,	0	},
 {	"filepriv",	run_filepriv,	NULL,	NULL,	SH_ARGV	},
 {	"runas",	run_runas,	NULL,	NULL,	SH_ARGV	},
+{	"cat",		run_cat,	NULL,	NULL,	SH_ARGV	},
 {	"gensym",	run_gensym,	NULL,	NULL,	0	},
 {	"uid2login",	run_uid2login,	NULL,	NULL,	0	},
 {	"login2uid",	run_login2uid,	NULL,	NULL,	0	},
@@ -257,7 +244,8 @@ run_gensym(argc, argv)
 	return 0;
 }
 
-static void
+extern void free_gensym __((void));
+void
 free_gensym()
 {
 	int i;
@@ -434,614 +422,6 @@ run_cadddr(avl, il)
 	return car(il);
 }
 
-static RETSIGTYPE sig_exit __((int));
-static RETSIGTYPE
-sig_exit(sig)
-int sig;
-{
-	if (canexit) {
-#ifdef	MALLOC_TRACE
-		dbfree(); zshfree();
-#endif	/* MALLOC_TRACE */
-		die(0, "signal");
-	}
-	mustexit = 1;
-	/* no need to reenable signal in USG, once will be enough */
-}
-
-static int gothup = 0;
-
-RETSIGTYPE
-sig_hup(sigarg)
-int sigarg;
-{
-	gothup = 1;
-	/* fprintf(stderr,"HUP\n"); */
-	SIGNAL_HANDLE(SIGHUP, sig_hup);
-}
-
-static void dohup __((int));
-static void
-dohup(sig)
-int sig;
-{
-	gothup = 0;
-	if (traps[SIGHUP] != NULL)
-		eval(traps[SIGHUP], "trap", NULL, NULL);
-}
-
-/*
- * Run the Router in Daemon mode.
- */
-
-/*
- * STABILITY option will make the router process incoming messages in
- * arrival (modtime) order, instead of randomly determined by position
- * in the router directory.  The scheme is to read in all the names,
- * and stat the files.  It would be possible to reuse the stat information
- * later, but I'm not convinced it is worthwhile since the later stat is
- * an fstat().  On the other hand, if we used splay tree insertion to
- * sort the entries, then the stat buffer could easily be reused in
- * makeLetter().
- *
- * SECURITY WARNING: iff makeLetter does not stat again, then there is
- * a window of opportunity for a Bad Guy to remove a file after it has
- * been stat'ed (with root privs), and before it has been processed.
- * This can be avoided partially by sticky-bitting the router directory,
- * and entirely by NOT saving the stat information we get here.
- */
-
-struct de {
-	int		f_name;
-	time_t		mtime;
-};
-
-static int decmp __((const void *, const void *));
-static int
-decmp(a, b)
-     const void *a, *b;
-{
-	register const struct de *aa = (const struct de *)a;
-	register const struct de *bb = (const struct de *)b;
-
-	return bb->mtime - aa->mtime;
-}
-
-static int desize, nbsize;
-static struct de *dearray = NULL;
-static char *nbarray = NULL;
-
-static void rd_initstability __((void));
-static void
-rd_initstability()
-{
-	desize = 1;	/* max. number of directory entries */
-	nbsize = 1;
-	dearray = (struct de *)emalloc(desize * sizeof (struct de));
-	nbarray = (char *)emalloc(nbsize * sizeof (char));
-}
-
-static void rd_endstability __((void));
-static void
-rd_endstability()
-{
-	if (dearray != NULL)
-		free((char *)dearray);
-	if (nbarray != NULL)
-		free((char *)nbarray);
-}
-
-static int
-run_doit(argc, argv)
-	int argc;
-	const char *argv[];
-{
-	const char *filename;
-	char *sh_memlevel = getlevel(MEM_SHCMD);
-	int r;
-	const char *av[3];
-
-	if (argc != 2) {
-	  fprintf(stderr, "Usage: %s <filename>\n", argv[0]);
-	  return 1;
-	}
-
-	filename = argv[1];
-
-	/* Do one file, return value is 0 or 1,
-	   depending on actually doing something
-	   on a file */
-
-	gensym = 1;
-	av[0] = "process"; /* I think this needs to be here */
-	av[1] = filename;
-	av[2] = NULL;
-	r = s_apply(2, av); /* "process" filename (within  rd_doit() ) */
-	free_gensym();
-
-	setlevel(MEM_SHCMD,sh_memlevel);
-
-	return r;
-}
-
-static int rd_doit __((const char *filename, const char *dirs));
-static int
-rd_doit(filename, dirs)
-	const char *filename, *dirs;
-{
-	/* Do one file, return value is 0 or 1,
-	   depending on actually doing something
-	   on a file */
-
-#ifdef	USE_ALLOCA
-	char *buf;
-#else
-	static char *buf = NULL;
-	static u_int blen = 0;
-#endif
-	const char *av[3];
-	char *p;
-	int len;
-	char pathbuf[512];
-	char *sh_memlevel = getlevel(MEM_SHCMD);
-	int thatpid;
-	struct stat stbuf;
-
-	*pathbuf = 0;
-	if (*dirs) {	/* If it is in alternate dir, move to primary one,
-			   and process there! */
-	  strcpy(pathbuf,dirs);
-	  strcat(pathbuf,"/");
-	}
-	strcat(pathbuf,filename);
-
-	len = strlen(filename);
-	thatpid = 0;
-	p = strchr(filename, '-');
-	if (p != NULL) {
-	  /* message file is "inode-pid" */
-	  thatpid = atoi(p+1);
-
-#if 0 /* very old thing, may harm at Solaris 2.6 ! */
-	  if (thatpid < 10) {	/* old-style locking.. */
-	    thatpid = 0;
-	  }
-#endif
-	  /* Probe it!
-	     Does the process exist ? */
-	  if (thatpid && kill(thatpid,0)==0 && thatpid != router_id) {
-	    /*
-	     * an already locked message file,
-	     * belonging to another process
-	     */
-	    if (*dirs) {
-	      fprintf(stderr,
-		      "** BUG **: %s%s not in primary router directory!\n",
-		      dirs,filename);
-	    }
-	    return 0;
-	    /*
-	     * This should not happen anywhere but at
-	     * primary router directory.  If  xxxx-nn
-	     * format file exists anywhere else, it is
-	     * a bug time!
-	     */
-	  }
-	}
-	if (strncmp(filename,"core",4) != 0 &&
-	    (p == NULL || thatpid != router_id)) {
-	  /* Not a core file, and ...
-	     not already in format of 'inode-pid' */
-	  /* If the pid did exist, we do not touch on that file,
-	     on the other hand, we need to rename the file now.. */
-#ifdef	USE_ALLOCA
-	  buf = (char*)alloca(len+16);
-#else
-	  if (blen == 0) {
-	    blen = len+16;
-	    buf = (char *)malloc(len+16);
-	  }
-	  while (len + 12 > blen) {
-	    blen = 2 * blen;
-	    buf = (char *)realloc(buf, blen);
-	  }
-#endif
-	  /* Figure out its inode number */
-	  if (stat(pathbuf,&stbuf) != 0) return 0; /* Failed ?  Well, skip it */
-	  if (!S_ISREG(stbuf.st_mode))   return 0; /* Not a regular file ??   */
-
-	  sprintf(buf, "%ld-%d", (long)stbuf.st_ino, router_id);
-
-	  if (eqrename(pathbuf, buf) < 0)
-	    return 0;		/* something is wrong, erename() complains.
-				   (some other process picked it ?) */
-	  filename = buf;
-	  /* message file is now "file-#" and belongs to this process */
-	}
-
-#ifdef	MALLOC_TRACE
-	mal_contents(stdout);
-#endif
-
-	gensym = 1;
-	av[0] = "process"; /* I think this needs to be here */
-	av[1] = filename;
-	av[2] = NULL;
-	s_apply(2, av); /* "process" filename (within  rd_doit() ) */
-	free_gensym();
-
-	setlevel(MEM_SHCMD,sh_memlevel);
-
-#ifdef MALLOC_TRACE
-	mal_contents(stdout);
-#endif
-
-	return 1;
-}
-
-static int rd_stability __((DIR *dirp, const char *dirs));
-static int
-rd_stability(dirp,dirs)
-	DIR *dirp;
-	const char *dirs;
-{
-	int deindex, nbindex, did_cnt;
-	int namelen;
-	struct dirent *dp;
-	struct stat statbuf;
-	char pathbuf[512]; /* Enough ? */
-
-	deindex = 0;
-	nbindex = 0;
-	/* collect the file names */
-	while ((dp = readdir(dirp)) != NULL) {
-		if (mustexit)
-			break;
-		if (dp->d_name[0] == '.')
-			continue;
-		/* Handle only files beginning with number -- plus "core"-
-		   files.. */
-		if (!(dp->d_name[0] >= '0' && dp->d_name[0] <= '9') &&
-		    strncmp(dp->d_name,"core",4) != 0)
-			continue;
-
-		/* See that the file is a regular file! */
-		sprintf(pathbuf,"%s%s%s", dirs, *dirs?"/":"", dp->d_name);
-		if (lstat(pathbuf,&statbuf) != 0) continue; /* ??? */
-		if (!S_ISREG(statbuf.st_mode)) continue; /* Hmm..  */
-
-		if (deindex >= desize) {
-			desize *= 2;
-			dearray =
-			  (struct de *)realloc((char *)dearray,
-					       desize * sizeof (struct de));
-		}
-
-		namelen = strlen(dp->d_name); /* Not everybody has d_namlen.. */
-
-		while (nbindex + namelen + 1 >= nbsize) {
-			nbsize *= 2;
-			nbarray =
-			  (char *)realloc(nbarray,
-					  nbsize * sizeof (char));
-		}
-
-		/*
-		 * The explicit alloc is done because alloc/dealloc
-		 * of such small chunks should not affect fragmentation
-		 * too much, and allocating large chunks would require
-		 * management code and might still do bad things with
-		 * the malloc algorithm.
-		 */
-		dearray[deindex].mtime = statbuf.st_mtime;
-		dearray[deindex].f_name = nbindex;
-		memcpy(nbarray + nbindex, dp->d_name, namelen+1);
-		nbindex += namelen + 1;
-
-		++deindex;
-	}
-	if (mustexit) {
-		return deindex;
-	}
-
-	qsort((void *)dearray, deindex, sizeof dearray[0], decmp);
-
-	did_cnt = 0;
-	while (deindex-- > 0) {
-		if (mustexit)
-			break;
-		if (gothup) 
-			dohup(0);
-		did_cnt += rd_doit(nbarray + dearray[deindex].f_name, dirs);
-
-		/* Maybe only process few files out of the low-priority
-		   subdirs, so we can go back and see if any higher-priority
-		   jobs have been created */
-		if ((*dirs) &&
-		    ((routerdirloops) && (routerdirloops == did_cnt)))
-		  break;
-
-	}
-	return did_cnt;
-}
-
-
-static int rd_instability __((DIR *dirp, char *dirs));
-static int
-rd_instability(dirp, dirs)
-	DIR *dirp;
-	char *dirs;
-{
-	struct dirent *dp;
-	int did_cnt = 0;
-	struct stat statbuf;
-	char pathbuf[512];
-
-	while ((dp = readdir(dirp)) != NULL) {
-		if (mustexit)
-			break;
-		if (gothup)
-			dohup(0);
-
-		/* Handle only files beginning with number -- plus "core"-
-		   files.. */
-		if (!(dp->d_name[0] >= '0' && dp->d_name[0] <= '9') &&
-		    strncmp(dp->d_name,"core",4) != 0)
-			continue;
-
-		/* See that the file is a regular file! */
-		sprintf(pathbuf,"%s%s%s", dirs, *dirs?"/":"", dp->d_name);
-		if (stat(pathbuf,&statbuf) != 0) continue; /* ??? */
-		if (!S_ISREG(statbuf.st_mode)) continue; /* Hmm..  */
-
-		did_cnt += rd_doit(dp->d_name, dirs);
-
-		/* Only process one file out of the low-priority subdirs,
-		   so we can go back and see if any higher-priority
-		   jobs have been created */
-		if (*dirs)
-			break;
-
-	}
-	return did_cnt;
-}
-
-
-static int
-run_stability(argc, argv)
-	int argc;
-	const char *argv[];
-{
-	switch (argc) {
-	case 1:
-		printf("%s %s\n", argv[0], stability ? "on" : "off");
-		break;
-	case 2:
-		if (strcmp(argv[1], "on") == 0) {
-			real_stability = 1;
-		} else if (strcmp(argv[1], "off") == 0) {
-			real_stability = 0;
-		} else {
-	default:
-			fprintf(stderr, "Usage: %s [ on | off ]\n", argv[0]);
-			return 1;
-		}
-		break;
-	}
-	return 0;
-}
-
-int
-run_daemon(argc, argv)
-	int argc;
-	const char *argv[];
-{
-#define ROUTERDIR_CNT 30
-	DIR *dirp[ROUTERDIR_CNT];  /* Lets say we have max 30 router dirs.. */
-	char *dirs[ROUTERDIR_CNT];
-	int did_cnt, i;
-	char *s, *rd, *routerdirs = getzenv("ROUTERDIRS");
-	char pathbuf[256];
-	memtypes oval = stickymem;
-	struct stat stb;
-
-	router_id = getpid();
-
-	SIGNAL_HANDLE(SIGTERM, sig_exit);	/* mustexit = 1 */
-	for (i=0; i<ROUTERDIR_CNT; ++i) {
-		dirp[i] = NULL; dirs[i] = NULL;
-	}
-	/* dirp[0] = opendir("."); */	/* assert dirp != NULL ... */
-#if 0
-#ifdef BSD
-	dirp[0]->dd_size = 0;	/* stupid Berkeley bug workaround */
-#endif
-#endif
-	stickymem = MEM_MALLOC;
-	dirs[0] = strnsave("",1);
-	if (routerdirs) {
-		/* Store up secondary router dirs! */
-		rd = routerdirs;
-		for (i = 1; i < ROUTERDIR_CNT && *rd; ) {
-			s = strchr(rd,':');
-			if (s)  *s = 0;
-			sprintf(pathbuf,"../%s",rd);
-			/* strcat(pathbuf,"/"); */
-
-			if (stat(pathbuf,&stb) == 0  && S_ISDIR(stb.st_mode)) {
-#if 0
-#ifdef	BSD
-			  dirp[i]->dd_size = 0;
-#endif
-#endif
-			  dirs[i] = strdup(pathbuf);
-			  ++i;
-			}
-			if (s)
-			  *s = ':';
-			if (s)
-			  rd = s+1;
-			else
-			  break;
-		}
-	}
-	setfreefd();
-	if (stability)
-		rd_initstability();
-	stickymem = oval;
-	did_cnt = 0;
-	i = -1;
-	for (; !mustexit ;) {
-		++i;	/* Increment it */
-		/* The last of the alternate dirs ?  Reset.. */
-		if (i >= ROUTERDIR_CNT || dirs[i] == NULL) {
-			i = 0;
-
-			canexit = 1;
-		/*
-		 * If a shell signal interrupts us here, there is
-		 * potential for problems knowing which file descriptors
-		 * are free.  One could add a setfreefd() to the trap
-		 * routine, in that case.
-		 */
-			sleep(sweepintvl);
-			canexit = 0;
-		}
-
-		/*
-		 * We would like to do a seekdir()/rewinddir()
-		 * instead of opendir()/closedir()  inside this
-		 * loop to avoid allocating and freeing a chunk
-		 * of memory all the time.  This can lead to
-		 * memory fragmentation and thus growing VM.
-		 *
-		 * However several systems do NOT guarantee that
-		 * rewinddir() will find any new data from the
-		 * system...
-		 */
-		/* rewinddir(dirp[i]); */	/* some system w/o this ? */
-
-#if 0
-#ifdef	BUGGY_CLOSEDIR
-		/*
-		 * Major serious bug time here;  some closedir()'s
-		 * free dirp before referring to dirp->dd_fd. GRRR.
-		 * XX: remove this when bug is eradicated from System V's.
-		 */
-		close(dirp[i]->dd_fd);
-#endif
-		closedir(dirp[i]);
-#endif
-		dirp[i] = opendir(dirs[i][0] == 0 ? "." : dirs[i]);
-		did_cnt = 0;
-		if (dirp[i] != NULL) {
-		  if (stability)
-		    did_cnt = rd_stability(dirp[i],dirs[i]);
-		  else
-		    did_cnt = rd_instability(dirp[i],dirs[i]);
-		}
-
-		if (stability != real_stability) {
-			stability = real_stability;
-			if (stability == 0)
-				rd_endstability();
-			else
-				rd_initstability();
-		}
-#if 1
-#ifdef	BUGGY_CLOSEDIR
-		/*
-		 * Major serious bug time here;  some closedir()'s
-		 * free dirp before referring to dirp->dd_fd. GRRR.
-		 * XX: remove this when bug is eradicated from System V's.
-		 */
-		close(dirp[i]->dd_fd);
-#endif
-		closedir(dirp[i]);
-		dirp[i] = NULL;
-#endif
-
-		if (mustexit)
-			break;
-
-		/* Alter router directory.  If processed directory
-		   had any job, reset the  index  to the begin.   */
-		if (did_cnt > 0)
-			i = -1;
-	}
-	for (i=0; dirp[i] != NULL && i < ROUTERDIR_CNT; ++i) {
-#if 0
-#ifdef	BUGGY_CLOSEDIR
-		/*
-		 * Major serious bug time here;  some closedir()'s
-		 * free dirp before referring to dirp->dd_fd. GRRR.
-		 * XX: remove this when bug is eradicated from System V's.
-		 */
-		close(dirp[i]->dd_fd);
-#endif
-		closedir(dirp[i]);
-#endif
-		free(dirs[i]);
-	}
-	return 0;
-}
-
-/*
- * Based on the name of a message file, figure out what to do with it.
- */
-
-struct protosw {
-	const char *pattern;
-	const char *function;
-} psw[] = {
-/*{	"[0-9]*.x400",		"x400"		}, */
-/*{	"[0-9]*.fax",		"fax"		}, */
-/*{	"[0-9]*.uucp",		"uucp"		}, */
-{	"[0-9]*",		"rfc822"	},
-};
-
-
-static int
-run_process(argc, argv)
-	int argc;
-	const char *argv[];
-{
-	struct protosw *pswp;
-	char *file;
-	int r;
-	char *sh_memlevel = getlevel(MEM_SHCMD);
-
-	if (argc != 2 || argv[1][0] == '\0') {
-		fprintf(stderr, "Usage: %s messagefile\n", argv[0]);
-		return PERR_USAGE;
-	}
-#ifdef	USE_ALLOCA
-	file = (char*)alloca(strlen(argv[1])+1);
-#else
-	file = (char*)emalloc(strlen(argv[1])+1);
-#endif
-	strcpy(file, argv[1]);
-
-	r = 0;	/* by default, ignore it */
-	for (pswp = &psw[0]; pswp < &psw[(sizeof psw / sizeof psw[0])]; ++pswp)
-		if (strmatch(pswp->pattern, file)) {
-			printf("process %s %s\n", pswp->function, file);
-			argv[0] = pswp->function;
-			r = s_apply(argc, argv); /* process-by-FUNC filename */
-			printf("done with %s\n", file);
-			if (r)
-				printf("status %d\n", r);
-			break;
-		}
-
-#ifndef	USE_ALLOCA
-	free(file);
-#endif
-	setlevel(MEM_SHCMD,sh_memlevel);
-
-	return r;
-}
-
 
 /*
  * Print a list of the members of a group.
@@ -2131,6 +1511,7 @@ run_filepriv(argc, argv)
 	memcpy(dir, file, cp - file);
 	dir[cp - file] = '\0';
 
+	/* Use stat(2), we fold symlinks to their final files */
 	if (stat(dir, &stbuf) < 0 || !S_ISDIR(stbuf.st_mode)) {
 		fprintf(stderr, "%s: not a directory: \"%s\"!\n",
 			argv0, dir);
@@ -2193,6 +1574,41 @@ run_runas(argc, argv)
 	return r;
 }
 
+static int
+run_cat(argc, argv)
+	int argc;
+	const char *argv[];
+{
+	FILE *fp;
+	char buf[8192];
+	int i;
+	struct stat stbuf;
+
+	if (argc < 2) {
+		fprintf(stderr, "Usage: %s [filenames ...]\n", argv[0]);
+		return 1;
+	}
+
+	for ( ;argv[1] != NULL; ++argv) {
+	  /* Must be a regular file (via a symlink, though!), no
+	     pipes, sockets, devices... */
+	  if (stat(argv[1], &stbuf) == 0 && S_ISREG(stbuf.st_mode) &&
+	      (fp = fopen(argv[1], "r"))) {
+	    for (;!ferror(fp) && !feof(fp);) {
+	      i = fread(buf, 1, sizeof(buf), fp);
+	      if (i > 0) {
+		int j = 0;
+		while (j < i)
+		  j += fwrite(buf + j, 1, i - j, stdout);
+	      } else
+		break;
+	    }
+	    fclose(fp);
+	  }
+	}
+	return 0;
+}
+
 
 static int
 run_uid2login(argc, argv)
