@@ -45,6 +45,7 @@
 #ifdef __linux__
 #define __USE_BSD 1	/* Linux headers ... Auch..  <endian.h> */
 #endif
+#include <arpa/inet.h>
 #include <arpa/nameser.h>
 #include <resolv.h>
 #include <string.h>
@@ -161,7 +162,7 @@ struct qtypes {
 	{	0,		0		}
 };
 
-const char *zh_errlist[] = {
+static const char *zh_errlist[] = {
 	"Error 0",
 	"unknown host/not findable at all",	/* 1 - HOST_NOT_FOUND	*/
 	"host name lookup/try again",		/* 2 - TRY_AGAIN	*/
@@ -169,6 +170,18 @@ const char *zh_errlist[] = {
 	"no address associated with name"	/* 4 - NO_ADDRESS	*/
 };
 
+static const char *res_respcodes[] = {
+	"NOERROR",
+	"FORMERR",
+	"SERVFAIL",
+	"NXDOMAIN",
+	"NOTIMP",
+	"REFUSED",
+	"Resp6",
+	"Resp7",
+	"Resp8"
+};
+
 static conscell * getmxrr    __((const char *, const char *, time_t *, int));
 static conscell * getcrrtype __((const char *, int, time_t *, int));
 static conscell * getrrtypec __((const char *, int, time_t *, int));
@@ -281,7 +294,7 @@ search_res(sip)
 		    "search_res: deferred: %s: %s (%d) error\n",
 		    host, qtp->typename, h_errno);
 
-	  sprintf(buf, "NS:%s/%s", host, qtp->typename);
+	  sprintf(buf, "NS:%.500s/%.20s", host, qtp->typename);
 	  v_set(DEFER, buf);
 	}
 	return rval;
@@ -310,11 +323,17 @@ getmxrr(host, localhost, ttlp, depth)
 	CUC *eom, *cp;
 	querybuf buf, answer;
 	int n, qlen, nmx, i, ancount, qdcount, maxpref;
-	conscell *lhead, *l, *tmp;
+	conscell *lhead, *l;
 	struct mxdata mx[20];
 	char hbuf[MAXNAME], realname[MAXNAME];
 	int type;
 	time_t ttl, maxttl;
+	GCVARS1;
+
+	nmx = 0;
+	maxttl = ttl = 0;
+	maxpref = -1;
+	lhead = l = NULL;
 
 	if (depth > 4) {
 	  fprintf(stderr,
@@ -364,21 +383,10 @@ getmxrr(host, localhost, ttlp, depth)
 				strcpy(h_errhost, host);
 				h_errno = TRY_AGAIN;
 				return NULL;
-#ifdef	OLDJEEVES
-			/*
-			 * Jeeves (TOPS-20 server) still does not
-			 * support MX records.  For the time being,
-			 * we must accept FORMERRs as the same as
-			 * NOERROR.
-			 */
-			case FORMERR:
-#endif	/* OLDJEEVES */
 			case NOERROR:
 				/* if we got this, then ancount == 0! */
 				return NULL /*getrrtypec(host, T_A, ttlp)*/;
-#ifndef	OLDJEEVES
 			case FORMERR:
-#endif	/* !OLDJEEVES */
 			case NOTIMP:
 			case REFUSED:
 				strcpy(h_errhost, host);
@@ -394,10 +402,7 @@ getmxrr(host, localhost, ttlp, depth)
 #else	/* !defined(BIND_VER) || (BIND_VER < 473) */
 		cp += dn_skip((CUC*)cp) + QFIXEDSZ;
 #endif	/* defined(BIND_VER) && (BIND_VER >= 473) */
-	nmx = 0;
-	maxttl = ttl = 0;
 	realname[0] = '\0';
-	maxpref = -1;
 	/* assert: stickymem == MEM_MALLOC;  for storing RHS of MX RR's */
 	while (--ancount >= 0 && cp < eom && nmx < (sizeof mx/sizeof mx[0])) {
 		n = dn_expand((CUC*)&answer, (CUC*)eom, (CUC*)cp,
@@ -432,7 +437,7 @@ getmxrr(host, localhost, ttlp, depth)
 		if (n < 0)
 			break;
 		cp += n;
-		mx[nmx].host = (char *)strsave((char*)hbuf);
+		mx[nmx].host = (char *)strdup((char*)hbuf);
 		if (localhost != NULL && CISTREQ(hbuf, localhost))
 		    if ((maxpref < 0) || (maxpref > (int)mx[nmx].pref))
 			maxpref = mx[nmx].pref;
@@ -467,8 +472,10 @@ getmxrr(host, localhost, ttlp, depth)
 #endif	/* RFC974 */
 	/* determine how many are left, and their max ttl */
 	n = 0;
-	lhead = l = NULL;
+	GCPRO1(lhead);
 	for (i = 0; i < nmx; ++i) {
+		int slen;
+		char *s;
 		if (mx[i].host == NULL)
 			continue;
 		++n; /* found one! */
@@ -477,20 +484,28 @@ getmxrr(host, localhost, ttlp, depth)
 		if (D_bind || _res.options & RES_DEBUG)
 			fprintf(stderr, "search_res: %s: mx[%d] = %s\n",
 				host, n, mx[i].host);
+		slen = strlen(mx[i].host);
+		s = dupnstr(mx[i].host, slen);
 		if (lhead == NULL)
-			lhead = l = newstring(mx[i].host);
+			lhead = l = newstring(s,slen);
 		else {
-			cdr(l) = newstring(mx[i].host);
+			cdr(l) = newstring(s,slen);
 			l = cdr(l);
 		}
+		if (mx[i].host) free(mx[i].host);
+		mx[i].host = NULL;
 	}
+	if (lhead)
+		lhead = ncons(lhead);
+	UNGCPRO1;
+
 	if (D_bind || _res.options & RES_DEBUG)
 		fprintf(stderr, "search_res: %s: %d valid MX RR's\n", host, n);
 	if (n == 0) /* MX's exist, but their WKS's show no TCP smtp service */
 		return NULL;
 	else if (maxttl > 0)
 		*ttlp = maxttl;
-	return ncons(lhead);
+	return lhead;
 }
 
 static conscell *
@@ -523,7 +538,7 @@ getcrrtype(host, rrtype, ttlp, depth)	/* getrrtypec() with completion */
 	}
 	for (i = 0, domain = _res.dnsrch[i];
 	     h_errno == 0 && (domain = _res.dnsrch[i]) != NULL; ++i) {
-		sprintf(buf, "%s.%s", host, domain);
+		sprintf(buf, "%.200s.%.300s", host, domain);
 		rval = getrrtypec(buf, rrtype, ttlp, depth+1);
 		if (rval != NULL)
 			return rval;
@@ -539,17 +554,17 @@ getrrtypec(host, rrtype, ttlp, depth)
 	time_t *ttlp;
 	int depth;
 {
-	conscell *lhead, *l, *tmp;
+	conscell *lhead, *l;
 	char *s;
 	HEADER *hp;
 	CUC *eom, *cp, *nextcp;
 	querybuf buf, answer;
 	int qlen, n, ancount, qdcount, ok, first;
-	struct in_addr inaddr;
 	time_t maxttl, ttl;
 	int type;
 	char nbuf[BUFSIZ];
 	char hb[MAXNAME];
+	GCVARS1;
 
 	if (depth > 4) {
 		fprintf(stderr,
@@ -594,10 +609,11 @@ getrrtypec(host, rrtype, ttlp, depth)
 	 * We don't care about errors here, only if we got an answer
 	 */
 	if (ancount == 0) {
+		int slen; char *s;
 		if (D_bind || _res.options & RES_DEBUG)
 			fprintf(stderr,
-				"search_res: rcode=%d, ancount=%d, rrtype=%d\n",
-				hp->rcode, ancount, rrtype);
+				"search_res: rcode=%s, ancount=%d, rrtype=%d\n",
+				res_respcodes[hp->rcode], ancount, rrtype);
 		if (rrtype == T_CNAME && hp->rcode == NOERROR) {
 		  if (qdcount > 0 && strchr(host, '.') == NULL) {
 		    cp = (CUC*) &answer + sizeof(HEADER);
@@ -606,13 +622,20 @@ getrrtypec(host, rrtype, ttlp, depth)
 		      if (hb[0] == '\0') {
 			hb[0] = '.'; hb[1] = '\0';
 		      }
-		      return newstring(strsave(hb));
+		      slen = strlen(hb);
+		      s = dupnstr(hb, slen);
+		      return newstring(s, slen);
 		    }
 		  }
-		  return newstring(strsave(host));
+		  slen = strlen(host);
+		  s = dupnstr(host, slen);
+		  return newstring(s, slen);
+		}
+		if (rrtype == T_WKS) { /* absence of WKS means YES ... */
+		  slen = strlen(host);
+		  s = dupnstr(host, slen);
+		  return newstring(s, slen);
 		}
-		if (rrtype == T_WKS) /* absence of WKS means YES ... */
-			return newstring(strsave(host));
 		return NULL;
 	}
 	cp = (CUC *)&answer + sizeof(HEADER);
@@ -626,7 +649,9 @@ getrrtypec(host, rrtype, ttlp, depth)
 	ok = rrtype != T_WKS;
 	maxttl = 0;
 	l = NULL;
-	for (lhead = NULL; --ancount >= 0 && cp < eom; cp = nextcp) {
+	lhead = NULL;
+	GCPRO1(lhead);
+	for (; --ancount >= 0 && cp < eom; cp = nextcp) {
 		n = dn_expand((CUC*)&answer, (CUC*)eom, (CUC*)cp,
 			      (void*)nbuf, sizeof(nbuf));
 		if (n < 0)
@@ -673,8 +698,14 @@ getrrtypec(host, rrtype, ttlp, depth)
 
 			*ttlp = maxttl;
 			n = (*cp) & 0xFF;
-			if (0 < n && n < sizeof hb)
-			  return newstring(strnsave((const char *)cp+1, n));
+			if (0 < n && n < (int)sizeof(hb)) {
+			  int slen; char *s;
+			  UNGCPRO1;
+			  *(char*)(cp+1+n) = 0;
+			  slen = strlen(cp+1);
+			  s = dupnstr(cp+1, slen);
+			  return newstring(s, slen);
+			}
 			break;
 
 		case T_WKS:
@@ -692,8 +723,11 @@ getrrtypec(host, rrtype, ttlp, depth)
 			if (*cp++ == IPPROTO_TCP) {	/* check protocol */
 			  if (cp + (IPPORT_SMTP/8) < nextcp &&
 			      *(cp+(IPPORT_SMTP/8)) & (0x80>>IPPORT_SMTP%8)) {
+			    int slen = strlen(hb);
+			    char *s = dupnstr(hb, slen);
 			    *ttlp = maxttl;
-			    return newstring(strsave(hb));
+			    UNGCPRO1;
+			    return newstring(s, slen);
 			  }
 			}
 			continue;
@@ -702,10 +736,14 @@ getrrtypec(host, rrtype, ttlp, depth)
 		case T_AAAA:
 			*ttlp = maxttl;
 			if (rrtype == T_ANY) {
-			  return newstring(strsave(host));
+			  int slen = strlen(host);
+			  char *s = dupnstr(host, slen);
+			  UNGCPRO1;
+			  return newstring(s, slen);
 			} else {
 			  char tb[80];
 			  const char *ss;
+			  int slen;
 
 #if defined(AF_INET6) && defined(INET6)
 			  if (type == T_AAAA)
@@ -713,11 +751,12 @@ getrrtypec(host, rrtype, ttlp, depth)
 			  else
 #endif
 			    ss = inet_ntop(AF_INET, cp, tb, sizeof(tb));
-			  s = strsave(ss);
+			  slen = strlen(ss);
+			  s = dupnstr(ss, slen);
 			  if (lhead == NULL)
-			    lhead = l = newstring(s);
+			    lhead = l = newstring(s, slen);
 			  else {
-			    cdr(l) = newstring(s);
+			    cdr(l) = newstring(s, slen);
 			    l = cdr(l);
 			  }
 			}
@@ -753,7 +792,12 @@ getrrtypec(host, rrtype, ttlp, depth)
 				/* chase it down */
 				getrrtypec(nbuf, rrtype, ttlp, depth+1);
 			*ttlp = maxttl;
-			return newstring(strsave(nbuf));
+			UNGCPRO1;
+			{
+			  int slen = strlen(nbuf);
+			  char *s = dupnstr(nbuf, slen);
+			  return newstring(s, slen);
+			}
 
 		case T_SOA:
 		case T_NS:
@@ -772,7 +816,10 @@ getrrtypec(host, rrtype, ttlp, depth)
 
 		case T_MX:
 			if (rrtype == T_ANY) {
-			    return newstring(strsave(host));
+			  int slen = strlen(host);
+			  char *s = dupnstr(host, slen);
+			  UNGCPRO1;
+			  return newstring(s, slen);
 			}
 		default:
 			fprintf(stderr,"search_res: getrrtypec: non-processed RR type for host='%s' query=%d, result=%d\n",host,rrtype,type);
@@ -781,11 +828,16 @@ getrrtypec(host, rrtype, ttlp, depth)
 		}
 	}
 	if (lhead)
-		return ncons(lhead);
+		lhead = ncons(lhead);
+	UNGCPRO1;
+	if (lhead)
+		return lhead;
 	if (ok) {
 		const char *cs = first ? host : hb;
+		int slen = strlen(cs);
+		char *s = dupnstr(cs, slen);
 		*ttlp = maxttl;
-		return newstring(strsave(cs));
+		return newstring(s, slen);
 	}
 	return NULL;
 }
