@@ -2,7 +2,7 @@
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
  *
- *	Copyright 1996-1997 Matti Aarnio
+ *	Copyright 1996-1999 Matti Aarnio
  */
 
 /* LINTLIBRARY */
@@ -12,7 +12,11 @@
 #ifdef HAVE_FCNTL_H
 # include <fcntl.h>
 #endif
-#include <db.h>
+#if defined(HAVE_DB_185_H) && !defined(HAVE_DB_OPEN2)
+# include <db_185.h>
+#else
+# include <db.h>
+#endif
 #include <sys/file.h>
 #include "search.h"
 #include "io.h"
@@ -29,14 +33,15 @@ extern int deferit;
  */
 
 void
-close_bhash(sip)
+close_bhash(sip,comment)
 	search_info *sip;
+	const char *comment;
 {
 	DB *db;
 	struct spblk *spl = NULL;
 	spkey_t symid;
 
-	if (sip->file == NULL || spt_files->symbols == NULL)
+	if (sip->file == NULL)
 		return;
 	symid = symbol_lookup_db(sip->file, spt_files->symbols);
 	if ((spkey_t)0 != symid)
@@ -46,7 +51,11 @@ close_bhash(sip)
 	spl = sp_lookup(symid, spt_files);
 	if (spl == NULL || (db = (DB *)spl->data) == NULL)
 		return;
+#ifdef HAVE_DB_OPEN2
+	(db->close)(db,0);
+#else
 	(db->close)(db);
+#endif
 	symbol_free_db(sip->file, spt_files->symbols);
 	sp_delete(spl, spt_files);
 }
@@ -59,24 +68,29 @@ open_bhash(sip, flag, comment)
 	int flag;
 	const char *comment;
 {
-	DB *db;
+	DB *db = NULL;
 	struct spblk *spl;
 	spkey_t symid;
 	int i;
 
 	if (sip->file == NULL)
 		return NULL;
-	if (spt_files == NULL)
-	  spt_files = sp_init();
-	if (spt_files->symbols == NULL)
-	  spt_files->symbols = sp_init();
+
 	symid = symbol_db(sip->file, spt_files->symbols);
 	spl = sp_lookup(symid, spt_files);
-	if (spl != NULL && flag == O_RDWR && spl->mark != O_RDWR)
-		close_bhash(sip);
+	if (spl != NULL && flag != spl->mark)
+		close_bhash(sip,"open_bhash");
 	if (spl == NULL || (db = (DB *)spl->data) == NULL) {
 		for (i = 0; i < 3; ++i) {
+#ifdef HAVE_DB_OPEN2
+		  int err;
+		  db = NULL;
+		  err = db_open(sip->file, DB_HASH,
+				DB_NOMMAP|((flag == O_RDONLY) ? DB_RDONLY:DB_CREATE),
+				0644, NULL, NULL, &db);
+#else
 		  db = dbopen(sip->file, flag, 0, DB_HASH, NULL);
+#endif
 		  if (db != NULL)
 		    break;
 		  sleep(1); /* Open failed, retry after a moment */
@@ -90,8 +104,10 @@ open_bhash(sip, flag, comment)
 		}
 		if (spl == NULL)
 			sp_install(symid, (void *)db, flag, spt_files);
-		else
+		else {
 			spl->data = (void *)db;
+			spl->mark = flag;
+		}
 	}
 	return db;
 }
@@ -107,31 +123,40 @@ search_bhash(sip)
 {
 	DB *db;
 	DBT val, key;
-	conscell *tmp;
-	struct spblk *spl;
 	int retry, rc;
-	spkey_t symid;
-	char *us;
 
 	retry = 0;
+#if 0
 reopen:
+#endif
 	db = open_bhash(sip, O_RDONLY, "search_bhash");
 	if (db == NULL)
 	  return NULL; /* Huh! */
 
-	key.data = (char*)sip->key;
+	memset(&val, 0, sizeof(val));
+	memset(&key, 0, sizeof(key));
+
+	key.data = (void*)sip->key;
 	key.size = strlen(sip->key) + 1;
+#ifdef HAVE_DB_OPEN2
+	rc = (db->get)(db, NULL, &key, &val, 0);
+#else
 	rc = (db->get)(db, &key, &val, 0);
+#endif
 	if (rc != 0) {
+
+#if 0 /* SleepyCat's DB 2.x leaks memory mappings when opening...
+	 at least the version at glibc 2.1.1 */
+
 		if (!retry && rc < 0) {
-			close_bhash(sip);
+			close_bhash(sip,"search_bhash");
 			++retry;
 			goto reopen;
 		}
+#endif
 		return NULL;
 	}
-	us = strnsave(val.data, val.size);
-	return newstring(us);
+	return newstring(dupnstr(val.data, val.size), val.size);
 }
 
 
@@ -151,11 +176,23 @@ add_bhash(sip, value)
 	db = open_bhash(sip, O_RDWR, "add_bhash");
 	if (db == NULL)
 		return EOF;
-	key.data = (char*)sip->key;
+
+	memset(&val, 0, sizeof(val));
+	memset(&key, 0, sizeof(key));
+
+	key.data = (void*)sip->key;
 	key.size = strlen(sip->key) + 1;
-	val.data = (char*)value;
+
+	val.data = (void*)value;
 	val.size = strlen(value)+1;
+#ifdef HAVE_DB_OPEN2
+	rc = (db->put)(db, NULL, &key, &val, 0);
+	/* Emulate BSD DB 1.85 behaviour */
+	if (rc != 0)
+	  rc = -1;
+#else
 	rc = (db->put)(db, &key, &val, 0);
+#endif
 	if (rc < 0) {
 		++deferit;
 		v_set(DEFER, DEFER_IO_ERROR);
@@ -181,9 +218,16 @@ remove_bhash(sip)
 	db = open_bhash(sip, O_RDWR, "remove_bhash");
 	if (db == NULL)
 		return EOF;
-	key.data = (char*)sip->key;
+
+	memset(&key, 0, sizeof(key));
+
+	key.data = (void*)sip->key;
 	key.size = strlen(sip->key) + 1;
+#ifdef HAVE_DB_OPEN2
+	rc = (db->del)(db, NULL, &key, 0);
+#else
 	rc = (db->del)(db, &key, 0);
+#endif
 	if (rc < 0) {
 		++deferit;
 		v_set(DEFER, DEFER_IO_ERROR);
@@ -206,11 +250,46 @@ print_bhash(sip, outfp)
 	DB *db;
 	DBT key, val;
 	int rc;
+#ifdef HAVE_DB_OPEN2
+	DBC *curs;
 
 	db = open_bhash(sip, O_RDONLY, "print_bhash");
 	if (db == NULL)
 		return;
 
+#ifdef HAVE_DB_CURSOR4
+	rc = (db->cursor)(db, NULL, &curs, 0);
+#else
+	rc = (db->cursor)(db, NULL, &curs);
+#endif
+	memset(&val, 0, sizeof(val));
+	memset(&key, 0, sizeof(key));
+
+	if (rc == 0 && curs)
+	  rc = (curs->c_get)(curs, &key, &val, DB_FIRST);
+	for ( ; rc == 0 ; ) {
+		if (val.data == NULL)
+			continue;
+		if (*(char*)val.data == '\0')
+			fprintf(outfp, "%s\n", key.data);
+		else
+			fprintf(outfp, "%s\t%s\n", key.data, val.data);
+
+		memset(&val, 0, sizeof(val));
+		memset(&key, 0, sizeof(key));
+
+		rc = (curs->c_get)(curs, &key, &val, DB_NEXT);
+	}
+	(curs->c_close)(curs);
+#else
+
+	db = open_bhash(sip, O_RDONLY, "print_bhash");
+	if (db == NULL)
+		return;
+
+	memset(&val, 0, sizeof(val));
+	memset(&key, 0, sizeof(key));
+
 	rc = (db->seq)(db, &key, &val, R_FIRST);
 	for ( ; rc == 0 ; ) {
 		if (val.data == NULL)
@@ -219,8 +298,13 @@ print_bhash(sip, outfp)
 			fprintf(outfp, "%s\n", key.data);
 		else
 			fprintf(outfp, "%s\t%s\n", key.data, val.data);
+
+		memset(&val, 0, sizeof(val));
+		memset(&key, 0, sizeof(key));
+
 		rc = (db->seq)(db, &key, &val, R_NEXT);
 	}
+#endif
 	fflush(outfp);
 }
 
@@ -238,16 +322,55 @@ count_bhash(sip, outfp)
 	int cnt = 0;
 	int rc;
 
+#ifdef HAVE_DB_OPEN2
+	DBC *curs;
+
 	db = open_bhash(sip, O_RDONLY, "count_bhash");
+
 	if (db != NULL) {
-	  rc = (db->seq)(db, &key, &val, R_FIRST);
+#ifdef HAVE_DB_CURSOR4
+	  rc = (db->cursor)(db, NULL, &curs, 0);
+#else
+	  rc = (db->cursor)(db, NULL, &curs);
+#endif
+
+	  memset(&val, 0, sizeof(val));
+	  memset(&key, 0, sizeof(key));
+
+	  if (rc == 0 && curs)
+	    rc = (curs->c_get)(curs, &key, &val, DB_FIRST);
 	  while (rc == 0) {
-	    if (val.data == NULL)
+	    if (val.data == NULL) /* ???? When this would happen ? */
 	      continue;
 	    ++cnt;
+
+	    memset(&val, 0, sizeof(val));
+	    memset(&key, 0, sizeof(key));
+
+	    rc = (curs->c_get)(curs, &key, &val, DB_NEXT);
+	  }
+	}
+	(curs->c_close)(curs);
+#else
+	db = open_bhash(sip, O_RDONLY, "count_bhash");
+	if (db != NULL) {
+
+	  memset(&val, 0, sizeof(val));
+	  memset(&key, 0, sizeof(key));
+
+	  rc = (db->seq)(db, &key, &val, R_FIRST);
+	  while (rc == 0) {
+	    if (val.data == NULL) /* ???? When this would happen ? */
+	      continue;
+	    ++cnt;
+
+	    memset(&val, 0, sizeof(val));
+	    memset(&key, 0, sizeof(key));
+
 	    rc = (db->seq)(db, &key, &val, R_NEXT);
 	  }
 	}
+#endif
 	fprintf(outfp,"%d\n",cnt);
 	fflush(outfp);
 }
@@ -265,6 +388,7 @@ owner_bhash(sip, outfp)
 {
 	DB *db;
 	struct stat stbuf;
+	int fd;
 
 	db = open_bhash(sip, O_RDONLY, "owner_bhash");
 	if (db == NULL)
@@ -272,7 +396,12 @@ owner_bhash(sip, outfp)
 
 	/* There are more timing hazards, when the internal fd is not
 	   available for probing.. */
-	if (fstat((db->fd)(db), &stbuf) < 0) {
+#ifdef HAVE_DB_OPEN2
+	(db->fd)(db, &fd);
+#else
+	fd = (db->fd)(db);
+#endif
+	if (fstat(fd, &stbuf) < 0) {
 		fprintf(stderr, "owner_bhash: cannot fstat(\"%s\")!\n",
 				sip->file);
 		return;
@@ -289,13 +418,18 @@ modp_bhash(sip)
 	struct stat stbuf;
 	struct spblk *spl;
 	spkey_t symid;
-	int rval;
+	int rval, fd;
 
 	db = open_bhash(sip, O_RDONLY, "owner_bhash");
 	if (db == NULL)
 		return 0;
 
-	if (fstat((db->fd)(db), &stbuf) < 0) {
+#ifdef HAVE_DB_OPEN2
+	(db->fd)(db, &fd);
+#else
+	fd = (db->fd)(db);
+#endif
+	if (fstat(fd, &stbuf) < 0) {
 		fprintf(stderr, "modp_bhash: cannot fstat(\"%s\")!\n",
 				sip->file);
 		return 0;
@@ -307,11 +441,11 @@ modp_bhash(sip)
 	spl = sp_lookup(symid, spt_modcheck);
 	if (spl != NULL) {
 		rval = ((long)stbuf.st_mtime != (long)spl->data ||
-			(long)stbuf.st_nlink != (long)spl->mark);
+			(long)stbuf.st_nlink != 1);
 	} else
 		rval = 0;
 	sp_install(symid, (void *)((long)stbuf.st_mtime),
-		   stbuf.st_nlink, spt_modcheck);
+		   (long)stbuf.st_nlink, spt_modcheck);
 	return rval;
 }
 #endif	/* HAVE_DB */
