@@ -8,6 +8,7 @@
 #include "mailer.h"
 #include <ctype.h>
 #include <string.h>
+
 #include "search.h"
 #include "io.h"
 #include "libz.h"
@@ -30,12 +31,12 @@ struct headerinfo mandatory_hdrs[] = {
 { "cc",			AddressList,	Recipient,	normal		},
 { "from",		AMailboxList,	Sender,		normal		},
 { "message-id",		MessageID,	nilUserType,	normal		},
-{ "reply-to",		AMailboxList,	Sender,		normal		},
+{ "reply-to",		AddressList,	Sender,		normal		},
 { "resent-bcc",		Addresses,	Recipient,	Resent		},
 { "resent-cc",		AddressList,	Recipient,	Resent		},
 { "resent-from",	AMailboxList,	Sender,		Resent		},
 { "resent-message-id",	MessageID,	nilUserType,	Resent		},
-{ "resent-reply-to",	AMailboxList,	Sender,		Resent		},
+{ "resent-reply-to",	AddressList,	Sender,		Resent		},
 { "resent-sender",	Mailbox,	Sender,		Resent		},
 { "resent-to",		AddressList,	Recipient,	Resent		},
 { "sender",		Mailbox,	Sender,		normal		},
@@ -52,11 +53,11 @@ struct headerinfo optional_hdrs[] = {
 { "encrypted",		Encrypted,	nilUserType,	normal		},
 { "errors-to",		AddressList,	Sender,		normal		},
 { "obsoletes",		MessageIDList,	nilUserType,	normal		},
-#if 0
+#if 0 /* Don't parse these .. */
+{ "received",		Received,	nilUserType,	normal		},
 { "keywords",		PhraseList,	nilUserType,	normal		},
 { "references",		References,	nilUserType,	normal		},
 { "in-reply-to",	References,	nilUserType,	normal		},
-{ "received",		Received,	nilUserType,	normal		},
 #endif
 };
 
@@ -74,7 +75,7 @@ struct headerinfo envelope_hdrs[] = {
 { "identinfo",	nilHeaderSemantics,	nilUserType,	eIdentinfo	},
 { "loginname",		UserAtDomain,	nilUserType,	ePrettyLogin	},
 { "notaryret",	nilHeaderSemantics,	nilUserType,	eNotaryRet	},
-{ "rcvdfrom",		UserAtDomain,	nilUserType,	eRcvdFrom	},
+{ "rcvdfrom",		DomainName,	nilUserType,	eRcvdFrom	},
 { "to",			AddressList,	Recipient,	eTo		},
 { "todsn",	nilHeaderSemantics,	Recipient,	eToDSN		},
 { "user",		Mailbox,	nilUserType,	eUser		},
@@ -119,7 +120,7 @@ static HeaderSemantics
 semname2enum(name)
 	const char *name;
 {
-	int i;
+	unsigned int i;
 
 	for (i = 0; i < (sizeof hdrsemtable / sizeof hdrsemtable[0]); ++i) {
 	  if (cistrcmp(name, hdrsemtable[i].name) == 0)
@@ -129,11 +130,11 @@ semname2enum(name)
 }
 
 static const char * semenum2name __((HeaderSemantics d));
-const char *
+static const char *
 semenum2name(d)
 	HeaderSemantics d;
 {
-	int i;
+	unsigned int i;
 
 	for (i = 0; i < (sizeof hdrsemtable / sizeof hdrsemtable[0]); ++i) {
 	  if (d == hdrsemtable[i].semantics)
@@ -146,7 +147,7 @@ void
 init_header()
 {
 	struct headerinfo rh, *rhp;
-	int i;
+	unsigned int i;
 	spkey_t symid;
 
 	if (spt_headers == NULL)           spt_headers           = sp_init();
@@ -249,8 +250,9 @@ search_header(sip)
 {
 	struct sptree *hdb;
 	struct headerinfo *rhp;
-	conscell *tmp;
 	char buf[1024];
+	int slen;
+	char *s;
 
 	hdb = open_header(sip);
 	if (hdb == NULL)
@@ -263,7 +265,9 @@ search_header(sip)
 		rhp->user_type  == Sender    ? "Sender"    :
 		(rhp->user_type == Recipient ? "Recipient" : ""),
 		rhp->class      == Resent    ? "Resent"    : "");
-	return newstring(strsave(buf));
+	slen = strlen(buf);
+	s = dupnstr(buf, slen);
+	return newstring(s, slen);
 }
 
 /*
@@ -281,8 +285,9 @@ hdfreedata(spl)
 }
 
 void
-close_header(sip)
+close_header(sip,comment)
 	search_info *sip;
+	const char *comment;
 {
 	struct sptree *hdb;
 
@@ -313,7 +318,10 @@ add_header(sip, cvalue)
 {
 	struct sptree *hdb;
 	struct spblk *spl;
-	char *cp, *value, *vbuf;
+	char *cp, *value;
+#ifndef USE_ALLOCA
+	char *vbuf;
+#endif
 	char *wcp;
 	char *lcbuf;
 	struct headerinfo rh, *rhp;
@@ -513,9 +521,9 @@ static int
 hdprintdata(spl)
 	struct spblk *spl;
 {
-	struct headerinfo *rhp;
+	const struct headerinfo *rhp;
 
-	rhp = (struct headerinfo *)spl->data;
+	rhp = (const struct headerinfo *)spl->data;
 	fprintf(pcfp, "%-16s\t%s:%s:%s (%s)\n", pname(spl->key),
 		semenum2name(rhp->semantics),
 		rhp->user_type  == Sender    ? "Sender"    :
