@@ -22,7 +22,6 @@
 #include "zsyslog.h"
 
 #include "prototypes.h"
-#include "libz.h"
 #include "libsh.h"
 
 #ifndef _IOFBF
@@ -56,21 +55,27 @@ static conscell	*find_errto __((conscell *list));
 
 #define dprintf	if (D_sequencer) printf
 
-#define	QCHANNEL(x)	(x)->string
-#define	QHOST(x)	(cdr(x))->cstring
-#define	QUSER(x)	(cddr(x))->cstring
-#define	QATTRIBUTES(x)	(cdr(cddr(x)))->string
+#define	QCHANNEL(x)	(x)
+#define	QHOST(x)	cdr(x)
+#define	QUSER(x)	cddr(x)
+#define	QATTRIBUTES(x)	cdr(cddr(x))
 
 
 conscell *
 makequad()
 {
-	conscell *l, *tmp;
+	conscell *l;
+	GCVARS1;
 
-	l = conststring(NULL);
-	cdr(l) = conststring(NULL);
-	cddr(l) = conststring(NULL);
-	cdr(cddr(l)) = conststring(NULL);
+	l = NULL;
+	GCPRO1(l);
+
+	l = conststring(NULL, 0);
+	cdr(l) = conststring(NULL, 0);
+	cddr(l) = conststring(NULL, 0);
+	cdr(cddr(l)) = conststring(NULL, 0);
+
+	UNGCPRO1;
 	return l;
 }
 
@@ -79,13 +84,24 @@ struct envelope *qate;
 int
 iserrmessage()
 {
-	return (qate != NULL
-		&& qate->e_from_trusted != NULL
-		&& (QCHANNEL(qate->e_from_trusted) == NULL
-		    || CISTREQ(QCHANNEL(qate->e_from_trusted), "error")));
+	return (qate != NULL && qate->e_from_trusted != NULL &&
+		(QCHANNEL(qate->e_from_trusted)->cstring == NULL ||
+		 CISTREQ(QCHANNEL(qate->e_from_trusted)->cstring, "error")));
 }
 
 
+static int zunlink __((const char *));
+static int zunlink(path)
+     const char *path;
+{
+  int rc;
+  while ((rc = unlink(path)) < 0 && (errno == EBUSY || errno == EINTR))
+    ;
+  return rc;
+}
+
+
+
 
 /*
  * Apply RFC822 parsing and processing to the message in the file in argv[1].
@@ -96,11 +112,12 @@ run_rfc822(argc, argv)
 	int argc;
 	const char *argv[];
 {
-	struct envelope *e;
+	struct envelope *e; /* TWO GC PROTECTABLE ITEMS! */
 	const char *file;
-	char buf[8196];
-	int status, errflg;
+	char buf[8196];   /* FILE* buffer, will be released at fclose() */
+	int status = PERR_OK, errflg;
 	memtypes oval;
+	GCVARS2;
 
 	errflg = 0;
 #if 0
@@ -132,6 +149,9 @@ run_rfc822(argc, argv)
 	stickymem = MEM_TEMP;	/* per-message space */
 
 	e = (struct envelope *)tmalloc(sizeof (struct envelope));
+	/* XXX: If this tmalloc() fails, we crash! */
+	memset(e, 0, sizeof(*e)); /* Lots of pointers, etc here! */
+	GCPRO2(e->e_from_trusted, e->e_from_resolved);
 	qate = e;
 
 	if ((e->e_fp = fopen(file, "r")) == NULL) {
@@ -198,7 +218,7 @@ run_rfc822(argc, argv)
 	if (D_final > 0)
 		dumpInfo(e);
 	if (status != PERR_CTRLFILE && status != PERR_DEFERRED && !savefile)
-		(void) unlink(file); /* SILENT! */
+		(void) zunlink(file); /* SILENT! */
 	if (e->e_fp != NULL)
 		fclose(e->e_fp);
 	tfree(MEM_TEMP);
@@ -206,6 +226,7 @@ run_rfc822(argc, argv)
 #ifdef	XMEM
 	mal_contents(stdout);
 #endif	/* XMEM */
+	UNGCPRO2;
 	return status;
 }
 
@@ -237,45 +258,45 @@ makeLetter(e, octothorp)
 	}
 	e->e_localtime = *(localtime(&(e->e_statbuf.st_mtime)));
 #ifdef	HAVE_ST_BLKSIZE
-	initline((int)e->e_statbuf.st_blksize);
+	initzline((int)e->e_statbuf.st_blksize);
 #else	/* !HAVE_ST_BLKSIZE */
-	initline(4096);
+	initzline(4096);
 #endif	/* !HAVE_ST_BLKSIZE */
 
 	inheader = 0;
-	while ((n = getline(e->e_fp)) > !octothorp) {
+	while ((n = zgetline(e->e_fp)) > !octothorp) {
 		/* We do kludgy processing things in case the input
 		   does have a CRLF at the end of the line.. */
 		if (n > 1 &&
-		    linebuf[n-2] == '\r' &&
-		    linebuf[n-1] == '\n') {
+		    zlinebuf[n-2] == '\r' &&
+		    zlinebuf[n-1] == '\n') {
 			--n;
-			linebuf[n-1] = '\n';
+			zlinebuf[n-1] = '\n';
 			if (n <= !octothorp) break;
 		}
 		/* Ok, now we can proceed with the original agenda.. */
-		i = hdr_status(linebuf, linebuf, n, octothorp);
+		i = hdr_status(zlinebuf, zlinebuf, n, octothorp);
 		if (i > 0) {		/* a real message header */
-			if (octothorp && linebuf[0] == '#')
+			if (octothorp && zlinebuf[0] == '#')
 				continue;
 			/* record start of headers for posterity */
 			if (!inheader)
-			  e->e_hdrOffset = lineoffset(e->e_fp)-n;
+			  e->e_hdrOffset = zlineoffset(e->e_fp)-n;
 			/* cons up a new one at top of header list */
-			h = makeHeader(spt_headers, linebuf, i);
+			h = makeHeader(spt_headers, zlinebuf, i);
 			h->h_next = e->e_headers;
 			e->e_headers = h;
-			h->h_lines = makeToken(linebuf+i+1, n-i-2);
+			h->h_lines = makeToken(zlinebuf+i+1, n-i-2);
 			h->h_lines->t_type = Line;
 			++inheader;
 		} else if (i == 0) {	/* a continuation line */
-			if (inheader && linebuf[0] == ':') {
+			if (inheader && zlinebuf[0] == ':') {
 				optsave(FYI_ILLHEADER, e);
 				/* cons up a new one at top of header list */
 				h = makeHeader(spt_headers,"X-Null-Field",12);
 				h->h_next = e->e_headers;
 				e->e_headers = h;
-				h->h_lines = makeToken(linebuf+i+1, n-i-2);
+				h->h_lines = makeToken(zlinebuf+i+1, n-i-2);
 				h->h_lines->t_type = Line;
 			} else if (inheader && n > 1) {
 				/* append to the header we just saw */
@@ -283,30 +304,30 @@ makeLetter(e, octothorp)
 				if (e->e_headers == NULL) {
 				  /* Wow, continuation without previous header! */
 				  /* It must be body.. */
-				  repos_getline(e->e_fp,
-						lineoffset(e->e_fp) - n);
+				  repos_zgetline(e->e_fp,
+						 zlineoffset(e->e_fp) - n);
 				  break;
 				}
 				t = e->e_headers->h_lines;
 				while (t->t_next != NULL)
 					t = t->t_next;
-				t->t_next = makeToken(linebuf, n-1);
+				t->t_next = makeToken(zlinebuf, n-1);
 				t->t_next->t_type = Line;
 			} else if (!octothorp) {
 				return PERR_BADCONTINUATION;
 			}
 		} else if (!inheader		/* envelope information */
-			   && (*(cp=linebuf-i) == ' ' || *cp == '\t'
+			   && (*(cp=zlinebuf-i) == ' ' || *cp == '\t'
 				|| *cp == '\n')) {
 			HeaderSemantics osem;
 
-			if (octothorp && linebuf[0] == '#')
+			if (octothorp && zlinebuf[0] == '#')
 				continue;
 			/* cons up a new one at top of envelope header list */
-			h = makeHeader(spt_eheaders, linebuf, -i);
+			h = makeHeader(spt_eheaders, zlinebuf, -i);
 			h->h_next = e->e_eHeaders;
 			e->e_eHeaders = h;
-			h->h_lines = makeToken(linebuf-i+1, n > 1-i ? n+i-2: 0);
+			h->h_lines = makeToken(zlinebuf-i+1, n > 1-i ? n+i-2: 0);
 			h->h_lines->t_type = Line;
 			switch (h->h_descriptor->class) {
 
@@ -316,7 +337,7 @@ makeLetter(e, octothorp)
 				fprintf(stderr,
 					"%s: unknown envelope header (class %d): ",
 					progname, h->h_descriptor->class);
-				fwrite((char *)linebuf, sizeof (char), -i, stderr);
+				fwrite((char *)zlinebuf, sizeof (char), -i, stderr);
 				putc('\n', stderr);
 				h->h_contents = hdr_scanparse(e, h, octothorp, 0);
 				h->h_stamp = hdr_type(h);
@@ -332,7 +353,7 @@ makeLetter(e, octothorp)
 			case eEnvEnd:		/* more elegance breaks */
 				inheader = 1;
 				/* record start of headers for posterity */
-				e->e_hdrOffset = lineoffset(e->e_fp);
+				e->e_hdrOffset = zlineoffset(e->e_fp);
 				e->e_eHeaders = h->h_next;
 				break;
 			default:		/* a real envelope header */
@@ -345,7 +366,7 @@ makeLetter(e, octothorp)
 			   We play as if it was the end of the headers, and
 			   start of the body.  Keep the first faulty input
 			   around by reseeking into its start.		    */
-			repos_getline(e->e_fp, lineoffset(e->e_fp) - n);
+			repos_zgetline(e->e_fp, zlineoffset(e->e_fp) - n);
 			break;
 		}
 	}
@@ -372,7 +393,7 @@ makeLetter(e, octothorp)
 	}
 	e->e_headers = ph;
 	/* record the start of the message body for posterity */
-	e->e_msgOffset = lineoffset(e->e_fp);
+	e->e_msgOffset = zlineoffset(e->e_fp);
 
 	return PERR_OK;
 }
@@ -515,9 +536,9 @@ squirrel(e, keyw, text)
 	char *path;
 
 #ifndef	USE_ALLOCA
-	path = emalloc(5+sizeof(POSTMANDIR)+10+strlen(keyw));
+	path = (char*)emalloc(5+sizeof(POSTMANDIR)+10+strlen(keyw));
 #else
-	path = alloca(5+sizeof(POSTMANDIR)+10+strlen(keyw));
+	path = (char*)alloca(5+sizeof(POSTMANDIR)+10+strlen(keyw));
 #endif
 	sprintf(path, "../%s/%d.%s",
 		POSTMANDIR, (int)e->e_statbuf.st_ino, keyw);
@@ -622,6 +643,18 @@ FILE *mfp;
 	}
 }
 
+static int nullhost __((conscell *cs));
+static int
+nullhost(cs)
+	conscell *cs;
+{
+	const char *s = cs->cstring;
+
+	return (s == NULL || *s == '\0' || strcmp(s, "-") == 0);
+	/* actually we should also check for localhostness, but lets not
+	   get carried away... */
+}
+
 /* Send the message back to originator with user-friendly chastizing errors */
 
 static void
@@ -648,9 +681,9 @@ reject(e, msgfile)
 	 * is pointing at, into recipient addresses.
 	 */
 #ifndef	USE_ALLOCA
-	path = emalloc(3+strlen(mailshare)+strlen(FORMSDIR)+strlen(msgfile));
+	path = (char*)emalloc(3+strlen(mailshare)+strlen(FORMSDIR)+strlen(msgfile));
 #else
-	path = alloca(3+strlen(mailshare)+strlen(FORMSDIR)+strlen(msgfile));
+	path = (char*)alloca(3+strlen(mailshare)+strlen(FORMSDIR)+strlen(msgfile));
 #endif
 	sprintf(path, "%s/%s/%s", mailshare, FORMSDIR, msgfile);
 	fp = fopen(path, "r");
@@ -761,23 +794,23 @@ defer(e, why)
 	}
 	if (stbuf.st_ino > 0) {
 #ifdef	USE_ALLOCA
-	  path = alloca(5+sizeof(DEFERREDDIR)+6+strlen(why));
+	  path = (char*)alloca(5+sizeof(DEFERREDDIR)+6+strlen(why));
 #else
-	  path = emalloc(5+sizeof(DEFERREDDIR)+6+strlen(why));
+	  path = (char*)emalloc(5+sizeof(DEFERREDDIR)+6+strlen(why));
 #endif
 	  sprintf(path,"../%s/%d.%s", DEFERREDDIR, (int)stbuf.st_ino, why);
 	} else {
 #ifdef	USE_ALLOCA
-	  path = alloca(5+sizeof(DEFERREDDIR)+strlen(e->e_file)+2+strlen(why));
+	  path = (char*)alloca(5+sizeof(DEFERREDDIR)+strlen(e->e_file)+2+strlen(why));
 #else
-	  path = emalloc(5+sizeof(DEFERREDDIR)+strlen(e->e_file)+2+strlen(why));
+	  path = (char*)emalloc(5+sizeof(DEFERREDDIR)+strlen(e->e_file)+2+strlen(why));
 #endif
 	  sprintf(path, "../%s/%s.%s", DEFERREDDIR, e->e_file, why);
 	}
 	s = path;
 	while ((s = strchr(s, ' '))) /* remap blanks to '-':es.. */
 	  *s = '-';
-	unlink(path);
+	zunlink(path);
 
 	/* try renameing every few seconds for a minute */
 	for (i = 0; erename(e->e_file, path) < 0 && i < 10 && errno != ENOENT;++i)
@@ -785,7 +818,7 @@ defer(e, why)
 	if (i >= 10) {
 	  zsyslog((LOG_ALERT, "cannot defer %s (%m)", e->e_file));
 	  fprintf(stderr, "cannot defer %s\n", e->e_file);
-	  /* XX: make sure file is not unlink()ed in main() */
+	  /* XX: make sure file is not zunlink()ed in main() */
 	} else {
 	  zsyslog((LOG_NOTICE, "%s; %sing deferred", e->e_file,
 		   (deferit || deferuid) ? "schedul" : "rout"));
@@ -810,8 +843,11 @@ mkSender(e, name, flag)
 	struct addr *pp, *qp, **ppp;
 	conscell *l;
 	int didbracket;
-	struct spblk *spl;
-	
+	GCVARS1;
+
+	l = NULL;
+	GCPRO1(l);
+
 	sh = (struct header *)tmalloc(sizeof (struct header));
 	sh->h_pname = e->e_resent ? "Resent-From" : "From";
 	sh->h_descriptor = senderDesc();
@@ -838,7 +874,12 @@ mkSender(e, name, flag)
 			if (pp->p_next == NULL)
 				break;
 		}
-	} else if ((spl = sp_lookup(symbol(name), spt_fullnamemap)) != NULL) {
+	} else {
+	  spkey_t spk = symbol_lookup_db(name, spt_fullnamemap->symbols);
+	  struct spblk *spl = NULL;
+	  if (spk)
+	    spl = sp_lookup(spk, spt_fullnamemap);
+	  if (spk && spl != NULL && (char*)(spl->data) != NULL) {
 		*ppp = (struct addr *)tmalloc(sizeof (struct addr));
 		pp = *ppp;
 		ppp = &pp->p_next;
@@ -846,6 +887,7 @@ mkSender(e, name, flag)
 					 strlen((char *)spl->data));
 		pp->p_tokens->t_type = Word;
 		pp->p_type = aPhrase;
+	  }
 	}
 	if (sh->h_contents.a->a_tokens != NULL) {	/* 'Full Name <' */
 		*ppp = (struct addr *)tmalloc(sizeof (struct addr));
@@ -866,7 +908,7 @@ mkSender(e, name, flag)
 			l = pickaddress(l);
 			flag = (QUSER(l) == NULL || !nullhost(QHOST(l)));
 			if (!flag)
-				flag = !CISTREQ(QUSER(l), name);
+			  flag = !CISTREQ(QUSER(l)->cstring, name);
 		} else
 			flag = 1;
 	} else
@@ -921,6 +963,7 @@ mkSender(e, name, flag)
 	sh->h_contents = hdr_scanparse(e, sh, 0, 0);
 #endif
 	/* X: optimize: save this in a hashtable somewhere */
+	UNGCPRO1;
 	return sh;
 }
 
@@ -983,10 +1026,15 @@ mkTrace(e)
 	} else
 		th->h_contents.r->r_with = NULL;
 	/* id */
-	na = (struct addr *)tmalloc(sizeof (struct addr));
-	na->p_tokens = makeToken(e->e_file, strlen(e->e_file));
-	na->p_next = NULL;
-	na->p_type = anAddress;	/* really a message id */
+	{
+	  char taspid[32];
+	  taspoolid(taspid, e->e_statbuf.st_mtime, (long)e->e_statbuf.st_ino);
+
+	  na = (struct addr *)tmalloc(sizeof (struct addr));
+	  na->p_tokens = makeToken(taspid, strlen(taspid));
+	  na->p_next = NULL;
+	  na->p_type = anAddress;	/* really a message id */
+	}
 	ap = (struct address *)tmalloc(sizeof (struct address));
 	ap->a_tokens = na;
 	ap->a_next = NULL;
@@ -1032,15 +1080,6 @@ rwalloc(rwpp)
 	return p;
 }
 
-int
-nullhost(s)
-	const char *s;
-{
-	return (s == NULL || *s == '\0' || strcmp(s, "-") == 0);
-	/* actually we should also check for localhostness, but lets not
-	   get carried away... */
-}
-
 conscell *
 pickaddress(l)
 	conscell *l;
@@ -1055,7 +1094,7 @@ pickaddress(l)
 	for (p = NULL, la = car(l); la != NULL && LIST(la) ; la = cdr(la)) {
 		/* AND addresses; i.e. exploded address list or one address */
 		for (lx = car(la); lx != NULL && LIST(lx) ; lx = cdr(lx)) {
-			if (STRING(cdar(lx)) && nullhost(cdar(lx)->string)) {
+			if (STRING(cdar(lx)) && nullhost(cdar(lx))) {
 				p = lx;
 				break;
 			}
@@ -1073,15 +1112,15 @@ thesender(e, a)
 	struct envelope *e;
 	struct address *a;
 {
-	conscell *l;
+	conscell *l; /* Var life ends after return.. no GC protection */
 
 	l = router(a, e->e_statbuf.st_uid, "sender");
 	if (l == NULL)
 		return 0;
 	e->e_from_resolved = pickaddress(l);
 
-	if (QUSER(e->e_from_resolved) == NULL
-	    || !nullhost(QHOST(e->e_from_resolved)))
+	if (QUSER(e->e_from_resolved)->cstring == NULL ||
+	    !nullhost(QHOST(e->e_from_resolved)))
 		return 0;
 	/*
 	 * We assume here, that local mailbox id's correspond to user
@@ -1093,9 +1132,35 @@ thesender(e, a)
 	 * only done if the message file wasn't created by a trusted user,
 	 * so in practise this is less of a problem than it might appear to be.
 	 */
-	return CISTREQ(QUSER(e->e_from_resolved),uidpwnam(e->e_statbuf.st_uid));
+	return CISTREQ(QUSER(e->e_from_resolved)->cstring,
+		       uidpwnam(e->e_statbuf.st_uid));
 }
 
+extern conscell *rwmappend __((conscell *, conscell *, conscell *));
+conscell *rwmappend(rwmroot,info,errtop)
+     conscell *rwmroot, *info, *errtop;
+{
+  conscell *p = NULL;
+  GCVARS4;
+
+  GCPRO4(p, rwmroot, info, errtop);
+
+  if (info) {
+    p = ncons(info);
+    cdr(p) = rwmroot;
+    rwmroot = p;
+  }
+  if (errtop) {
+    p = ncons(info);
+    cdr(p) = rwmroot;
+    rwmroot = p;
+  }
+
+  UNGCPRO4;
+  return rwmroot;
+}
+
+
 
 /*
  * The sequencer takes care of doing the right things with the right headers
@@ -1109,14 +1174,15 @@ sequencer(e, file)
 	struct envelope *e;
 	const char *file;
 {
-	struct header  *h, *ph, *nh, *oh, *msgidh, **hp;
+	struct header  *h, *ph, *nh = NULL, *oh, *msgidh, **hp;
 	struct address *a, *ap;
 	struct addr    *p = NULL;
 	char	       *ofpname, *path, *qpath;
 	conscell       *l, *routed_addresses, *sender, *to;
+	conscell       *rwmchain;
 	struct rwmatrix *rwhead, *nsp, *rwp = NULL, *rcp = NULL;
 	token822   *t = NULL;
-	int   idnumber, nxor, i;
+	int   idnumber = 0, nxor = 0, i, slen;
 	int   def_uid, header_error, perr;
 	FILE	       *ofp, *vfp;
 	int		ofperrors = 0;
@@ -1127,6 +1193,7 @@ sequencer(e, file)
 	const char     *fromaddr = "?from?";
 	const char     *msgidstr = "?msgid?";
 	const char     *smtprelay = NULL;
+	GCVARS5;
 
 	deferuid = 0;
 
@@ -1155,23 +1222,41 @@ sequencer(e, file)
 	    smtprelay = sr;
 	  }
 	} else {
-	  /* Ok, we don't trust it!  In fact we might overrule the data
-	     that the message writer coded in... */
 	  const char *s, *ps;
 	  char *ts;
 	  int totlen;
 
-	  dprintf("Message has 'rcvdfrom' envelope header, but we don't trust it!\n");
-	  s = uidpwnam(e->e_statbuf.st_uid);
-	  ps = "";
-	  if (h) ps = h->h_lines->t_pname;
-	  totlen = 10 + strlen(s) + 20 + (h ? strlen(ps) + 10 : 0);
-	  /* ts = strnsave("", totlen); */
-	  ts = tmalloc(totlen); /* Alloc the space */
-	  if (h) {
-	    sprintf(ts, "rcvdfrom %s@localhost (fake: %s)", s, h->h_lines->t_pname);
-	  } else
-	    sprintf(ts, "rcvdfrom %s@localhost", s);
+	  if (h != NULL) {
+
+	    /* Ok, we don't trust it!  In fact we might overrule the data
+	       that the message writer coded in... */
+
+	    dprintf("Message has 'rcvdfrom' envelope header, but we don't trust it!\n");
+	    s = uidpwnam(e->e_statbuf.st_uid);
+	    ps = "";
+	    totlen = 10 + strlen(s) + 60;
+	    if (h) {
+	      ps = h->h_lines->t_pname;
+	      totlen += strlen(ps) + 10;
+	    }
+	    /* ts = strnsave("", totlen); */
+	    ts = tmalloc(totlen); /* Alloc the space */
+	    if (h) {
+	      sprintf(ts, "rcvdfrom STDIN (from localhost user: '%s' uid#%ld fake: %s)",
+		      s, (long) e->e_statbuf.st_uid, h->h_lines->t_pname);
+	    } else
+	      sprintf(ts, "rcvdfrom STDIN (from localhost user: '%s' uid#%ld)",
+		      s, (long) e->e_statbuf.st_uid);
+
+	  } else {
+
+	    /* No "rcvdfrom" envelope header */
+	    s = uidpwnam(e->e_statbuf.st_uid);
+	    totlen = 60 + strlen(s);
+	    ts = tmalloc(totlen);
+	    sprintf(ts, "rcvdfrom STDIN (from localhost user: '%s', uid#%ld)",
+		    s, (long) e->e_statbuf.st_uid);
+	  }
 
 	  h = makeHeader(spt_eheaders, ts, 8);
 	  h->h_next = e->e_eHeaders;
@@ -1284,6 +1369,10 @@ sequencer(e, file)
 	    && h->h_contents.a->a_pname != NULL) {
 		notaryret = h->h_contents.a->a_pname;
 	}
+
+	rwmchain = l = routed_addresses = sender = to = NULL;
+	GCPRO5(l, routed_addresses, sender, to, rwmchain);
+
 	FindEnvelope(eFrom);
 	if (h == NULL && e->e_trusted) {
 		/* Perhaps  'channel error' ??? */
@@ -1294,8 +1383,11 @@ sequencer(e, file)
 			    && (p = ap->a_tokens) != NULL
 			    && p->p_tokens != NULL) {
 				t = p->p_tokens;
+				slen = TOKENLEN(t);
+				l = cdr(QCHANNEL(e->e_from_trusted));
 				QCHANNEL(e->e_from_trusted) =
-					strnsave(t->t_pname, TOKENLEN(t));
+				  newstring(dupnstr(t->t_pname, slen), slen);
+				cdr(QCHANNEL(e->e_from_trusted)) = l;
 			} else {
 				/*
 				 * No origination channel, or channel
@@ -1306,8 +1398,10 @@ sequencer(e, file)
 		}
 	}
 
-	if (deferuid)
+	if (deferuid) {
+	  UNGCPRO5;
 	  return PERR_DEFERRED;
+	}
 
 	if (h == NULL) {
 		dprintf("A sender was NOT specified in the envelope\n");
@@ -1381,10 +1475,13 @@ sequencer(e, file)
 			    && (p = ap->a_tokens) != NULL
 			    && p->p_tokens != NULL) {
 				t = p->p_tokens;
+				slen = TOKENLEN(t);
+				l = cdr(QCHANNEL(e->e_from_trusted));
 				QCHANNEL(e->e_from_trusted) =
-					strnsave(t->t_pname, TOKENLEN(t));
+				  newstring(dupnstr(t->t_pname, slen), slen);
+				cdr(QCHANNEL(e->e_from_trusted)) = l;
 			}
-			if (QCHANNEL(e->e_from_trusted) == NULL) {
+			if (QCHANNEL(e->e_from_trusted)->cstring == NULL) {
 				/*
 				 * the mailer is supposed to know about
 				 * all valid channel identifiers. Gripe.
@@ -1393,14 +1490,26 @@ sequencer(e, file)
 			}
 		}
 		FindEnvelope(eRcvdFrom);
-		if (h != NULL)		/* a previous host was specified */
-			QHOST(e->e_from_trusted) = h->h_contents.a->a_pname;
+		if (h != NULL && h->h_contents.a && h->h_contents.a->a_pname) {
+			/* a previous host was specified */
+			slen = strlen(h->h_contents.a->a_pname);
+			l = cdr(QHOST(e->e_from_trusted));
+			QHOST(e->e_from_trusted) =
+			  newstring(dupnstr(h->h_contents.a->a_pname, slen),slen);
+			cdr(QHOST(e->e_from_trusted)) = l;
+		}
 		FindEnvelope(eUser);
-		if (h != NULL)		/* a previous user was specified */
-			QUSER(e->e_from_trusted) = h->h_contents.a->a_pname;
-		if (QCHANNEL(e->e_from_trusted) == NULL
-		    || QHOST(e->e_from_trusted) == NULL
-		    || QUSER(e->e_from_trusted) == NULL) {
+		if (h != NULL && h->h_contents.a->a_pname) {
+			/* a previous user was specified */
+			slen = strlen(h->h_contents.a->a_pname);
+			l = cdr(QHOST(e->e_from_trusted));
+			QHOST(e->e_from_trusted) =
+			  newstring(dupnstr(h->h_contents.a->a_pname, slen),slen);
+			cdr(QHOST(e->e_from_trusted)) = l;
+		}
+		if (QCHANNEL(e->e_from_trusted)->cstring == NULL
+		    || QHOST(e->e_from_trusted)->cstring == NULL
+		    || QUSER(e->e_from_trusted)->cstring == NULL) {
 			FindEnvelope(eFrom);
 			/* X: assert h != NULL */
 			if (h == NULL || h->h_stamp == BadHeader) {
@@ -1417,6 +1526,8 @@ sequencer(e, file)
 			}
 			if (h == NULL)
 				abort(); /* Failed to make sender header */
+
+			/* This conscell lifetime is limited.. */
 			l = router(h->h_contents.a,
 				   e->e_statbuf.st_uid, "sender");
 			if (l == NULL) {
@@ -1441,13 +1552,13 @@ sequencer(e, file)
 			/* local user */
 			FindEnvelope(eExternal);
 			if (h != NULL || (e->e_statbuf.st_mode & 022)) {
-				optsave(FYI_BREAKIN, e);
-			} else if (QUSER(e->e_from_resolved) != NULL)
-				def_uid =
-					login_to_uid(QUSER(e->e_from_resolved));
-			else if (QUSER(e->e_from_trusted) != NULL)
-				def_uid =
-					login_to_uid(QUSER(e->e_from_trusted));
+			  optsave(FYI_BREAKIN, e);
+			} else if (QUSER(e->e_from_resolved)->cstring != NULL)
+			  def_uid =
+			    login_to_uid(QUSER(e->e_from_resolved)->cstring);
+			else if (QUSER(e->e_from_trusted)->cstring != NULL)
+			  def_uid =
+			    login_to_uid(QUSER(e->e_from_trusted)->cstring);
 		}
 	} else {
 		dprintf("We know sender is local and one of the peons\n");
@@ -1490,17 +1601,31 @@ sequencer(e, file)
 		  set_pname(e, nh, "Sender");
 		}
 	}
-	if (QCHANNEL(e->e_from_trusted) == NULL)
-		QCHANNEL(e->e_from_trusted) = QCHANNEL(e->e_from_resolved);
-	if (QHOST(e->e_from_trusted) == NULL)
-		QHOST(e->e_from_trusted) = QHOST(e->e_from_resolved);
-	if (QUSER(e->e_from_trusted) == NULL)
-		QUSER(e->e_from_trusted) = QUSER(e->e_from_resolved);
-	if (QATTRIBUTES(e->e_from_trusted) == NULL)
-		QATTRIBUTES(e->e_from_trusted)=QATTRIBUTES(e->e_from_resolved);
+	if (QCHANNEL(e->e_from_trusted)->cstring == NULL) {
+	  l = cdr(QCHANNEL(e->e_from_trusted));
+	  QCHANNEL(e->e_from_trusted) = copycell(QCHANNEL(e->e_from_resolved));
+	  cdr(QCHANNEL(e->e_from_trusted)) = l;
+	}
+	if (QHOST(e->e_from_trusted)->cstring == NULL) {
+	  l = cdr(QHOST(e->e_from_trusted));
+	  QHOST(e->e_from_trusted) = copycell(QHOST(e->e_from_resolved));
+	  cdr(QHOST(e->e_from_trusted)) = l;
+	}
+	if (QUSER(e->e_from_trusted)->cstring == NULL) {
+	  l = cdr(QUSER(e->e_from_trusted));
+	  QUSER(e->e_from_trusted) = copycell(QUSER(e->e_from_resolved));
+	  cdr(QUSER(e->e_from_trusted)) = l;
+	}
+	if (QATTRIBUTES(e->e_from_trusted)->cstring == NULL) {
+	  l = cdr(QATTRIBUTES(e->e_from_trusted));
+	  QATTRIBUTES(e->e_from_trusted) = copycell(QATTRIBUTES(e->e_from_resolved));
+	  cdr(QATTRIBUTES(e->e_from_trusted)) = l;
+	}
 
-	if (deferuid)
+	if (deferuid) {
+	  UNGCPRO5;
 	  return PERR_DEFERRED;
+	}
 
 	dprintf("Recipient determination\n");
 	FindEnvelope(eTo);
@@ -1508,6 +1633,7 @@ sequencer(e, file)
 		dprintf("No recipient(s) specified in the envelope\n");
 		if (header_error) {
 			dprintf("Due to header error, we ignore message\n");
+			UNGCPRO5;
 			return PERR_HEADER;
 		}
 		ph = e->e_eHeaders;
@@ -1529,29 +1655,32 @@ sequencer(e, file)
 			e->e_eHeaders = nh;
 		}
 		dprintf("Are we supposed to be psychic?\n");
-		if (ph == e->e_eHeaders)
-			return PERR_NORECIPIENTS;
+		if (ph == e->e_eHeaders) {
+		  UNGCPRO5;
+		  return PERR_NORECIPIENTS;
+		}
 	}
 
 	dprintf("Nuke Bcc/Return-Path/X-Orcpt/X-Envid headers, if any\n");
-	oh = NULL;
-	for (h = e->e_headers; h != NULL; oh = h, h = h->h_next) {
-	  if (h->h_descriptor->hdr_name == NULL)
-	    continue;
-	  if ((h->h_descriptor->class == normal &&
-	       (CISTREQ(h->h_descriptor->hdr_name,"bcc") ||
-		CISTREQ(h->h_descriptor->hdr_name,"return-path") ||
-		CISTREQ(h->h_descriptor->hdr_name,"x-orcpt")     ||
-		CISTREQ(h->h_descriptor->hdr_name,"x-envid")        ))
-	      || (h->h_descriptor->class == Resent &&
-		  (CISTREQ(h->h_descriptor->hdr_name,"resent-BCC")         ||
-		   CISTREQ(h->h_descriptor->hdr_name,"resent-return-path") ||
-		   CISTREQ(h->h_descriptor->hdr_name,"resent-x-orcpt")     ||
-		   CISTREQ(h->h_descriptor->hdr_name,"resent-x-envid")    ))) {
-	    if (oh == NULL)
-	      e->e_headers = h->h_next;
-	    else
-	      oh->h_next = h->h_next;
+	hp = & e->e_headers;
+	while (*hp != NULL) {
+	  h = *hp;
+	  if (h->h_descriptor->hdr_name != NULL &&
+	      ((h->h_descriptor->class == normal &&
+		(CISTREQ(h->h_descriptor->hdr_name,"bcc") ||
+		 CISTREQ(h->h_descriptor->hdr_name,"return-path") ||
+		 CISTREQ(h->h_descriptor->hdr_name,"x-orcpt")     ||
+		 CISTREQ(h->h_descriptor->hdr_name,"x-envid")        ))
+	       || (h->h_descriptor->class == Resent &&
+		   (CISTREQ(h->h_descriptor->hdr_name,"resent-BCC")         ||
+		    CISTREQ(h->h_descriptor->hdr_name,"resent-return-path") ||
+		    CISTREQ(h->h_descriptor->hdr_name,"resent-x-orcpt")     ||
+		    CISTREQ(h->h_descriptor->hdr_name,"resent-x-envid")  )))) {
+	    /* Skip this one */
+	    *hp = h->h_next;
+	  } else {
+	    /* No header dropping here */
+	    hp = & h->h_next;
 	  }
 	}
 
@@ -1570,7 +1699,7 @@ sequencer(e, file)
 		}
 		dprintf("Emit warning headers\n");
 		for (ph = NULL, h = e->e_headers; h != NULL; ph=h,h=h->h_next) {
-			if (h->h_stamp == BadHeader) {
+			if (h->h_stamp == BadHeader && do_hdr_warning) {
 				if (ph == NULL)
 					e->e_headers = hdr_warning(h);
 				else
@@ -1582,22 +1711,57 @@ sequencer(e, file)
 	FindHeader("from",e->e_resent);
 	if (h == NULL) {
 		FindEnvelope(eFrom);
+		oh = h;
 		if (h == NULL) {	/* panic... can't happen... */
 			dprintf(" none available!\n");
 			optsave(FYI_NOSENDER, e);
 		} else {
-			/* assume we only care about local users */
-			nh = (struct header *)tmalloc(sizeof (struct header));
-			*nh = *h;
-			set_pname(e, nh, "From");
-			FindHeader("to",e->e_resent);
-			if (h == NULL) {
-			  for (h = e->e_headers; h != NULL; h = h->h_next)
-			    if (CISTREQ(h->h_pname, "subject"))
-			      break;
+			FindEnvelope(eChannel);
+			if (h != NULL) {
+			  if ((ap = h->h_contents.a) != NULL
+			      && (p = ap->a_tokens) != NULL
+			      && (p->p_tokens != NULL)) {
+			    t = p->p_tokens;
+			    if (TOKENLEN(t) == 5 &&
+				strncmp(t->t_pname,"error",5)==0) {
+
+			      char *ss = (char*)tmalloc(10);
+			      strcpy(ss, "From: <>");
+
+			      nh = makeHeader(spt_headers, ss, 4);
+			      nh->h_lines = makeToken(ss+5,3);
+			      nh->h_lines->t_type = Line;
+			      nh->h_contents = hdr_scanparse(e, nh, 0, 0);
+			      nh->h_stamp = hdr_type(nh);
+			      FindHeader("to",e->e_resent);
+			      if (h == NULL) {
+				for (h = e->e_headers; h != NULL; h = h->h_next)
+				  if (CISTREQ(h->h_pname, "subject"))
+				    break;
+			      }
+			      InsertHeader(h, nh);
+			      dprintf(" nope (added)!\n");
+			    }
+			  }
+			} else {
+
+			  /* We make a copy of eFrom envelope header
+			     to be our new  From:  header */
+
+			  nh = makeHeader(spt_headers, "From", 4);
+			  nh->h_lines    = oh->h_lines;
+			  nh->h_contents = oh->h_contents;
+			  nh->h_stamp    = hdr_type(nh);
+			  
+			  FindHeader("to",e->e_resent);
+			  if (h == NULL) {
+			    for (h = e->e_headers; h != NULL; h = h->h_next)
+			      if (CISTREQ(h->h_pname, "subject"))
+				break;
+			  }
+			  InsertHeader(h, nh);
+			  dprintf(" nope (added)!\n");
 			}
-			InsertHeader(h, nh);
-			dprintf(" nope (added)!\n");
 		}
 	} else
 		dprintf(" yup!\n");
@@ -1641,7 +1805,6 @@ sequencer(e, file)
 	incoming-rewriting rules for the originating channel.
 #endif
 	dprintf("Route recipient addresses\n");
-	routed_addresses = NULL;
 
 	if (errors_to) free(errors_to);
 	errors_to = NULL;
@@ -1677,24 +1840,24 @@ sequencer(e, file)
 				continue;
 
 			if (routed_addresses == NULL) {
-				conscell *tmp;
 				routed_addresses = ncons(car(l));
 			} else {
 				cdr(s_last(car(l))) = car(routed_addresses);
 				car(routed_addresses) = car(l);
 			}
 			/* freecell(l) */
+			l = NULL;
+		}
+		if (deferuid) {
+		  UNGCPRO5;
+		  return PERR_DEFERRED;
 		}
-		if (deferuid)
-			return PERR_DEFERRED;
 	}
 
 	dprintf("Crossbar to be applied to all (sender,routed-address) pairs\n");
 
-	{
-	  conscell *tmp;
-	  sender = ncons(e->e_from_trusted);
-	}
+	sender = ncons(e->e_from_trusted);
+
 #if 0
 	if (LIST(sender) && LIST(car(sender)))
 		sender = caar(sender);
@@ -1702,6 +1865,7 @@ sequencer(e, file)
 
 	if (routed_addresses == NULL)	/* they were probably all deferred */ {
 		printf("No routed addresses -> deferred\n");
+		UNGCPRO5;
 		return PERR_DEFERRED;
 	}
 
@@ -1711,7 +1875,8 @@ sequencer(e, file)
 		++idnumber, nxor = 0;
 		for (to = car(l); to != NULL; to = cdr(to)) {
 			conscell *x, *tmp, *errto, *gg;
-			conscell *sender1;
+			conscell *sender1, *z;
+			GCVARS6;
 
 			/* secondlevel is XORs */
 #if 0
@@ -1724,14 +1889,28 @@ sequencer(e, file)
 #endif
 
 			if ((x = crossbar(sender, to)) == NULL)
-				continue;
-#if 0
-			printf("crossbar returns: ");
-			s_grind(x, stdout);
-			putchar('\n');
-#endif
+			  continue;
+
+	/*
+	 * We expect to see something like
+	 * (rewrite (fc fh fu) (tc th tu)) or
+	 * ((address-rewrite header-rewrite) (fc fh fu) (tc th tu))
+	 * back from the crossbar function.
+	 */
+
+			errto = gg = sender1 = tmp = z = NULL;
+			GCPRO6(x, errto, gg, sender1, tmp, z);
+
+
+			if (D_sequencer) {
+			  printf("crossbar returns: ");
+			  s_grind(x, stdout);
+			  putchar('\n');
+			}
+			
 			++nxor;
 			tmp   = copycell(car(x));
+
 			cdr(tmp) = NULL;
 			sender1 = cdar(x);
 			gg = cdr(cddr(car(sender1))); /* recipient attrib! */
@@ -1741,11 +1920,14 @@ sequencer(e, file)
 
 			/* Rewriter level: */
 			for (rwp = rwhead; rwp != NULL; rwp = rwp->next)
-				if (s_equal(rwp->info, tmp))
-					break;
+			  if (s_equal(rwp->info, tmp)) {
+			    break;
+			  }
 			if (rwp == NULL) {
 				rwp = rwalloc(&rwhead);
 				rwp->info  = tmp;	/* rewritings */
+				rwp->errto = NULL;
+				rwmchain = rwmappend(rwmchain,tmp,NULL);
 			}
 			/* else the 'tmp' leaks for a moment ? */
 			
@@ -1761,6 +1943,7 @@ sequencer(e, file)
 				nsp = rwalloc(&rwp->down);
 				nsp->info  = sender1;	/* new sender */
 				nsp->errto = errto;
+				rwmchain = rwmappend(rwmchain,sender1,errto);
 			}
 #if 0
 			else {
@@ -1776,6 +1959,7 @@ sequencer(e, file)
 			rcp->info = tmp;		/* new recipient */
 			rcp->urw.number = idnumber;
 			rcp->errto = errto;
+			rwmchain = rwmappend(rwmchain,tmp,errto);
 #if 0
 			for (rwp = rwhead; rwp != NULL; rwp = rwp->next) {
 				printf("**");
@@ -1798,6 +1982,7 @@ sequencer(e, file)
 				}
 			}
 #endif
+			UNGCPRO6;
 		}
 		if (nxor <= 1) {
 			--idnumber;
@@ -1806,8 +1991,10 @@ sequencer(e, file)
 		}
 	}
 
-	if (deferuid)
-		return PERR_DEFERRED;
+	if (deferuid) {
+	  UNGCPRO5;
+	  return PERR_DEFERRED;
+	}
 
 	isSenderAddr = isRecpntAddr = 0;
 	dprintf("Make sure Date, From, To, are in the header\n");
@@ -1831,11 +2018,15 @@ sequencer(e, file)
 		*hp = NULL;
 	}
  
-	if (deferuid)
+	if (deferuid) {
+	  UNGCPRO5;
 	  return PERR_DEFERRED;
+	}
 
-	if (rwhead == NULL)
+	if (rwhead == NULL) {
+	  UNGCPRO5;
 	  return PERR_NORECIPIENTS;
+	}
 
 	dprintf("Emit specification to the transport system\n");
 #ifdef	USE_ALLOCA
@@ -1844,12 +2035,13 @@ sequencer(e, file)
 	ofpname = (char *)emalloc((u_int)(strlen(file)+9));
 #endif
 	/* Guaranteed unique within this machine */
-	sprintf(ofpname,".%s.%d",file,getpid());
+	sprintf(ofpname,".%s.%d", file, (int)getpid());
 	if ((ofp = fopen(ofpname, "w+")) == NULL) {
 #ifndef USE_ALLOCA
 		free(ofpname);
 #endif
 		printf("Creation of control file failed\n");
+		UNGCPRO5;
 		return PERR_CTRLFILE;
 	}
 	setvbuf(ofp, vbuf, _IOFBF, sizeof vbuf);
@@ -1887,6 +2079,9 @@ sequencer(e, file)
 	} else
 		vfp = NULL;
 
+	fprintf(ofp, "%c%c0x%08lx\n",
+		_CF_FORMAT, _CFTAG_NORMAL, (long)_CF_FORMAT_KNOWN_SET);
+
 	fprintf(ofp, "%c%c%s\n",
 		_CF_MESSAGEID, _CFTAG_NORMAL, file);
 	fprintf(ofp, "%c%c%d\n",
@@ -2012,7 +2207,8 @@ sequencer(e, file)
 			ofperrors |= ferror(ofp);
 			if (ofperrors) break; /* Sigh.. */
 
-			if (!iserrmessage()  &&  nsp->errto != NULL) {
+			if (!iserrmessage() && nsp->errto &&
+			    nsp->errto->string) {
 				/* print envelope sender address */
 				putc(_CF_ERRORADDR, ofp);
 				putc(_CFTAG_NORMAL, ofp);
@@ -2026,6 +2222,11 @@ sequencer(e, file)
 
 			fprintf(ofp, "%c%c", _CF_SENDER, _CFTAG_NORMAL);
 			fromaddr = prctladdr(nsp->info, ofp, _CF_SENDER, "sender");
+			if (!fromaddr) {
+			  UNGCPRO5;
+			  goto bad_addrdata;
+			}
+			
 			putc('\n', ofp);
 			if (vfp != NULL) {
 				fprintf(vfp, "%c%c",
@@ -2041,37 +2242,41 @@ sequencer(e, file)
 				if (ofperrors) break; /* Sigh.. */
 
 				if (rcp->urw.number > 0)
-					putc(_CF_XORECIPIENT, ofp);
+				  putc(_CF_XORECIPIENT, ofp);
 				else
-					putc(_CF_RECIPIENT, ofp);
+				  putc(_CF_RECIPIENT, ofp);
 				putc(_CFTAG_NORMAL, ofp);
-				fprintf(ofp,"%*s",_CFTAG_RCPTPIDSIZE,"");
+				fprintf(ofp,"%*s",(_CFTAG_RCPTPIDSIZE +
+						   _CFTAG_RCPTDELAYSIZE),"");
 				if (rcp->urw.number > 0)
 					fprintf(ofp, "%d ", rcp->urw.number);
-				prctladdr(rcp->info, ofp,
-					  _CF_RECIPIENT, "recipient");
+				if (! prctladdr(rcp->info, ofp,
+						_CF_RECIPIENT, "recipient")) {
+				  UNGCPRO5;
+				  goto bad_addrdata;
+				}
+				
 				putc('\n', ofp);
 				++nrcpts;
 				/* DSN data output ! */
 				prdsndata(rcp->info, ofp,
 					  _CF_RCPTNOTARY, "recipient");
 				if (vfp != NULL) {
-					if (rcp->urw.number > 0)
-						fprintf(vfp, "%c%c%d ",
-							_CF_XORECIPIENT,
-							_CFTAG_NORMAL,
-							rcp->urw.number);
-					else
-						fprintf(vfp, "%c%c%*s",
-							_CF_RECIPIENT,
-							_CFTAG_NORMAL,
-							_CFTAG_RCPTPIDSIZE,"");
-					prctladdr(rcp->info, vfp,
-						  _CF_RECIPIENT, "recipient");
-					putc('\n', vfp);
-					/* DSN data output ! */
-					prdsndata(rcp->info, vfp,
-						  _CF_RCPTNOTARY, "recipient");
+				  if (rcp->urw.number > 0)
+				    putc(_CF_XORECIPIENT, vfp);
+				  else
+				    putc(_CF_RECIPIENT, vfp);
+				  putc(_CFTAG_NORMAL, vfp);
+				  fprintf(vfp,"%*s",(_CFTAG_RCPTPIDSIZE +
+						     _CFTAG_RCPTDELAYSIZE),"");
+				  if (rcp->urw.number > 0)
+				    fprintf(vfp, "%d ", rcp->urw.number);
+				  prctladdr(rcp->info, vfp,
+					    _CF_RECIPIENT, "recipient");
+				  putc('\n', vfp);
+				  /* DSN data output ! */
+				  prdsndata(rcp->info, vfp,
+					    _CF_RCPTNOTARY, "recipient");
 				}
 			}
 		}
@@ -2107,6 +2312,8 @@ sequencer(e, file)
 			putc('\n', vfp);
 	}
 
+	UNGCPRO5;
+
 	if (vfp != NULL) {
 		fprintf(vfp, "router done processing %s\n", file);
 		fflush(vfp);
@@ -2123,9 +2330,9 @@ sequencer(e, file)
 	}
 	/* all is nirvana -- link the input file to somewhere safe */
 #ifdef	USE_ALLOCA
-	qpath = alloca(5+strlen(QUEUEDIR)+strlen(file));
+	qpath = (char*)alloca(5+strlen(QUEUEDIR)+strlen(file));
 #else
-	qpath = emalloc(5+strlen(QUEUEDIR)+strlen(file));
+	qpath = (char*)emalloc(5+strlen(QUEUEDIR)+strlen(file));
 #endif
 	sprintf(qpath, "../%s/%s", QUEUEDIR, file);
 	fflush(ofp);
@@ -2135,8 +2342,8 @@ sequencer(e, file)
 	ofperrors |= ferror(ofp);
 
 	if ((fclose(ofp) != 0) || ofperrors || (erename(file, qpath) != 0)) {
-	  unlink(qpath);
-	  unlink(ofpname);
+	  zunlink(qpath);
+	  zunlink(ofpname);
 #ifndef	USE_ALLOCA
 	  free(ofpname);
 	  free(qpath);
@@ -2145,17 +2352,17 @@ sequencer(e, file)
 	}
 
 #ifndef USE_ALLOCA
-	path = emalloc(5+strlen(TRANSPORTDIR)+strlen(file));
+	path = (char*)emalloc(5+strlen(TRANSPORTDIR)+strlen(file));
 #else
 	/* This actually reallocs more space from stack, but then it
 	   is just stack space and will disappear.. */
-	path = alloca(5+strlen(TRANSPORTDIR)+strlen(file));
+	path = (char*)alloca(5+strlen(TRANSPORTDIR)+strlen(file));
 #endif
 	sprintf(path, "../%s/%s", TRANSPORTDIR, file);
-	unlink(path);	/* Should actually always fail.. */
+	zunlink(path);	/* Should actually always fail.. */
 	if (erename(ofpname, path) < 0) {
-	  unlink(qpath);
-	  unlink(path);
+	  zunlink(qpath);
+	  zunlink(path);
 #ifndef	USE_ALLOCA
 	  free(ofpname);
 	  free(qpath);
@@ -2164,8 +2371,9 @@ sequencer(e, file)
 	  return PERR_CTRLFILE;
 	}
 
-	rtsyslog(e->e_statbuf.st_mtime, path, fromaddr, smtprelay,
-		 (int) e->e_statbuf.st_size, nrcpts, msgidstr);
+	rtsyslog(e->e_statbuf.st_mtime, (long)e->e_statbuf.st_ino,
+		 fromaddr, smtprelay, (int) e->e_statbuf.st_size,
+		 nrcpts, msgidstr);
 
 #ifndef	USE_ALLOCA
 	free(ofpname);
@@ -2175,11 +2383,22 @@ sequencer(e, file)
 
 	/* we did it! */
 	return PERR_OK;
+
+ bad_addrdata:;
+
+	/* Bad data in routing results, can't do it now, must defer! */
+
+	fclose(ofp);
+	zunlink(ofpname); /* created file is thrown away.. */
+#ifndef	USE_ALLOCA
+	free(ofpname);
+#endif
+	return PERR_CTRLFILE;
 }
 
 static const char *
 prctladdr(info, fp, cfflag, comment)
-	conscell *info;
+	conscell *info; /* No conscell allocs down here */
 	FILE *fp;
 	int cfflag;
 	const char *comment;
@@ -2188,6 +2407,7 @@ prctladdr(info, fp, cfflag, comment)
 	register conscell *l, *x;
 	const char *user = "?user?";
 	const char *channel = NULL;
+	const char *privilege = NULL, *p;
 
 	/* We print the quad of  channel/host/user/privilege  information
 	   with this routine, and we return pointer to the user info.
@@ -2231,6 +2451,8 @@ prctladdr(info, fp, cfflag, comment)
 		} else
 		  fprintf(fp, "%s", x->string);
 	      }
+	      if (i == 4 && x->string != NULL)
+		privilege = x->string;
 	      if (i == 3 && x->string != NULL)
 		user = x->string;
 	      if (i == 1 && x->string != NULL)
@@ -2238,11 +2460,25 @@ prctladdr(info, fp, cfflag, comment)
 	    }
 	    if (cdr(l))
 	      putc(' ', fp);
-	  } else
+	  } else if (channel && strcmp(channel,"error")!=0)
 	    fprintf(stderr, "Malformed %s\n", comment);
 	}
 	if ((cfflag == _CF_SENDER) && channel && strcmp(channel,"error") == 0)
 		user = ""; /* error channel source address -> no user! */
+
+	if (!privilege) {
+	  fprintf(stderr, "Malformed (missing) %s privilege data!\n", comment);
+	  return NULL;
+	}
+	p = privilege;
+	if (*p == '-') ++p;
+	while ('0' <= *p && *p <= '9') ++p;
+	if (*p) {
+	  fprintf(stderr, "Malformed ('%s') %s privilege data!\n",
+		  privilege, comment);
+	  return NULL;
+	}
+
 	return user;
 }
 
@@ -2254,7 +2490,7 @@ prdsndata(info, fp, cfflag, comment)
 	const char *comment;
 {
 	int i = 0;
-	register conscell *l, *x = NULL;
+	register conscell *l, *x = NULL; /* No allocs down here */
 
 	for (l = car(info); l != NULL; l = cdr(l)) {
 	  ++i;
@@ -2282,7 +2518,7 @@ prdsndata(info, fp, cfflag, comment)
 
 static conscell *
 find_errto(info)
-conscell *info;
+     conscell *info;	/* No allocs under here */
 {
 	register conscell *x = NULL;
 
