@@ -26,7 +26,7 @@
 
 #include "prototypes.h"
 
-const char *postoffice = NULL;
+extern const char *postoffice; /* At libzmailer.a: mail.c */
 
 #if 0
 extern struct group *getgrnam __();
@@ -49,6 +49,7 @@ const char * logfn;
 time_t	now;
 
 extern memtypes stickymem;
+extern conscell *s_value;
 
 int	mustexit = 0;
 int	canexit = 0;
@@ -59,13 +60,17 @@ int	origoptind;
 int	savefile = 0;
 const char * zshopts = "-O";
 int	nosyslog = 1;
+int	routerdirloops = 0;
+int	do_hdr_warning = 0;
+int	workermode = 0;
+int	nrouters = 1;
 
 int
 main(argc, argv)
 	int	    argc;
 	const char *argv[];
 {
-	int c, errflg, daemonflg, killflg, interactiveflg, tac, nrouters;
+	int c, errflg, daemonflg, killflg, interactiveflg, tac;
 	int version;
 	long offout, offerr;
 	char *config, *cp;
@@ -79,6 +84,18 @@ main(argc, argv)
 	setgroups(0, NULL);
 #endif
 
+#if 1 /* LINE BUFFERED */
+
+	setvbuf(stdout, (char *)NULL, _IOLBF, 0);
+	setvbuf(stderr, (char *)NULL, _IOLBF, 0);
+
+#else /* NOT BUFFERED AT ALL */
+
+	setvbuf(stdout, (char *)NULL, _IONBF, 0);
+	setvbuf(stderr, (char *)NULL, _IONBF, 0);
+
+#endif
+
 	progname = strrchr(argv[0], '/');
 	if (progname == NULL)
 		progname = argv[0];
@@ -92,7 +109,7 @@ main(argc, argv)
 
 
 	while (1) {
-		c = getopt(argc, (char*const*)argv, "m:n:dikf:o:t:L:P:sSV");
+		c = getopt(argc, (char*const*)argv, "m:n:dikf:o:t:L:P:r:sSVwW");
 		if (c == EOF)
 			break;
 	  
@@ -159,12 +176,24 @@ main(argc, argv)
 		case 'V':
 			version = 1;
 			break;
+		case 'w':
+			workermode = 1;
+			break;
+		case 'W':
+			do_hdr_warning = !do_hdr_warning;
+			break;
+		case 'r':
+			routerdirloops = atoi(optarg);
+			if (routerdirloops < 0)
+				routerdirloops = 0;
+			break;
 		case '?':
 		default:
 			errflg++;
 			break;
 		}
 	}
+
 	if (errflg || (interactiveflg && daemonflg)) {
 		fprintf(stderr,
 			"Usage: %s [ -dikV -n #routers -t traceflag -f configfile -L logfile -P postoffice]\n",
@@ -204,16 +233,11 @@ main(argc, argv)
 					  + strlen(progname)));
 		sprintf((char*)logfn, "%s/%s", logdir, progname);
 	}
-	/* setvbuf(stdout, (char *)NULL, _IOLBF, 0);
-	   setvbuf(stderr, (char *)NULL, _IOLBF, 0); */
-	setvbuf(stdout, (char *)NULL, _IONBF, 0);
-	setvbuf(stderr, (char *)NULL, _IONBF, 0);
 
 	if (logfn != NULL) {
 		/* loginit is a signal handler, so can't pass log */
-		if (loginit(0) < 0)	/* do setlinebuf() there */
+		if (loginit(SIGHUP) < 0) /* do setlinebuf() there */
 			die(1, "log initialization failure");
-		signal(SIGHUP, sig_hup); /* close and reopen log files */
 	} else
 		signal(SIGHUP, SIG_IGN); /* no surprises please */
 
@@ -228,6 +252,7 @@ main(argc, argv)
 		/* lax, no NULL guard on end of tav */
 		run_trace(++tac, tav);
 	}
+
 	stickymem = MEM_PERM;
 
 	initialize(config, argc - c, &argv[c]);
@@ -236,9 +261,9 @@ main(argc, argv)
 	offout = ftell(stdout);
 	offerr = ftell(stderr);
 
-#ifdef	MALLOC_TRACE
+#ifdef MALLOC_TRACE
 	mal_leaktrace(1);
-#endif	/* MALLOC_TRACE */
+#endif /* MALLOC_TRACE */
 
 	if (daemonflg) {
 		if (chdir(postoffice) < 0 || chdir(ROUTERDIR) < 0)
@@ -251,6 +276,7 @@ main(argc, argv)
 		  fprintf(stderr, "%s: daemon not started.\n", progname);
 		  die(1, "errors during startup");
 		}
+
 		if (tac == 0)		/* leave worldy matters behind */
 		  detach();
 		printf("%s: router daemon (%s)\n\tstarted at %s\n",
@@ -261,7 +287,7 @@ main(argc, argv)
 		    fprintf(stderr,"router can't create pidfile ?? Disk full ??\n");
 		    exit(2);
 		  }
-		if (nrouters > 1) {
+		if (nrouters > 0) {
 		  int pgrp, ppid = getpid();
 #ifndef	GETPGRP_VOID		/* We assume getpgrp() and setpgrp() calling
 				   conventions match in the machine.. */
@@ -275,19 +301,13 @@ main(argc, argv)
 		    fprintf(stderr, "%s: daemon not started.\n", progname);
 		    die(1, "capability error during startup");
 		  }
-		  for (router_id = 2; router_id <= nrouters; ++router_id){
-		    if (fork() <= 0) /* child or error */
-		      break;
-		    /* parent */
-		  }
-		  printf("%s: number %d started as pid %d\n",
-			 progname, router_id, getpid());
-		  router_id = getpid();
 		}
-		else
-		  router_id = getpid();
-		zopenlog("router", LOG_PID, LOG_MAIL);
+		router_id = getpid();
 	}
+
+	/* Each (sub-)process does openlog() all by themselves */
+	zopenlog("router", LOG_PID, LOG_MAIL);
+
 	if (c < argc) {
 	  savefile = 1;
 	  /*
@@ -320,8 +340,8 @@ write(30, "\n", 1);
 	dbfree();
 	zshfree();
 #endif	/* MALLOC_TRACE */
-	if (mustexit)
-	  die(0, "signal");
+	/* if (mustexit)
+	   die(0, "signal"); */
 #ifdef	MALLOC_TRACE
 	die(0, "malloc trace");
 #endif	/* MALLOC_TRACE */
@@ -353,6 +373,8 @@ initialize(configfile, argc, argv)
 	while (argc-- > 0)
 		av[ac++] = *argv++;
 	av[ac] = NULL;
+
+	staticprot(&s_value);
 	zshinit(ac, av);
 
 	/* add builtin router functions to list of builtin shell functions */
@@ -418,7 +440,7 @@ login_to_uid(name)
 	struct passwd *pw;
 	uid_t uid;
 	char buf[BUFSIZ];
-	char *cp, *fn;
+	char *cp;
 	struct spblk *spl;
 
 	spl = lookup_incoresp(name, spt_loginmap);
@@ -432,14 +454,14 @@ login_to_uid(name)
 		} else {
 			uid = pw->pw_uid;
 			cp = strsave(pw->pw_name);
-			sp_install(uid, cp, 0, spt_uidmap);
+			sp_install(uid, cp, 0L, spt_uidmap);
 			fullname(pw->pw_gecos,buf,sizeof buf,pw->pw_name);
 			add_incoresp(cp, buf, spt_fullnamemap);
 		}
 		addd_incoresp(name, (void*)((long)uid), spt_loginmap);
 		stickymem = oval;
 	} else
-		uid = (int)(spl->data);
+		uid = (long)(spl->data);
 	return uid;
 }
 
@@ -452,9 +474,9 @@ uidpwnam(uid)
 	struct spblk *spl;
 	char buf[BUFSIZ];
 
-	spl = sp_lookup((u_int)uid, spt_uidmap);
+	spl = sp_lookup((u_long)uid, spt_uidmap);
 	if (spl == NULL) {
-		pw = getpwuid(uid);
+		pw = getpwuid((uid_t)uid);
 		if (pw == NULL) {
 			/* memory shall be temporary in
 			   its nature for this data! */
@@ -468,7 +490,7 @@ uidpwnam(uid)
 			addd_incoresp(pw->pw_name, (void*)((long)(pw->pw_uid)), spt_loginmap);
 			fullname(pw->pw_gecos,buf, sizeof buf, pw->pw_name);
 			add_incoresp(pw->pw_name, buf, spt_fullnamemap);
-			sp_install((u_int)uid, cp, 0, spt_uidmap);
+			sp_install((u_int)uid, cp, 0L, spt_uidmap);
 			stickymem = oval;
 		}
 	} else
@@ -572,53 +594,7 @@ logit(file, id, from, to)
 	const char *file, *id, *from, *to;
 {
 	int flen, baselen;
-#if 0
-	char c;
 
-	if (id == NULL)
-		id = file;
-	baselen = strlen(file) + strlen(id) + 4;
-	c = '\0';
-	while (baselen + strlen(from) > MAXSAFESIZE) {
-		/* Wonderful software we're dealing with here... */
-		c = *(from+MAXSAFESIZE-baselen);
-		*(from+MAXSAFESIZE-baselen) = '\0';
-		printf("%s: file: %s %s...\n", id, file, from);
-		if (!nosyslog)
-		  zsyslog((LOG_INFO, "%s: file: %s %s...", id, file, from));
-		from += MAXSAFESIZE-baselen;
-		*from = c;
-		*--from = '.';
-		*--from = '.';
-		*--from = '.';
-	}
-	flen = strlen(from);
-	while (baselen + flen + strlen(to) > MAXSAFESIZE) {
-		c = *(to+MAXSAFESIZE-baselen-flen);
-		*(to+MAXSAFESIZE-baselen-flen) = '\0';
-		if (flen > 0)
-			printf("%s: file: %s %s =>%s\n", id, file, from, to);
-		else
-			printf("%s: file: %s %s\n", id, file, to);
-		if (!nosyslog)
-		  if (flen > 0)
-		    zsyslog((LOG_INFO, "%s: file: %s %s =>%s",
-			     id, file, from, to));
-		  else
-		    zsyslog((LOG_INFO, "%s: file: %s %s", id, file, to));
-		to += MAXSAFESIZE-baselen-flen;
-		*to = c;
-		*--to = '.';
-		*--to = '.';
-		*--to = '.';
-		flen = 0;
-	}
-	if (flen > 0) {
-	  printf("%s: file: %s %s =>%s\n", id, file, from, to);
-	  if (!nosyslog)
-	    zsyslog((LOG_INFO, "%s: file: %s %s =>%s", id, file, from, to));
-	}
-#else
 	if (id == NULL)
 		id = file;
 	baselen = strlen(file) + strlen(id) + 4;
@@ -632,13 +608,13 @@ logit(file, id, from, to)
 		 id, file, from, to);
 	else
 	  printf("%.200s: file: %.150s %.250s\n", id, file, to);
-	if (!nosyslog)
+	if (!nosyslog) {
 	  if (flen > 0)
 	    zsyslog((LOG_INFO, "%.200s: file: %.150s %.200s => %.200s",
 		     id, file, from, to));
 	  else
 	    zsyslog((LOG_INFO, "%.200s: file: %.150s %.200s", id, file, to));
-#endif
+	}
 }
 
 const char *
