@@ -8,7 +8,7 @@
  */
 
 #include "hostenv.h"
-#include <stdio.h>
+#include <sfio.h>
 #include "scheduler.h"
 #include "prototypes.h"
 
@@ -28,19 +28,22 @@ doagenda()
 	do {
 	  didsomething = 0;
 	  if (verbose)
-	    printf("curitem %lx curitem->wakeup %lu now %d\n", (long)ncuritem,
-		   ncuritem != NULL ? ncuritem->wakeup : 0, (int)now);
+	    sfprintf(sfstdout,"curitem %p curitem->wakeup %lu now %d\n",
+		     ncuritem, ncuritem ? ncuritem->wakeup : 0, (int)now);
 
 	  /* thread_head -chain should be in time order, thus the while-loop
 	     should be traversed only once, because  thread_start kicks
 	     right away (or reschedules) */
 
-	  while (ncuritem != NULL /* && ncuritem->wakeup <= now */ ) {
+	  while (ncuritem  && ncuritem->wakeup <= now  ) {
 	    /* Object pointed by  ncuritem  may disappear due to
 	       expiration at thread_start() .. */
 	    nncuritem = ncuritem->nexttr;
+	    /* Not running, and wakeup in past ? */
 	    if (ncuritem->proc == NULL && ncuritem->wakeup <= now)
-	      didsomething += thread_start(ncuritem);
+	      /* Try to start it! */
+	      while (thread_start(ncuritem, 0))
+		didsomething += 1;
 
 	    ncuritem = nncuritem;
 	  }
@@ -72,16 +75,19 @@ const char *turnarg;
 	struct web *wp;
 	char *cp = strchr(turnarg,' ');
 	struct thread *ncuritem, *nncuritem;
+	spkey_t spk;
 
 	/* caller has done 'strlower()' to our input.. */
 	if (cp) *cp++ = 0;
 	/* If cp is non-null, some additional args were present.. */
 
-	spl = sp_lookup(symbol((void*)turnarg), spt_mesh[L_HOST]);
-	if (spl == NULL || (wp = (struct web *)spl->data) == NULL) {
+	spk = symbol_lookup_db((void*)turnarg, spt_mesh[L_HOST]->symbols);
+	spl = sp_lookup(spk, spt_mesh[L_HOST]);
+	if (spl == NULL || spl->data == NULL) {
 	  /* Not found, nothing to do.. */
 	  return;
 	}
+	wp = (struct web *)spl->data;
 	
 	ncuritem = thread_head;
 	while (ncuritem != NULL) {
@@ -91,7 +97,7 @@ const char *turnarg;
 
 	  if (wp == ncuritem->whost && ncuritem->proc == NULL) {
 	    ncuritem->wakeup = 0; /* Force its starttime! */
-	    thread_start(ncuritem);
+	    thread_start(ncuritem, 1);
 	    /* We MAY get multiple matches, though it is unlikely.. */
 	  }
 	  ncuritem = nncuritem;
