@@ -8,23 +8,23 @@
  */
 
 /* stdout and stderr end up in qlogdir/progname when scheduler is a daemon */
-char	*qlogdir = "/usr/spool/log";	/* overridden by LOGDIR zenvariable */
+const char * qlogdir = "/usr/spool/log"; /* overridden by LOGDIR zenvariable */
 
 /* suffix of scheduler configuration file, i.e. MAILSHARE/progname.qcf_suffix */
-char	*qcf_suffix = "conf";
+const char * qcf_suffix = "conf";
 
 /* default directory for transport agent commands, as in MAILBIN/qdefaultdir */
 /* This value is duplicated in several Makefiles in the distribution */
-char	*qdefaultdir = "ta";
+const char * qdefaultdir = "ta";
 
 /* output file for queue status command */
-char	*qoutputfile = "/usr/tmp/.mailq.text";
+const char * qoutputfile = "/usr/tmp/.mailq.text";
 
 /* command parameter that is replaced with a host name */
-char	*replhost = "$host";
+const char * replhost = "$host";
 
 /* command parameter that is replaced with a channel name */
-char	*replchannel = "$channel";
+const char * replchannel = "$channel";
 
 /* unprivileged user id */
 int	nobody = -2;
@@ -32,9 +32,12 @@ int	nobody = -2;
 /* directory scanning interval in seconds */
 int	sweepinterval = 10;
 
-/* At most 10 new childs per second -- The 'R'-option can be used to increase this */
-int	forkrate_limit = 10;
+/* At most 30 new childs per second -- The 'R'-option can be used to increase this */
+int	forkrate_limit = 30;
 
 /* Set if you don't want to return full content with error report! */
 int	never_full_content = 0;
 
+/* If we use ZMailer malloc library in debug mode, BAD-PTR looks like
+   following value.. */
+void *BADPTR = (void*) 0x5555555555555555L;
