@@ -4,7 +4,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1995
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1999
  */
 
 /*
@@ -22,11 +22,14 @@
 #include <sys/stat.h>
 #include <sys/param.h>
 #include <pwd.h>
+#include <stdlib.h>
 #include "mail.h"
 #include "scheduler.h"
-#include "malloc.h"
+#include "zmalloc.h"
 #include "mailer.h"
 
+#include "md5.h"
+
 #ifdef HAVE_DIRENT_H
 # include <dirent.h>
 #else /* not HAVE_DIRENT_H */
@@ -47,9 +50,13 @@
 #include <netdb.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
+#include <arpa/inet.h>
 #include <fcntl.h>
 #include <sys/file.h>
 #endif	/* HAVE_SOCKET */
+#ifdef HAVE_SYS_UN_H
+#include <sys/un.h>
+#endif
 
 #ifdef	MALLOC_TRACE
 struct conshell *envarlist = NULL;
@@ -73,6 +80,11 @@ extern char *strchr(), *strrchr();
 const char	*progname;
 const char	*postoffice;
 
+static char *port = NULL;
+
+char * v2username = "nobody";
+char * v2password = "nobody";
+
 int	debug, verbose, summary, user, status, onlyuser, nonlocal, schedq;
 int	sawcore, othern;
 
@@ -82,6 +94,8 @@ extern char *optarg;
 extern int   optind;
 char	path[MAXPATHLEN];
 
+char *host = NULL;
+
 int
 main(argc, argv)
 	int argc;
@@ -89,14 +103,7 @@ main(argc, argv)
 {
 	int fd, c, errflg, eval;
 	struct passwd *pw;
-#ifdef	AF_INET
-	short port = 0;
-	struct in_addr naddr;
-	char *host = NULL;
-	struct hostent *hp = NULL, he;
-	struct sockaddr_in sad;
-	struct servent *serv = NULL;
-#else	/* !AF_INET */
+#ifndef	AF_INET
 	char *rendezvous = NULL;
 	FILE *fp;
 	struct stat stbuf;
@@ -119,13 +126,9 @@ main(argc, argv)
 	    if (verbose == 0)
 	      ++verbose;
 	    break;
-#ifdef	AF_INET
+#if defined(AF_INET) || defined(AF_UNIX)
 	  case 'p':
-	    if ((port = (short)atoi(optarg)) <= 0) {
-	      fprintf(stderr, "%s: illegal port: %s\n", progname, optarg);
-	      ++errflg;
-	    } else
-	      port = htons(port);
+	    port = optarg;
 	    break;
 #else  /* !AF_INET */
 	  case 'r':
@@ -196,118 +199,163 @@ main(argc, argv)
 	  postoffice = POSTOFFICE;
 
 	sprintf(path, "%s/%s", postoffice, PID_SCHEDULER);
+
 	errno = 0;
-#ifdef	AF_INET
-	nonlocal = 0; /* Claim it to be: "localhost" */
 
-	if (status < 2 || summary) {
+#if defined(AF_UNIX) && defined(HAVE_SYS_UN_H)
+	if (port && *port == '/') {
+	  struct sockaddr_un sad;
 
-	  if (port == 0 && (serv = getservbyname("mailq", "tcp")) == NULL) {
-	    fprintf(stderr, "%s: cannot find 'mailq' tcp service\n", progname);
-	    exit(EX_TEMPFAIL);
-	  } else if (port == 0)
-	    port = serv->s_port;
-
-	  if (host == NULL) {
-	    host = getzenv("MAILSERVER");
-	    if ((host == NULL || *host == '\n')
-		&& (host = whathost(path)) == NULL) {
-	      if (status > 0) {
-		host = "127.0.0.1"; /* "localhost" */
-		nonlocal = 0;
-	      } else {
-		if (whathost(postoffice)) {
-		  fprintf(stderr, "%s: %s is not active", progname, postoffice);
-		  fprintf(stderr, " (\"%s\" does not exist)\n", path);
-		} else
-		  fprintf(stderr, "%s: cannot find postoffice host\n", progname);
-		exit(EX_OSFILE);
-	      }
-	    }
+	  if (status) {
+	    checkrouter();
+	    checkscheduler();
+	    if (status > 1 && !summary)
+	      exit(0);
 	  }
-	  hp = gethostbyname(host);
-	  if (hp == NULL) {
-	    naddr.s_addr = inet_addr(host);
-	    if (naddr.s_addr != -1) {
-	      naddr.s_addr = ntohl(naddr.s_addr);
-	      hp = gethostbyaddr((void*)&naddr, sizeof(naddr), AF_INET);
-	      if (hp == NULL){
-		hp = &he;
-		he.h_name = host;
-		he.h_length = 4;
-	      }
-	    } else if (hp == NULL) {
-	      fprintf(stderr, "%s: cannot find address of %s\n", progname, host);
-	      exit(EX_UNAVAILABLE);
-	    }
-	  }
-	  if (strcmp(host, "localhost") != 0 &&
-	      strcmp(host, "127.0.0.1") != 0) {
-	    printf("[%s]\n", hp->h_name);
-	    nonlocal = 1;
-	  } else
-	    nonlocal = 0;	/* "localhost" is per default a "local" */
-	}
-	if (status) {
-	  checkrouter();
-	  checkscheduler();
-	  if (status > 1 && !summary)
-	    exit(0);
-	}
-	/* try grabbing a port */
-	fd = socket(AF_INET, SOCK_STREAM, 0);
-	if (fd < 0) {
-	  fprintf(stderr, "%s: ", progname);
-	  perror("socket");
-	  exit(EX_UNAVAILABLE);
-	}
-	sad.sin_family = AF_INET;
-	if (hp == &he)
-	  sad.sin_addr.s_addr = htonl(naddr.s_addr);
-	else {
-	  char *addr;
 
-	  eval = EFAULT;
-	  hp_init(hp);
-	  while ((addr = *hp_getaddr()) != NULL) {
-	    sad.sin_addr.s_addr = 0;
-	    sad.sin_port = htons((short)0);
-	    if (bind(fd, (void*)&sad, sizeof sad) < 0) {
-	      fprintf(stderr, "%s: ", progname);
-	      perror("bind");
-	      exit(EX_UNAVAILABLE);
-	    }
-	    sad.sin_port = port;
-	    memcpy((void*)&sad.sin_addr, addr, hp->h_length);
-	    if (connect(fd, (void*)&sad, sizeof sad) < 0) {
-	      eval = errno;
-	      fprintf(stderr, "%s: connect failed to %s", progname,
-		      dottedquad(&sad.sin_addr));
-	      if ((addr = *hp_nextaddr()) != NULL) {
-		memcpy((char *)&sad.sin_addr, addr, hp->h_length);
-		fprintf(stderr, ", trying %s...\n", dottedquad(&sad.sin_addr));
-	      } else
-		putc('\n', stderr);
-	      close(fd);
-	      if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0){
-		fprintf(stderr, "%s: ", progname);
-		perror("socket");
-		exit(EX_UNAVAILABLE);
-	      }
-	      continue;
-	    }
-	    break;
-	  }
-	  if (*hp_getaddr() == NULL) {
+	  /* try grabbing a port */
+	  fd = socket(PF_UNIX, SOCK_STREAM, 0);
+	  if (fd < 0) {
 	    fprintf(stderr, "%s: ", progname);
-	    errno = eval;
-	    perror("connect");
-	    fprintf(stderr, "%s: unable to contact scheduler on %s\n",
-		    progname, hp->h_name);
+	    perror("socket");
 	    exit(EX_UNAVAILABLE);
 	  }
+
+	  sad.sun_family = AF_UNIX;
+	  strncpy(sad.sun_path, port, sizeof(sad.sun_path));
+	  sad.sun_path[ sizeof(sad.sun_path) ] = 0;
+
+	  if (connect(fd, (void*)&sad, sizeof sad) < 0) {
+	    fprintf(stderr,"%s: connect failed to path: '%s'\n",progname,sad.sun_path);
+	    exit(EX_UNAVAILABLE);
+	  }
+
+	  docat((char *)NULL, fd);
+	}
+#endif
+#ifdef	AF_INET
+	if (!port || (port && *port != '/')) {
+
+	  struct in_addr naddr;
+	  struct hostent *hp = NULL, he;
+	  struct sockaddr_in sad;
+	  struct servent *serv = NULL;
+
+	  int portnum = 174;
+	  nonlocal = 0; /* Claim it to be: "localhost" */
+
+	  if (status < 2 || summary) {
+
+	    if (port && isdigit(*port)) {
+	      portnum = atol(port);
+	    } else if (port == NULL &&
+		       (serv = getservbyname(port ? port : "mailq", "tcp")) == NULL) {
+
+	      fprintf(stderr,"%s: cannot find 'mailq' tcp service\n",progname);
+
+	    } else if (port == 0)
+	      
+	      portnum = ntohs(serv->s_port);
+
+	    if (host == NULL) {
+	      host = getzenv("MAILSERVER");
+	      if ((host == NULL || *host == '\n')
+		  && (host = whathost(path)) == NULL) {
+		if (status > 0) {
+		  host = "127.0.0.1"; /* "localhost" */
+		  nonlocal = 0;
+		} else {
+		  if (whathost(postoffice)) {
+		    fprintf(stderr, "%s: %s is not active", progname, postoffice);
+		    fprintf(stderr, " (\"%s\" does not exist)\n", path);
+		  } else
+		    fprintf(stderr, "%s: cannot find postoffice host\n", progname);
+		  exit(EX_OSFILE);
+		}
+	      }
+	    }
+	    hp = gethostbyname(host);
+	    if (hp == NULL) {
+	      if (inet_pton(AF_INET, host, &naddr.s_addr) == 1) {
+		hp = gethostbyaddr((void*)&naddr, sizeof(naddr), AF_INET);
+		if (hp == NULL){
+		  hp = &he;
+		  he.h_name = host;
+		  he.h_length = 4;
+		}
+	      } else if (hp == NULL) {
+		fprintf(stderr, "%s: cannot find address of %s\n", progname, host);
+		exit(EX_UNAVAILABLE);
+	      }
+	    }
+	    if (strcmp(host, "localhost") != 0 &&
+		strcmp(host, "127.0.0.1") != 0) {
+	      fprintf(stdout,"[%s]\n", hp->h_name);
+	      nonlocal = 1;
+	    } else
+	      nonlocal = 0;	/* "localhost" is per default a "local" */
+	  }
+	  if (status) {
+	    checkrouter();
+	    checkscheduler();
+	    if (status > 1 && !summary)
+	      exit(0);
+	  }
+	  /* try grabbing a port */
+	  fd = socket(PF_INET, SOCK_STREAM, 0);
+	  if (fd < 0) {
+	    fprintf(stderr, "%s: ", progname);
+	    perror("socket");
+	    exit(EX_UNAVAILABLE);
+	  }
+	  sad.sin_family = AF_INET;
+	  if (hp == &he)
+	    sad.sin_addr.s_addr = htonl(naddr.s_addr);
+	  else {
+	    char *addr;
+
+	    eval = EFAULT;
+	    hp_init(hp);
+	    while ((addr = *hp_getaddr()) != NULL) {
+	      sad.sin_addr.s_addr = 0;
+	      sad.sin_port = htons((short)0);
+	      if (bind(fd, (void*)&sad, sizeof sad) < 0) {
+		fprintf(stderr, "%s: ", progname);
+		perror("bind");
+		exit(EX_UNAVAILABLE);
+	      }
+	      sad.sin_port = htons(portnum);
+	      memcpy((void*)&sad.sin_addr, addr, hp->h_length);
+	      if (connect(fd, (void*)&sad, sizeof sad) < 0) {
+		eval = errno;
+		fprintf(stderr, "%s: connect failed to %s", progname,
+			dottedquad(&sad.sin_addr));
+		if ((addr = *hp_nextaddr()) != NULL) {
+		  memcpy((char *)&sad.sin_addr, addr, hp->h_length);
+		  fprintf(stderr, ", trying %s...\n", dottedquad(&sad.sin_addr));
+		} else
+		  putc('\n', stderr);
+		close(fd);
+		if ((fd = socket(PF_INET, SOCK_STREAM, 0)) < 0){
+		  fprintf(stderr, "%s: ", progname);
+		  perror("socket");
+		  exit(EX_UNAVAILABLE);
+		}
+		continue;
+	      }
+	      break;
+	    }
+	    if (*hp_getaddr() == NULL) {
+	      fprintf(stderr, "%s: ", progname);
+	      errno = eval;
+	      perror("connect");
+	      fprintf(stderr, "%s: unable to contact scheduler on %s\n",
+		      progname, hp->h_name);
+	      exit(EX_UNAVAILABLE);
+	    }
+	  }
+	  docat((char *)NULL, fd);
 	}
-	docat((char *)NULL, fd);
 #else	/* !AF_INET */
 	if (strcmp(host, "localhost") == 0 ||
 	    strcmp(host, "127.0.0.1") == 0) {
@@ -376,22 +424,28 @@ docat(file, fd)
 	const char *file;
 	int fd;
 {
-	char buf[BUFSIZ];
-	int n;
-	FILE *fp = NULL;
+	FILE *fpi = NULL, *fpo = NULL;
 
-	if (fd < 0 && (fp = fopen(file, "r")) == NULL) {
+	if (fd < 0 && (fpi = fopen(file, "r")) == NULL) {
 	  fprintf(stderr, "%s: %s: %s\n", progname, file, strerror(errno));
 	  exit(EX_OSFILE);
 	  /* NOTREACHED */
-	} else if (fd >= 0)
-	  fp = fdopen(fd, "r");
-	if (debug)
-	  while ((n = fread(buf, sizeof buf[0], sizeof buf, fp)) > 0)
+	} else if (fd >= 0) {
+	  fpi = fdopen(fd, "r");
+	  fpo = fdopen(fd, "w");
+	}
+#if 0
+	if (debug && fpi) {
+	  char buf[BUFSIZ];
+	  int n;
+	  while ((n = fread(buf, 1, sizeof buf, fpi)) > 0)
 	    fwrite(buf, sizeof buf[0], n, stdout);
-	else
-	  report(fp);
-	fclose(fp);
+	} else
+#endif
+	  if (fpi && fpo)
+	    report(fpi, fpo);
+	if (fpi) fclose(fpi);
+	if (fpo) fclose(fpo);
 }
 
 /*
@@ -430,7 +484,7 @@ checkrouter()
 	close(dirp->dd_fd);
 #endif
 	closedir(dirp);
-	printf("%d entr%s in router queue: ", n, n != 1 ? "ies" : "y");
+	fprintf(stdout,"%d entr%s in router queue: ", n, n != 1 ? "ies" : "y");
 
 	if (nonlocal)
 	  r = -2;
@@ -444,23 +498,23 @@ checkrouter()
 	  if (fstat(FILENO(fp), &pidbuf) < 0) {
 	    fprintf(stderr, "\n%s: fstat: %s", progname, strerror(errno));
 	  } else if (stat(path, &corebuf) == 0 && pidbuf.st_mtime < corebuf.st_mtime)
-	    printf("core dumped\n");
+	    fprintf(stdout,"core dumped\n");
 	  else
-	    printf("no daemon\n");
+	    fprintf(stdout,"no daemon\n");
 	  fclose(fp);
 	  break;
 	case EX_OK:
 	  if (n)
-	    printf("processing\n");
+	    fprintf(stdout,"processing\n");
 	  else
-	    printf("idle\n");
+	    fprintf(stdout,"idle\n");
 	  fclose(fp);
 	  break;
 	case -2:
-	  printf("non-local\n");
+	  fprintf(stdout,"non-local\n");
 	  break;
 	default:
-	  printf("never started\n");
+	  fprintf(stdout,"never started\n");
 	  break;
 	}
 
@@ -485,7 +539,7 @@ checkrouter()
 #endif
 	closedir(dirp);
 	if (n)
-	  printf("%d message%s deferred\n", n, n != 1 ? "s" : "");
+	  fprintf(stdout,"%d message%s deferred\n", n, n != 1 ? "s" : "");
 }
 
 int countfiles __((const char *));
@@ -558,7 +612,7 @@ checkscheduler()
 
 	n = countfiles(path);
 
-	printf("%d message%s in transport queue: ",
+	fprintf(stdout,"%d message%s in transport queue: ",
 	       n, n != 1 ? "s" : "");
 
 	if (nonlocal)
@@ -568,27 +622,27 @@ checkscheduler()
 
 	switch (r) {
 	case EX_UNAVAILABLE:
-	  printf("no scheduler daemon");
+	  fprintf(stdout,"no scheduler daemon");
 	  fclose(fp);
 	  break;
 	case EX_OK:
 	  if (n == 0)
-	    printf("idle");
+	    fprintf(stdout,"idle");
 	  else
-	    printf("working");
+	    fprintf(stdout,"working");
 	  break;
 	case -2:
-	  printf("non-local");
+	  fprintf(stdout,"non-local");
 	  break;
 	default:
-	  printf("never started");
+	  fprintf(stdout,"never started");
 	  if (n > 0)
-	    printf(" \"%s/%s\" polluted", postoffice, TRANSPORTDIR);
+	    fprintf(stdout," \"%s/%s\" polluted", postoffice, TRANSPORTDIR);
 	  break;
 	}
 	if (sawcore)
-	  printf(" (core exists)");
-	printf("\n");
+	  fprintf(stdout," (core exists)");
+	fprintf(stdout,"\n");
 
 }
 
@@ -622,20 +676,56 @@ isalive(pidfil, pidp, fpp)
 #define	LEN_MAGIC_PREAMBLE	(sizeof MAGIC_PREAMBLE - 1)
 #define	VERSION_ID		"zmailer 1.0"
 #define	VERSION_ID2		"zmailer 2.0"
-#define	GETLINE(buf,fp)		\
-	if (fgets(buf, bufsize, fp) == NULL) {\
-		fprintf(stderr, "%s: no input from scheduler\n", progname);\
-		buf[0] = '\0';\
-		return 0;\
-	}\
-	{ char *s; if ((s = strchr(buf,'\n'))) *s = '\0'; }
-				
+
+static int _getline(buf, bufsize, bufspace, fp)
+     char **buf;
+     int *bufsize;
+     int *bufspace;
+     FILE *fp;
+{
+  int c;
+
+  if (!*buf) {
+    *bufsize = 0;
+    *bufspace = 110;
+    *buf = malloc(*bufspace+3);
+  }
+
+  while ((c = fgetc(fp)) != EOF) {
+    if (c == '\n')
+      break;
+
+    if (*bufsize >= *bufspace) {
+      *bufspace *= 2;
+      *buf = realloc(*buf, *bufspace+3);
+    }
+    (*buf)[*bufsize] = c;
+    *bufsize += 1;
+  }
+  (*buf)[*bufsize] = 0;
+
+  if (c == EOF && *bufsize != 0) {
+    fprintf(stderr, "%s: no input from scheduler\n", progname);
+    (*buf)[0] = '\0';
+    return -1;
+  }
+
+  if (debug && *buf)
+    fprintf(stderr, "- %s\n",*buf);
+
+  return 0; /* Got something */
+}
+
+
+#define GETLINE(buf, bufsize, bufspace, fp) _getline(&buf, &bufsize, &bufspace, fp)
+
 
 const char *names[SIZE_L+2];
 
 #define	L_VERTEX	SIZE_L
 #define L_END		SIZE_L+1
-struct sptree *spt_ids[SIZE_L+2];
+struct sptree *spt_ids [SIZE_L+2];
+struct sptree *spt_syms[SIZE_L+2];
 
 #define	EQNSTR(a,b)	(!strncmp(a,b,strlen(b)))
 
@@ -651,18 +741,17 @@ parse(fp)
 	register int	i;
 	u_long	list, key;
 	struct spblk *spl;
-	int bufsize = 8192*32; /* Sometimes the input has LONG lines.. */
-	char  *buf, *ocp;
+	int bufsize, bufspace;
+	char  *buf = NULL, *ocp;
 
 	names[L_CTLFILE] = "Vertices:";
 	names[L_HOST]    = "Hosts:";
 	names[L_CHANNEL] = "Channels:";
 	names[L_END]     = "End:";
 
-
-	buf = emalloc(bufsize);
-
-	GETLINE(buf,fp);
+	bufsize = 0;
+	if (GETLINE(buf,bufsize,bufspace,fp))
+	  return 0;
 
 	if (EQNSTR(buf, MAGIC_PREAMBLE) &&
 	    EQNSTR(buf+LEN_MAGIC_PREAMBLE, VERSION_ID2))
@@ -673,17 +762,37 @@ parse(fp)
 	  fprintf(stderr, "%s: version mismatch, input is \"%s\".\n", progname, buf);
 	  return 0;
 	}
-	GETLINE(buf,fp);
+
+	if (schedq) {
+	  /* We ignore the classical mailq data, just read it fast */
+	  while (1) {
+	    bufsize = 0;
+	    if (GETLINE(buf, bufsize, bufspace, fp))
+	      return 1; /* EOF ? */
+	    if (memcmp(buf,"End:",4) == 0)
+	      return 1;
+	  }
+	  /* NOT REACHED */
+	}
+
+	bufsize = 0;
+	if (GETLINE(buf,bufsize,bufspace,fp))
+	  return 0;
 	if (!EQNSTR(buf, names[L_CTLFILE]))
 	  return 0;
 	list = L_CTLFILE;
-	spt_ids[L_CTLFILE] = sp_init();
-	spt_ids[L_VERTEX ] = sp_init();
-	spt_ids[L_CHANNEL] = sp_init();
-	spt_ids[L_HOST   ] = sp_init();
+	spt_ids [L_CTLFILE] = sp_init();
+	spt_ids [L_VERTEX ] = sp_init();
+	spt_ids [L_CHANNEL] = sp_init();
+	spt_ids [L_HOST   ] = sp_init();
+	spt_syms[L_CTLFILE] = sp_init();
+	spt_syms[L_VERTEX ] = sp_init();
+	spt_syms[L_CHANNEL] = sp_init();
+	spt_syms[L_HOST   ] = sp_init();
 	while (1) {
 
-	  if (fgets(buf, bufsize-1, fp) == NULL)
+	  bufsize = 0;
+	  if (GETLINE(buf, bufsize, bufspace, fp))
 	    break;
 
 	  switch ((int)list) {
@@ -711,7 +820,11 @@ parse(fp)
 	    while (isascii(*cp) && !isspace(*cp))
 	      ++cp;
 	    *cp++ = '\0';
-	    spl = sp_lookup(symbol(ocp), spt_ids[L_CTLFILE]);
+if (debug)
+  fprintf(stderr," - '%s'\n",ocp);
+
+	    spl = sp_lookup(symbol_db(ocp,spt_syms[L_CTLFILE]),
+			    spt_ids[L_CTLFILE]);
 	    if (spl == NULL || (cfp = (struct ctlfile *)spl->data) == NULL) {
 	      cfp = (struct ctlfile *)emalloc(sizeof (struct ctlfile));
 	      memset((void*)cfp,0,sizeof(struct ctlfile));
@@ -724,7 +837,8 @@ parse(fp)
 	      cfp->id = 0;
 	      cfp->mid = strsave(ocp);
 	      cfp->mark = 0;
-	      sp_install(symbol(ocp), (void *)cfp, 0, spt_ids[L_CTLFILE]);
+	      sp_install(symbol_db(ocp,spt_syms[L_CTLFILE]),
+			 (void *)cfp, 0, spt_ids[L_CTLFILE]);
 	    }
 	    while (*cp == ' ' || *cp == '\t')
 	      ++cp;
@@ -732,6 +846,10 @@ parse(fp)
 	    while ('0' <= *cp && *cp <= '9')
 	      ++cp;
 	    *cp++ = '\0';
+
+if (debug)
+  fprintf(stderr," - '%s'\n",ocp);
+
 	    if ((i = atoi(ocp)) < 1) {
 	      fprintf(stderr, "%s: bad number of addresses: '%s'\n", progname, ocp);
 	      break;
@@ -748,6 +866,10 @@ parse(fp)
 		++cp;
 	      *cp = '\0';
 	      v->index[i++] = atol(ocp);
+
+if (debug)
+  fprintf(stderr," - '%s'\n",ocp);
+
 	    }
 	    while (*cp != '\0' && *cp != '\n' && *cp != '#')
 	      ++cp;
@@ -757,6 +879,10 @@ parse(fp)
 		++cp;
 	      *cp = '\0';
 	      v->message = strsave(ocp);
+
+if (debug)
+  fprintf(stderr," - '%s'\n",ocp);
+
 	    } else
 	      v->message = NULL;
 	    v->next[L_CTLFILE] = cfp->head;
@@ -796,15 +922,20 @@ parse(fp)
 	    }
 	    *cp++ = '\0';
 
+if (debug)
+  fprintf(stderr," - '%s'\n",buf);
+
+
 	    /* Look for channel/host identifier splay-tree */
-	    spl = sp_lookup(symbol(buf), spt_ids[list]);
+	    spl = sp_lookup(symbol_db(buf,spt_syms[list]), spt_ids[list]);
 	    if (spl == NULL || (w = (struct web *)spl->data) == NULL) {
 	      w = (struct web *)emalloc(sizeof (struct web));
 	      memset((void*)w,0,sizeof(struct web));
 	      w->name = strsave(buf);
 	      w->kids = 0;
 	      w->link = w->lastlink = NULL;
-	      sp_install(symbol(buf), (void *)w, 0, spt_ids[list]);
+	      sp_install(symbol_db(buf,spt_syms[list]),
+			 (void *)w, 0, spt_ids[list]);
 	    }
 	    while (*cp == ' ' || *cp == '\t')
 	      ++cp;
@@ -812,11 +943,18 @@ parse(fp)
 	    /* Pick each vertex reference */
 
 	    ++cp;		/* skip the first '>' */
-	    while (*cp != '\0' && isascii(*cp) && isdigit(*cp)) {
+	    while (*cp && isascii(*cp) && isdigit(*cp)) {
+	      int c;
 	      ocp = cp;
-	      while (isascii(*cp) && isdigit(*cp))
+	      while (*cp && isascii(*cp) && isdigit(*cp))
 		++cp;
-	      *cp++ = '\0';
+	      c = *cp;
+	      *cp = '\0';
+	      if (c) ++cp;
+
+if (debug)
+  fprintf(stderr," - '%s'\n",ocp);
+
 	      spl = sp_lookup((u_long)atol(ocp), spt_ids[L_VERTEX]);
 	      if (spl == NULL || (v = (struct vertex *)spl->data)==NULL) {
 		fprintf(stderr, "%s: unknown key %s\n", progname, ocp);
@@ -858,7 +996,7 @@ repscan(spl)
 	  if (vv->ngroup == 0)
 	    continue;
 	  if (!onlyuser)
-	    printf("%s/%s:\n", w->name, vv->orig[L_HOST]->name);
+	    fprintf(stdout,"%s/%s:\n", w->name, vv->orig[L_HOST]->name);
 	  else
 	    flag = 0;
 	  filecnt = 0;
@@ -877,18 +1015,18 @@ repscan(spl)
 	      }
 	      close(fd);
 	      if (flag == 0)
-		printf("%s/%s:\n", w->name, vv->orig[L_HOST]->name);
+		fprintf(stdout,"%s/%s:\n", w->name, vv->orig[L_HOST]->name);
 	    }
 	    if (!summary) {
 	      flag = 1;
-	      printf("\t%s", v->cfp->mid);
+	      fprintf(stdout,"\t%s", v->cfp->mid);
 	      if (v->ngroup > 1)
-		printf("/%d", v->ngroup);
-	      putchar(':');
+		fprintf(stdout,"/%d", v->ngroup);
+	      fprintf(stdout,":");
 	      if (v->message)
-		printf("\t%s\n", v->message);
+		fprintf(stdout,"\t%s\n", v->message);
 	      else
-		putchar('\n');
+		fprintf(stdout,"\n");
 	      if (verbose)
 		printaddrs(v);
 	    } else {
@@ -916,66 +1054,132 @@ repscan(spl)
 	    v->ngroup = 0;
 	  }
 	  if (summary == 1 && !onlyuser) {
-	    printf("\t  %d file%s, ", (int)filecnt, filecnt>1 ? "s":"");
+	    fprintf(stdout,"\t  %d file%s, ", (int)filecnt, filecnt>1 ? "s":"");
 	    if (filesizesum == 0)
-	      printf("no file size info available\n");
+	      fprintf(stdout,"no file size info available\n");
 	    else
-	      printf("%ld bytes total, %d bytes average\n",
-		     filesizesum, (int)(filesizesum/filecnt));
+	      fprintf(stdout,"%ld bytes total, %d bytes average\n",
+		      filesizesum, (int)(filesizesum/filecnt));
 	  }
 	  if (summary > 1 && !onlyuser) {
-	    printf("\t  %d file%s\n", (int)filecnt, filecnt>1 ? "s":"");
+	    fprintf(stdout,"\t  %d file%s\n", (int)filecnt, filecnt>1 ? "s":"");
 	  }
 	}
 	return 0;
 }
 
-void query2 __((FILE *));
-void query2(fp)
-	FILE *fp;
+void query2 __((FILE *, FILE*));
+void query2(fpi, fpo)
+	FILE *fpi, *fpo;
 {
-	int  len;
-	char buf[512];
+	int  len, i;
+	int bufsize = 0;
+	int bufspace = 0;
+	char *challenge = NULL;
+	char *buf = NULL;
+	MD5_CTX CTX;
+	unsigned char digbuf[16];
+
+	/* XX: Authenticate the query - get challenge */
+	bufsize = 0;
+	if (GETLINE(challenge, bufsize, bufspace, fpi))
+	  return;
+
+	MD5Init(&CTX);
+	MD5Update(&CTX, challenge, strlen(challenge));
+	MD5Update(&CTX, v2password, strlen(v2password));
+	MD5Final(digbuf, &CTX);
+	
+	fprintf(fpo, "AUTH %s ", v2username);
+	for (i = 0; i < 16; ++i) fprintf(fpo,"%02x",digbuf[i]);
+	fprintf(fpo, "\n");
+	if (fflush(fpo) || ferror(fpo)) {
+	    perror("login to scheduler command interface failed");
+	    return;
+	}
+
+	bufsize = 0;
+	if (GETLINE(buf, bufsize, bufspace, fpi))
+	    return;
+
+	if (*buf != '+') {
+	  fprintf(stdout,"User '%s' not accepted to server '%s'; err='%s'\n",
+		  v2username, host ? host : "<NO-HOST-?>", buf+1);
+	  return;
+	}
 
 	if (schedq) {
-	  if (schedq > 1)
-	    strcpy(buf,"SHOW-QUEUE CONDENCED\n");
+
+	  if (schedq > 2)
+	    strcpy(buf,"SHOW SNMP\n");
+	  else if (schedq > 1)
+	    strcpy(buf,"SHOW QUEUE SHORT\n");
 	  else
-	    strcpy(buf,"SHOW-QUEUE THREADS\n");
+	    strcpy(buf,"SHOW QUEUE THREADS\n");
+
 	  len = strlen(buf);
-	  if (write(fileno(fp),buf,len) != len) {
+
+	  if (fwrite(buf,1,len,fpo) != len || fflush(fpo)) {
 	    perror("write to scheduler command interface failed");
 	    return;
 	  }
-	  while (!feof(fp) && !ferror(fp)) {
-	    int c = getc(fp);
-	    if (c == EOF) break;
-	    putc(c,stdout);
+
+	  bufsize = 0;
+	  if (GETLINE(buf, bufsize, bufspace, fpi))
+	    return;
+
+	  if (*buf != '+') {
+
+	    fprintf(stdout,"Scheduler response: '%s'\n",buf);
+
+	  } else {
+
+	    for (;;) {
+	      bufsize = 0;
+	      if (GETLINE(buf, bufsize, bufspace, fpi))
+		break;
+	      if (buf[0] == '.' && buf[1] == 0)
+		break;
+	      /* Do leading dot duplication suppression */
+	      fprintf(stdout,"%s\n",((*buf == '.') ? buf+1 : buf));
+	    }
+
 	  }
 	} else {
-	  printf("Sorry, scheduler with protocol version 2 can't give results without -Q option\n");
+	  fprintf(stdout,"Sorry, scheduler with protocol version 2 can't give results without -Q option\n");
 	}
 }
 
 void
-report(fp)
-	FILE *fp;
+report(fpi,fpo)
+	FILE *fpi, *fpo;
 {
-	int rc = parse(fp);
+	int rc = parse(fpi);
 	if (rc == 0)
 	  return;
 	if (rc == 2) {
-	  query2(fp);
+	  query2(fpi,fpo);
+	  return;
 	}
 	if (schedq) {
-	  char linebuf[4000];
-	  while (!feof(fp) && !ferror(fp)) {
-	    if (fgets(linebuf,sizeof(linebuf),fp) == NULL)
+	  /* Old-style processing */
+	  int prevc = -1;
+	  int linesuppress = 0;
+	  while (!ferror(fpi)) {
+	    int c = getc(fpi);
+	    if (c == EOF)
 	      break;
-	    if (schedq > 1 && linebuf[0] == ' ')
-	      continue;
-	    fputs(linebuf,stdout);
+	    if (prevc == '\n') {
+	      linesuppress = 0;
+	      if (c == ' ' && schedq > 1)
+		linesuppress = 1;
+	      fflush(stdout);
+	    }
+	    if (!linesuppress)
+	      putc(c,stdout);
+	    prevc = c;
 	  }
+	  fflush(stdout);
 	  return;
 	}
 
@@ -983,12 +1187,12 @@ report(fp)
 	sp_scan(repscan, (struct spblk *)NULL, spt_ids[L_CHANNEL]);
 	if (!r_i) {
 	  if (onlyuser)
-	    printf("No user messages found\n");
+	    fprintf(stdout,"No user messages found\n");
 	  else
 	    if (schedq == 0)
-	      printf("Transport queue is empty -- or scheduler uses -Q -mode\n");
+	      fprintf(stdout,"Transport queue is empty -- or scheduler uses -Q -mode\n");
 	    else
-	      printf("Transport queue is empty\n");
+	      fprintf(stdout,"Transport queue is empty\n");
 	}
 }
 
@@ -1004,25 +1208,25 @@ printaddrs(v)
 	if (v->cfp->contents == NULL) {
 	  sprintf(path, "%s/%s/%s", postoffice, TRANSPORTDIR, v->cfp->mid);
 	  if ((fd = open(path, O_RDONLY, 0)) < 0) {
-#if	0
-	    printf("\t\t%s: %s\n", path, strerror(errno));
+#if 0
+	    fprintf(stdout,"\t\t%s: %s\n", path, strerror(errno));
 #endif
 	    return;
 	  }
 	  if (fstat(fd, &stbuf) < 0) {
-	    printf("\t\tfstat(%s): %s\n", path, strerror(errno));
+	    fprintf(stdout,"\t\tfstat(%s): %s\n", path, strerror(errno));
 	    close(fd);
 	    return;
 	  }
 	  v->cfp->contents = malloc((u_int)stbuf.st_size);
 	  if (v->cfp->contents == NULL) {
-	    printf("\t\tmalloc(%d): out of memory!\n", (int)stbuf.st_size);
+	    fprintf(stdout,"\t\tmalloc(%d): out of memory!\n", (int)stbuf.st_size);
 	    close(fd);
 	    return;
 	  }
 	  errno = 0;
 	  if (read(fd, v->cfp->contents, stbuf.st_size) < stbuf.st_size){
-	    printf("\t\tread(%d): %s\n", (int)stbuf.st_size,
+	    fprintf(stdout,"\t\tread(%d): %s\n", (int)stbuf.st_size,
 		   errno == 0 ? "failed" : strerror(errno));
 	    close(fd);
 	    return;
@@ -1050,44 +1254,44 @@ printaddrs(v)
 	    if (stat(path, &stbuf) == 0) {
 	      /* overload offset[] to be size of message */
 	      v->cfp->offset[0] = stbuf.st_size;
-	      v->cfp->ctime     = stbuf.st_ctime;
+	      v->cfp->mtime     = stbuf.st_mtime;
 	    } else {
 	      v->cfp->offset[0] = 0;
-	      v->cfp->ctime     = 0;
+	      v->cfp->mtime     = 0;
 	    }
 	  }
 	}
 	if (summary)
 	  return;
 	if (v->cfp->logident)
-	  printf("\t  id\t%s", v->cfp->logident);
+	  fprintf(stdout,"\t  id\t%s", v->cfp->logident);
 	if (verbose > 1 && v->cfp->offset[0] > 0) {
-	  long dt = now - v->cfp->ctime;
-	  int fields = 2;
-	  printf(", %ld bytes, age ", v->cfp->offset[0]);
-	  /* age (now-ctime) printout */
+	  long dt = now - v->cfp->mtime;
+	  int fields = 3;
+	  fprintf(stdout,", %ld bytes, age ", (long)v->cfp->offset[0]);
+	  /* age (now-mtime) printout */
 	  if (dt > (24*3600)) {	/* Days */
-	    printf("%dd", (int)(dt /(24*3600)));
+	    fprintf(stdout,"%dd", (int)(dt /(24*3600)));
 	    dt %= (24*3600);
 	    --fields;
 	  }
 	  if (dt > 3600) {
-	    printf("%dh",(int)(dt/3600));
+	    fprintf(stdout,"%dh",(int)(dt/3600));
 	    dt %= 3600;
 	    --fields;
 	  }
 	  if (dt > 60 && fields > 0) {
-	    printf("%dm",(int)(dt/60));
+	    fprintf(stdout,"%dm",(int)(dt/60));
 	    dt %= 60;
 	    --fields;
 	  }
 	  if (fields > 0) {
-	    printf("%ds",(int)dt);
+	    fprintf(stdout,"%ds",(int)dt);
 	  }
 	}
-	putchar('\n');
+	fprintf(stdout,"\n");
 	if (v->cfp->mark > 0)
-	  printf("\t  from\t%s\n", v->cfp->contents + v->cfp->mark);
+	  fprintf(stdout,"\t  from\t%s\n", v->cfp->contents + v->cfp->mark);
 	for (i = 0; i < v->ngroup; ++i) {
 	  cp = v->cfp->contents + v->index[i] + 2;
 	  if (*cp == ' ' || (*cp >= '0' && *cp <= '9'))
@@ -1117,10 +1321,9 @@ printaddrs(v)
 	      *cp = '\0';
 	    }
 	  }
-	  putchar('\t');
+	  fprintf(stdout,"\t");
 	  if (i == 0)
-	    printf("  to");
-	  putchar('\t');
-	  puts(ocp);
+	    fprintf(stdout,"  to");
+	  fprintf(stdout,"\t%s\n",ocp);
 	}
 }
