@@ -1,6 +1,6 @@
 /*	Prototypes of ZMailer Scheduler component routines	*/
 /*
- *	Copyright Matti Aarnio <mea@nic.funet.fi> 1995
+ *	Copyright Matti Aarnio <mea@nic.funet.fi> 1995-1999
  */
 
 #ifndef __
@@ -52,6 +52,10 @@ extern const char *replhost;
 extern const char *replchannel;
 extern int  nobody;
 extern int  sweepinterval;
+extern char *mq2authfile;
+extern char *mailqsock;
+extern void *BADPTR;
+
 
 /* msgerror.c */
 extern void msgerror __((struct vertex *vp, long offset, const char *message));
@@ -97,36 +101,41 @@ extern void free_cfp_memory __((struct ctlfile *cfp));
 extern int  vtxredo __((struct spblk *spl));
 extern char *timestring __((void));
 extern time_t now;
-extern FILE *statuslog;
-extern void resync_file __((const char *filename));
+#ifdef _SFIO_H
+extern Sfio_t *statuslog;
+extern Sfio_t * vfp_open __((struct ctlfile *));
+#endif
+extern void resync_file __((struct procinfo *proc, const char *filename));
 extern int thread_count_recipients __((void));
 extern time_t mytime __((time_t *));
 extern int dq_insert __((void*, long, const char*, int));
 extern int in_dirscanqueue __((void *, long));
 extern const char *cfpdirname __((int));
 
+extern struct MIB_MtaEntry *MIBMtaEntry;
+
 /* threads.c */
 extern struct thread *thread_head, *thread_tail;
-extern struct threadgroup *create_threadgroup __((struct config_entry *cep, struct web *wc, struct web *wh, int withhost, void (*ce_fillin)__((struct threadgroup *, struct config_entry *)) ));
 extern void  delete_threadgroup __((struct threadgroup *thgp));
 extern void  thread_linkin __((struct vertex *cp, struct config_entry *cep, int cfgid, void (*ce_fillin)__((struct threadgroup *, struct config_entry *)) ));
 extern struct web *web_findcreate __((int flag, const char *s));
 extern void        unweb __((int flag, struct web *wp));
-extern int   thread_start __((struct thread *thr));
+extern int   thread_start __((struct thread *thr, int queue_only_too));
 extern int   idle_cleanup __((void));
-extern void  pick_next_vertex __((struct procinfo *proc, int ok, int justfree));
-extern void  thread_report __((FILE *));
+extern int   pick_next_vertex __((struct procinfo * proc));
+extern int   pick_next_thread __((struct procinfo * proc));
+#ifdef _SFIO_H
+extern void  thread_report __((Sfio_t *, int));
+#endif
 extern int   idleprocs;
-extern void  web_disentangle __((struct vertex *vp, int ok));
+extern void  web_detangle __((struct vertex *vp, int ok));
 extern void  reschedule __((struct vertex *vp, int factor, int index));
 extern void  thread_reschedule __((struct thread *, time_t, int index));
 
 /* transport.c */
 extern struct procinfo *cpids;
 extern int  numkids;
-extern void feed_child      __((struct procinfo *cpidp));
-extern int flush_child      __((struct procinfo *cpidp));
-extern void idle_child      __((struct procinfo *cpidp));
+extern void ta_hungry       __((struct procinfo *cpidp));
 extern int  start_child     __((struct vertex *vhead, struct web *channel, struct web *host));
 extern void shutdown_kids   __(( void ));
 extern RETSIGTYPE sig_chld __((int sig));
@@ -148,12 +157,23 @@ extern void expire __((struct vertex *, int));
 /* wantconn.c */
 extern int wantconn __(( int sock, const char *prgname ));
 
+/* mailq.c */
 extern int isalive __(( const char *pidfil, int *pidp, FILE **fpp ));
 extern void docat __(( const char *file, int fd ));
 extern void printaddrs __(( struct vertex *v ));
 extern void checkrouter __(( void ));
 extern void checkscheduler __(( void ));
-extern void report __(( FILE *fp ));
+extern void report __(( FILE *fpi, FILE *fpo ));
 
 /* Transport library */
-/* extern int lockaddr __((int, char *, long, int, int)); */
+/* extern int lockaddr __((int, char *, int, int, int)); */
+
+/* mq2.c */
+extern int mq2add_to_mask __((fd_set *rdmaskp, fd_set *wrmaskp, int));
+extern void mq2_register __((int fd));
+extern void mq2_areinsets __((fd_set *rdmaskp, fd_set *wrmaskp));
+extern int  mq2_puts __((struct mailq *, char *s));
+extern int  mq2_putc __((struct mailq *, int c));
+
+/* mq2auth.c */
+extern void mq2auth __((struct mailq *, char *));
