@@ -4,52 +4,66 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1995
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1999
  */
 
 #include "hostenv.h"
-#include <stdio.h>
+#include <sfio.h>
 #include <ctype.h>
 #include <pwd.h>
 #include <grp.h>
+#include <unistd.h>
 #include "scheduler.h"
 #include "prototypes.h"
 #include "mail.h"
+#ifdef HAVE_SYS_RESOURCE_H
+#include <sys/resource.h>
+#endif
 
 #include "libz.h"
 
 static void celink __((struct config_entry *, struct config_entry **, struct config_entry **));
-static int readtoken __((FILE *fp, char *buf, int buflen, int *linenump));
+static int readtoken __((Sfio_t *fp, char *buf, int buflen, int *linenump));
 static u_int parse_intvl __((char *string));
+static int paramparse __((char *line));
 
-static int rc_command  __((char *key, char *arg, struct config_entry *ce));
-static int rc_expform  __((char *key, char *arg, struct config_entry *ce));
-static int rc_expiry   __((char *key, char *arg, struct config_entry *ce));
-static int rc_group    __((char *key, char *arg, struct config_entry *ce));
-static int rc_interval __((char *key, char *arg, struct config_entry *ce));
-static int rc_maxchannel __((char *key, char *arg, struct config_entry *ce));
-static int rc_maxring  __((char *key, char *arg, struct config_entry *ce));
-static int rc_maxta    __((char *key, char *arg, struct config_entry *ce));
-static int rc_idlemax  __((char *key, char *arg, struct config_entry *ce));
-static int rc_retries  __((char *key, char *arg, struct config_entry *ce));
-static int rc_user     __((char *key, char *arg, struct config_entry *ce));
-static int rc_skew     __((char *key, char *arg, struct config_entry *ce));
-static int rc_bychannel __((char *key, char *arg, struct config_entry *ce));
-static int rc_ageorder __((char *key, char *arg, struct config_entry *ce));
-static int rc_queueonly __((char *key, char *arg, struct config_entry *ce));
-static int rc_deliveryform __((char *key, char *arg, struct config_entry *ce));
-static int rc_overfeed __((char *key, char *arg, struct config_entry *ce));
+#define RCKEYARGS __((char *key, char *arg, struct config_entry *ce))
+
+static int rc_command		RCKEYARGS;
+static int rc_expform		RCKEYARGS;
+static int rc_expiry		RCKEYARGS;
+static int rc_group		RCKEYARGS;
+static int rc_interval		RCKEYARGS;
+static int rc_maxchannel	RCKEYARGS;
+static int rc_maxring		RCKEYARGS;
+static int rc_maxta		RCKEYARGS;
+static int rc_maxthr		RCKEYARGS;
+static int rc_idlemax		RCKEYARGS;
+static int rc_retries		RCKEYARGS;
+static int rc_user		RCKEYARGS;
+static int rc_skew		RCKEYARGS;
+static int rc_bychannel		RCKEYARGS;
+static int rc_ageorder		RCKEYARGS;
+static int rc_queueonly		RCKEYARGS;
+static int rc_deliveryform	RCKEYARGS;
+static int rc_overfeed		RCKEYARGS;
+static int rc_priority		RCKEYARGS;
+static int rc_nice		RCKEYARGS;
+static int rc_syspriority	RCKEYARGS;
+static int rc_sysnice		RCKEYARGS;
 
 extern int errno;
 
-static const char *mailbin = NULL;
-
 extern struct group *getgrnam();
 extern struct passwd *getpwnam();
 
 struct config_entry *default_entry = NULL;
 struct config_entry *rrcf_head     = NULL;
 
+/* where the  MAILQv2  authentication dataset file is ? */
+char * mq2authfile = NULL;
+
+
 static struct rckeyword {
 	const char	*name;
 	int		(*parsef)();
@@ -66,6 +80,7 @@ static struct rckeyword {
 {	"maxring",		rc_maxring	},	/* number */
 {	"maxrings",		rc_maxring	},	/* number */
 {	"maxta",		rc_maxta	},	/* number */
+{	"maxthr",		rc_maxthr	},	/* number */
 {	"maxtransport",		rc_maxta	},	/* number */
 {	"maxtransports",	rc_maxta	},	/* number */
 {	"user",			rc_user		},	/* number */
@@ -75,6 +90,10 @@ static struct rckeyword {
 {	"ageorder",		rc_ageorder	},	/* boolean */
 {	"queueonly",		rc_queueonly	},	/* boolean */
 {	"overfeed",		rc_overfeed	},	/* number */
+{	"priority",		rc_priority	},	/* number */
+{	"nice",			rc_nice		},	/* number */
+{	"syspriority",		rc_syspriority	},	/* number */
+{	"sysnice",		rc_sysnice	},	/* number */
 {	NULL,			0		}
 };
 
@@ -87,7 +106,9 @@ defaultconfigentry(ce,defaults)
 	if (defaults && ce != defaults) {
 	  /* Use the configurations script defaults.. */
 	  ce->next = NULL;
+#if 0
 	  ce->mark = 0;
+#endif
 
 	  ce->interval		= defaults->interval;
 	  ce->idlemax		= defaults->idlemax;
@@ -99,6 +120,7 @@ defaultconfigentry(ce,defaults)
 	  ce->flags		= defaults->flags;
 	  ce->maxkids		= defaults->maxkids;
 	  ce->maxkidChannel	= defaults->maxkidChannel;
+	  ce->maxkidThread	= defaults->maxkidThread;
 	  ce->maxkidThreads	= defaults->maxkidThreads;
 	  ce->argv		= defaults->argv;
 	  ce->nretries		= defaults->nretries;
@@ -106,6 +128,7 @@ defaultconfigentry(ce,defaults)
 	  ce->skew		= defaults->skew;
 	  ce->deliveryform	= defaults->deliveryform;
 	  ce->overfeed		= defaults->overfeed;
+	  ce->priority		= defaults->priority;
 	} else if (defaults == NULL) {
 	  /* Compile these defaults in.. Only for the "*" / "* / *" entry.. */
 	  ce->next	= NULL;
@@ -121,6 +144,7 @@ defaultconfigentry(ce,defaults)
 	  ce->flags	= 0;
 	  ce->maxkids	= -1;
 	  ce->maxkidChannel = -1;
+	  ce->maxkidThread  =  1;
 	  ce->maxkidThreads = -1;
 	  ce->argv	= NULL;
 	  ce->nretries	= 0;
@@ -128,6 +152,7 @@ defaultconfigentry(ce,defaults)
 	  ce->skew	= 5;
 	  ce->deliveryform = NULL;
 	  ce->overfeed	= 0;
+	  ce->priority  = 0; /* nice(0) -- no change */
 	}
 }
 
@@ -139,47 +164,55 @@ vtxprint(vp)
 	struct config_entry *ce = &(vp->thgrp->ce);
 
 	if (vp->orig[L_CHANNEL] != NULL && vp->orig[L_HOST] != NULL)
-	  printf("%s/%s", vp->orig[L_CHANNEL]->name, vp->orig[L_HOST]->name);
+	  sfprintf(sfstdout, "%s/%s", vp->orig[L_CHANNEL]->name,
+		   vp->orig[L_HOST]->name);
 	else
-	  printf("%s/%s", ISS(ce->channel), ISS(ce->host));
-	printf(" 0x%lx\n",		(unsigned long)ce);
-	printf("\tinterval %d\n",	(int)ce->interval);
-	printf("\tidlemax %d\n",	ce->idlemax);
-	printf("\texpiry %d\n",		(int)ce->expiry);
-	printf("\texpiryform %s\n",	ISS(ce->expiryform));
-	printf("\tdeliveryform %s\n",	ISS(ce->deliveryform));
-	printf("\tuid %d\n",		ce->uid);
-	printf("\tgid %d\n",		ce->gid);
-	printf("\tcommand %s\n",	ISS(ce->command));
-	printf("\tflags:");
+	  sfprintf(sfstdout, "%s/%s", ISS(ce->channel), ISS(ce->host));
+	sfprintf(sfstdout," %p\n",		ce);
+	sfprintf(sfstdout,"\tinterval %d\n",	(int)ce->interval);
+	sfprintf(sfstdout,"\tidlemax %d\n",	ce->idlemax);
+	sfprintf(sfstdout,"\texpiry %d\n",		(int)ce->expiry);
+	sfprintf(sfstdout,"\texpiryform %s\n",	ISS(ce->expiryform));
+	sfprintf(sfstdout,"\tdeliveryform %s\n",	ISS(ce->deliveryform));
+	sfprintf(sfstdout,"\tuid %d\n",		ce->uid);
+	sfprintf(sfstdout,"\tgid %d\n",		ce->gid);
+	sfprintf(sfstdout,"\tcommand %s\n",	ISS(ce->command));
+	sfprintf(sfstdout,"\tflags:");
 	if (ce->flags == 0)
-	  printf(" (none)");
+	  sfprintf(sfstdout," (none)");
 	else {
-	  if (ce->flags & CFG_BYCHANNEL) printf(" BYCHANNEL");
-	  if (ce->flags & CFG_WITHHOST)  printf(" WITHHOST");
-	  if (ce->flags & CFG_AGEORDER)  printf(" AGEORDER");
+	  if (ce->flags & CFG_BYCHANNEL) sfprintf(sfstdout," BYCHANNEL");
+	  if (ce->flags & CFG_WITHHOST)  sfprintf(sfstdout," WITHHOST");
+	  if (ce->flags & CFG_AGEORDER)  sfprintf(sfstdout," AGEORDER");
 	}
-	printf("\n");
-	printf("\tmaxkids %d\n",	ce->maxkids);
-	printf("\tmaxkidChannel %d\n",	ce->maxkidChannel);
-	printf("\tmaxkidThreads %d\n",	ce->maxkidThreads);
-	printf("\toverfeed %d\n",	ce->overfeed);
+	sfprintf(sfstdout,"\n");
+	sfprintf(sfstdout,"\tmaxkids %d\n",	ce->maxkids);
+	sfprintf(sfstdout,"\tmaxkidChannel %d\n",	ce->maxkidChannel);
+	sfprintf(sfstdout,"\tmaxkidThread  %d\n",	ce->maxkidThread);
+	sfprintf(sfstdout,"\tmaxkidThreads %d\n",	ce->maxkidThreads);
+	sfprintf(sfstdout,"\toverfeed %d\n",	ce->overfeed);
+
+	if (ce->priority >= 80)
+	  sfprintf(sfstdout,"\tpriority %d\n",	ce->priority - 100);
+	else
+	  sfprintf(sfstdout,"\tnice %d\n",		ce->priority);
+
 	if (ce->argv != NULL) {
 	  for (i = 0; ce->argv[i] != NULL; ++i)
-	    printf("\targv[%d] = %s\n", i, ce->argv[i]);
+	    sfprintf(sfstdout,"\targv[%d] = %s\n", i, ce->argv[i]);
 	}
-	printf("\tnretries %d\n", ce->nretries);
+	sfprintf(sfstdout,"\tnretries %d\n", ce->nretries);
 	if (ce->nretries > 0) {
-	  printf("\tretries = (");
+	  sfprintf(sfstdout,"\tretries = (");
 	  for (i = 0; i < ce->nretries ; ) {
-	    printf("%d", ce->retries[i]);
+	    sfprintf(sfstdout,"%d", ce->retries[i]);
 	    ++i;
 	    if (i < ce->nretries)
-	      printf(" ");
+	      sfprintf(sfstdout," ");
 	  }
-	  printf(")\n");
+	  sfprintf(sfstdout,")\n");
 	}
-	printf("\tskew %d\n", ce->skew);
+	sfprintf(sfstdout,"\tskew %d\n", ce->skew);
 }
 
 static void
@@ -209,11 +242,13 @@ celink(ce, headp, tailp)
 	    ce->flags		= (*tailp)->flags;
 	    ce->maxkids		= (*tailp)->maxkids;
 	    ce->maxkidChannel	= (*tailp)->maxkidChannel;
+	    ce->maxkidThread	= (*tailp)->maxkidThread;
 	    ce->maxkidThreads	= (*tailp)->maxkidThreads;
 	    ce->argv		= (*tailp)->argv;
 	    ce->nretries	= (*tailp)->nretries;
 	    ce->retries		= (*tailp)->retries;
 	    ce->overfeed	= (*tailp)->overfeed;
+	    ce->priority	= (*tailp)->priority;
 	  }
 	}
 }
@@ -227,22 +262,31 @@ readconfig(file)
 	struct config_entry *ce, *head, *tail;
 	struct rckeyword *rckp;
 	struct vertex v;
-	FILE *fp;
+	Sfio_t *fp;
 	int linenum = 0;
 
 	ce = head = tail = NULL;
 	errflag = 0;
 
-	if ((fp = fopen(file, "r")) == NULL) {
-	  fprintf(stderr, "%s: %s: %s\n",
-		  progname, file, strerror(errno));
+	if ((fp = sfopen(NULL, file, "r")) == NULL) {
+	  sfprintf(sfstderr, "%s: %s: %s\n",
+		   progname, file, strerror(errno));
 	  return NULL;
 	}
 	while ((n = readtoken(fp, line, sizeof line, &linenum)) != -1) {
 	  if (verbose)
-	    printf("read '%s' %d\n",  line, n);
+	    sfprintf(sfstdout, "read '%s' %d\n",  line, n);
 	  if (n == 1) {
-	    /* Selector entry */
+	    /* Selector entry - or "PARAM" */
+	    if (cistrncmp(line,"PARAM",5) == 0) {
+	      if (paramparse(line+5)) {
+		sfprintf(sfstderr, "%s: illegal syntax at %s:%d\n",
+			progname, file, linenum);
+		++errflag;
+	      }
+	      continue;;
+	    }
+
 	    if (ce != NULL)
 	      celink(ce, &head, &tail);
 	    ce = (struct config_entry *)emalloc(sizeof (struct config_entry));
@@ -290,26 +334,30 @@ readconfig(file)
 		  *cp = '\0';
 	      }
 	    }
-	    ce->mark = 0;
+
+	    if (ce)
+	      ce->mark = 0;
+
 	    for (rckp = &rckeys[0]; rckp->name != NULL ; ++rckp)
 	      if (cistrcmp(rckp->name, line) == 0) {
 		errflag += (*rckp->parsef)(line, a, ce);
 		break;
 	      }
 	    if (rckp->name == NULL) {
-	      fprintf(stderr,
+	      sfprintf(sfstderr,
 		      "%s: unknown keyword %s in %s:%d\n",
 		      progname, line, file, linenum);
 	      ++errflag;
 	    }
 	  } else {
-	    fprintf(stderr, "%s: illegal syntax\n", progname);
+	    sfprintf(sfstderr, "%s: illegal syntax at %s:%d\n",
+		    progname, file, linenum);
 	    ++errflag;
 	  }
 	}
 	if (ce != NULL)
 	  celink(ce, &head, &tail);
-	fclose(fp);
+	sfclose(fp);
 	if (verbose) {
 	  struct threadgroup tg;
 	  v.orig[L_CHANNEL] = v.orig[L_HOST] = NULL;
@@ -324,7 +372,7 @@ readconfig(file)
 
 static int
 readtoken(fp, buf, buflen, linenump)
-	FILE *fp;
+	Sfio_t *fp;
 	char *buf;
 	int buflen, *linenump;
 {
@@ -335,7 +383,7 @@ readtoken(fp, buf, buflen, linenump)
 
 redo_readtoken:
 	if (lp == NULL) {
-	  if (fgets(line, sizeof line, fp) == NULL)
+	  if (cfgets(line, sizeof line, fp) < 0)
 	    return -1;
 	  *linenump += 1;
 	  lp = line;
@@ -369,8 +417,8 @@ redo_readtoken:
 	    ++elp;
 	    while (*elp != '"' && *elp != '\0') {
 	      if (*elp == '\\' && *(elp+1) == '\n') {
-		if (fgets(elp, sizeof line - (elp - line), fp) == NULL) {
-		  fprintf(stderr,
+		if (cfgets(elp, sizeof line - (elp - line), fp) < 0) {
+		  sfprintf(sfstderr,
 			  "%s: bad continuation line\n",
 			  progname);
 		  return -1;
@@ -379,7 +427,7 @@ redo_readtoken:
 	      ++elp;
 	    }
 	    if (*elp == '\0') {
-	      fprintf(stderr,
+	      sfprintf(sfstderr,
 		      "%s: missing end-quote in: %s\n",
 		      progname, line);
 	      return -1;
@@ -439,7 +487,7 @@ rereadconfig(head, file)
 {
 	struct config_entry *ce, *nce, *head2;
 
-	fprintf(stderr,
+	sfprintf(sfstderr,
 		"%s: reread configuration file: %s\n", progname, file);
 	/* free all the old config file entries */
 	for (ce = head; ce != NULL; ce = nce) {
@@ -496,19 +544,7 @@ static int rc_command(key, arg, ce)
 	char *cp, **av, *argv[100];
 	int j;
 
-	if (*arg != '/') {
-
-	  if (mailbin == NULL)
-	    mailbin = getzenv("MAILBIN");
-	  if (mailbin == NULL)
-	    mailbin = MAILBIN;
-
-	  ce->command = emalloc(strlen(mailbin)+1+strlen(qdefaultdir)
-				+1+strlen(arg)+1);
-	  sprintf(ce->command,
-		  "%s/%s/%s", mailbin, qdefaultdir, arg);
-	} else
-	  ce->command = strsave(arg);
+	ce->command = strsave(arg);
 	j = 0;
 	for (cp = ce->command; *cp != '\0' && isascii(*cp);) {
 	  argv[j++] = cp;
@@ -538,6 +574,10 @@ static int rc_command(key, arg, ce)
 	  for (av = &ce->argv[0]; *av != NULL; ++av)
 	    if (strcmp(*av, replchannel) == 0) {
 	      ce->flags |= CFG_BYCHANNEL;
+	      if (ce->channel && strtok(ce->channel,"[{*?") != NULL)
+		ce->flags |= CFG_WITHHOST; /* Well, sort of..
+					      Channel-part is wild-card, thus
+					      it is also very restrictive.. */
 	      break;
 	    }
 	}
@@ -569,7 +609,7 @@ static int rc_group(key, arg, ce)
 	if (isascii(*arg) && isdigit(*arg))
 	  ce->gid = atoi(arg);
 	else if ((gr = getgrnam(arg)) == NULL) {
-	  fprintf(stderr, "%s: unknown group: %s\n", progname, arg);
+	  sfprintf(sfstderr, "%s: unknown group: %s\n", progname, arg);
 	  return 1;
 	} else
 	  ce->gid = gr->gr_gid;
@@ -602,6 +642,62 @@ static int rc_overfeed(key, arg, ce)
 	return 0;
 }
 
+static int rc_priority(key, arg, ce)
+	char *key, *arg;
+	struct config_entry *ce;
+{
+	if (sscanf(arg,"%d",&ce->priority) != 1 ||
+	    ce->priority < -20 || ce->priority > 19) {
+	  sfprintf(sfstderr, "%s: Bad UNIX priority value, acceptable in range: -20..19; input=\"%s\"\n", progname, arg);
+	  return 1;
+	}
+	ce->priority += 100;
+	return 0;
+}
+
+static int rc_nice(key, arg, ce)
+	char *key, *arg;
+	struct config_entry *ce;
+{
+	if (sscanf(arg,"%d",&ce->priority) != 1 ||
+	    ce->priority < -40 || ce->priority > 39) {
+	  sfprintf(sfstderr, "%s: Bad UNIX nice offset value, acceptable in range: -40..39; input=\"%s\"\n", progname, arg);
+	  return 1;
+	}
+	return 0;
+}
+
+static int rc_syspriority(key, arg, ce)
+	char *key, *arg;
+	struct config_entry *ce;
+{
+	int i;
+	if (sscanf(arg,"%d",&i) != 1 ||
+	    i < -20 || i > 19) {
+	  sfprintf(sfstderr, "%s: Bad UNIX priority value, acceptable in range: -20..19; input=\"%s\"\n", progname, arg);
+	  return 1;
+	}
+#ifdef HAVE_SETPRIORITY
+	/* PRIO_PROCESS depends likely of  HAVE_SYS_RESOURCE_H */
+	setpriority(PRIO_PROCESS, 0, i);
+#endif
+	return 0;
+}
+
+static int rc_sysnice(key, arg, ce)
+	char *key, *arg;
+	struct config_entry *ce;
+{
+	int i;
+	if (sscanf(arg,"%d",&i) != 1 ||
+	    i < -40 || i > 39) {
+	  sfprintf(sfstderr, "%s: Bad UNIX nice offset value, acceptable in range: -40..39; input=\"%s\"\n", progname, arg);
+	  return 1;
+	}
+	nice(i);
+	return 0;
+}
+
 static int rc_maxchannel(key, arg, ce)
 	char *key, *arg;
 	struct config_entry *ce;
@@ -634,6 +730,16 @@ static int rc_maxta(key, arg, ce)
 	return 0;
 }
 
+static int rc_maxthr(key, arg, ce)
+	char *key, *arg;
+	struct config_entry *ce;
+{
+	ce->maxkidThread = atoi(arg);
+	if (ce->maxkidThread <= 0)
+	  ce->maxkidThread = 1;
+	return 0;
+}
+
 static int rc_retries(key, arg, ce)
 	char *key, *arg;
 	struct config_entry *ce;
@@ -656,7 +762,7 @@ static int rc_retries(key, arg, ce)
 	  if (i > 0)
 	    arr[j++] = i;
 	  else {
-	    fprintf(stderr,
+	    sfprintf(sfstderr,
 		    "%s: not a numeric factor: %s\n",
 		    progname, d);
 	    return 1;
@@ -672,7 +778,7 @@ static int rc_retries(key, arg, ce)
 	  memcpy((char *)ce->retries, (char *)&arr[0], sizeof (int) * j);
 	  ce->nretries = j;
 	} else {
-	  fprintf(stderr, "%s: empty retry factor list\n", progname);
+	  sfprintf(sfstderr, "%s: empty retry factor list\n", progname);
 	  return 1;
 	}
 	return 0;
@@ -687,7 +793,7 @@ static int rc_user(key, arg, ce)
 	if (isascii(*arg) && isdigit(*arg))
 	  ce->uid = atoi(arg);
 	else if ((pw = getpwnam(arg)) == NULL) {
-	  fprintf(stderr, "%s: unknown user: %s\n", progname, arg);
+	  sfprintf(sfstderr, "%s: unknown user: %s\n", progname, arg);
 	  return 1;
 	} else
 	  ce->uid = pw->pw_uid;
@@ -701,7 +807,7 @@ static int rc_skew(key, arg, ce)
 	int v;
 
 	if (!isascii(*arg) || !isdigit(*arg) || (v = atoi(arg)) < 1) {
-	  fprintf(stderr, "%s: bad skew value: %s\n", progname, arg);
+	  sfprintf(sfstderr, "%s: bad skew value: %s\n", progname, arg);
 	  return 1;
 	}
 	ce->skew = v;
@@ -739,3 +845,50 @@ static int rc_queueonly(key, arg, ce)
 	ce->flags |= CFG_QUEUEONLY;
 	return 0;
 }
+
+extern int mailqmode;
+
+static int paramparse(line)
+	char *line;
+{
+	char *s, *a = NULL;
+
+	if ((s = strchr(line, '=')) != NULL) {
+	  char *p = s-1;
+	  *s = '\0';
+	  while (p >= line && (*p == ' ' || *p == '\t'))
+	    *p-- = '\0';
+	  a = s+1;
+	  while (*a == ' ' || *a == '\t') ++a;
+	  if (*a == '"') {
+	    ++a;
+	    s = a;
+	    while (*s && *s != '"') {
+	      if (*s == '\\' && s[1] != 0)
+		++s;
+	      ++s;
+	    }
+	    if (*s)
+	      *s = '\0';
+	  }
+	}
+
+	if (cistrcmp(line,"authfile")==0 && a) {
+	  if (mq2authfile)
+	    free(mq2authfile);
+	  mq2authfile = strsave(a);
+
+	  if (mq2authfile && access(mq2authfile,R_OK)==0)
+	    mailqmode = 2;
+
+	  return 0;
+	}
+
+	if (cistrcmp(line,"mailqsock")==0 && a) {
+	  if (mailqsock)
+	    free(mailqsock);
+	  mailqsock = strsave(a);
+	  return 0;
+	}
+	return 1;
+}
