@@ -12,6 +12,7 @@
  */
 
 #include <stdio.h>
+#include <sfio.h>
 #include <sys/param.h>
 #include "hostenv.h"
 #include <ctype.h>
@@ -86,7 +87,7 @@ const char * pidfile = PID_SCHEDULER;
 const char * mailshare;
 const char * logfn;
 const char * statusfn;
-FILE    * statuslog = NULL;
+Sfio_t     * statuslog = NULL;
 int	slow_shutdown = 0;
 extern int readsockcnt; /* from transport.c: mux() */
  
@@ -118,9 +119,10 @@ char * procselect = NULL;	/* Non-null defines  channel/host specifier
 				   prevent running anything else, and also
 				   prevent running error processing, or
 				   job-specifier deletions. */
-char * procselhost = NULL;	/* Just spliced out 'host'-part of the above */
+char *  procselhost = NULL;	/* Just spliced out 'host'-part of the above */
 extern int forkrate_limit;	/* How many forks per second ? */
 int	mailqmode = 1;		/* ZMailer v1.0 mode on mailq */
+char *  mailqsock = NULL;
 
 static int vtxprep_skip      = 0;
 static int vtxprep_skip_any  = 0;
@@ -128,7 +130,7 @@ static int vtxprep_skip_lock = 0;
 static time_t next_dirscan     = 0;
 static time_t next_idlecleanup = 0;
 static struct sptree *dirscan_mesh = NULL;
-static int newents_limit = 200;
+static int newents_limit = 400;
 extern int never_full_content; /* on conf.c */
 
 #include "memtypes.h"
@@ -174,36 +176,36 @@ int sig;
 	int flags;
 
 	if (logfn != NULL) {
-	  fflush(stdout);
-	  rewind(stdout);
-	  fflush(stderr);
-	  rewind(stderr);
-	  if (freopen(logfn, "a", stdout) != stdout
-	      || dup2(FILENO(stdout), FILENO(stderr)) < 0) {	/* sigh */
+	  sfsync(sfstdout);
+	  sfseek(sfstdout, 0, 0);
+	  sfsync(sfstderr);
+	  sfseek(sfstderr, 0, 0);
+	  if (sfopen(sfstdout, logfn, "a") != sfstdout
+	      || dup2(sffileno(sfstdout), sffileno(sfstderr)) < 0) {	/* sigh */
 	    /* XX: stderr might be closed at this point... */
-	    fprintf(stderr, "%s: cannot open log: %s, errno=%d\n", progname, logfn, errno);
+	    sfprintf(sfstderr, "%s: cannot open log: %s, errno=%d\n", progname, logfn, errno);
 	    return -1;
 	  }
 #if	defined(F_SETFL) && defined(O_APPEND)
-	  flags = fcntl(FILENO(stdout), F_GETFL, 0);
+	  flags = fcntl(sffileno(sfstdout), F_GETFL, 0);
 	  flags |= O_APPEND;
-	  fcntl(FILENO(stdout), F_SETFL, flags);
+	  fcntl(sffileno(sfstdout), F_SETFL, flags);
 #endif	/* F_SETFL */
 #if defined(F_SETFD)
-	  fcntl(FILENO(stdout), F_SETFD, 1); /* close-on-exec */
+	  fcntl(sffileno(sfstdout), F_SETFD, 1); /* close-on-exec */
 #endif
-	  setvbuf(stdout, (char *)NULL, _IOLBF, 0);
-	  setvbuf(stderr, (char *)NULL, _IOLBF, 0);
+	  sfset(sfstdout, SF_LINE, 1);
+	  sfset(sfstderr, SF_LINE, 1);
 	}
 	if (statusfn != NULL && statuslog != NULL) {
-	  fflush(statuslog);
-	  if (freopen(statusfn, "a", statuslog) != statuslog) {
-	    fprintf(stderr,"%s: cannot open statuslog: %s, errno=%d\n", progname, statusfn, errno);
+	  sfsync(statuslog);
+	  if (sfopen(statuslog, statusfn, "a") != statuslog) {
+	    sfprintf(sfstderr,"%s: cannot open statuslog: %s, errno=%d\n", progname, statusfn, errno);
 	    return -1;
 	  }
-	  setvbuf(statuslog, (char *)NULL, _IOLBF, 0);
+	  sfset(statuslog, SF_WHOLE, 1);
 #if defined(F_SETFD)
-	  fcntl(FILENO(statuslog), F_SETFD, 1); /* close-on-exec */
+	  fcntl(sffileno(statuslog), F_SETFD, 1); /* close-on-exec */
 #endif
 	}
 	SIGNAL_HANDLE(SIGHUP, (RETSIGTYPE(*)__((int))) loginitsched);
@@ -233,19 +235,26 @@ static int syncweb __((struct dirqueue *dq));
 int global_maxkids = 1000;
 time_t now;
 
-FILE * vfp_open __((struct ctlfile *));
-FILE * vfp_open(cfp)
+Sfio_t * vfp_open(cfp)
 struct ctlfile *cfp;
 {
-	FILE *vfp;
+	Sfio_t *vfp;
+	int fd;
+
+	if (!cfp->vfpfn) return NULL;
+
+	/* Open the vfp *ONLY* if the logging file exists,
+	   and can be written to.  If the file does not
+	   exist, no logging shall happen! */
 
 	setreuid(0, cfp->uid);
-	vfp = fopen(cfp->vfpfn, "a+");
+	fd = open(cfp->vfpfn, O_WRONLY|O_APPEND, 0);
 	setreuid(0, 0);
+	if (fd < 0) return NULL; /* Can't open it! */
+	vfp = sfnew(NULL, NULL, 0, fd, SF_WRITE|SF_APPEND|SF_LINE);
 	if (!vfp) return NULL; /* Failure to open */
 
-	setvbuf(vfp, (char *)NULL, _IOLBF, 0);   /* Set LINEBUFFERING */
-	fseek(vfp, (off_t)0, SEEK_END);
+	sfseek(vfp, (Sfoff_t)0, SEEK_END);
 	return vfp;
 }
 
@@ -262,25 +271,33 @@ struct ctlfile *cfp;
 	/* Delete from memory */
 
 	if (cfp->vfpfn) {
-	  FILE *vfp = vfp_open(cfp);
+	  Sfio_t *vfp = vfp_open(cfp);
 	  if (vfp) {
-	    fprintf(vfp,
-		    "ordered deletion of task file from scheduler memory (%s)\n",
-		    cfp->mid);
-	    fclose(vfp);
+	    sfprintf(vfp,
+		     "ordered deletion of task file from scheduler memory (%s)\n",
+		     cfp->mid);
+	    sfclose(vfp);
 	  }
 	}
 
-	if (cfp->head != NULL) {
-	  for (vp = cfp->head; vp != NULL; vp = nvp) {
-	    nvp = vp->next[L_CTLFILE];
-	    MIBMtaEntry->mtaStoredRecipients -= vp->ngroup;
-	    vp->ngroup = 0;
-	    unvertex(vp,-1,1); /* Don't unlink()! Just free()! */
-	  }
-	} else {
-	  free_cfp_memory(cfp);
+	/* Throw all way, if no vertices.. */
+	if (!cfp->head) {
+	  /* This should *not* be happening.. */
+	  unctlfile(cfp, 1);
+	  return;
 	}
+
+	for (vp = cfp->head; vp != NULL; vp = nvp) {
+	  /* Stash the next pointer, because the CFP content
+	     may become trashed at the end of the run. */
+	  nvp = cfp->head->next[L_CTLFILE];
+
+	  MIBMtaEntry->mtaStoredRecipients     -= vp->ngroup;
+	  MIBMtaEntry->mtaReceivedRecipientsSc -= vp->ngroup;
+	  vp->ngroup = 0;
+	  unvertex(vp,1,1); /* Don't unlink()! Just free()! */
+	}
+	/* CFP freeup happens at last vertex's unvertex() */
 }
 
 static void cfp_free(cfp, spl)
@@ -289,7 +306,7 @@ struct spblk   *spl;
 {
 	/* Delete from the  spt_mesh[]  */
 
-	if (spl == NULL)
+	if (spl == NULL && cfp->id)
 	  spl = sp_lookup((u_long)(cfp->id), spt_mesh[L_CTLFILE]);
 	if (spl != NULL)
 	  sp_delete(spl, spt_mesh[L_CTLFILE]);
@@ -327,7 +344,7 @@ struct ctlfile *cfp;
 }
 
 
-struct dirqueue dirqb;
+struct dirqueue dirqb = { 0, };
 struct dirqueue *dirq = &dirqb;
 
 const char **ArgvSave;
@@ -337,6 +354,110 @@ extern int main __((int, const char **));
 static struct config_entry *cehead = NULL;
 
 
+/* #define GLIBC_MALLOC_DEBUG__ */
+#ifdef GLIBC_MALLOC_DEBUG__ /* memory allocation debugging with GLIBC */
+
+#include <malloc.h> /* GLIBC malloc.h ! */
+
+/* Global variables used to hold underlaying hook values.  */
+static void *(*old_malloc_hook) (size_t, const void * );
+static void *(*old_realloc_hook) (void *, size_t, const void *);
+static void (*old_free_hook) (void*, const void *);
+static void *(*old_memalign_hook) (size_t, size_t, const void *);
+
+/* Prototypes for our hooks.  */
+static void *my_malloc_hook  (size_t, const void*);
+static void *my_realloc_hook (void *,size_t, const void*);
+static void  my_free_hook    (void*, const void*);
+static void *my_memalign_hook  (size_t, size_t, const void*);
+     
+static void *
+my_malloc_hook (size_t size, const void *CALLER)
+{
+  void *result;
+  /* Restore all old hooks */
+  __malloc_hook = old_malloc_hook;
+  __free_hook   = old_free_hook;
+  /* Call recursively */
+  result = malloc (size);
+  /* Save underlaying hooks */
+  old_malloc_hook = __malloc_hook;
+  old_free_hook   = __free_hook;
+  /* `printf' might call `malloc', so protect it too. */
+  sfprintf(sfstderr,"# malloc (%u) returns %p @%p\n",
+	   (unsigned int) size, result, CALLER);
+  /* Restore our own hooks */
+  __malloc_hook = my_malloc_hook;
+  __free_hook = my_free_hook;
+  return result;
+}
+
+static void *
+my_realloc_hook (void *ptr, size_t size, const void *CALLER)
+{
+  void *result;
+  /* Restore all old hooks */
+  __realloc_hook = old_realloc_hook;
+  __malloc_hook = old_malloc_hook;
+  __free_hook = old_free_hook;
+  /* Call recursively */
+  result = realloc (ptr, size);
+  /* Save underlaying hooks */
+  old_realloc_hook = __realloc_hook;
+  old_malloc_hook  = __malloc_hook;
+  old_free_hook    = __free_hook;
+  /* `printf' might call `malloc', so protect it too. */
+  sfprintf(sfstderr,"# realloc (%p,%u) returns %p @%p\n", ptr, (unsigned int) size, result, CALLER);
+  /* Restore our own hooks */
+  __realloc_hook = my_realloc_hook;
+  __malloc_hook  = my_malloc_hook;
+  __free_hook    = my_free_hook;
+  return result;
+}
+
+static void *
+my_memalign_hook (size_t align, size_t size, const void *CALLER)
+{
+  void *result;
+  /* Restore all old hooks */
+  __memalign_hook = old_memalign_hook;
+  __malloc_hook = old_malloc_hook;
+  __free_hook = old_free_hook;
+  /* Call recursively */
+  result = memalign (align, size);
+  /* Save underlaying hooks */
+  old_memalign_hook = __memalign_hook;
+  old_malloc_hook  = __malloc_hook;
+  old_free_hook    = __free_hook;
+  /* `printf' might call `malloc', so protect it too. */
+  sfprintf(sfstderr,"# memalign (%u,%u) returns %p @%p\n",
+	   (unsigned)align, (unsigned)size, result, CALLER);
+  /* Restore our own hooks */
+  __memalign_hook = my_memalign_hook;
+  __malloc_hook  = my_malloc_hook;
+  __free_hook    = my_free_hook;
+  return result;
+}
+     
+static void
+my_free_hook (void *ptr, const void *CALLER)
+{
+  /* Restore all old hooks */
+  __malloc_hook = old_malloc_hook;
+  __free_hook = old_free_hook;
+  /* Call recursively */
+  free (ptr);
+  /* Save underlaying hooks */
+  old_malloc_hook = __malloc_hook;
+  old_free_hook = __free_hook;
+  /* `printf' might call `free', so protect it too. */
+  sfprintf(sfstderr,"# freed pointer %p @%p\n", ptr, CALLER);
+  /* Restore our own hooks */
+  __malloc_hook = my_malloc_hook;
+  __free_hook = my_free_hook;
+}
+#endif
+
 int
 main(argc, argv)
 	int argc;
@@ -347,7 +468,20 @@ main(argc, argv)
 	int i, daemonflg, c, errflg, version, fd;
 	long offout, offerr;
 
-	char *t, *syslogflg = getzenv("SYSLOGFLG");
+	char *t, *syslogflg;
+
+#ifdef GLIBC_MALLOC_DEBUG__ /* memory allocation debugging with GLIBC */
+	old_malloc_hook = __malloc_hook;
+	__malloc_hook = my_malloc_hook;
+	old_memalign_hook = __memalign_hook;
+	__memalign_hook = my_memalign_hook;
+	old_realloc_hook = __realloc_hook;
+	__realloc_hook = my_realloc_hook;
+	old_free_hook = __free_hook;
+	__free_hook = my_free_hook;
+#endif
+
+	syslogflg = getzenv("SYSLOGFLG");
 	if (syslogflg == NULL)
 	  syslogflg = "";
 	t = syslogflg;
@@ -410,14 +544,14 @@ main(argc, argv)
 			break;
 		case 'l':
 			statusfn = optarg;
-			statuslog = fopen(statusfn,"a");
+			statuslog = sfopen(NULL, statusfn, "a");
 			if (!statuslog) {
 			  perror("Can't open statistics log file (-l)");
 			  exit(1);
 			}
-			setvbuf(statuslog, (char *)NULL, _IOLBF, 0);
+			sfset(statuslog, SF_LINE, 1);
 #if defined(F_SETFD)
-			fcntl(FILENO(statuslog), F_SETFD, 1);
+			fcntl(sffileno(statuslog), F_SETFD, 1);
 			/* close-on-exec */
 #endif
 			break;
@@ -431,7 +565,7 @@ main(argc, argv)
 		case 'M':
 			mailqmode = atoi(optarg);
 			if (mailqmode < 1 || mailqmode > 2) {
-			  fprintf(stderr,"scheduler: -M parameter is either 1, or 2\n");
+			  sfprintf(sfstderr,"scheduler: -M parameter is either 1, or 2\n");
 			  exit(EX_USAGE);
 			}
 			break;
@@ -448,7 +582,7 @@ main(argc, argv)
 			if (procselhost)
 			  *procselhost++ = 0;
 			else {
-			  fprintf(stderr,"scheduler: -p parameter is of form: channel/host\n");
+			  sfprintf(sfstderr,"scheduler: -p parameter is of form: channel/host\n");
 			  exit(64);
 			}
 			break;
@@ -495,9 +629,9 @@ main(argc, argv)
 	}
 
 	if (errflg) {
-	  fprintf(stderr,
-		  "Usage: %s [-dHisvV -M (1|2) -f configfile -L logfile -P postoffice -Q rendezvous]\n",
-		  progname);
+	  sfprintf(sfstderr,
+		   "Usage: %s [-dHisvV -M (1|2) -f configfile -L logfile -P postoffice -Q rendezvous]\n",
+		   progname);
 	  exit(128+errflg);
 	}
 	mailshare = getzenv("MAILSHARE");
@@ -518,8 +652,7 @@ main(argc, argv)
 	  SIGNAL_HANDLE(SIGHUP, (RETSIGTYPE(*)__((int))) loginitsched);
 	} else {
 	  SIGNAL_IGNORE(SIGHUP); /* no surprises please */
-	  setvbuf(stdout, (char *)NULL, _IOLBF, 0);
-	  setbuf(stderr,NULL);	/* [mea] No buffering on stdout.. */
+	  sfset(sfstdout, SF_LINE, 1);
 	}
 #ifdef USE_SIGREAPER
 # ifdef SIGCLD
@@ -542,10 +675,10 @@ main(argc, argv)
 	  prversion("scheduler");
 	  if (version)
 	    exit(0);
-	  putc('\n', stderr);
+	  sfputc(sfstderr, '\n');
 	}
-	offout = ftell(stdout);
-	offerr = ftell(stderr);
+	offout = sftell(sfstdout);
+	offerr = sftell(sfstderr);
 	if (config == NULL) {
 	  config = emalloc(3 + (u_int)(strlen(mailshare)
 				       + strlen(progname)
@@ -564,30 +697,30 @@ main(argc, argv)
 	  postoffice = POSTOFFICE;
 
 	if (chdir(postoffice) < 0 || chdir(TRANSPORTDIR) < 0)
-	  fprintf(stderr, "%s: cannot chdir to %s/%s.\n",
-		  progname, postoffice, TRANSPORTDIR);
+	  sfprintf(sfstderr, "%s: cannot chdir to %s/%s.\n",
+		   progname, postoffice, TRANSPORTDIR);
 
 	if (rendezvous == NULL && (rendezvous = getzenv("RENDEZVOUS")) == NULL)
 	  rendezvous = qoutputfile;
 	if (daemonflg) {
 	  /* X: check if another daemon is running already */
 	  if (!verbose
-	      && (offout < ftell(stdout) || offerr < ftell(stderr))) {
-	    fprintf(stderr, "%ld %ld %ld %ld\n", offout, ftell(stdout),
-		    offerr, ftell(stderr));
-	    fprintf(stderr, "%s: daemon not started.\n", progname);
+	      && (offout < sftell(sfstdout) || offerr < sftell(sfstderr))) {
+	    sfprintf(sfstderr, "%ld %ld %ld %ld\n", offout, sftell(sfstdout),
+		     offerr, sftell(sfstderr));
+	    sfprintf(sfstderr, "%s: daemon not started.\n", progname);
 	    die(1, "too many scheduler daemons");
 	    /* NOTREACHED */
 	  }
 	  detach();		/* leave worldy matters behind */
 	  mytime(&now);
-	  printf("%s: scheduler daemon (%s)\n\tpid %d started at %s\n",
-		 progname, Version, (int)getpid(), (char *)rfc822date(&now));
+	  sfprintf(sfstdout, "%s: scheduler daemon (%s)\n\tpid %d started at %s\n",
+		   progname, Version, (int)getpid(), (char *)rfc822date(&now));
 	}
 	/* Actually we want this to act as daemon,
 	   even when not in daemon mode.. */
 	if (killprevious(SIGTERM, pidfile) != 0) {
-	  printf("%s: Can't write my pid to a file ?? Out of diskspace ??\n",progname);
+	  sfprintf(sfstdout, "%s: Can't write my pid to a file ?? Out of diskspace ??\n",progname);
 	  die(1,"Can't write scheduler pid to a file!?");
 	  /* NOTREACHED */
 	}
@@ -607,8 +740,8 @@ main(argc, argv)
 	    cfp = schedule(fd, argv[optind], ino, 0);
 	    if (cfp == NULL) {
 	      if (verbose)
-		fprintf(stderr, "Nothing scheduled for %s!\n",
-			argv[optind]);
+		sfprintf(sfstderr, "Nothing scheduled for %s!\n",
+			 argv[optind]);
 	    } else
 	      eunlink(argv[optind]);
 	  }
@@ -632,7 +765,6 @@ main(argc, argv)
 	vtxprep_skip_lock = 0;
 	syncweb(dirq);
 
-	if (dlyverbose) verbose = dlyverbose;
 
 	canexit = 1;
 	SIGNAL_HANDLE(SIGTERM, sig_exit);	/* split */
@@ -652,12 +784,14 @@ main(argc, argv)
 	      break;
 	    queryipccheck();
 	  }
-	  fprintf(stderr,"Synchronous startup completed, messages: %d (%d skipped) recipients: %d\n",
-		  global_wrkcnt, startcount - global_wrkcnt, thread_count_recipients());
-	  fprintf(stderr,"***********************************************************************\n");
+	  sfprintf(sfstderr,"Synchronous startup completed, messages: %d (%d skipped) recipients: %d\n",
+		   global_wrkcnt, startcount - global_wrkcnt, thread_count_recipients());
+	  sfprintf(sfstderr,"***********************************************************************\n");
 	  syncstart = 0;
 	}
 
+	if (dlyverbose) verbose = dlyverbose;
+
 	do {
 	  time_t timeout;
 
@@ -686,10 +820,15 @@ main(argc, argv)
 	  }
 
 	  /* Submit one item from pre-scheduler-queue into the scheduler */
-	  if (dirq->wrksum > 100 ) {
-	    /* If more than 100 in queue, submit 20 in a burst.. */
-	    for (i=0; i < 20; ++i)
-	      syncweb(dirq);
+	  if (dirq->wrksum > 1 ) {
+	    /* If more things in queue, submit them up to 200 pieces, or
+	       until spent two seconds at it! */
+	    time_t now0 = now + 2;
+	    for (i=0; i < 200 && now > now0; ++i) {
+	      if (syncweb(dirq) < 0)
+		break; /* Out of queue! */
+	      mytime(&now);
+	    }
 	  }
 	  if (dirq->wrksum > 0)
 	    syncweb(dirq);
@@ -812,6 +951,8 @@ dq_insert(DQ, ino, file, delay)
 	struct dirstatname *dsn;
 	struct dirqueue *dq = DQ;
 
+	if (!ino) return 1; /* Well, actually it isn't, but we "makebelieve" */
+
 	mytime(&now);
 
 	if (dq == NULL)
@@ -826,13 +967,13 @@ dq_insert(DQ, ino, file, delay)
 
 	/* 
 	   if (DQ == NULL)
-	   fprintf(stderr,"dq_insert(NULL,ino=%ld, file='%s', delay=%d)\n",
+	   sfprintf(sfstderr,"dq_insert(NULL,ino=%ld, file='%s', delay=%d)\n",
 	   ino,file,delay);
 	 */
 
 	/* Is it already in the database ? */
 	if (sp_lookup((u_long)ino,dirscan_mesh) != NULL) {
-	  fprintf(stderr,"scheduler: tried to dq_insert(ino=%ld, file='%s') already in queue\n",ino, file);
+	  sfprintf(sfstderr,"scheduler: tried to dq_insert(ino=%ld, file='%s') already in queue\n",ino, file);
 	  return 1; /* It is! */
 	}
 
@@ -942,7 +1083,8 @@ static int dirqueuescan(dir, dq, subdirs)
 #endif
 
 	if (verbose && dir[0] == '.') {
-	  printf("dirqueuescan(dir='%s') ",dir); fflush(stdout);
+	  sfprintf(sfstdout, "dirqueuescan(dir='%s') ",dir);
+	  sfsync(sfstdout);
 	}
 
 	/* Some changes lately, open the dir and read it */
@@ -1019,7 +1161,7 @@ static int dirqueuescan(dir, dq, subdirs)
 	closedir(dirp);
 
 	if (verbose && dir[0] == '.')
-	  printf("wrksum=%d new=%d\n",dq->wrksum,newents);
+	  sfprintf(sfstdout,"wrksum=%d new=%d\n",dq->wrksum,newents);
 
 	return newents;
 }
@@ -1111,12 +1253,14 @@ static int syncweb(dq)
 }
 
 
-static int sync_cfps __((struct ctlfile *, struct ctlfile *));
-static int sync_cfps(oldcfp, newcfp)
+static int sync_cfps __((struct ctlfile *,struct ctlfile *,struct procinfo*));
+static int sync_cfps(oldcfp, newcfp, proc)
      struct ctlfile *oldcfp, *newcfp;
+     struct procinfo *proc;
 {
-	struct vertex *ovp, *nvp;
-	struct vertex *novp, *vp;
+	struct vertex *ovp,  *nvp;
+	struct vertex *novp, *nnvp;
+	int ovp_count = 0;
 
 	/* Scan both files thru their   vp->next[L_CTLFILE]  vertex chains.
 	   If oldcfp has things that newcfp does not have, remove those from
@@ -1132,41 +1276,136 @@ static int sync_cfps(oldcfp, newcfp)
 	   skenario..
 	*/
 
+	if (verbose) {
+	  sfprintf(sfstdout,"sync_cfps() OLDVTXES =");
+	  for (ovp = oldcfp->head; ovp; ovp = ovp->next[L_CTLFILE])
+	    sfprintf(sfstdout," %p{%p %p}", ovp, ovp->orig[L_CHANNEL], ovp->orig[L_HOST]);
+	  sfprintf(sfstdout,"\n");
+	  sfprintf(sfstdout,"sync_cfps() NEWVTXES =");
+	  for (nvp = newcfp->head; nvp; nvp = nvp->next[L_CTLFILE])
+	    sfprintf(sfstdout," %p{%p %p}", nvp, nvp->orig[L_CHANNEL], nvp->orig[L_HOST]);
+	  sfprintf(sfstdout,"\n");
+	}
+
 	ovp = oldcfp->head;
 	nvp = newcfp->head;
 
-	while (ovp != NULL) {
+
+	while (ovp != NULL && nvp != NULL) {
+	  ++ovp_count;
+
 	  /* Always prepare for removal of the ovp object..
 	     Pick the next-ovp pointer now */
-	  novp = ovp->next[L_CTLFILE];
+	  novp  = ovp->next[L_CTLFILE];
+	  nnvp  = nvp->next[L_CTLFILE];
 
 	  /* Does this exist also on NVP chain ? */
 
 #define VTXMATCH(ovp,nvp) (((ovp)->orig[L_CHANNEL] == (nvp)->orig[L_CHANNEL]) && ((ovp)->orig[L_HOST] == (nvp)->orig[L_HOST]))
 
+	  if (VTXMATCH(ovp, nvp)) {
+	    /* Verify/adjust OVP so that OVP and NVP have same
+	       address indexes in them. */
+	    int i, j, id;
+	    if (verbose) {
+	      sfprintf(sfstdout, " OLD ngroup=%d {", ovp->ngroup);
+	      for(i = 0; i < ovp->ngroup; ++i)
+		sfprintf(sfstdout, " %d", ovp->index[i]);
+	      sfprintf(sfstdout, " }\n");
+	      sfprintf(sfstdout, " NEW ngroup=%d {", nvp->ngroup);
+	      for(i = 0; i < nvp->ngroup; ++i)
+		sfprintf(sfstdout, " %d", nvp->index[i]);
+	      sfprintf(sfstdout, " }\n");
+	    }
+	    for (i = ovp->ngroup -1; i >= 0; --i) {
+	      id = ovp->index[i];
+	      for (j = nvp->ngroup -1; j >= 0; --j) {
+		if (nvp->index[j] == id)
+		  goto next_i;
+	      }
+	      /* ovp index elt not found in new set! */
+	      for (j = i+1; j < ovp->ngroup; ++j)
+		ovp->index[j-1] = ovp->index[j];
+	      ovp->ngroup -= 1;
+	      MIBMtaEntry->mtaStoredRecipients -= 1;
+	    next_i:;
+	    }
+	    if (ovp->ngroup <= 0)
+	      unvertex(ovp,-1,1); /* Don't unlink()! free() *just* ovp! */
+
+	    ovp = novp;
+	    nvp = nnvp;
+	    continue;
+	  }
+
 	  if (!VTXMATCH(ovp,nvp)) {
 
+	    struct vertex *vp;
+	    int i;
+
 	    /* Uugh... Does not match :-( */
 
 	    vp = ovp;
 
-	    while (vp && !VTXMATCH(vp,nvp)) {
+	    while (vp && !VTXMATCH(vp,nvp))
 	      vp = vp->next[L_CTLFILE];
-	    }
+
 	    if (vp == NULL) {
 	      /* New not in old at all ??? */
-	      return -1;
+	      if (verbose)
+		sfprintf(sfstdout," -- NEW vertex %p NOT in OLD set!\n", nvp);
+	      /* Ok, skip it, and retry with OLD vs. new NEW */
+	      nvp = nnvp;
+	      continue;
 	    }
-	    /* XX: All OVP instances before matching NVP
+
+	    /* All OVP instances before matching NVP
 	       are to be removed from OVP chains */
-	    /* XX: Adjust NOVP variable too */
-	  }
-	  if (VTXMATCH(ovp, nvp)) {
-	    /* XX: Verify that OVP and NVP have same amount
-	       of address indexes in them, adjust OVP if not. */
+	    i = 0;
+	    while (ovp && ovp != vp) {
+	      novp = ovp->next[L_CTLFILE];
+
+	      MIBMtaEntry->mtaStoredRecipients -= vp->ngroup;
+	      ovp->ngroup = 0;
+	      unvertex(ovp,-1,1); /* Don't unlink()! free() *just* ovp! */
+
+	      ++i;
+	      ++ovp_count;
+	      ovp = novp;
+	    }
+	    if (verbose)
+	      sfprintf(sfstdout," -- Deleted %d OLD vertices\n", i);
+	    /* Adjust NOVP variable too */
+	    novp = NULL;
+	    if (ovp)
+	      novp = ovp->next[L_CTLFILE];
 	  }
+
+	  ovp = novp;
+	  nvp = nnvp;
+	}
+
+	/* Ok,  'ovp' might be non-NULL while 'nvp' is already NULL */
+	while (ovp) {
+	  novp = ovp->next[L_CTLFILE];
+
+	  MIBMtaEntry->mtaStoredRecipients -= ovp->ngroup;
+	  ovp->ngroup = 0;
+	  unvertex(ovp,-1,1); /* Don't unlink()! free() *just* ovp! */
+	  /* Leaves CFP unharmed */
+
 	  ovp = novp;
 	}
+	
+
+	if (verbose)
+	  sfprintf(sfstdout," -- Synced %d OLD vertices\n", ovp_count);
+
+	oldcfp->rcpnts_failed = newcfp->rcpnts_failed;
+	oldcfp->haderror |= newcfp->haderror;
+
+	/* Now  oldcfp->head  might be NULL -> bad news.. */
+
 	return 0;
 }
 
@@ -1196,10 +1435,12 @@ void resync_file(proc, file)
 	/* Already in processing ? */
 	spl = sp_lookup((u_long)ino, spt_mesh[L_CTLFILE]);
 	if (spl == NULL) {
+#if 0
 	  if (!in_dirscanqueue(NULL,ino)) {
-	    printf("Resyncing file \"%s\" (ino=%ld)", file, ino);
-	    printf(" .. not in processing database\n");
+	    sfprintf(sfstdout,"Resyncing file \"%s\" (ino=%ld)", file, ino);
+	    sfprintf(sfstdout," .. not in processing database\n");
 	  }
+#endif
 	  /* Not (anymore) in processing! */
 	  return;
 	}
@@ -1211,10 +1452,20 @@ void resync_file(proc, file)
 	  sp_delete(spl, spt_mesh[L_CTLFILE]);
 	spl = NULL;
 
-	printf("Resyncing file \"%s\" (ino=%d) (of=%d ho='%s')",
-	       file, (int) ino, proc->overfed,
-	       (proc->ho ? proc->ho->name : "<NULL>"));
-	/* printf(" .. in processing db\n"); */
+	oldcfp->resynccount += 1;
+
+	sfprintf(sfstdout, "Resyncing file \"%s\" (ino=%d pid=%d of=%d ho='%s') reqcnt=%d\n",
+		 file, (int) ino, proc->pid, proc->overfed,
+		 (proc->ho ? proc->ho->name : "<NULL>"), oldcfp->resynccount);
+	/* sfprintf(sfstdout, " .. in processing db\n"); */
+
+	if (oldcfp->resynccount > MAXRESYNCS) {
+	  /* Sigh.. Throw everything away :-( */
+	  oldcfp->id = ino;
+	  cfp_free(oldcfp, NULL);
+	  sfprintf(sfstdout, " ... too many Resync attempts this way, throwing  it away...\n");
+	  return;
+	}
 
 	/* cfp_free()->unvertex()->unctlfile() will do reinsertion */
 	/* dq_insert(NULL,ino,file,31); */
@@ -1223,21 +1474,20 @@ void resync_file(proc, file)
 	fd = eopen(file, O_RDWR, 0);
 	if (fd < 0) {
 	  /* ???? */
-	  printf(" .. FILE OPEN FAILED!\n");
+	  sfprintf(sfstdout," .. FILE OPEN FAILED!\n");
 
 	  /* Delete it from memory */
 	  cfp_free(oldcfp, NULL);
 	  return;
 	}
 
-	newcfp = schedule(fd, file, ino, 1);
 
-#if 0 /* XX: not usable before  sync_cfps()  works! */
+	newcfp = schedule(fd, file, ino, 1); /* rereading */
 
 	if (newcfp != NULL) {
 	  /* ????  What ever, it succeeds, or it fails, all will be well */
 
-	  sync_cfps(oldcfp, newcfp);
+	  sync_cfps(oldcfp, newcfp, proc);
 	  newcfp->id = 0; /* Don't scramble spt_mesh[] below */
 
 	  spl = sp_lookup((u_long)ino, spt_mesh[L_CTLFILE]);
@@ -1246,30 +1496,26 @@ void resync_file(proc, file)
 	  oldcfp->id = ino;
 	  sp_install(oldcfp->id, (void *)oldcfp, 0, spt_mesh[L_CTLFILE]);
 
-	  printf(" .. resynced!\n");
+	  /* Delete it from memory */
+	  cfp_free0(newcfp);
+
+	  if (oldcfp->head == NULL) {
+	    cfp_free(oldcfp,spl);
+	    if (verbose)
+	      sfprintf(sfstdout," .. LOST in resync ??!\n");
+	  } else
+	    if (verbose)
+	      sfprintf(sfstdout," .. resynced!\n");
+
 	} else {
-	  printf(" .. NOT resynced!\n");
+
+	  if (verbose)
+	    sfprintf(sfstdout," .. NOT resynced!\n");
+
 	  /* Sigh.. Throw everything away :-( */
 	  oldcfp->id = ino;
 	  cfp_free(oldcfp, NULL);
 	}
-
-	/* Delete it from memory */
-	if (newcfp != NULL)
-	  cfp_free0(newcfp);
-
-#else
-
-	if (newcfp != NULL) {
-	  printf(" .. resynced!\n");
-	} else {
-	  printf(" .. NOT resynced!\n");
-	  /* Sigh.. Throw everything away :-( */
-	  oldcfp->id = ino;
-	}
-	
-	cfp_free0(oldcfp);
-#endif
 }
 
 
@@ -1291,11 +1537,12 @@ static struct ctlfile *schedule(fd, file, ino, reread)
 	struct vertex *vp;
 
 	/* read and process the control file */
-	if ((cfp = vtxprep(slurp(fd, ino), file, reread)) == NULL) {
+	cfp = vtxprep(slurp(fd, ino), file, reread);
+	if (cfp == NULL) {
 	  if (!vtxprep_skip) {	/* Unless skipped.. */
 	    eunlink(file);	/* everything here has been processed */
 	    if (verbose)
-	      printf("completed, unlink %s\n",file);
+	      sfprintf(sfstdout,"completed, unlink %s\n",file);
 	    return NULL;
 	  }
 	  vtxprep_skip_any += vtxprep_skip;
@@ -1304,26 +1551,28 @@ static struct ctlfile *schedule(fd, file, ino, reread)
 	if (cfp->head == NULL) {
 	  ++global_wrkcnt;
 	  ++MIBMtaEntry->mtaStoredMessages;
-	  MIBMtaEntry->mtaStoredRecipients     += cfp->rcpnts_work;
-	  MIBMtaEntry->mtaReceivedRecipientsSc += cfp->rcpnts_work;
 	  unctlfile(cfp, 0); /* Delete the file.
-				(decrements those counters too!) */
+				(decrements those counters above!) */
 	  return NULL;
 	}
-	for (vp = cfp->head; vp != NULL; vp = vp->next[L_CTLFILE]) {
-	  vtxdo(vp, cehead, file);
-	}
+
+	if (!reread)
+	  sp_install(cfp->id, (void *)cfp, 0, spt_mesh[L_CTLFILE]);
+	++MIBMtaEntry->mtaStoredMessages;
+	++global_wrkcnt;
+	
+	if (!reread)
+	  for (vp = cfp->head; vp != NULL; vp = vp->next[L_CTLFILE]) {
+	    /* Put into the schedules */
+	    vtxdo(vp, cehead, file);
+	  }
+
 	/* Now we have no more need for the contents in core */
 	if (cfp->contents != NULL) {
 	  free(cfp->contents);
 	  cfp->contents = NULL;
 	}
 
-	sp_install(cfp->id, (void *)cfp, 0, spt_mesh[L_CTLFILE]);
-	++MIBMtaEntry->mtaStoredMessages;
-	++global_wrkcnt;
-	MIBMtaEntry->mtaStoredRecipients     += cfp->rcpnts_work;
-	MIBMtaEntry->mtaReceivedRecipientsSc += cfp->rcpnts_work;
 	return cfp;
 }
 
@@ -1421,7 +1670,7 @@ slurp(fd, ino)
 		/* cfp->erroraddr = strsave(s+2); */
 	      } else if (*s == _CF_TURNME) {
 		/* Umm... it is a bit complex */
-		printf("A TURNME request for target '%s'\n",s+2);
+		sfprintf(sfstdout,"A TURNME request for target '%s'\n",s+2);
 	      }
 	    }
 	  }
@@ -1494,8 +1743,8 @@ lockverify(cfp,cp,verbflg)	/* Return 1 when lock process does not exist */
 	if (kill(lockpid,sig) != 0) return 1; /* PID does not exist, or
 					       other error.. */
 	if (verbflg)
-	  fprintf(stderr,"lockverify: Lock with PID=%d is active on %s:%s\n",
-		  lockpid, cfp->mid, cp+_CFTAG_RCPTPIDSIZE);
+	  sfprintf(sfstderr,"lockverify: Lock with PID=%d is active on %s:%s\n",
+		   lockpid, cfp->mid, cp+_CFTAG_RCPTPIDSIZE);
 	return 0;	/* Locking PID does exist.. */
 }
 
@@ -1530,6 +1779,7 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	int prevrcpt = -1;
 	int is_turnme = 0;
 	time_t wakeuptime;
+	long format = 0;
 
 	char fpath[128], path[128], path2[128];
 
@@ -1558,7 +1808,7 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	    /*
 	     * This can happen when we restart the scheduler, and
 	     * some previous transporter is still running.
-	     *
+	     * (and DEFINITELY during resync! when we simply ignore locks)
 	     */
 	    if (!lockverify(cfp, cp, !rereading)) {
 #if 0
@@ -1570,11 +1820,11 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	       */
 	      close(cfp->fd);	/* Was opened on  schedule() */
 	      if (cfp->vfpfn != NULL) {
-		FILE *vfp = vfp_open(cfp);
+		Sfio_t *vfp = vfp_open(cfp);
 		if (vfp) {
-		  fprintf(vfp,
-			  "New scheduler: Skipped a job-file because it is held locked by PID=%6.6s\n",cp+1);
-		  fclose(vfp);
+		  sfprintf(vfp,
+			   "scheduler: Skipped a job-file because it is held locked by PID=%6.6s\n",cp+1);
+		  sfclose(vfp);
 		}
 	      }
 	      cfp_free(cfp, NULL);
@@ -1610,17 +1860,35 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	    ++cfp->rcpnts_total;
 	    if (*cp == _CFTAG_NOTOK) {
 	      ++cfp->rcpnts_failed;
+	      ++MIBMtaEntry->mtaReceivedRecipientsSc,
+	      ++MIBMtaEntry->mtaTransmittedRecipientsSc;
 	      prevrcpt = -1;
 	    } else if (*cp != _CFTAG_OK) {
 	      ++cfp->rcpnts_work;
+	    } else { /* _CFTAG_OK */
+	      ++MIBMtaEntry->mtaReceivedRecipientsSc,
+	      ++MIBMtaEntry->mtaTransmittedRecipientsSc;
 	    }
 	  }
 	  if (*cp == _CFTAG_NORMAL ||
+	      (rereading && *cp == _CFTAG_LOCK) ||
 	      *cp == '\n' /* This appears for msg-header entries.. */ ) {
-	    --cp;
-	    switch (*cp) {
+	    ++cp;
+	    switch (cp[-2]) {
+	    case _CF_FORMAT:
+	      sscanf(cp,"%li",&format);
+	      if (format & (~_CF_FORMAT_KNOWN_SET)) {
+
+		sfprintf(sfstderr,"%s: ** FILE: '%s' has unknown/unsupported format set: 0x%08lx !\n",
+			 progname, file, format);
+
+		cfp_free(cfp, NULL);
+		free(offarr);
+		return NULL;
+	      }
+	      cfp->format = format;
+	      break;
 	    case _CF_SENDER:
-	      ++cp;
 	      while (*cp == ' ') ++cp;
 	      senderchannel = cp;
 	      while (*cp != 0 && *cp != ' ') ++cp; 
@@ -1646,14 +1914,15 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 						    offspc);
 	      }
 	      offarr[opcnt].offset = *lp + 2;
-	      cp += 2;
 	      strlower(cp);
-	      if (*cp == ' ' || (*cp >= '0' && *cp <= '9')) {
+	      if ((format & _CF_FORMAT_TA_PID) || *cp == ' ' ||
+		  (*cp >= '0' && *cp <= '9')) {
 		/* New PID locking scheme.. */
 		offarr[opcnt].offset += _CFTAG_RCPTPIDSIZE;
 		cp += _CFTAG_RCPTPIDSIZE;
 	      }
-	      if (*cp == ' ' || (*cp >= '0' && *cp <= '9')) {
+	      if ((format & _CF_FORMAT_DELAY1) || *cp == ' ' ||
+		  (*cp >= '0' && *cp <= '9')) {
 		/* Newer DELAY data slot - _CFTAG_RCPTDELAYSIZE bytes */
 		offarr[opcnt].delayslot = offarr[opcnt].offset;
 		offarr[opcnt].offset += _CFTAG_RCPTDELAYSIZE;
@@ -1668,12 +1937,16 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	      offarr[opcnt].notifyflg = NOT_FAILURE;
 	      prevrcpt = opcnt;
 	      ++opcnt;
+
+	      /* Account for all yet to be delivered recipients */
+	      ++MIBMtaEntry->mtaStoredRecipients;
+	      ++MIBMtaEntry->mtaReceivedRecipientsSc;
+
 	      break;
 	    case _CF_RCPTNOTARY:
 	      /* IETF-NOTARY-DRPT+NOTIFY DATA! */
 	      if (prevrcpt >= 0) {
 		offarr[prevrcpt].drptoffset = *lp + 2;
-		cp += 2;
 		/* Lets parse the input, we want NOTIFY= flags */
 		while (*cp != 0) {
 		  while (*cp == ' ' || *cp == '\t') ++cp;
@@ -1711,49 +1984,40 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	      break;
 	    case _CF_DSNRETMODE:
 	      if (cfp->dsnretmode) free(cfp->dsnretmode);
-	      cfp->dsnretmode = strsave(cp+2);
+	      cfp->dsnretmode = strsave(cp);
 	      break;
 	    case _CF_DSNENVID:
 	      if (cfp->envid) free(cfp->envid); /* shouldn't happen.. */
-	      cfp->envid = strsave(cp+2);
+	      cfp->envid = strsave(cp);
 	      break;
 	    case _CF_DIAGNOSTIC:
 	      cfp->haderror = 1;
 	      break;
 	    case _CF_MESSAGEID:
 	      if (cfp->mid != NULL) free(cfp->mid); /* shouldn't happen.. */
-	      cfp->mid = strsave(cp+2);
+	      cfp->mid = strsave(cp);
 	      break;
 	    case _CF_BODYOFFSET:
-	      cfp->msgbodyoffset = atoi(cp+2);
+	      cfp->msgbodyoffset = atoi(cp);
 	      break;
 	    case _CF_LOGIDENT:
 	      if (cfp->logident) free(cfp->logident); /* shouldn't happen..*/
-	      cfp->logident = strsave(cp+2);
+	      cfp->logident = strsave(cp);
 	      break;
 	    case _CF_ERRORADDR:
 	      if (cfp->erroraddr) free(cfp->erroraddr); /* could happen */
-	      cfp->erroraddr = strsave(cp+2);
+	      cfp->erroraddr = strsave(cp);
 	      break;
 	    case _CF_OBSOLETES:
-	      deletemsg(cp+2, cfp);
+	      deletemsg(cp, cfp);
 	      break;
 	    case _CF_VERBOSE:
-	      cfp->vfpfn = strsave(cp+2);
-
-	      {
-		FILE *vfp = vfp_open(cfp);
-		if (vfp != NULL && cfp->mid != NULL)
-		  fprintf(vfp,
-			  "scheduler processing %s\n",
-			  cfp->mid);
-		if (vfp) fclose(vfp);
-	      }
+	      cfp->vfpfn = strsave(cp);
 	      break;
 	    case _CF_TURNME:
-	      printf("TURNME: %s\n",cp+2);
-	      strlower(cp+2);
-	      turnme(cp+2);
+	      sfprintf(sfstdout,"TURNME: %s\n",cp);
+	      strlower(cp);
+	      turnme(cp);
 	      is_turnme = 1;
 	      break;
 	    }
@@ -1837,12 +2101,11 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	}
 	if (cfp->mid == NULL || cfp->logident == NULL ||
 	    estat(mfpath, &stbuf) < 0) {
-	  if (!is_turnme && cfp->vfpfn != NULL) {
-	    FILE *vfp = vfp_open(cfp);
+	  if (!is_turnme) {
+	    Sfio_t *vfp = vfp_open(cfp);
 	    if (vfp) {
-	      fprintf(vfp,
-		      "aborted due to missing information\n");
-	      fclose(vfp);
+	      sfprintf(vfp, "aborted due to missing information\n");
+	      sfclose(vfp);
 	    }
 	  }
 	  cfp_free(cfp, NULL);
@@ -1925,17 +2188,19 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 					  (i - svn - 1) * sizeof (int));
 	      vp = (struct vertex *)emalloc(alloc_size);
 	      memset((char*)vp, 0, alloc_size);
-	      vp->cfp          = cfp;
+	      vp->cfp             = cfp;
 	      vp->next[L_CTLFILE] = NULL;
 	      vp->prev[L_CTLFILE] = pvp;
 #if 0
-	      vp->message      = NULL;
-	      vp->retryindex   = 0;
-	      vp->nextitem     = NULL;
-	      vp->previtem     = NULL;
-	      vp->proc         = NULL;
-	      vp->attempts     = 0;
-	      vp->notary       = NULL;
+	      vp->thread          = NULL;
+	      vp->web[L_CTLFILE]  = NULL;
+	      vp->message         = NULL;
+	      vp->retryindex      = 0;
+	      vp->nextitem        = NULL;
+	      vp->previtem        = NULL;
+	      vp->proc            = NULL;
+	      vp->attempts        = 0;
+	      vp->notary          = NULL;
 #endif
 	      vp->ngroup       = i - svn;
 	      /* vp->sender       = strsave(offarr[svn].sender); */
@@ -2003,9 +2268,9 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	*l_echannel = *l_ehost = ' ';
 	/*
 	   for (vp = head; vp != NULL; vp = vp->next[L_CTLFILE]) {
-	     printf("--\n");
+	     sfprintf(sfstdout,"--\n");
 	     for (i = 0; i < vp->ngroup; ++i)
-	       printf("\t%s\n", cfp->contents+cfp->offset[vp->index[i]]);
+	       sfprintf(sfstdout,"\t%s\n", cfp->contents+cfp->offset[vp->index[i]]);
 	   }
 	*/
 	cfp->head = head;
@@ -2013,10 +2278,23 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	if (verbose) {
 	  int completed = cfp->rcpnts_total - cfp->rcpnts_work -
 			  cfp->rcpnts_failed;
-	  printf("vtxprep: msg %s rcptns total %d work %d failed %d done %d\n",
+	  sfprintf(sfstdout,"vtxprep: msg %s rcptns total %d work %d failed %d done %d\n  vertices:",
 		 cfp->mid, cfp->rcpnts_total, cfp->rcpnts_work,
 		 cfp->rcpnts_failed, completed );
+	  vp = cfp->head;
+	  for (vp = cfp->head; vp; vp = vp->next[L_CTLFILE]) {
+	    sfprintf(sfstdout," %p", vp);
+	  }
+	  sfprintf(sfstdout,"\n");
 	}
+
+	{
+	  Sfio_t *vfp = vfp_open(cfp);
+	  if (vfp != NULL && cfp->mid != NULL)
+	    sfprintf(vfp, "scheduler processing %s\n", cfp->mid);
+	  if (vfp) sfclose(vfp);
+	}
+
 	return cfp;
 }
 
@@ -2032,7 +2310,7 @@ static int vtxmatch(vp, tp)
 {
 	/* if the channel doesn't match, there's no hope! */
 	if (verbose>1)
-	  printf("ch? %s %s\n", vp->orig[L_CHANNEL]->name, tp->channel);
+	  sfprintf(sfstdout,"ch? %s %s\n", vp->orig[L_CHANNEL]->name, tp->channel);
 	if (tp->channel[0] == '*' && tp->channel[1] == '\0')
 	  return 0; /* Never match the defaults entry! */
 	if (!globmatch(tp->channel, vp->orig[L_CHANNEL]->name))
@@ -2045,7 +2323,7 @@ static int vtxmatch(vp, tp)
 	}
 
 	if (verbose>1)
-	  printf("host %s %s\n", vp->orig[L_HOST]->name, tp->host);
+	  sfprintf(sfstdout,"host %s %s\n", vp->orig[L_HOST]->name, tp->host);
 
 	return 1;
 }
@@ -2125,14 +2403,14 @@ static void vtxdo(vp, cehdr, path)
 	  }
 	}
 	if (n == 0) {
-	  fprintf(stderr, "%s: no pattern matched %s/%s address\n",
-		  progname, vp->orig[L_CHANNEL]->name, vp->orig[L_HOST]->name);
+	  sfprintf(sfstderr, "%s: no pattern matched %s/%s address\n",
+		   progname, vp->orig[L_CHANNEL]->name,vp->orig[L_HOST]->name);
 	  /* XX: memory leak here? */
 	  return;
 	}
 	if (verbose)
-	  printf("Matched %dth config entry with: %s/%s\n", cnt,
-		 vp->orig[L_CHANNEL]->name, vp->orig[L_HOST]->name);
+	  sfprintf(sfstdout, "Matched %dth config entry with: %s/%s\n", cnt,
+		   vp->orig[L_CHANNEL]->name, vp->orig[L_HOST]->name);
 
 	/* set default values */
 	if (tp->expiry > 0)
@@ -2329,7 +2607,7 @@ struct timeserver {
 #else
 	time_t	time_sec;
 #endif
-#define MAPSIZE 16*1024
+#define MAPSIZE 16*1024 /* Should work at all systems ? */
 };
 struct timeserver *timeserver_segment = NULL;
 static void init_timeserver()
@@ -2341,25 +2619,27 @@ static void init_timeserver()
 	int fd = -1;
 	char blk[1024];
 	int i;
-	FILE *fp = tmpfile();
+	Sfio_t *fp = sftmp(0); /* Create the backing file fairly reliably. */
 
-	if (fp) {
-	  fd = dup(fileno(fp));
-	  fclose(fp);
+	if (fp)
+	  fd = sffileno(fp);
+	if (fd < 0) {
+	  perror("init_timeserver() didn't create tempfile ??");
+	  return; /* Brr! */ 
 	}
-	if (fd < 0) return; /* Brr! */ 
 
 	memset(blk,0,sizeof(blk));
 	for (i=0; i < MAPSIZE; i += sizeof(blk))
-	  write(fd,blk,sizeof(blk));
-	lseek(fd,(off_t)0,0);
+	  sfwrite(fp, blk, sizeof(blk));
+	sfsync(fp);
+	sfseek(fp, 0, 0);
 
 #ifndef MAP_FILE
 # define MAP_FILE 0 /* SunOS 4.1 does not have this */
 #endif
 	timeserver_segment = (void*)mmap(NULL, MAPSIZE, PROT_READ|PROT_WRITE,
 					 MAP_FILE|MAP_SHARED, fd, 0);
-	close(fd);
+	sfclose(fp);
 	
 #else
 #ifndef MAP_VARIABLE
@@ -2396,8 +2676,8 @@ static void init_timeserver()
 	  struct timeval tv;
 	  int rc;
 
-	  tv.tv_sec = 1;
-	  tv.tv_usec = 0;
+	  tv.tv_sec = 0;
+	  tv.tv_usec = 300000;
 
 	  gettimeofday(&timeserver_segment->tv, NULL);
 
@@ -2436,7 +2716,17 @@ time_t *timep;
 	  return t;
 	}
 #endif
+#ifdef HAVE_GETTIMEOFDAY
+	{
+	  struct timeval tv;
+	  gettimeofday(&tv,NULL);
+	  if (timep)
+	    *timep = tv.tv_sec;
+	  return tv.tv_sec;
+	}
+#else
 	return time(timep); /* The classical old version.. */
+#endif
 }
 
 const char *
