@@ -5,14 +5,24 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1995
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1999
  */
 
 #define USE_SIGREAPER /* DO Use SIGCLD-driven reaper.. */
 
 #include "hostenv.h"
-#include "malloc.h"
+#include "zmalloc.h"
 #include "splay.h"
+#include "shmmib.h"
+#include "servauth.h"
+
+/* Some "forward declarations" */
+struct config_entry;
+struct procinfo;
+struct web;
+struct thread;
+struct threadgroup;
+struct vertex;
 
 struct config_entry {
 	struct config_entry *next;
@@ -38,9 +48,11 @@ struct config_entry {
 	int	maxkids;	/* run command only if # TA's running < this */
 	int	maxkidChannel;	/* run only if # TA's running channel < this */
 	int	maxkidThreads;	/*run only if # TA's running thread-ring<this*/
+	int	maxkidThread;	/*run only if # TA's running thread   < this */
 	int	idlemax;	/* max time to keep idle ta-procs around     */
 	int	skew;		/* retry skew parameter			     */
 	int	mark;		/* non-0 if we started a TA the last time    */
+	int	priority;	/* Scheduling priority			     */
 	int	overfeed;	/* How much overfeeding instead of sync processing ? */
 	char	**argv;		/* execv parameters for the command	     */
 	int	nretries;	/* number of retry factors known	     */
@@ -63,9 +75,11 @@ struct ctlfile {
 	int	fd;		/* a file descriptor pointing at the file    */
 	char	*vfpfn;		/* a filename for verbose logging of mail    */
 	uid_t	uid;		/* the owner of the control file (= msg file)*/
-	time_t	ctime;		/* when the original file arrived	     */
+	time_t	mtime;		/* original msg file's mtime (~ arrival)     */
 	time_t	envctime;	/* when the transport file was created	     */
 	int	haderror;	/* some errors/diagnostics need processing   */
+	int	resynccount;	/* How many resync requests ?		     */
+#define MAXRESYNCS 30
 	struct vertex	*head;		/* head of the list of groups	     */
 	int	nlines;		/* number of lines/entries in the file	     */
 	char	*erroraddr;	/* error address(es)			     */
@@ -83,10 +97,12 @@ struct ctlfile {
 	int	rcpnts_work;	/* .. yet to deliver ?			     */
 	int	mark;		/* flag used by selector() to pass filenames */
 	int	msgbodyoffset;	/* size of original headers to skip on errrpt*/
-	long	offset[1];	/* array of nlines byte offsets into the file*/
+	int	format;		/* Message format version -- _CF_FORMAT data */
+	int	offset[1];	/* array of nlines byte offsets into the file*/
 };
 
 struct threadgroup {
+	int		groupid;	/* Unique id */
 	int		threads;	/* Number of threads in the group   */
 	int		transporters;	/* Number of transporters running   */
 	int		idlecnt;	/* Number of idle transporters	    */
@@ -95,17 +111,21 @@ struct threadgroup {
 	struct web	*whost;		/* Host identity web		    */
 	char		*hostpatt;
 	int		withhost;	/* When set, wchan matters.	    */
-	struct thread	*thread;	/* Any of the threads in the group  */
-	struct threadgroup *next, *prev; /* Ring of thread groups	    */
+	struct thread	*thread;	/* First of the thread in the group */
+	struct thread	*thrtail;	/* Last of the threads in the group */
+	struct threadgroup *nextthg;    /* Ring of thread groups	    */
+	struct threadgroup *prevthg;
 	struct config_entry *cep;	/* Pointer to a config database     */
 	struct config_entry ce;		/* consed scheduler config file entry*/
 };
 
 struct thread {
+	long		threadid;	/* Unique id */
 	time_t		wakeup;		/* When to wake up ?		    */
 	int		attempts;	/* How many times activated ?	    */
 	int		retryindex;	/* when, what ?			    */
 	char		*channel, *host; /* documenting */
+	char		*pending;	/* reason for pending		    */
 	struct web	*wchan;		/* Web of CHANNELs		    */
 	struct web	*whost;		/* Web of HOSTs			    */
 	struct thread	*nexttr;	/* Next one in threads queue	    */
@@ -113,40 +133,71 @@ struct thread {
 	struct thread	*nextthg;	/* Next one in thread GROUP	    */
 	struct thread	*prevthg;	/* previous one..		    */
 	struct threadgroup *thgrp;	/* our group-leader		    */
-	struct vertex	*vertices;	/* First one of the vertices	    */
-	struct vertex	*lastvertex;	/* Last one of the vertices	    */
+	struct vertex	*thvertices;	/* First one of the thread vertices */
+	struct vertex	*lastthvertex;	/* Last one of the thread vertices  */
 	struct procinfo	*proc;		/* NULL or ptr to xport proc	    */
+	int		thrkids;	/* Number of procs at this thread   */
 	int		jobs;		/* How many items in this thread    */
+	struct vertex   *nextfeed;	/* vertex within that thread	    */
+					/* feed_child() forwards nextfeed   */
 };
 
 
 struct web {
 	char		*name;		/* name of the L_? thingy	    */
-	int		linkcnt;
+	int		linkcnt;	/* How many usage instances	    */
 	int		kids;	/* how many transport agents running for me */
 	struct vertex	*link;		/* points at group of addresses     */
 	struct vertex	*lastlink;	/* for efficiency at link_in()	    */
 };
 
+typedef enum {
+  CFSTATE_LARVA = 1,		/* The first feed of the thread		*/
+  CFSTATE_STUFFING = 2,		/* More feeds for the thread		*/
+  CFSTATE_FINISHING = 3,	/* end of thread, waiting reports	*/
+  CFSTATE_IDLE = 4,		/* Idle state				*/
+  CFSTATE_ERROR = 0		/* Error encountered			*/
+} TASTATE;
+
+  /* State changes:  fork() -> (1) --> (2) -+-> (3) -+-> (4) -+-> death
+                                ^       ^   |        |        |
+                                |       |-<-|        v        v
+                                |---<------------<---|--<-----|
+  */
+
 struct procinfo {
 	pid_t	pid;		/* Process-id				*/
 	int	reaped;
 	int	tofd;		/* tell transporter job data thru this	*/
-	int	hungry;		/* State of reports from child		*/
-	time_t	hungertime;	/* .. when presented			*/
-	int	fed;
+	int	waitstat;	/* What previously called WAIT told..   */
+
+	TASTATE	state;		/* Child-Feed State Machine state	*/
+
+	time_t	hungertime;	/* .. when last state change		*/
 	int	overfed;	/* Now many jobs fed to it over the normal 1?*/
 	time_t	feedtime;	/* .. when fed				*/
+
 	struct web *ch;		/* Web of CHANNELs			*/
 	struct web *ho;		/* Web of HOSTs				*/
-	struct thread *thread;	/* The thread we are processing		*/
-	struct vertex *vertex;	/* vertex within that thread		*/
-	struct procinfo *next;	/* next one in IDLE queue		*/
+				/* Set at channel creation, removed at
+				   reclaim(), modified at thread_start()
+				   when using IDLE queue.		*/
+
+
 	struct threadgroup *thg; /* The thread-ring we are in		*/
+
+	struct thread *pthread;	/* The thread we are processing		*/
+				/* ta_hungry() forwards pthread		*/
+
+	struct procinfo *pnext;	/* next one of procs in idle/thread	*/
+	struct procinfo *pprev;	/* prev one of procs in idle/thread	*/
+
 	char	*carryover;	/* Long responces..			*/
 	int	cmdlen;		/* buffer content size			*/
-	char	cmdbuf[1000];	/* outgoing pipe leftovers..		*/
-	char	cmdline[1000];	/* Approximation of the execl() params	*/
+	int	cmdspc;		/* buffer size				*/
+	char	*cmdbuf;	/* outgoing pipe leftovers..		*/
+	char	*cmdline;	/* Approximation of the execl() params	*/
+	int	cmdlspc;	/* cmdline buffer size			*/
 };
 
 /* Stores the offset indices of all addresses that have same channel and host*/
@@ -162,8 +213,8 @@ struct vertex {
 	struct vertex	*nextitem;	/* next in list of scheduled vertices*/
 	struct vertex	*previtem;	/* prev in list of scheduled vertices*/
 	char		*message;	/* some text associated with node    */
-	long		headeroffset;	/* Message headers for this rcpt     */
-	long		drptoffset;	/* IETF-NOTARY DRPT  data	     */
+	int		headeroffset;	/* Message headers for this rcpt     */
+	int		drptoffset;	/* IETF-NOTARY DRPT  data	     */
 	char		*notary;	/* IETF Notary report data	     */
 	int		notaryflg;	/* IETF DSN notary control flags     */
 #define NOT_NEVER   001
@@ -172,11 +223,45 @@ struct vertex {
 #define NOT_FAILURE 010
 	int		ce_pending;	/* pending on what ?		     */
 	time_t		ce_expiry;	/* when this vertex expires ?        */
-	struct procinfo *proc;		/* current TA process description    */
 	int		attempts;	/* count of number of TA invocations */
 	int		retryindex;	/* cur index into ce->retries array  */
 	time_t		wakeup;		/* time to wake up and run this      */
+	time_t		lastfeed;	/* When the last feed was ?	     */
 	char		*sender;	/* Message Sender/error recipient    */
 	int		ngroup;		/* number of addresses in group      */
 	int		index[1];	/* index of cfp->offset for group    */
 };
+
+
+/* mailq iterator state -- non-forking reporter mode */
+
+struct mailq; /* forward declarator */
+
+struct mailq {
+	struct mailq	*nextmailq;
+	int		auth;		/* what can do */
+	int		fd;		/* FD for I/O (nonblocking-IO) */
+
+	char		*challenge;
+
+	int		inbufspace;	/* Raw input buffer stuff */
+	int		inbufsize;
+	int		inbufcount;
+	char		*inbuf;
+
+	int		inplinespace;	/* Split into lines */
+	int		inplinesize;
+	char		*inpline;
+
+	int		outbufspace;	/* Output buffer */
+	int		outbufsize;
+	int		outbufcount;
+	int		outcol;
+	char		*outbuf;
+};
+
+#define MQ2MODE_SNMP	0x0001
+#define MQ2MODE_QQ	0x0002
+#define MQ2MODE_FULL	0x0004
+#define MQ2MODE_ETRN	0x0008
+#define MQ2MODE_KILL	0x0010
