@@ -5,11 +5,11 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1996
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1999
  */
 
 #include "hostenv.h"
-#include <stdio.h>
+#include <sfio.h>
 #include <ctype.h>
 #include <sys/param.h>
 #include <sys/stat.h>
@@ -26,8 +26,6 @@ static int ctlowner __((struct ctlfile *));
 static void vtxupdate __((struct vertex *, int, int));
 static void expaux __((struct vertex *, int, const char *));
 
-extern FILE *vfp_open __((struct ctlfile *));
-
 #if 0
 extern int ermdir();
 extern int rmdir();
@@ -53,7 +51,7 @@ extern char *procselect;
 
 /* dispatch table for diagnostic types */
 
-#define DARGS __((struct vertex *, long, long, long, const char*, const char*))
+#define DARGS __((struct procinfo *, struct vertex *, long, long, long, const char*, const char*))
 
 static int u_ok       DARGS ;
 static int u_ok2      DARGS ;
@@ -77,6 +75,7 @@ static struct diagcodes {
 		{	NULL,		NULL		}
 };
 
+
 void
 update(fd, diagnostic)
 	int fd;
@@ -93,143 +92,94 @@ update(fd, diagnostic)
 
 	if (*diagnostic == 0) {
 	  /* Lone newline.. old-style indications from the transporter */
-	  if (proc->tofd >= 0 &&
-	      proc->hungry == 0) {
-	    proc->hungry = 1;
-	    mytime(&now);
-	    proc->hungertime = now;
-	    ++hungry_childs;
-	  }
-	  /* Things are known to DIE from under us! */
-	  /* .. propably not a good idea to try to pick any next.. */
-	  if (proc->overfed > 0)
-	    proc->overfed -= 1;
-	  pick_next_vertex(proc, 1, 0);
-	  if (proc->hungry)
-	    feed_child(proc);
-	  flush_child(proc);
-	  return;
-	}
-	if (*diagnostic != '#') { /* Not debug diagnostic message */
-	  inum = atol(diagnostic);
-	  if ((cp = strchr(diagnostic, '/')) == NULL) {
-	    fprintf(stderr, "%s Misformed diagnostic: %s\n",
-		    timestring(), diagnostic);
-	    return;
-	  }
-	  offset = atol(++cp);
-	  if ((cp = strchr(cp, '\t')) == NULL) {
-	    fprintf(stderr, "%s Misformed diagnostic: %s\n",
-		    timestring(), diagnostic);
-	    return;
-	  }
-	  notary = ++cp;
-	  if ((cp = strchr(cp, '\t')) == NULL) {
-	    fprintf(stderr, "%s Misformed diagnostic: %s\n",
-		    timestring(), diagnostic);
-	    return;
-	  }
-	  *cp = 0; /* Trailing TAB after notary string, NULL it */
-	  type = ++cp;
-	  while (*cp != '\0' && isascii(*cp) && !isspace(*cp))
-	    ++cp;
-	  if (*cp == '\0') {
-	    message = NULL;
-	  } else {
-	    *cp++ = '\0';
-	    message = cp;
-	  }
-	  if (verbose)
-	    printf("diagnostic: %ld/%ld\t%s\t%s\n",inum,offset,notary,type);
-
-	  if ((vp = findvertex(inum, offset, &index)) == NULL)
-	    return;
-
-	  if (vp->notary != NULL)
-	    free(vp->notary);
-	  vp->notary = NULL;
-	  if (*notary)
-	    vp->notary = strsave(notary);
-
-	  /* Select function by the type name: ok/error/retryat/deferred */
-
-	  for (dcp = &diags[0]; dcp->name != NULL; ++dcp) {
-	    /* XX: replace strcmp() with cistrcmp() ??  Should not need,
-	       unless something is wrong with the transporters. */
-	    if (strcmp(dcp->name, type) == 0) {
-	      (dcp->fcn)(vp, index, inum, offset, notary, message);
-	      break;
-	    }
-	  }
-	  if (dcp->name == NULL)
-	    fprintf(stderr, "%s Unknown diagnostic type ignored: %s\n",
-		    timestring(), type);
+	  ta_hungry(proc);
 	  return;
 	}
 
-	/* Now (*diagnostic == '#') */
-	if (strncmp(diagnostic,"#hungry",7)==0) {
-	  /* This is an "actor model" behaviour,
-	     where actor tells, when it needs a new
-	     job to be fed to it. */
-	  if (proc->tofd   >= 0) {
-	    proc->hungry = 1;
-	    mytime(&proc->hungertime);
-	    if (proc->overfed > 0)
-	      /* It was overfed, decrement that counter first.. */
-	      proc->overfed -= 1;
-	    if (!proc->overfed) {
-	      ++hungry_childs;
-	      /* Unless it is still overfed,
-		 Pick next, and feed it! */
-	      pick_next_vertex(proc, 1, 0);
-	      /* While we have a thread, and things to feed.. */
-	      while (!proc->fed && proc->thread) {
-		if (proc->hungry)
-		  feed_child(proc);
-		if (proc->fed)
-		  proc->overfed += 1;
-		else
-		  break; /* Huh! Feed/flush failure! */
-		/* See if we should, and can feed more! */
-		if (proc->thg == NULL ||
-		    proc->pid == 0    ||
-		    proc->thread == NULL)
-		  break;	/* No new active threads/vertices/proc.. */
-		if (proc->overfed >= proc->thg->ce.overfeed)
-		  break;	/* if the limit is zero, don't overfeed ever.*/
-		/* Ok, increment the counter, and loop back.. */
-		proc->hungry = 1; /* Simulate hunger.. */
-		proc->fed = 1; /* tell them, it has been fed.. */
-		pick_next_vertex(proc, 1, 0);
-		/* If it got next,  ``proc->fed'' is now zero.. */
-	      }
-	      flush_child(proc);
-	      proc->hungry = 0; /* ... satiated.. */
-	    } else {
-	      if (verbose)
-		printf("... child pid %d overfed=%d\n",
-		       proc->pid,proc->overfed);
-	    }
-	  } else
-	    if (verbose)
-	      printf("'#hungry' from child without forward-channel\n");
-	  return;
-	} /* end of '#hungry' processing */
-	if (strncmp(diagnostic,"#resync",7)==0) {
-	  /* The transporter has noticed that the scheduler
-	     gave it a job spec, which does not have anything
-	     left for processing, it is time for the scheduler
-	     to recheck the job file. */
-	  char *p, *s = diagnostic + 7;
-	  while (*s == ' ' || *s == '\t') ++s;
-	  p = strchr(s,'\n');
-	  if (p) *p = 0; /* newline AFTER filename */
-	  if (*s != 0)
-	    resync_file(s);
+	if (*diagnostic == '#') {
+	  /* Now (*diagnostic == '#') */
+
+	  if (strncmp(diagnostic,"#hungry",7)==0) {
+	    ta_hungry(proc);
+	    return;
+	  } /* end of '#hungry' processing */
+
+	  if (strncmp(diagnostic,"#resync",7)==0) {
+	    /* The transporter has noticed that the scheduler
+	       gave it a job spec, which does not have anything
+	       left for processing, it is time for the scheduler
+	       to recheck the job file. */
+	    char *p, *s = diagnostic + 7;
+	    while (*s == ' ' || *s == '\t') ++s;
+	    p = strchr(s,'\n');
+	    if (p) *p = 0; /* newline AFTER filename */
+	    if (*s != 0)
+	      resync_file(proc, s);
+	    return;
+	  }
+
+	  sfprintf(sfstderr, "%s DBGdiag: %s\n", timestring(), diagnostic);
 	  return;
 	}
-	fprintf(stderr, "%s DBGdiag: %s\n", timestring(), diagnostic);
+
+	/* Not debug diagnostic message */
+
+	inum = atol(diagnostic);
+	if ((cp = strchr(diagnostic, '/')) == NULL) {
+	  sfprintf(sfstderr, "%s Misformed diagnostic1: %s\n",
+		   timestring(), diagnostic);
+	  return;
+	}
+	offset = atol(++cp);
+	if ((cp = strchr(cp, '\t')) == NULL) {
+	  sfprintf(sfstderr, "%s Misformed diagnostic2: %s\n",
+		   timestring(), diagnostic);
+	  return;
+	}
+	notary = ++cp;
+	if ((cp = strchr(cp, '\t')) == NULL) {
+	  sfprintf(sfstderr, "%s Misformed diagnostic3: %s\n",
+		   timestring(), diagnostic);
+	  return;
+	}
+	*cp = 0; /* Trailing TAB after notary string, NULL it */
+	type = ++cp;
+	while (*cp != '\0' && isascii(*cp) && !isspace(*cp))
+	  ++cp;
+	if (*cp == '\0') {
+	  message = NULL;
+	} else {
+	  *cp++ = '\0';
+	  message = cp;
+	}
+	if (verbose)
+	  sfprintf(sfstdout,"diagnostic: %ld/%ld\t%s\t%s\n",
+		   inum, offset, notary, type);
+
+	if ((vp = findvertex(inum, offset, &index)) == NULL)
+	  return;
+
+	if (vp->notary != NULL)
+	  free(vp->notary);
+	vp->notary = NULL;
+	if (*notary)
+	  vp->notary = strsave(notary);
+
+	/* Select function by the type name: ok/error/retryat/deferred */
+
+	for (dcp = &diags[0]; dcp->name != NULL; ++dcp) {
+	  /* XX: replace strcmp() with cistrcmp() ??  Should not need,
+	     unless something is wrong with the transporters. */
+	  if (strcmp(dcp->name, type) == 0) {
+	    (dcp->fcn)(proc, vp, index, inum, offset, notary, message);
+	    break;
+	  }
+	}
+
+	if (dcp->name == NULL)
+	  sfprintf(sfstderr, "%s Unknown diagnostic type ignored: %s\n",
+		   timestring(), type);
+
 	return;
 }
 
@@ -242,28 +192,28 @@ unctlfile(cfp, no_unlink)
 	struct ctlfile *cfp;
 	int no_unlink;
 {
-	struct spblk *spl;
-	char	path[MAXPATHLEN+1];
-
-	if (cfp->dirind > 0)
-	  sprintf(path, "%s/%s", cfpdirname(cfp->dirind), cfp->mid);
-	else
-	  strcpy(path, cfp->mid);
-
 	if (!no_unlink && !procselect) {
+	  char	path[MAXPATHLEN+1];
+
+	  if (cfp->dirind > 0)
+	    sprintf(path, "%s/%s", cfpdirname(cfp->dirind), cfp->mid);
+	  else
+	    strcpy(path, cfp->mid);
+
 	  reporterrs(cfp);
 
 	  if (do_syslog) {
-	    char taspid[30],fnam[30];
-	    sprintf(fnam,"%lu",(long)cfp->id);
-	    taspoolid(taspid, sizeof(taspid), cfp->ctime, fnam);
+	    char taspid[30];
+	    taspoolid(taspid, cfp->mtime, (long)cfp->id);
 	    zsyslog((LOG_INFO, "%s: complete (total %d recepients, %d failed)",
 		     taspid, cfp->rcpnts_total, cfp->rcpnts_failed));
 	  }
+	  ++MIBMtaEntry->mtaTransmittedMessagesSc;
 
 	  eunlink(path);
 	  if (verbose)
-	    printf("%s: unlink %s (mid=0x%p)", cfp->logident, path, cfp->mid);
+	    sfprintf(sfstdout,"%s: unlink %s (mid=%p)",
+		     cfp->logident, path, cfp->mid);
 
 	  if (cfp->dirind > 0)
 	    sprintf(path, "../%s/%s/%s",
@@ -274,104 +224,94 @@ unctlfile(cfp, no_unlink)
 
 	  eunlink(path);
 	  if (verbose)
-	    printf("   and %s/\n", path);
+	    sfprintf(sfstdout, "   and %s/\n", path);
 
 	  if (cfp->vfpfn != NULL) {
-	    FILE *vfp = vfp_open(cfp);
+	    Sfio_t *vfp = vfp_open(cfp);
 	    if (vfp) {
-	      fprintf(vfp, "scheduler done processing %s\n", cfp->mid);
-	      fclose(vfp);
+	      sfprintf(vfp, "scheduler done processing %s\n", cfp->mid);
+	      sfclose(vfp);
 	    }
 	  }
 	} else {
+#if 0
 	  /* We will LOOSE this from the schedules -- add info about
 	     it into the indirscanqueue -- at the tail... */
-	  dq_insert(NULL, atol(cfp->mid), path, 30);
+	  if (cfp->id)
+	    dq_insert(NULL, cfp->id, path, 30);
+#endif
 	}
-	spl = sp_lookup(cfp->id, spt_mesh[L_CTLFILE]);
-	if (spl != NULL)
-	  sp_delete(spl, spt_mesh[L_CTLFILE]);
-	free_cfp_memory(cfp);
+
+	if (cfp->id != 0) {
+	  struct spblk *spl;
+	  spl = sp_lookup(cfp->id, spt_mesh[L_CTLFILE]);
+	  if (spl != NULL)
+	    sp_delete(spl, spt_mesh[L_CTLFILE]);
+	}
+
 	--global_wrkcnt;
+	--MIBMtaEntry->mtaStoredMessages;
+	free_cfp_memory(cfp);
 }
 
 /* unvertex() .. the ``vp'' CAN'T be invalid.. */
+/* As a side-effect; THR where this VTX is one of members *may*
+   become deleted in case the VTX (kids-)count goes to zero.. */
 void unvertex(vp, justfree, ok)
 	register struct vertex *vp;
 	int justfree, ok;
 {
-	int	i, removeme;
+	int	i;
 
 	if (vp->ngroup > 0)
 	  return;
 
-	if (verbose && justfree < 0)
-	  fprintf(stderr,
-		  "unvertex(vtx=0x%p (thr=0x%p proc=0x%p, ng=%d) ,%d,%d)\n",
-		  vp,vp->thread,vp->proc,vp->ngroup,justfree,ok);
+	if (verbose)
+	  sfprintf(sfstderr,
+		   "unvertex(vtx=%p (thr=%p, ng=%d) ,%d,%d)\n",
+		   vp, vp->thread, vp->ngroup, justfree, ok);
 
-	if (vp->thread != NULL &&
-	    vp->thread->vertices == vp && vp->thread->proc &&
-	    vp->thread->proc->vertex == vp) {
-	    /* Whoops, we are the active first on thread.. */
-	  if (vp->proc == NULL)
-	    vp->proc = vp->thread->proc;
-	  vp->proc->vertex = vp;
+	if (vp->thread && vp->thread->proc) {
+	  /* Somebody here, move it elsewere! */
+	  /* This MAY happen -- expire() hits when some
+	     thread is coming into processing...            */
+	  if (vp->thread->nextfeed == vp)
+	    vp->thread->nextfeed = vp->nextitem;
 	}
-
-	if (vp->proc && vp->proc->vertex == vp) {
-	  vp->proc->fed     = 1; /* Mark it fed just in case.. */
-	  vp->proc->overfed = 0; /* .. and clear this .. */
-	  /* Pick next, but don't feed it (yet)! */
-	  pick_next_vertex(vp->proc, ok, justfree);
-	  if (vp->proc && vp->proc->vertex == vp){
-	    fprintf(stderr,
-		    "unvertex(vtx=0x%p,%d,%d) failed to pick_next_vertex()!\n",
-		    vp,justfree,ok);
-	    /* We may become called with child feeder yet unflushed;
-	       shall we kill the kid ? (pick_next_vertex won't change
-	       vertex then..) */
-	    /* abort(); */
-	  }
-	}
-	if (vp->thread != NULL &&
-	    vp->thread->proc && vp->thread->proc->vertex == vp) {
-	  /* XX: This is actually vestigal from somewhere, and
-	         should not occur at all.. */
-	  vp->thread->proc->vertex = NULL;
-	}
-
 	for (i = 0; i < SIZE_L; ++i) {
-	  if (vp->next[i] != NULL)
+
+	  if (vp->next[i])
 	    vp->next[i]->prev[i] = vp->prev[i];
-	  if (vp->prev[i] != NULL)
+	  if (vp->prev[i])
 	    vp->prev[i]->next[i] = vp->next[i];
+
 	  if (i == L_CTLFILE)
 	    continue;
-	  removeme = 0;
+
 	  if (vp->orig[i]->link == vp)
-	    if ((vp->orig[i]->link = vp->next[i]) == NULL)
-	      removeme = 1;
+	    vp->orig[i]->link = vp->next[i];
+
 	  if (vp->orig[i]->lastlink == vp)
-	    if ((vp->orig[i]->lastlink = vp->prev[i]) == NULL)
-	      removeme = 1;
-	  if (removeme) {
-	    vp->orig[i]->linkcnt -= 1;
-	    unweb(i, vp->orig[i]);
-	    vp->orig[i] = NULL;
-	  }
+	    vp->orig[i]->lastlink = vp->prev[i];
+
+	  vp->orig[i]->linkcnt -= 1;
+	  unweb(i, vp->orig[i]);
 	}
 
-	if (vp->cfp->head == vp)
-	  if ((vp->cfp->head = vp->next[L_CTLFILE]) == NULL) 
+	if (vp->cfp->head == vp) {
+	  vp->cfp->head = vp->next[L_CTLFILE];
+	  if (vp->cfp->head == NULL && justfree >= 0)
 	    unctlfile(vp->cfp, justfree);
+	}
 
-	web_disentangle(vp, ok); /* does also unthread() */
+	web_detangle(vp, ok); /* does also unthread() */
 
 	if (vp->message != NULL) free(vp->message);
 	if (vp->notary  != NULL) free(vp->notary);
 	/* if (vp->sender != NULL) free(vp->sender); */ /* XX: cache !! ?? */
 	free((char *)vp);
+
+	return;
 }
 
 static struct vertex *findvertex(inum, offset, idx)
@@ -389,8 +329,8 @@ static struct vertex *findvertex(inum, offset, idx)
 	if (spl == NULL || (cfp = (struct ctlfile *)spl->data) == NULL) {
 	  /* It may have been kicked into input queue */
 	  if (!in_dirscanqueue(NULL,(long)inum))
-	    fprintf(stderr, "%s: cannot find control file for %ld!\n",
-		    progname, inum);
+	    sfprintf(sfstderr, "%s: cannot find control file for %ld!\n",
+		     progname, inum);
 	  return NULL;
 	}
 	for (i = 0; i < cfp->nlines; ++i) {
@@ -400,18 +340,18 @@ static struct vertex *findvertex(inum, offset, idx)
 	  }
 	}
 	if (i >= cfp->nlines) {
-	  fprintf(stderr,
-		  "%s: unknown address offset %ld in control file %ld!\n",
-		  progname, offset, inum);
+	  sfprintf(sfstderr,
+		   "%s: unknown address offset %ld in control file %ld!\n",
+		   progname, offset, inum);
 	  return NULL;
 	}
 	for (vp = cfp->head; vp != NULL; vp = vp->next[L_CTLFILE])
 	  for (i = 0; i < vp->ngroup; ++i)
 	    if (vp->index[i] == *idx)
 	      return vp;
-	fprintf(stderr,
-		"%s: multiple processing of address at %ld in control file %ld!\n",
-		progname, offset, inum);
+	sfprintf(sfstderr,
+		 "%s: multiple processing of address at %ld in control file %ld!\n",
+		 progname, offset, inum);
 	return NULL;
 }
 
@@ -434,9 +374,9 @@ static int ctlowner(cfp)
 	if (cfp->mid == NULL)
 	  abort(); /* calling-convention error! */
 #ifdef USE_ALLOCA
-	path = alloca(5+strlen(cfp->mid)+sizeof QUEUEDIR+8);
+	path = (char*)alloca(5+strlen(cfp->mid)+sizeof QUEUEDIR+8);
 #else
-	path = emalloc(5+strlen(cfp->mid)+sizeof QUEUEDIR+8);
+	path = (char*)emalloc(5+strlen(cfp->mid)+sizeof QUEUEDIR+8);
 #endif
 	if (cfp->dirind > 0)
 	  sprintf(path, "../%s/%s/%s",
@@ -493,9 +433,11 @@ deletemsg(msgid, curcfp)
 	 * this strange way of doing the unlink.
 	 */
 	while (cfp->head->next[L_CTLFILE] != NULL) {
+	  MIBMtaEntry->mtaStoredRecipients -= cfp->head->ngroup;
 	  cfp->head->ngroup = 0;
 	  unvertex(cfp->head,0,1);
 	}
+	MIBMtaEntry->mtaStoredRecipients -= cfp->head->ngroup;
 	cfp->head->ngroup = 0;
 	unvertex(cfp->head,0,1);
 }
@@ -595,6 +537,9 @@ static void vtxupdate(vp, index, ok)
 	  if (vp->index[i] == index) {
 	    /* remove us from the vertex indices */
 	    vp->ngroup -= 1;
+
+	    MIBMtaEntry->mtaStoredRecipients -= 1;
+
 	    /* compact the index array */
 	    for (++i; i <= vp->ngroup; ++i)
 	      vp->index[i-1] = vp->index[i];
@@ -605,20 +550,21 @@ static void vtxupdate(vp, index, ok)
 	  }
 }
 
-static void logstat __((FILE *, struct vertex *, const char *));
+static void logstat __((struct vertex *, const char *));
 
-static void logstat(fp,vp,reason)
-	FILE *fp;
+static void logstat(vp,reason)
 	struct vertex *vp;
 	const char *reason;
 {
+	if (!statuslog) return;
+
 	mytime(&now);
-	fprintf(fp,"%ld %s %ld %ld %s %s/%s\n",
-		(long)vp->cfp->ctime, vp->cfp->mid,
-		(long)(vp->cfp->envctime - vp->cfp->ctime),
-		(long)(now - vp->cfp->envctime), reason,
-		vp->orig[L_CHANNEL]->name,vp->orig[L_HOST]->name);
-	fflush(fp);
+	sfprintf(statuslog, "%ld %s %ld %ld %s %s/%s\n",
+		 (long)vp->cfp->mtime, vp->cfp->mid,
+		 (long)(vp->cfp->envctime - vp->cfp->mtime),
+		 (long)(now - vp->cfp->envctime), reason,
+		 vp->orig[L_CHANNEL]->name,vp->orig[L_HOST]->name);
+	sfsync(statuslog);
 }
 
 
@@ -637,15 +583,15 @@ static void expaux(vp, index, buf)
 	  }
 
 	/* Log something into the scheduler log */
-	fprintf(stderr, "%s %s: %s/%s from %s %s\n", timestring(), progname,
-		vp->orig[L_CHANNEL]->name, vp->orig[L_HOST]->name,
-		vp->cfp->erroraddr == NULL ? "?" : vp->cfp->erroraddr, buf);
+	sfprintf(sfstderr, "%s %s: %s/%s from %s %s\n", timestring(), progname,
+		 vp->orig[L_CHANNEL]->name, vp->orig[L_HOST]->name,
+		 vp->cfp->erroraddr == NULL ? "?" : vp->cfp->erroraddr, buf);
 
-	if (statuslog)
-	  logstat(statuslog,vp,"expire");
+	logstat(vp,"expire");
 
 	/* Delete this vertex from scheduling datasets */
 	vtxupdate(vp, index, 0);
+	++MIBMtaEntry->mtaTransmittedRecipientsSc;
 }
 
 void
@@ -664,13 +610,13 @@ expire(vp, index)
 		  "failed",
 		  "5.4.7 (unspecified timeout failure)",
 		  "smtp; 500 (Expired after ");
-	  saytime((u_long)(vp->ce_expiry - vp->cfp->ctime), buf, 0);
+	  saytime((u_long)(vp->ce_expiry - vp->cfp->mtime), buf, 0);
 	  strcat(buf,")\001");
 	  vp->notary = strsave(buf);
 	}
 
 	strcpy(buf, "expired after ");
-	saytime((u_long)(vp->ce_expiry - vp->cfp->ctime), buf, 0);
+	saytime((u_long)(vp->ce_expiry - vp->cfp->mtime), buf, 0);
 
 	if (vp->message != NULL && *(vp->message) != '\0') {
 	  emsg = emalloc(strlen(buf) + strlen(vp->message) + strlen(fmt));
@@ -680,7 +626,7 @@ expire(vp, index)
 
 	if (index < 0) {
 	  /* Expire from the LAST index to the first, this way
-	     we won't do the mistake of referring indixes after
+	     we won't do the mistake of referring indices after
 	     they have been deleted.. */
 	  for (i = vp->ngroup -1; i >= 0; --i)
 	    expaux(vp, vp->index[i], emsg);
@@ -693,21 +639,25 @@ expire(vp, index)
 
 
 /*ARGSUSED*/
-static int u_ok(vp, index, inum, offset, notary, message)
-	struct vertex *vp;
-	long	index, inum, offset;
-	const char	*notary;
-	const char	*message;
+static int u_ok(proc, vp, index, inum, offset, notary, message)
+     struct procinfo *proc;
+     struct vertex *vp;
+     long   index, inum, offset;
+     const char	*notary;
+     const char	*message;
 {
 	if (verbose)
-	  fprintf(stderr,"%s: %ld/%ld/%s/ok %s\n", vp->cfp->logident, inum,
-		  offset, notary, message ? message : "-");
+	  sfprintf(sfstderr,"%s: %ld/%ld/%s/ok %s\n", vp->cfp->logident, inum,
+		   offset, notary, message ? message : "-");
 #if 0
-	if (vp->cfp->vfp != NULL && vp->cfp->contents != NULL) {
-	  fseek(vp->cfp->vfp, (off_t)0, SEEK_END);
-	  fprintf(vp->cfp->vfp, "%s: ok %s\n",
-		  vp->cfp->contents + offset + 2 + _CFTAG_RCPTPIDSIZE,
-		  message == NULL ? "(sent)" : message);
+	if (vp->cfp->contents != NULL) {
+	  Sfio_t *vfp = vfp_open(vp->cfp);
+	  if (vfp) {
+	    sfprintf(vfp, "%s: ok %s\n",
+		     vp->cfp->contents + offset + 2 + _CFTAG_RCPTPIDSIZE,
+		     message == NULL ? "(sent)" : message);
+	    sfclose(vfp);
+	  }
 	}
 #endif
 	if (vp->notaryflg & NOT_SUCCESS) {
@@ -715,94 +665,106 @@ static int u_ok(vp, index, inum, offset, notary, message)
 	  msgerror(vp, offset, message);
 	}
 
-	if (statuslog)
-	  logstat(statuslog,vp,"ok");
+	logstat(vp,"ok");
 
 	/* Delete this vertex from scheduling datasets */
 	vtxupdate(vp, index, 1);
+	++MIBMtaEntry->mtaTransmittedRecipientsSc;
 	return 1;
 }
 
 /*ARGSUSED*/
-static int u_ok2(vp, index, inum, offset, notary, message)
-	struct vertex *vp;
-	long	index, inum, offset;
-	const char	*notary;
-	const char	*message;
+static int u_ok2(proc, vp, index, inum, offset, notary, message)
+     struct procinfo *proc;
+     struct vertex *vp;
+     long   index, inum, offset;
+     const char	*notary;
+     const char	*message;
 {
 	if (vp->notaryflg & NOT_SUCCESS) {
 	  vp->cfp->haderror = 1; /* The transporter logged it for us! */
 	}
 	if (verbose)
-	  fprintf(stderr,"%s: %ld/%ld/%s/ok2 %s\n", vp->cfp->logident, inum,
-		  offset, notary, message ? message : "-");
+	  sfprintf(sfstderr,"%s: %ld/%ld/%s/ok2 %s\n", vp->cfp->logident, inum,
+		   offset, notary, message ? message : "-");
 #if 0
-	if (vp->cfp->vfp != NULL && vp->cfp->contents != NULL) {
-	  fseek(vp->cfp->vfp, (off_t)0, SEEK_END);
-	  fprintf(vp->cfp->vfp, "%s: ok %s\n",
-		  vp->cfp->contents + offset + 2 + _CFTAG_RCPTPIDSIZE,
-		  message == NULL ? "(sent)" : message);
+	if (vp->cfp->contents != NULL) {
+	  Sfio_t *vfp = vfp_open(vp->cfp);
+	  if (vfp) {
+	    sfprintf(vfp, "%s: ok2 %s\n",
+		     vp->cfp->contents + offset + 2 + _CFTAG_RCPTPIDSIZE,
+		     message == NULL ? "(sent)" : message);
+	    sfclose(vfp);
+	  }
 	}
 #endif
-	if (statuslog)
-	  logstat(statuslog,vp,"ok2");
+	logstat(vp,"ok2");
 
 	/* Delete this vertex from scheduling datasets */
 	vtxupdate(vp, index, 1);
+	++MIBMtaEntry->mtaTransmittedRecipientsSc;
 	return 1;
 }
 
 /*ARGSUSED*/
-static int u_ok3(vp, index, inum, offset, notary, message)
-	struct vertex *vp;
-	long	index, inum, offset;
-	const char	*notary;
-	const char	*message;
+static int u_ok3(proc, vp, index, inum, offset, notary, message)
+     struct procinfo *proc;
+     struct vertex *vp;
+     long   index, inum, offset;
+     const char	*notary;
+     const char	*message;
 {
 	/* Success, but the transporter was able to relay the DSN info
 	   to another system, thus no diagnostics here! */
 	if (verbose)
-	  fprintf(stderr,"%s: %ld/%ld/%s/ok3 %s\n", vp->cfp->logident, inum,
-		  offset, notary, message ? message : "-");
+	  sfprintf(sfstderr,"%s: %ld/%ld/%s/ok3 %s\n", vp->cfp->logident, inum,
+		   offset, notary, message ? message : "-");
 #if 0
-	if (vp->cfp->vfp != NULL && vp->cfp->contents != NULL) {
-	  fseek(vp->cfp->vfp, (off_t)0, SEEK_END);
-	  fprintf(vp->cfp->vfp, "%s: ok %s\n",
-		  vp->cfp->contents + offset + 2 + _CFTAG_RCPTPIDSIZE,
-		  message == NULL ? "(sent)" : message);
+	if (vp->cfp->contents != NULL) {
+	  Sfio_t *vfp = vfp_open(vp->cfp);
+	  if (vfp) {
+	    sfprintf(vfp, "%s: ok3 %s\n",
+		     vp->cfp->contents + offset + 2 + _CFTAG_RCPTPIDSIZE,
+		     message == NULL ? "(sent)" : message);
+	    sfclose(vfp);
+	  }
 	}
 #endif
-	if (statuslog)
-	  logstat(statuslog,vp,"ok3");
+	logstat(vp,"ok3");
 
 	/* Delete this vertex from scheduling datasets */
 	vtxupdate(vp, index, 1);
+	++MIBMtaEntry->mtaTransmittedRecipientsSc;
 	return 1;
 }
 
 
-static int u_deferred(vp, index, inum, offset, notary, message)
-	struct vertex *vp;
-	long	index, inum, offset;
-	const char	*notary;
-	const char	*message;
+static int u_deferred(proc, vp, index, inum, offset, notary, message)
+     struct procinfo *proc;
+     struct vertex *vp;
+     long   index, inum, offset;
+     const char	*notary;
+     const char	*message;
 {
-	/* fprintf(stderr,"%s: %ld/%ld/%s/deferred %s\n", vp->cfp->logident,
+	/* sfprintf(sfstderr,"%s: %ld/%ld/%s/deferred %s\n", vp->cfp->logident,
 	   inum, offset, notary, message ? message : "-"); */
 	if (message != NULL) {
 	  if (vp->message != NULL)
 	    free(vp->message);
-	  /* fprintf(stderr, "add message '%s' to node %s/%s\n",
+	  /* sfprintf(sfstderr, "add message '%s' to node %s/%s\n",
 	     message, vp->orig[L_CHANNEL]->name,
 	     vp->orig[L_HOST]->name); */
 	  vp->message = strsave(message);
 	}
 #if 0
-	if (vp->cfp->vfp != NULL && vp->cfp->contents != NULL) {
-	  fseek(vp->cfp->vfp, (off_t)0, SEEK_END);
-	  fprintf(vp->cfp->vfp, "%s: deferred %s\n",
-		  vp->cfp->contents + offset + 2 + _CFTAG_RCPTPIDSIZE,
-		  message == NULL ? "(unknown)" : message);
+	if (vp->cfp->contents != NULL) {
+	  Sfio_t *vfp = vfp_open(vp->cfp);
+	  if (vfp) {
+	    sfprintf(vfp, "%s: deferred %s\n",
+		     vp->cfp->contents + offset + 2 + _CFTAG_RCPTPIDSIZE,
+		     message == NULL ? "(sent)" : message);
+	    sfclose(vfp);
+	  }
 	}
 #endif
 	/*
@@ -814,65 +776,73 @@ static int u_deferred(vp, index, inum, offset, notary, message)
 	return 1;
 }
 
-static int u_error(vp, index, inum, offset, notary, message)
-	struct vertex  *vp;
-	long		index, inum, offset;
-	const char	*notary;
-	const char	*message;
+static int u_error(proc, vp, index, inum, offset, notary, message)
+     struct procinfo *proc;
+     struct vertex *vp;
+     long   index, inum, offset;
+     const char	*notary;
+     const char	*message;
 {
 	if (message == NULL)
 	  message = "(unknown)";
-	fprintf(stderr,
-		"%s %s: %ld/%ld/%s/error %s\n", timestring(),
-		vp->thgrp->ce.command, inum, offset, notary, message);
+	sfprintf(sfstderr,
+		 "%s %s: %ld/%ld/%s/error %s\n", timestring(),
+		 vp->thgrp->ce.command, inum, offset, notary, message);
 
 	if (!procselect && vp->notaryflg & NOT_FAILURE)
 	  msgerror(vp, offset, message);
 #if 0
-	if (vp->cfp->vfp != NULL && vp->cfp->contents != NULL) {
-	  fseek(vp->cfp->vfp, (off_t)0, SEEK_END);
-	  fprintf(vp->cfp->vfp, "%s: error %s\n",
-		  vp->cfp->contents + offset + 2 + _CFTAG_RCPTPIDSIZE,
-		  message);
+	if (vp->cfp->contents != NULL) {
+	  Sfio_t *vfp = vfp_open(vp->cfp);
+	  if (vfp) {
+	    sfprintf(vfp, "%s: error %s\n",
+		     vp->cfp->contents + offset + 2 + _CFTAG_RCPTPIDSIZE,
+		     message == NULL ? "(sent)" : message);
+	    sfclose(vfp);
+	  }
 	}
 #endif
-	if (statuslog)
-	  logstat(statuslog,vp,"error");
+	logstat(vp,"error");
 
 	/* Delete this vertex from scheduling datasets */
 	vtxupdate(vp, index, 0);
+	++MIBMtaEntry->mtaTransmittedRecipientsSc;
 	return 1;
 }
 
 /* A variant where the TRANSPORT AGENT has logged the report into
    the file! */
-static int u_error2(vp, index, inum, offset, notary, message)
-	struct vertex *vp;
-	long	index, inum, offset;
-	const char	*notary;
-	const char	*message;
+static int u_error2(proc, vp, index, inum, offset, notary, message)
+     struct procinfo *proc;
+     struct vertex *vp;
+     long   index, inum, offset;
+     const char	*notary;
+     const char	*message;
 {
 	if (message == NULL)
 	  message = "(unknown)";
-	fprintf(stderr,
+	sfprintf(sfstderr,
 		"%s %s: %ld/%ld/%s/error2 %s\n", timestring(),
 		vp->thgrp->ce.command, inum, offset, notary, message);
 
 	/* We don't need to log it! */
 	vp->cfp->haderror = 1; /* Mark it into the incore dataset */
 #if 0
-	if (vp->cfp->vfp != NULL && vp->cfp->contents != NULL) {
-	  fseek(vp->cfp->vfp, (off_t)0, SEEK_END);
-	  fprintf(vp->cfp->vfp, "%s: error %s\n",
-		  vp->cfp->contents + offset + 2 + _CFTAG_RCPTPIDSIZE,
-		  message);
+	if (vp->cfp->contents != NULL) {
+	  Sfio_t *vfp = vfp_open(vp->cfp);
+	  if (vfp) {
+	    sfprintf(vfp, "%s: error2 %s\n",
+		     vp->cfp->contents + offset + 2 + _CFTAG_RCPTPIDSIZE,
+		     message == NULL ? "(sent)" : message);
+	    sfclose(vfp);
+	  }
 	}
 #endif
-	if (statuslog)
-	  logstat(statuslog,vp,"error");
+	logstat(vp,"error");
 
 	/* Delete this vertex from scheduling datasets */
 	vtxupdate(vp, index, 0);
+	++MIBMtaEntry->mtaTransmittedRecipientsSc;
 	return 1;
 }
 
@@ -881,16 +851,47 @@ static int u_error2(vp, index, inum, offset, notary, message)
  * specify relative (w/ leading +) or absolute (w/o leading +) retry time.
  */
 
-static int u_retryat(vp, index, inum, offset, notary, message)
-	struct vertex *vp;
-	long	index, inum, offset;
-	const char	*notary;
-	const char	*message;
+static int u_retryat(proc, vp, index, inum, offset, notary, message)
+     struct procinfo *proc;
+     struct vertex *vp;
+     long   index, inum, offset;
+     const char	*notary;
+     const char	*message;
 {
 	time_t	retrytime;
 	long    dtvalue;
 	const char * cp;
 
+	/* If a message gets a "retryat" signal, kick this thread at
+	   next  "#hungry"  into FINISHING state */
+
+	if (verbose)
+	  sfprintf(sfstdout,
+		   "RETRYAT: proc=%p (S=%d OF=%d tofd=%d) vp=%p[%d] message='%s'\n",
+		   proc, (int)proc->state, proc->overfed, proc->tofd, vp, index, message);
+
+#if 0
+	if ((proc->state   == CFSTATE_LARVA) &&
+	    (proc->overfed == 1) &&
+	    (proc->tofd    >= 0))
+	  proc->state = CFSTATE_FINISHING;
+	/* That previous state change would mean a fork()ed sub-program
+	   has spontaneously decided to produce  retryat  state... */
+#endif
+
+	if ((proc->state   == CFSTATE_STUFFING) &&
+	    (proc->tofd    >= 0)) {
+
+	  /* Fed work-entry caused 'retryat' to occur; stop
+	     feeding and move to the finishing state before
+	     continuing anywhere... */
+
+	  /* sfprintf(sfstderr,
+	     "%% u_retryat(proc=%p) -> CFSTATE_FINISHING\n", proc); */
+
+	  proc->state = CFSTATE_FINISHING;
+	}
+
 	if (*message == '+')
 	  ++message;
 	dtvalue = 0;
@@ -907,17 +908,20 @@ static int u_retryat(vp, index, inum, offset, notary, message)
 	if (message != NULL) {
 	  if (vp->message != NULL)
 	    free(vp->message);
-	  /* fprintf(stderr, "add message '%s' to node %s/%s\n",
+	  /* sfprintf(sfstderr, "add message '%s' to node %s/%s\n",
 	     message, vp->orig[L_CHANNEL]->name,
 	     vp->orig[L_HOST]->name); */
 	  vp->message = strsave(message);
 	}
 #if 0
-	if (vp->cfp->vfp != NULL && vp->cfp->contents != NULL) {
-	  fseek(vp->cfp->vfp, (off_t)0, SEEK_END);
-	  fprintf(vp->cfp->vfp, "%s: retryat %d %s\n",
-		  vp->cfp->contents + offset + 2 + _CFTAG_RCPTPIDSIZE,
-		  (int)retrytime, message == NULL ? "(unknown)" : message);
+	if (vp->cfp->contents != NULL) {
+	  Sfio_t *vfp = vfp_open(vp->cfp);
+	  if (vfp) {
+	    sfprintf(vfp, "%s: retryat %d %s\n",
+		     vp->cfp->contents + offset + 2 + _CFTAG_RCPTPIDSIZE,
+		     (int)retrytime, message == NULL ? "(unknown)" : message);
+	    sfclose(vfp);
+	  }
 	}
 #endif
 	/*
@@ -926,12 +930,8 @@ static int u_retryat(vp, index, inum, offset, notary, message)
 	 * the time is already in the future should help out.
 	 */
 
-	/* ``vp'' might become expired by  thread_reschedule() .. */
-	if (vp->proc && vp->proc->vertex == vp)
-	  /* Pick next, but don't feed it (yet)! */
-	  pick_next_vertex(vp->proc, 0, 0);
-
-	thread_reschedule(vp->thread, retrytime, index);
+	if (vp->thread != NULL)
+	  thread_reschedule(vp->thread, retrytime, index);
 
 	return 1;
 }
