@@ -1,7 +1,7 @@
 /*
  *   mx_client_verify() -- subroutine for ZMailer smtpserver
  *
- *   By Matti Aarnio <mea@nic.funet.fi> 1997
+ *   By Matti Aarnio <mea@nic.funet.fi> 1997-1999
  */
 
 #include "hostenv.h"
@@ -45,7 +45,9 @@
 #include "mail.h"
 #include "zsyslog.h"
 #include "ta.h"
-/* #include "malloc.h" */
+#include "zmalloc.h"
+#include "libz.h"
+#include "libc.h"
 
 #if	defined(TRY_AGAIN) && defined(HAVE_RESOLVER)
 #define	BIND		/* Want BIND (named) nameserver support enabled */
@@ -68,6 +70,10 @@
 #endif	/* !BIND_VER */
 #endif	/* BIND */
 
+#ifndef	T_TXT
+# define T_TXT 16	/* Text strings */
+#endif
+
 extern int h_errno;
 extern int res_mkquery(), res_send(), dn_skipname(), dn_expand();
 
@@ -85,22 +91,26 @@ typedef union {
 } querybuf;
 
 
+#include "policytest.h"
+
 extern int debug;
+static char * txt_buf = NULL;
 
 static int
 dnsmxlookup(host, depth, mxmode, qtype)
 	const char *host;
 	int depth;
 	int mxmode;
+	int qtype;
 {
 	HEADER *hp;
 	msgdata *eom, *cp;
 	querybuf qbuf, answer;
 	msgdata buf[8192], realname[8192];
-	int qlen, n, i, j, ancount, qdcount, maxpref;
+	int qlen, n, i, ancount, qdcount, maxpref;
 	u_short type;
-	int saw_cname = 0;
-	int ttl;
+	int saw_cname = 0, had_mx_record = 0;
+	/* int ttl; */
 	struct addrinfo req, *ai;
 
 	if (depth == 0)
@@ -112,11 +122,11 @@ dnsmxlookup(host, depth, mxmode, qtype)
 
 	if (debug) {
 	  if (qtype == T_TXT)
-	    printf("TXT-lookup for domain: '%s'\n", host);
+	    printf("000- TXT-lookup for domain: '%s'\n", host);
 	  else if (mxmode)
-	    printf("MX-Verify: Look MX for host '%s'\n", host);
+	    printf("000- MX-Verify: Look MX for host '%s'\n", host);
 	  else
-	    printf("DNS-Verify: Look MX, or Addr for host '%s'\n", host);
+	    printf("000- DNS-Verify: Look MX, or Addr for host '%s'\n", host);
 	}
 
 	qlen = res_mkquery(QUERY, host, C_IN, qtype, NULL, 0, NULL,
@@ -148,21 +158,10 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	    return -EX_NOHOST;
 	  case SERVFAIL:
 	    return -EX_TEMPFAIL;
-#ifdef OLDJEEVES
-	    /*
-	     * Jeeves (TOPS-20 server) still does not
-	     * support MX records.  For the time being,
-	     * we must accept FORMERRs as the same as
-	     * NOERROR.
-	     */
-	  case FORMERR:
-#endif
 	  case NOERROR:
 	    goto perhaps_address_record;
 
-#ifndef OLDJEEVES
 	  case FORMERR:
-#endif
 	  case NOTIMP:
 	  case REFUSED:
 	    return -EX_NOPERM;
@@ -228,27 +227,27 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	    i = _getaddrinfo_((const char*)buf, "0", &req, &ai,
 			      debug ? stdout : NULL);
 	    if (debug)
-	      printf("  getaddrinfo('%s','0') -> r=%d, ai=%p\n",buf,i,ai);
+	      printf("000-  getaddrinfo('%s','0') -> r=%d, ai=%p\n",buf,i,ai);
 #endif
 
 	    if (debug)
-	      printf("  getaddrinfo('%s') yields %d\n", buf, i);
+	      printf("000-  getaddrinfo('%s') yields %d\n", buf, i);
 	    
 	    if (i != 0)
 	      continue;		/* Well well.. spurious! */
 
 	    if (!mxmode) /* Accept if found ANYTHING! */ {
-	      if (debug) printf("  ... accepted!\n");
+	      if (debug) printf("000-  ... accepted!\n");
 	      freeaddrinfo(ai);
 	      return 1;
 	    }
 	  
 	    {
 	      struct addrinfo *ai2 = ai;
-	      int i = 1, rc;
+	      int k = 1, rc;
 	    
-	      for ( ; ai2 != NULL; ai2 = ai2->ai_next, ++i ) {
-#if 0
+	      for ( ; ai2 != NULL; ai2 = ai2->ai_next, ++k ) {
+#if 1
 		if (debug) {
 		  struct sockaddr * sa = ai2->ai_addr;
 		  char buf[60];
@@ -270,23 +269,28 @@ dnsmxlookup(host, depth, mxmode, qtype)
 		rc = matchmyaddress(ai2->ai_addr);
 		if (rc == 1) {
 		  if (debug)
-		    printf("  ADDRESS MATCH!\n");
+		    printf("000-   ADDRESS MATCH!\n");
 		  freeaddrinfo(ai);
 		  return 1; /* Found a match! */
 		} else
 		  if (debug)
-		    printf("  matchmyaddress() yields: %d\n", rc);
+		    printf("000-   matchmyaddress() yields: %d\n", rc);
 	      }
 	      if (debug)
-		printf("  No address match among %d address!\n",i);
+		printf("000-   No address match among %d address!\n",i);
 	    }
 	    freeaddrinfo(ai);
-#if 0
-	    return 2; /* We have SOME date, but no match on ourselves! */
-#endif
+	    had_mx_record = 1;
 	  } /* ===== END OF MX DATA PROCESING ========= */
 
 	  if (type == T_TXT) {
+	    int len = (*cp++) & 0xFF; /* 0..255 chars */
+	    if (txt_buf != NULL)
+	      free(txt_buf);
+	    txt_buf = emalloc(len+1);
+	    memcpy(txt_buf, cp, len);
+	    txt_buf[len] = '\0';
+	    return 1; /* OK! */
 	  }
 
 	} /* ===== END OF DNS ANSWER PROCESSING ======= */
@@ -295,8 +299,11 @@ dnsmxlookup(host, depth, mxmode, qtype)
 	if (saw_cname)
 	  return dnsmxlookup((void *)realname, depth+1, mxmode, qtype);
 
+	if (had_mx_record && mxmode)
+	    return 2; /* We have SOME date, but no match on ourselves! */
+
 perhaps_address_record:
-	if (qtype == T_MX && !mxmode) {
+	if (qtype == T_MX) {
 	  /* No MX, perhaps A ? */
 	  memset(&req, 0, sizeof(req));
 	  req.ai_socktype = SOCK_STREAM;
@@ -313,22 +320,33 @@ perhaps_address_record:
 	  i = _getaddrinfo_((const char*)host, "0", &req, &ai, debug ? stdout : NULL);
 #endif
 	  if (debug)
-	    printf("  getaddrinfo('%s','0') -> r=%d, ai=%p\n",host,i,ai);
+	    printf("000-   perhaps A? getaddrinfo('%s','0') -> r=%d, ai=%p\n",host,i,ai);
 	  if (i != 0) /* Found nothing! */
-	    return 0;
-#if 0
+	    return i;
+
+	  i = matchmyaddresses(ai);
+#if 1
 	  /* With this we can refuse to accept any message with
 	     source domain pointing back to loopback ! */
-	  if (matchmyaddresses(ai) == 2) {
+	  if (i == 2) {
 	    /* Loopback ! */
 	    freeaddrinfo(ai);
 	    return 0;
 	  }
 #endif
+	  if (i == 0 && mxmode) {
+	    freeaddrinfo(ai);
+	    return 2; /* Didn't find our local address in client-MX-mode */
+	  }
+
 	  freeaddrinfo(ai);
-	  return 1; /* Found SOME address */
+	  return 1; /* Found any address, or in client-MX-mode,
+		       a local address! */
 	}
 
+	if (mxmode)
+	  return 2; /* Not found, had no MX data either */
+	
 	return 0; /* Not found! */
 }
 
@@ -340,7 +358,7 @@ perhaps_address_record:
 
 int mx_client_verify(retmode, domain, alen)
      int retmode, alen;
-     char *domain;
+     const char *domain;
 {
 	char hbuf[2000];
 	int rc;
@@ -351,10 +369,13 @@ int mx_client_verify(retmode, domain, alen)
 	strncpy(hbuf, domain, alen);
 	hbuf[alen] = 0; /* Chop off the trailers from the name */
 
-	rc = dnsmxlookup(hbuf, 1, 1, T_MX);
+	rc = dnsmxlookup(hbuf, 0, 1, T_MX);
 
 	if (rc == 1) return 0; /* Found! */
 
+	if (rc == -EX_TEMPFAIL) {
+	  return -104;
+	}
 	if (retmode == '+') {
 	  if (rc == -EX_NOHOST ||
 	      rc == -EX_UNAVAILABLE)
@@ -371,9 +392,9 @@ int mx_client_verify(retmode, domain, alen)
 
 int sender_dns_verify(retmode, domain, alen)
      int retmode, alen;
-     char *domain;
+     const char *domain;
 {
-	char hbuf[2000];
+	char hbuf[500];
 	int rc;
 
 	if (alen >= sizeof(hbuf)-2)
@@ -382,13 +403,22 @@ int sender_dns_verify(retmode, domain, alen)
 	strncpy(hbuf, domain, alen);
 	hbuf[alen] = 0; /* Chop off the trailers from the name */
 
-	rc = dnsmxlookup(hbuf, 1, 0, T_MX);
+	rc = dnsmxlookup(hbuf, 0, 0, T_MX);
+
+	if (debug)
+	  printf("000- dnsmxlookup() did yield: %d, retmode='%c'\n",
+		 rc,retmode);
 
 	if (rc == 1) return 0; /* Found! */
 
+	if (rc == -EX_TEMPFAIL) {
+	  return -104;
+	}
 	if (retmode == '+') {
-	  if (rc == -EX_NOHOST ||
-	      rc == -EX_UNAVAILABLE)
+	  if (rc == -EX_NOHOST      ||
+	      rc == -EX_UNAVAILABLE ||
+	      rc == EAI_NODATA      ||
+	      rc == EAI_NONAME)
 	    return -2; /* Definitely hard errors */
 	  if (rc == 2)
 	    return -103;
@@ -402,44 +432,76 @@ int sender_dns_verify(retmode, domain, alen)
 
 int client_dns_verify(retmode, domain, alen)
      int retmode, alen;
-     char *domain;
+     const char *domain;
 {
 	return sender_dns_verify(retmode, domain, alen);
 }
 
-int rbl_dns_test(ipv4addr, msgp)
-     u_char *ipv4addr;
+int rbl_dns_test(ipaf, ipaddr, rbldomain, msgp)
+     const int ipaf;
+     const u_char *ipaddr;
+     char *rbldomain;
      char **msgp;
 {
-	char hbuf[2000], *s;
-	int rc;
+	char hbuf[2000], *s, *suf;
+	/* int hspc; */
 
-	sprintf (hbuf, "%d.%d.%d.%d.rbl.maps.vix.com.",
-		 ipv4addr[3], ipv4addr[2], ipv4addr[1], ipv4addr[0]);
+	if (ipaf == P_K_IPv4) {
+	  sprintf(hbuf, "%d.%d.%d.%d.",
+		  ipaddr[3], ipaddr[2], ipaddr[1], ipaddr[0]);
 
+	} else { /* Ok, the other variant is IPv6 ... */
 
-	if (debug)
-	  printf("looking up DNS object: %s\n", hbuf);
+	  int i;
+	  for (i = 15; i >= 0; --i) {
+	    sprintf(hbuf + ((15-i) << 2),
+		    "%x.%x.", ipaddr[i] & 0x0F, (ipaddr[i] >> 4) & 0x0F);
+	  }
+	  strcpy(hbuf+32,"ip6."); /* Fixed length of hex nybbles */
+	}
 
-	/* XX: Should verify that the named object has A record: 127.0.0.2 */
+	suf = hbuf + strlen(hbuf);
+	/* hspc = sizeof(hbuf)-strlen(hbuf)-1; */
 
-	if (gethostbyname(hbuf) != NULL) {
+	while (*rbldomain) {
+	  /* "rbldomain" is possibly a COLON-demarked set of
+	     domain names:  rbl.maps.vix.com:dul.maps.vix.com
+	     which isn't so easy to read, but ... */
+	  s = strchr(rbldomain, ':');
+	  if (s) *s = 0;
+	  if (strcmp(rbldomain,"+") == 0)
+	    strcpy (suf, "rbl.maps.vix.com");
+	  else
+	    strcpy (suf, rbldomain);
 
-	/* XX: Could - perhaps - use:  dnsmxlookup(hbuf, 0, 0, T_TXT) ?
-	   QQ: How to return the TXT entry ? */
-
-	  sprintf (hbuf, "Blackholed - see <URL:http://maps.vix.com/cgi-bin/lookup?%d.%d.%d.%d>",
-		   ipv4addr[3], ipv4addr[2], ipv4addr[1], ipv4addr[0]);
-
-	  if (*msgp != NULL)
-	    free(*msgp);
-
-	  *msgp = strdup(hbuf);
+	  if (s) {
+	    *s = ':';
+	    rbldomain = s+1;
+	  } else {
+	    rbldomain += strlen(rbldomain);
+	  }
 
 	  if (debug)
-	    printf("looking up DNS object: %s\n", hbuf);
+	    printf("000- looking up DNS A object: %s\n", hbuf);
 
-	  return -1;
+	  type(NULL,0,NULL, "Looking up DNS A object: %s", hbuf);
+
+	  if (gethostbyname(hbuf) != NULL) {
+	    /* XX: Should verify that the named object has A record: 127.0.0.2
+	       D'uh.. alternate dataset has A record: 127.0.0.3 */
+
+	    /* Ok, then lookup for the TXT entry too! */
+	    if (debug)
+	      printf("000- looking up DNS TXT object: %s\n", hbuf);
+
+	    if (dnsmxlookup(hbuf, 0, 0, T_TXT) == 1) {
+	      if (*msgp != NULL)
+		free(*msgp);
+	      *msgp = strdup(txt_buf);
+	    }
+	    return -1;
+	  }
 	}
+
 	return 0;
 }
