@@ -113,6 +113,8 @@ int X_settrrc = 9;
 #endif				/* USE_TRANSLATION */
 int strict_protocol = 0;
 int mustexit = 0;
+int configuration_ok = 0;
+int gotalarm;
 
 char logtag[16];
 
@@ -133,6 +135,7 @@ const char *m454 = "4.5.4";
 const char *m471 = "4.7.1";
 const char *m513 = "5.1.3";
 const char *m517 = "5.1.7";
+const char *m530 = "5.3.0";
 const char *m534 = "5.3.4";
 const char *m540 = "5.4.0";
 const char *m543 = "5.4.3";
@@ -175,7 +178,7 @@ int MaxParallelConnections = 800; /* Total number of childs allowed */
 int MaxErrorRecipients = 10;	/* Max number of recipients for a message
 				   that has a "box" ( "<>" ) as its source
 				   address. */
-int percent_accept = 0;
+int percent_accept = -1;
 
 int maxloadavg = 999;		/* Maximum load-average that is tolerated
 				   with smtp-server actively receiving..
@@ -206,6 +209,7 @@ int auth_ok = 0;
 int ehlo_ok = 1;
 int etrn_ok = 1;
 int starttls_ok = 0;
+int msa_mode = 0;
 char *tls_cert_file = NULL;
 char *tls_key_file  = NULL;
 char *tls_CAfile    = NULL;
@@ -219,8 +223,16 @@ int log_rcvd_whoson = 0;
 int log_rcvd_ident  = 0;
 int log_rcvd_authuser = 0;
 int log_rcvd_tls_mode = 0;
-int log_rcvd_tls_ccert = 0;
+int log_rcvd_tls_peer = 0;
 int auth_login_without_tls = 0;
+char *smtpauth_via_pipe = NULL;
+Usockaddr bindaddr;
+int bindaddr_set    = 0;
+u_short bindport = 0;
+int bindport_set = 0;
+int use_tcpwrapper = 0;
+int tarpit_initial = 0;
+int tarpit_exponent = 0;
 
 #ifndef	IDENT_TIMEOUT
 #define	IDENT_TIMEOUT	5
@@ -293,7 +305,7 @@ int insecure;
 			progname, fname, strerror(errno));
 	} else {
 	    logfp = fdopen(fd, "a");
-	    /* XX: */ setvbuf(logfp, NULL, _IOLBF, BUFSIZ);
+	    setvbuf(logfp, NULL, _IOFBF, BUFSIZ);
 	    /* Line-buffered */
 	}
 #ifndef HAVE_ALLOCA
@@ -304,7 +316,8 @@ int insecure;
 	logfp = NULL;
 }
 
-extern int main __((int, char **));
+int main __((int, char **));
+
 int main(argc, argv)
 int argc;
 char **argv;
@@ -312,17 +325,21 @@ char **argv;
     int inetd, errflg, raddrlen, s, msgfd, version, i;
     const char *mailshare;
     char path[1024];
-    u_short port = 0;
-    int port_set = 0;
     int force_ipv4 = 0;
     int localsocksize;
     char *cfgpath = NULL;
+    char *pidfile = PID_SMTPSERVER;
+    int pidfile_set = 0;
     SmtpState SS;
     int childpid, sameipcount, childcnt;
     time_t now;
+    char *t, *syslogflg;
 
 
-    char *t, *syslogflg = getzenv("SYSLOGFLG");
+    setvbuf(stdout, NULL, _IOFBF, 8192);
+    setvbuf(stderr, NULL, _IOLBF, 8192);
+
+    syslogflg = getzenv("SYSLOGFLG");
     if (syslogflg == NULL)
       syslogflg = "";
     t = syslogflg;
@@ -330,9 +347,10 @@ char **argv;
       if (*t == 's' || *t == 'S')
 	break;
     }
-    smtp_syslog = (*t != '\0');
+    smtp_syslog = *t;
 
     memset(&SS, 0, sizeof(SS));
+    memset(&bindaddr, 0, sizeof(bindaddr));
     SS.mfp = NULL;
     SS.style = "ve";
     SS.state = Hello;
@@ -363,15 +381,15 @@ char **argv;
 #ifndef __STDC__
 #if defined(AF_INET6) && defined(INET6)
 #ifdef USE_TRANSLATION
-		       "?46aBC:d:ighl:np:L:M:P:R:s:S:VvwX8"
+		       "?46aBC:d:ighl:np:I:L:M:P:R:s:S:VvwX8"
 #else /* xlate */
-		       "?46aBC:d:ighl:np:L:M:P:R:s:S:Vvw"
+		       "?46aBC:d:ighl:np:I:L:M:P:R:s:S:Vvw"
 #endif /* xlate */
 #else /* INET6 */
 #ifdef USE_TRANSLATION
-		       "?4aBC:d:ighl:np:L:M:P:R:s:S:VvwX8"
+		       "?4aBC:d:ighl:np:I:L:M:P:R:s:S:VvwX8"
 #else
-		       "?4aBC:d:ighl:np:L:M:P:R:s:S:Vvw"
+		       "?4aBC:d:ighl:np:I:L:M:P:R:s:S:Vvw"
 #endif /* xlate */
 #endif /* INET6 */
 #else /* __STDC__ */
@@ -382,7 +400,7 @@ char **argv;
 #endif
 		       "aBC:d:ighl:n"
 		       "p:"
-		       "L:M:P:R:s:S:Vvw"
+		       "I:L:M:P:R:s:S:Vvw"
 #ifdef USE_TRANSLATION
 		       "X8"
 #endif /* USE_TRANSLATION */
@@ -451,14 +469,18 @@ char **argv;
 	    else
 	      style = strdup(optarg);
 	    break;
+	case 'I':		/* PID file */
+	    pidfile = optarg;
+	    pidfile_set = 1;
+	    break;
 	case 'L':		/* Max LoadAverage */
 	    maxloadavg = atoi(optarg);
 	    if (maxloadavg < 1)
 		maxloadavg = 10;	/* Humph.. */
 	    break;
 	case 'p':
-	    port = htons(atoi(optarg));
-	    port_set = 1;
+	    bindport = htons(atoi(optarg));
+	    bindport_set = 1;
 	    break;
 	case 'R':		/* router binary used for verification */
 	    routerprog = strdup(optarg);
@@ -506,7 +528,8 @@ char **argv;
 		"Usage: %s [-46aBignVvw]\
  [-C cfgfile] [-s xx] [-L maxLoadAvg]\
  [-M SMTPmaxsize] [-R rtrprog] [-p port#]\
- [-P postoffice] [-l logfile] [-S 'local'|'remote']\n"
+ [-P postoffice] [-l logfile] [-S 'local'|'remote']\
+ [-I pidfile]\n"
 #else /* __STDC__ */
 		"Usage: %s [-4"
 #if defined(AF_INET6) && defined(INET6)
@@ -518,7 +541,8 @@ char **argv;
 #endif
 		"] [-C cfgfile] [-s xx] [-L maxLoadAvg]"
 		" [-M SMTPmaxsize] [-R rtrprog] [-p port#]"
-		" [-P postoffice] [-l logfile] [-S 'local'|'remote']\n"
+		" [-P postoffice] [-l logfile] [-S 'local'|'remote']"
+		" [-I pidfile]\n"
 #endif /* __STDC__ */
 		, progname);
 	exit(1);
@@ -556,15 +580,26 @@ char **argv;
       memset(&SS.raddr, 0, raddrlen);
       if (getpeername(SS.inputfd, (struct sockaddr *) &SS.raddr, &raddrlen))
 	netconnected_flg = 0;
-      else
+      else {
+	/* Got a peer name (it is a socket) */
 	netconnected_flg = 1;
+	if (SS.raddr.v4.sin_family != AF_INET
+#ifdef AF_INET6
+	    && SS.raddr.v4.sin_family != AF_INET6
+#endif
+	    )
+	  /* well, but somebody uses socketpair(2)  which is
+	     an AF_UNIX thing and sort of full-duplex pipe(2)... */
+	  netconnected_flg = 0;
+      }
 
       strcpy(SS.rhostname, "stdin");
       SS.rport = -1;
       SS.ihostaddr[0] = '\0';
       sprintf(SS.ident_username, "uid#%d@localhost", (int)getuid());
 
-      s_setup(&SS, FILENO(stdin), stdout);
+      /* INTERACTIVE */
+      s_setup(&SS, FILENO(stdin), FILENO(stdout));
       smtpserver(&SS, 0);
 
     } else
@@ -604,9 +639,12 @@ char **argv;
 	}
 	zopenlog("smtpserver", LOG_PID, LOG_MAIL);
 
-	s_setup(&SS, FILENO(stdin), stdout);
+	if (netconnected_flg)
+	  s_setup(&SS, FILENO(stdin), FILENO(stdin));
+	else
+	  s_setup(&SS, FILENO(stdin), FILENO(stdout));
 
-	if (ident_flag != 0)
+	if (ident_flag != 0 && !daemon_flg)
 	    setrfc1413ident(&SS);
 	else
 	    strcpy(SS.ident_username, "IDENT-NOT-QUERIED");
@@ -650,9 +688,9 @@ char **argv;
 	if (postoffice == NULL
 	    && (postoffice = getzenv("POSTOFFICE")) == NULL)
 	  postoffice = POSTOFFICE;
-	if (!port_set) {
+	if (pidfile_set || (!bindport_set && !bindaddr_set)) {
 	  /* Kill possible previous smtpservers now! */
-	  if (killprevious(SIGTERM, PID_SMTPSERVER) != 0) {
+	  if (killprevious(SIGTERM, pidfile) != 0) {
 	    fprintf(stderr,
 		    "%s: Can't write my pidfile!  Disk full ?\n",
 		    progname);
@@ -669,28 +707,28 @@ char **argv;
 	   If we are not explicitely told to use IPv6 only, we will try
 	   here to use IPv6, and if successfull, register it!  */
 	if (!use_ipv6 && !force_ipv4) {
-	  s = socket(AF_INET6, SOCK_STREAM, 0 /* IPPROTO_IPV6 */ );
+	  s = socket(PF_INET6, SOCK_STREAM, 0 /* IPPROTO_IPV6 */ );
 	  if (s >= 0) {
 	    use_ipv6 = 1;	/* We can do it! */
 	    close(s);
 	  }
 	}
 	if (use_ipv6) {
-	  s = socket(AF_INET6, SOCK_STREAM, 0 /* IPPROTO_IPV6 */ );
+	  s = socket(PF_INET6, SOCK_STREAM, 0 /* IPPROTO_IPV6 */ );
 #if 0
 	  if (s < 0) {	/* Fallback to the IPv4 mode .. */
-	    s = socket(AF_INET, SOCK_STREAM, 0 /* IPPROTO_IP   */ );
+	    s = socket(PF_INET, SOCK_STREAM, 0 /* IPPROTO_IP   */ );
 	    use_ipv6 = 0;
 	  }
 #endif
 	} else
-	  s = socket(AF_INET, SOCK_STREAM, 0 /* IPPROTO_IP   */ );
+	  s = socket(PF_INET, SOCK_STREAM, 0 /* IPPROTO_IP   */ );
 #else
-	s = socket(AF_INET, SOCK_STREAM, 0);
+	s = socket(PF_INET, SOCK_STREAM, 0);
 #endif
 	if (s < 0) {
 	  fprintf(stderr,
-		  "%s: socket(AF_INET%s, SOCK_STREAM): %s\n",
+		  "%s: socket(PF_INET%s, SOCK_STREAM): %s\n",
 		  progname, (use_ipv6 ? "6" : ""), strerror(errno));
 	  exit(1);
 	}
@@ -730,17 +768,17 @@ char **argv;
 	    exit(1);
 	  }
 #endif
-	if (port <= 0) {
+	if (bindport <= 0) {
 	  struct servent *service;
 #ifdef	IPPORT_SMTP
-	  port = htons(IPPORT_SMTP);
+	  bindport = htons(IPPORT_SMTP);
 #endif				/* !IPPORT_SMTP */
 	  if ((service = getservbyname("smtp", "tcp")) == NULL) {
 	    fprintf(stderr,
 		    "%s: no SMTP service entry, using default\n",
 		    progname);
 	  } else
-	    port = service->s_port;
+	    bindport = service->s_port;
 	}
 #if defined(AF_INET6) && defined(INET6)
 	if (use_ipv6) {
@@ -749,8 +787,10 @@ char **argv;
 	  memset(&si6, 0, sizeof(si6));
 	  si6.sin6_family = AF_INET6;
 	  si6.sin6_flowinfo = 0;
-	  si6.sin6_port = port;
+	  si6.sin6_port = bindport;
 	  memcpy( &si6.sin6_addr, zin6addrany, 16 );
+	  if (bindaddr_set && bindaddr.v6.sin6_family == AF_INET6)
+	    memcpy(&si6.sin6_addr, &bindaddr.v6.sin6_addr, 16);
 
 	  i = bind(s, (struct sockaddr *) &si6, sizeof si6);
 	  if (i < 0) {
@@ -766,7 +806,9 @@ char **argv;
 	    memset(&si4, 0, sizeof(si4));
 	    si4.sin_family = AF_INET;
 	    si4.sin_addr.s_addr = INADDR_ANY;
-	    si4.sin_port = port;
+	    si4.sin_port = bindport;
+	    if (bindaddr_set && bindaddr.v4.sin_family == AF_INET)
+	      memcpy(&si4.sin_addr, &bindaddr.v4.sin_addr, 4);
 
 	    i = bind(s, (struct sockaddr *) &si4, sizeof si4);
 	    if (i < 0) {
@@ -808,10 +850,10 @@ char **argv;
 	dup(0);
 	dup(0);			/* fd's 0, 1, 2 are in use again.. */
 
-	if (!port_set || port != htons(25)) {
+	if (pidfile_set || (!bindport_set && !bindaddr_set)) {
 	  sleep(3); /* Give a moment to possible previous server
 		       to die away... */
-	  killprevious(0, PID_SMTPSERVER);	/* deposit pid */
+	  killprevious(0, pidfile);	/* deposit pid */
 	}
 #if 1
 	pid = getpid();
@@ -958,7 +1000,7 @@ char **argv;
 	    }
 	    zopenlog("smtpserver", LOG_PID, LOG_MAIL);
 
-	    s_setup(&SS, msgfd, stdout);
+	    s_setup(&SS, msgfd, msgfd);
 
 	    if (ident_flag != 0)
 	      setrfc1413ident(&SS);
@@ -1049,7 +1091,7 @@ char **argv;
 	  }
 	}
 	/* Stand-alone server, kill the pidfile at the exit! */
-	killpidfile(PID_SMTPSERVER);
+	killpidfile(pidfile);
 	openlogfp(&SS, daemon_flg);
 	if (logfp != NULL) {
 	  char *cp;
@@ -1160,10 +1202,11 @@ static RETSIGTYPE
  timedout(sig)
 int sig;
 {
-    /* Return to the smtpserver's mail-program.
+    /* Return to the smtpserver's main-program.
        We are commiting a suicide, but we need
        data that exists only in that context... */
-    longjmp(jmpalarm, 1);
+    gotalarm = 1;
+    siglongjmp(jmpalarm, 1);
     _exit(253);			/* We did return ?!?! Boo!! */
 }
 
@@ -1261,7 +1304,7 @@ SmtpState *SS;
 }
 
 int s_getc(SS)
-SmtpState *SS;
+     SmtpState *SS;
 {
     int rc = 0;
 
@@ -1272,9 +1315,9 @@ SmtpState *SS;
     redo:
 	rc = Z_read(SS, SS->s_buffer, sizeof(SS->s_buffer));
 	if (rc < 0) {
-	  goto redo; /* XX: ??? some input-problem circumvention problem ?? */
 	  if (errno == EINTR || errno == EAGAIN)
-	    goto redo;
+	    if (!gotalarm)
+	      goto redo;
 	  /* Other results are serious errors -- maybe */
 	  SS->s_status = EOF;
 	  return EOF;
@@ -1306,7 +1349,7 @@ SmtpState *SS;
 #else
 	fcntl(SS->inputfd, F_SETFL, flags | FNDELAY);
 #endif
-	SS->s_bufread = read(SS->inputfd, SS->s_buffer, sizeof(SS->s_buffer));
+	SS->s_bufread = Z_read(SS, SS->s_buffer, sizeof(SS->s_buffer));
 	fcntl(SS->inputfd, F_SETFL, flags);
 	if (SS->s_bufread > 0)
 	    return SS->s_bufread;
@@ -1323,9 +1366,7 @@ char *buf, *cop, *cp;
 int buflen, *rcp;
 {
 	int c, co = -1;
-	int i, rc;
-
-	rc = -1;
+	int i = -1, rc = -1;
 
 	if (!pipeliningok || !s_hasinput(SS))
 	    typeflush(SS);
@@ -1339,11 +1380,10 @@ int buflen, *rcp;
 	alarm(SMTP_COMMAND_ALARM_IVAL);
 
 	/* Our own  fgets() -- gets also NULs, flags illegals.. */
-	i = 0;
 	--buflen;
 	while ((c = s_getc(SS)) != EOF && i < buflen) {
 	    if (c == '\n') {
-		buf[i++] = c;
+		buf[++i] = c;
 		break;
 	    } else if (co == '\r' && rc < 0)
 		rc = i;		/* Spurious CR on the input.. */
@@ -1355,10 +1395,10 @@ int buflen, *rcp;
 	    if (c != '\r' && c != '\t' &&
 		(c < 32 || c == 127) && rc < 0)
 		rc = i;
-	    buf[i++] = c;
+	    buf[++i] = c;
 	    co = c;
 	}
-	buf[i] = '\0';
+	buf[++i] = '\0';
 	alarm(0);		/* Cancel the alarm */
 
 	if (c == EOF && i == 0) {
@@ -1369,22 +1409,34 @@ int buflen, *rcp;
 	}
 
 	/* Zap the ending newline */
-	if (c  == '\n') buf[--i] = '\0';
+	if (c  == '\n') buf[i-1] = '\0';
 	/* Zap the possible preceeding \r */
-	if (co == '\r') buf[--i] = '\0';
+	if (co == '\r') buf[i-2] = '\0';
 
 	*cop = co;
 	*cp  = c;
 	*rcp = rc;
+
+	if (i >= buflen && c != EOF && c != '\n') {
+	  /* Huh, oversized input line ?? */
+	  while ((c = s_getc(SS)) != EOF && c != '\n')
+	    ;
+	  /* Input eaten until a NEWLINE, or EOF occurred at the input. */
+	}
+
 	return i;
 }
 
 
-void s_setup(SS, infd, outfp)
+void s_setup(SS, infd, outfd)
 SmtpState *SS;
 int infd;
-FILE *outfp;
 {
+
+    FILE *outfp = fdopen(outfd,"w");
+
+    setvbuf(outfp, NULL, _IOFBF, 8192);
+
     SS->inputfd = infd;
     SS->outfp = outfp;
     SS->s_status = 0;
@@ -1414,6 +1466,8 @@ int insecure;
 #endif
     SS->VerboseCommand = 0;
 
+    SS->tarpit = tarpit_initial;
+
     stashmyaddresses(NULL);
 
     pid = getpid();
@@ -1488,7 +1542,7 @@ int insecure;
 #endif				/* USE_TRANSLATION */
     }
 #ifdef HAVE_WHOSON_H
-    if (do_whoson) {
+    if (do_whoson && netconnected_flg) {
 	char buf[64];
 	buf[0]='\0';
 	if (SS->raddr.v4.sin_family == AF_INET) {
@@ -1512,6 +1566,10 @@ int insecure;
 #else
     policystatus     = policyinit(&policydb, &SS->policystate, 0);
 #endif
+    if (!netconnected_flg && policystatus < 0)
+      policystatus = 0; /* For internal - non-net-connected - mode
+			   lack of PolicyDB is no problem at all.. */
+
     SS->policyresult = policytestaddr(policydb, &SS->policystate,
 				      POLICY_SOURCEADDR,
 				      (void *) &SS->raddr);
@@ -1527,7 +1585,8 @@ int insecure;
     zopenlog("smtpserver", LOG_PID, LOG_MAIL);
 
 #ifdef HAVE_TCPD_H		/* TCP-Wrapper code */
-    if (wantconn(SS->inputfd, "smtp-receiver") == 0) {
+    if (use_tcpwrapper && netconnected_flg &&
+	wantconn(SS->inputfd, "smtp-receiver") == 0) {
 	zsyslog((LOG_WARNING, "refusing connection from %s:%d/%s",
 		 SS->rhostname, SS->rport, SS->ident_username));
 	type(SS, 421, NULL, "%s ZMailer Server %s WILL NOT TALK TO YOU at %s",
@@ -1540,14 +1599,15 @@ int insecure;
 
     if (SS->reject_net) {
 	char *msg = policymsg(policydb, &SS->policystate);
+	smtp_tarpit(SS);
 	if (msg != NULL) {
-	  type(SS, -553, NULL, "%s", msg);
+	  type(SS, -550, NULL, "%s", msg);
 	} else {
-	  type(SS, -553, NULL, "%s - You are on our reject-IP-address -list, GO AWAY!",
+	  type(SS, -550, NULL, "%s - You are on our reject-IP-address -list, GO AWAY!",
 	       SS->myhostname);
 	}
-	type(SS, -553, NULL, "If you feel we mistreat you, do contact us.");
-	type(SS, 553, NULL, "Ask HELP for our contact information.");
+	type(SS, -550, NULL, "If you feel we mistreat you, do contact us.");
+	type(SS, 550, NULL, "Ask HELP for our contact information.");
     } else
 #ifdef USE_TRANSLATION
 	if (hdr220lines[0] == NULL) {
@@ -1600,7 +1660,8 @@ int insecure;
 	int i;
 
 	i = s_gets(SS, buf, sizeof(buf), &rc, &co, &c );
-	if (i == 0)	/* EOF ??? */
+
+	if (i <= 0)	/* EOF ??? */
 	  break;
 
 	eobuf = &buf[i-1];	/* Buf end ptr.. */
@@ -1662,6 +1723,7 @@ int insecure;
 	for (cp = buf; isascii(*cp) && isalpha(*cp); ++cp)
 	    continue;
 	if (cp > buf + 8) {	/* "DEBUG" is longest of them.. */
+	    smtp_tarpit(SS);
 	    type(SS, 550, m552, "Syntax error");
 	    typeflush(SS);
 	    continue;
@@ -1679,6 +1741,8 @@ int insecure;
 
 	unknown_command:
 
+	    smtp_tarpit(SS);
+
 	    type(SS, 550, m552, "Unknown command '%s'", buf);
 	    zsyslog((LOG_WARNING,
 		     "unknown SMTP command '%s' from %s/%d",
@@ -1704,8 +1768,18 @@ int insecure;
 	if (SS->carp->cmd == BData     && ! chunkingok)
 	  goto unknown_command;
 
+	/* Lack of configuration is problem only with network connections */
+	if (netconnected_flg && !configuration_ok) {
+	  smtp_tarpit(SS);
+	  type(SS, -400, "4.7.0", "This SMTP server has not been configured!");
+	  typeflush(SS);
+	  zsyslog((LOG_EMERG, "smtpserver configuration missing!"));
+	  sleep(20);
+	  continue;
+	}
 	if (policystatus != 0 &&
 	    SS->carp->cmd != Quit && SS->carp->cmd != Help) {
+	  smtp_tarpit(SS);
 	  type(SS, -400, "4.7.0", "Policy database problem, code=%d", policystatus);
 	  type(SS,  400, "4.7.0", "With 'HELP' command you can get our contact information.");
 	  typeflush(SS);
@@ -1714,13 +1788,35 @@ int insecure;
 	  continue;
 	}
 	if (SS->reject_net && SS->carp->cmd != Quit && SS->carp->cmd != Help) {
-	    type(SS, -553, NULL, "You are on our reject-IP-address -list, GO AWAY!");
-	    type(SS, -553, NULL, "If you feel we mistreat you, do contact us.");
-	    type(SS, 553, NULL, "With 'HELP' command you can get out contact information.");
+	    smtp_tarpit(SS);
+	    type(SS, -550, NULL, "You are on our reject-IP-address -list, GO AWAY!");
+	    type(SS, -550, NULL, "If you feel we mistreat you, do contact us.");
+	    type(SS, 550, NULL, "With 'HELP' command you can get out contact information.");
 	    typeflush(SS);
 	    continue;
 	}
 
+
+	if ( msa_mode && SS->authuser == NULL &&
+	     !(SS->policystate.always_accept || SS->policystate.full_trust) ) {
+	  switch (SS->carp->cmd) {
+#ifdef HAVE_OPENSSL
+	  case StartTLS:
+#endif /* - HAVE_OPENSSL */
+	  case Auth:
+	  case Hello:
+	  case Hello2:
+	  case NoOp:
+	  case Reset:
+	  case Quit:
+	    break;
+	  default:
+	    type(SS, 530, m530, "Authentication required" );
+	    typeflush(SS);
+	    continue;
+	  }
+	}
+
 	switch (SS->carp->cmd) {
 	case Null:
 	    type(SS, 550, m550, "panic!");
@@ -1782,7 +1878,8 @@ int insecure;
 	    }
 	    if (SS->state != Hello)
 		SS->state = MailOrHello;
-	    type(SS, 250, m200, NULL);
+	    type(SS, 250, m200, "2.0.0 Reset processed, now waiting for MAIL command");
+	    SS->policyresult = 0; /* Clear this state too */
 	    typeflush(SS);
 	    break;
 	case Help:
@@ -2105,6 +2202,11 @@ type220headers(SS, identflg, xlatelang, curtime)
 {
     char *s, **hh = hdr220lines;
 
+    /* Below use of  fprintf()  for SS->outfp  channel is for
+       ensuring that  setvbuf( _IOFBF ) is honoured always.
+       It appears not to be so with  fputc() and putc(), it *may*
+       be so with  fputs()  -- Solaris 2.5.1 */
+
     for (; *hh ; ++hh) {
       char c = (hh[1] == NULL) ? ' ' : '-';
       
@@ -2128,30 +2230,30 @@ type220headers(SS, identflg, xlatelang, curtime)
 	  ++s;
 	  switch (*s) {
 	  case '%':
-	    putc('%',SS->outfp);
+	    fprintf(SS->outfp,"%%");
 	    if (logfp) putc('%',logfp);
 	    break;
 	  case 'H':
-	    fputs(SS->myhostname,SS->outfp);
+	    fprintf(SS->outfp,"%s",SS->myhostname);
 	    if (logfp) fputs(SS->myhostname,logfp);
 	    break;
 	  case 'I':
 	    if (identflg) {
-	      fputs("+IDENT",SS->outfp);
+	      fprintf(SS->outfp,"+IDENT");
 	      if (logfp) fputs("+IDENT",logfp);
 	    }
 	    break;
 	  case 'V':
-	    fputs(VersionNumb,SS->outfp);
+	    fprintf(SS->outfp,"%s",VersionNumb);
 	    if (logfp) fputs(VersionNumb,logfp);
 	    break;
 	  case 'T':
-	    fputs(curtime,SS->outfp);
+	    fprintf(SS->outfp,"%s",curtime);
 	    if (logfp) fputs(curtime,logfp);
 	    break;
 	  case 'X':
 	    if (!xlatelang) xlatelang = "";
-	    fputs(xlatelang,SS->outfp);
+	    fprintf(SS->outfp,"%s",xlatelang);
 	    if (logfp) fputs(xlatelang,logfp);
 	    break;
 	  default:
@@ -2159,14 +2261,13 @@ type220headers(SS, identflg, xlatelang, curtime)
 	    break;
 	  }
 	} else {
-	  putc(*s, SS->outfp);
+	  fprintf(SS->outfp,"%c",*s);
 	  if (logfp) putc(*s, logfp);
 	}
 	if (*s) ++s;
       }
-      putc('\r',SS->outfp);
-      putc('\n',SS->outfp);
-      if (logfp) putc('\n',logfp);
+      fprintf(SS->outfp,"\r\n");
+      if (logfp) fputc('\n',logfp);
     }
     fflush(SS->outfp);
     if (logfp) fflush(logfp);
@@ -2212,8 +2313,6 @@ va_dcl
 #endif
 
     s = inbuf + 3 + 1;
-    if (enhancedstatusok)
-      s += strlen(status);
 
     /* These are not always safe... but they should be ok
        if we are carrying  (char*)s or (int)s.. */
@@ -2227,8 +2326,10 @@ va_dcl
     if (multilinereplies) {
       if (enhancedstatusok) {
 	sprintf(buf, "%03d-%s ", abscode, status);
+	s += strlen(status) +1;
       } else { /* No status codes */
 	sprintf(buf, "%03d- ", abscode);
+	++s;
       }
       bp = buf + strlen(buf);
       while (s < rfc821_error_ptr && --maxcnt >= 0) {
@@ -2287,6 +2388,20 @@ SmtpState *SS;
 	strncpy(SS->ident_username, "IDENT-CALL-FAULT", MAXHOSTNAMELEN);
 }
 
+
+void smtp_tarpit(SS)
+     SmtpState *SS;
+{
+    if (SS->tarpit) {
+	if (SS->tarpit < 0 || SS->tarpit > 250)
+	    SS->tarpit = 250;
+	sleep(SS->tarpit);
+	SS->tarpit += (SS->tarpit * tarpit_exponent);
+    }
+}
+
+
+
 void header_to_mime(buf, lenptr, maxlen)
 char *buf;
 int *lenptr;
