@@ -32,7 +32,13 @@
 #define SMTPLINESIZE	8192
 
 #include "hostenv.h"
+
 #include <stdio.h>
+#ifndef FILE /* Some systems don't have this as a MACRO.. */
+# define FILE FILE
+#endif
+#include <sfio.h>
+
 #include "zmalloc.h"
 #include <sys/types.h>
 #include <ctype.h>
@@ -164,6 +170,7 @@ typedef struct {
     struct policystate policystate;
     int  policyresult, reject_net;
     int  postmasteronly;
+    int  tarpit;
     int  rport;
     char ihostaddr[sizeof("[ipv6.ffff:ffff:ffff:ffff:ffff:ffff:255.255.255.255]") + 8];
     Usockaddr raddr;
@@ -177,9 +184,9 @@ typedef struct {
     /* space, how much stuff in, where the output cursor is */
 #endif /* - HAVE_OPENSSL */
     char *tls_cipher_info;
-    char *tls_ccert_subject;
-    char *tls_ccert_issuer;
-    char *tls_ccert_fingerprint;
+    char *tls_peer_subject;
+    char *tls_peer_issuer;
+    char *tls_peer_fingerprint;
 
     int  s_bufread;
     int  s_readout;
@@ -220,6 +227,8 @@ extern char logtag[];
 extern long availspace;
 extern long minimum_availspace;
 extern long maxsize;
+extern int tarpit_initial;
+extern int tarpit_exponent;
 extern int MaxErrorRecipients;
 extern int TcpRcvBufferSize;
 extern int TcpXmitBufferSize;
@@ -242,15 +251,23 @@ extern int auth_ok;
 extern int ehlo_ok;
 extern int etrn_ok;
 extern int starttls_ok;
+extern int msa_mode;
 extern char *tls_cert_file, *tls_key_file, *tls_CAfile, *tls_CApath;
-extern int tls_loglevel, tls_enforce_tls, tls_ccert_vd;
-extern int tls_ask_cert, tls_req_cert;
+extern int tls_loglevel, tls_enforce_tls, tls_ccert_vd, tls_use_scache;
+extern int tls_ask_cert, tls_req_cert, tls_scache_timeout;
 extern int log_rcvd_whoson, log_rcvd_ident, log_rcvd_authuser;
-extern int log_rcvd_tls_mode, log_rcvd_tls_ccert;
+extern int log_rcvd_tls_mode, log_rcvd_tls_peer;
 extern int auth_login_without_tls;
+extern char *smtpauth_via_pipe;
 extern int strict_protocol;
 extern int rcptlimitcnt;
 extern int enable_router;
+extern int use_tcpwrapper;
+extern int configuration_ok;
+
+extern int bindaddr_set, bindport_set;
+extern u_short   bindport;
+extern Usockaddr bindaddr;
 
 extern const char *progname;
 extern int debug, skeptical, checkhelo, ident_flag, verbose;
@@ -308,6 +325,7 @@ extern const char *m513;
 extern const char *m515;
 extern const char *m517;
 extern const char *m518;
+extern const char *m530;
 extern const char *m534;
 extern const char *m540;
 extern const char *m543;
@@ -353,7 +371,7 @@ extern int kill __((pid_t, int));
 extern const char *rfc822atom __((const char *str));
 extern const char *xtext_string __((const char *str));
 
-extern void s_setup __((SmtpState * SS, int fd, FILE * fp));
+extern void s_setup __((SmtpState * SS, int infd, int outfd));
 extern int s_feof __((SmtpState * SS));
 extern int s_getc __((SmtpState * SS));
 extern int s_hasinput __((SmtpState * SS));
@@ -388,6 +406,7 @@ extern void smtp_expand __((SmtpState * SS, const char *buf, const char *cp));
 extern int  smtp_data   __((SmtpState * SS, const char *buf, const char *cp));
 extern int  smtp_bdata  __((SmtpState * SS, const char *buf, const char *cp));
 extern void add_to_toplevels __((char *str));
+extern void smtp_tarpit __((SmtpState * SS));
 
 extern void smtp_auth __((SmtpState * SS, const char *buf, const char *cp));
 
