@@ -9,7 +9,7 @@
 
 #else /* Not NDBM */
 
-#error Touch, supports only NDBM for now
+#error Tough, supports only NDBM for now
 
 #endif
 
