@@ -4,7 +4,6 @@
  */
 
 #include "hostenv.h"
-#include <stdio.h>
 #ifdef linux
 #define __USE_BSD 1
 #endif
@@ -13,16 +12,19 @@
 #include <sysexits.h>
 #include <sys/param.h>
 #include <sys/stat.h>
+
 #include "zmsignal.h"
-#include "malloc.h"
+#include "zmalloc.h"
 #include "zsyslog.h"
 
+#include "ta.h"
+
 #ifdef	HAVE_RESOLVER
 #include "netdb.h"
 #include <sys/socket.h>
 #include <netinet/in.h>
 #endif
-#include "ta.h"
+#include "libz.h"
 
 #if	defined(TRY_AGAIN) && defined(HAVE_RESOLVER)
 #define	BIND		/* Want BIND (named) nameserver support enabled */
@@ -128,20 +130,20 @@ extern char **environ;
 #ifndef strchr
 extern char *strchr(), *strrchr();
 #endif
-extern int   ranny();
 
 #ifdef	lint
 #undef	putc
 #define	putc	fputc
 #endif	/* lint */
 
+static char filename[MAXPATHLEN+8000];
+
 int
 main(argc, argv)
 	int argc;
 	char *argv[];
 {
 	const char *channel, *host;
-	char file[MAXPATHLEN+1];
 	int errflg, c, i;
 	struct ctldesc *dp;
 	RETSIGTYPE (*oldsig) __((int));
@@ -205,16 +207,16 @@ main(argc, argv)
 
 	  printf("#hungry\n");
 	  fflush(stdout);
-	  if (fgets(file, sizeof file, stdin) == NULL)
+	  if (fgets(filename, sizeof(filename), stdin) == NULL)
 	    break;
-	  if (strchr(file, '\n') == NULL) break; /* No ending '\n' !  Must
+	  if (strchr(filename, '\n') == NULL) break; /* No ending '\n' !  Must
 						    have been partial input! */
-	  if (strcmp(file, "#idle\n") == 0)
+	  if (strcmp(filename, "#idle\n") == 0)
 	    continue; /* Ah well, we can stay idle.. */
-	  if (emptyline(file, sizeof file))
+	  if (emptyline(filename, sizeof(filename)))
 	    break;
 
-	  s = strchr(file,'\t');
+	  s = strchr(filename,'\t');
 	  host = NULL;
 	  if (s != NULL) {
 	    /* Ignore the host part */
@@ -222,12 +224,12 @@ main(argc, argv)
 	    host = s;
 	  }
 
-	  dp = ctlopen(file, channel, host, &getout, NULL, NULL, NULL, NULL);
+	  dp = ctlopen(filename, channel, host, &getout, NULL, NULL, NULL, NULL);
 	  if (dp != NULL) {
 	    process(dp);
 	    ctlclose(dp);
 	  } else {
-	    printf("#resync %s\n",file);
+	    printf("#resync %s\n",filename);
 	    fflush(stdout);
 	  }
 	}
@@ -333,7 +335,8 @@ process(dp)
 	    }
 	  }
 
-	writeheaders(dp->recipients,mfp,"\n",0,0);
+	fwriteheaders(dp->recipients,mfp,"\n",0,0,NULL);
+	fprintf(mfp,"\n");
 
 	/* append message body itself */
 	while ((n = read(dp->msgfd, buf, sizeof buf)) > 0)
