@@ -28,7 +28,6 @@
  */
 
 #include "hostenv.h"
-#include <stdio.h>
 #include <ctype.h>
 #include <sys/stat.h>
 #include <sys/file.h>
@@ -37,23 +36,25 @@
 # include <fcntl.h>
 #endif
 #include <sysexits.h>
-#include "mail.h"
-#include "malloc.h"
-#include "libz.h"
+
 #include "ta.h"
 
-#ifdef	HAVE_MMAP
+#include "mail.h"
+#include "zmalloc.h"
+#include "libz.h"
+
+#if defined(HAVE_MMAP) && defined(TA_USE_MMAP)
 #include <sys/mman.h>
 #endif
 
-extern char *routermxes __((char *, struct address *));
+extern char *routermxes __((char *, struct taddress *));
 extern int errno;
 
 #ifndef strrchr
 extern char *strrchr();
 #endif
 
-static struct address *ctladdr __((char *cp));
+static struct taddress *ctladdr __((char *cp));
 
 
 #ifndef	MAXPATHLEN
@@ -66,10 +67,13 @@ ctlfree(dp,anyp)
 	struct ctldesc *dp;
 	void *anyp;
 {
-	unsigned long lowlim = (unsigned long) dp->contents;
-	unsigned long highlim = lowlim + dp->contentsize;
-	if ((unsigned long)anyp < lowlim ||
-	    (unsigned long)anyp > highlim)
+	void * lowlim  = (void *) dp->contents;
+	void * highlim = (void *)(((char*)lowlim) + dp->contentsize);
+#if 0
+	fprintf(stderr,"# ctlfree(%p) (%p,%p] @%p\n", anyp,
+		lowlim, highlim, __builtin_return_address(0));
+#endif
+	if (anyp < lowlim || anyp >= highlim)
 	  free(anyp);	/* It isn't within DP->CONTENTS data.. */
 }
 
@@ -79,18 +83,23 @@ ctlrealloc(dp,anyp,size)
 	void *anyp;
 	size_t size;
 {
-	unsigned long lowlim = (unsigned long) dp->contents;
-	unsigned long highlim = lowlim + dp->contentsize;
+	void * lowlim  = (void *) dp->contents;
+	void * highlim = (void *)(((char*)lowlim) + dp->contentsize);
 	void *anyp2;
+#if 0
+	fprintf(stderr,"# ctlrealloc(%p,%lu) (%p,%p] @%p\n", anyp, size,
+		lowlim, highlim, __builtin_return_address(0));
+#endif
+	/* If old one isn't our local thing, delete it! */
+	if (anyp < lowlim || anyp >= highlim)
+	  return realloc(anyp, size); /* realloc(); it isn't within DP->CONTENTS data.. */
 
-	if ((unsigned long)anyp < lowlim ||
-	    (unsigned long)anyp > highlim)
-	  return realloc(anyp,size);	/* It isn't within
-					   DP->CONTENTS data.. */
-	/* It is within the data, allocate a new storage.. */
+	/* Allocate a new storage.. */
 	anyp2 = (void*) malloc(size);
-	if (anyp2)
-	  memcpy(anyp2,anyp,size);
+	if (!anyp) return NULL;
+
+	memcpy(anyp2,anyp,size);
+
 	return anyp2;
 }
 
@@ -98,7 +107,7 @@ void
 ctlclose(dp)
 	struct ctldesc *dp;
 {
-	struct address *ap, *nextap;
+	struct taddress *ap, *nextap;
 	struct rcpt *rp, *nextrp;
 	char ***msghpp;
 
@@ -107,7 +116,7 @@ ctlclose(dp)
 	    continue;
 	  diagnostic(rp, EX_TEMPFAIL, 0, "address was left locked!!");
 	}
-#ifdef	HAVE_MMAP
+#if defined(HAVE_MMAP) && defined(TA_USE_MMAP)
 	if (dp->let_buffer != NULL)
 	  munmap((void*)dp->let_buffer, dp->let_end - dp->let_buffer);
 	dp->let_buffer = NULL;
@@ -119,15 +128,11 @@ ctlclose(dp)
 	  close(dp->ctlfd);
 	if (dp->msgfd >= 0)
 	  close(dp->msgfd);
-	if (dp->contents != NULL)
-	  free((void*)dp->contents);
-	if (dp->offset != NULL)
-	  free((char *)dp->offset);
 	for (ap = dp->senders; ap != NULL; ap = nextap) {
 	  nextap = ap->link;
 	  free((char *)ap);
 	}
-	dp->senders = NULL; /* XX: Excessive caution.. */
+	dp->senders = NULL;
 	for (rp = dp->recipients; rp != NULL; rp = nextrp) {
 	  nextrp = rp->next;
 	  ap = rp->addr;
@@ -136,7 +141,7 @@ ctlclose(dp)
 	  }
 	  free((char *)rp);
 	}
-	dp->recipients = NULL; /* XX: Excessive caution.. */
+	dp->recipients = NULL;
 	/* Free ALL dp->msgheader's, if they have been reallocated.
 	   Don't free on individual recipients, only on this global set.. */
 	for (msghpp = dp->msgheaders; msghpp &&  *msghpp; ++msghpp) {
@@ -146,7 +151,7 @@ ctlclose(dp)
 	  free(*msghpp);
 	}
 	free(dp->msgheaders);
-	dp->msgheaders = NULL; /* XX: Excessive caution.. */
+	dp->msgheaders = NULL;
 	for (msghpp = dp->msgheaderscvt; msghpp &&  *msghpp; ++msghpp) {
 	  char **msghp = *msghpp;
 	  for ( ; msghp && *msghp ; ++msghp )
@@ -154,17 +159,25 @@ ctlclose(dp)
 	  free(*msghpp);
 	}
 	free(dp->msgheaderscvt);
-	dp->msgheaders = NULL; /* XX: Excessive caution.. */
+	dp->msgheaderscvt = NULL;
+
+	if (dp->offset != NULL)
+	  free((void*)dp->offset);
+	dp->offset = NULL;
+
+	if (dp->contents != NULL)
+	  free((void*)dp->contents);
+	dp->contents = NULL;
 }
 
 
-static struct address *
+static struct taddress *
 ctladdr(cp)
 	char *cp;
 {
-	struct address *ap;
+	struct taddress *ap;
 
-	ap = (struct address *)emalloc(sizeof (struct address));
+	ap = (struct taddress *)malloc(sizeof (struct taddress));
 	if (ap == NULL)
 		return NULL;
 	ap->link = NULL;
@@ -202,7 +215,7 @@ ctladdr(cp)
 struct ctldesc *
 ctlopen(const char *file, const char *channel, const char *host,
 	int *exitflagp, int (*selectaddr)(const char *, const char *, void *),
-	void *saparam,  int (*matchrouter)(const char *, struct address *, void *),
+	void *saparam,  int (*matchrouter)(const char *, struct taddress *, void *),
 	void *mrparam)
 
 #else
@@ -213,24 +226,25 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 	int *exitflagp;
 	int (*selectaddr)  __((const char *, const char *, void *));
 	void *saparam;
-	int (*matchrouter) __((const char *, struct address *, void *));
+	int (*matchrouter) __((const char *, struct taddress *, void *));
 	void *mrparam;
 
 #endif
 {
 	register char *s, *contents;
-	char *mfpath;
+	char *mfpath, *delayslot;
 	int i, n;
-	struct address *ap;
+	struct taddress *ap;
 	struct rcpt *rp = NULL, *prevrp = NULL;
 	struct stat stbuf;
 	char ***msgheaders = NULL;
 	char ***msgheaderscvt = NULL;
-	int headers_cnt = 0;
-	int headers_spc = 0;
+	int headers_cnt;
+	int headers_spc;
 	int largest_headersize = 80; /* Some magic minimum.. */
 	char dirprefix[8];
 	int mypid = getpid();
+	long format = 0;
 
 	static struct ctldesc d; /* ONLY ONE OPEN AT THE TIME! */
 
@@ -262,8 +276,8 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 	  getwd(cwd);
 #endif
 	  sprintf(buf,
-		  "Cannot open control file \"%%s\" from \"%s\" for \"%%s/%%s\" as uid %d! (%%m)",
-		  cwd, geteuid());
+		  "Cannot open control file \"%%s\" from \"%s\" for \"%%s/%%s\" as uid %d!",
+		  cwd, (int)geteuid());
 	  errno = e;
 	  if (host == NULL)
 	    host = "-";
@@ -282,16 +296,20 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 	}
 	/* 4 is the minimum number of characters per line */
 	n = sizeof (long) * (stbuf.st_size / 4);
-	if ((d.contents = (s = emalloc((u_int)stbuf.st_size+1))) == NULL
-	    || (d.offset = (long *)emalloc((u_int)n)) == NULL) {
+	d.contents = contents = s = malloc((u_int)stbuf.st_size+1);
+	if (d.contents == NULL) {
+	  warning("Out of virtual memory!", (char *)NULL);
+	  exit(EX_SOFTWARE);
+	}
+	d.offset = (long *)malloc((u_int)n);
+	if (d.offset == NULL) {
 	  warning("Out of virtual memory!", (char *)NULL);
 	  exit(EX_SOFTWARE);
 	}
-	contents = s;
 
 	fcntl(d.ctlfd, F_SETFD, 1); /* Close-on-exec */
 
-#ifdef HAVE_MMAP
+#if defined(HAVE_MMAP) && defined(TA_USE_MMAP)
 #ifndef MAP_VARIABLE
 # define MAP_VARIABLE 0
 #endif
@@ -309,12 +327,14 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 	d.ctlmap = NULL;
 #endif
 	d.contentsize = (int) stbuf.st_size;
-	s[ d.contentsize ] = 0; /* Treat it as a long string.. */
-	if (read(d.ctlfd, s, d.contentsize) != d.contentsize) {
+	contents[ d.contentsize ] = 0; /* Treat it as a long string.. */
+	if (read(d.ctlfd, contents, d.contentsize) != d.contentsize) {
 	  warning("Wrong size read from control file \"%s\"! (%m)",
 		  file);
 	  free((void*)d.contents);
+	  d.contents = NULL;
 	  free((void*)d.offset);
+	  d.offset = NULL;
 	  close(d.ctlfd);
 	  return NULL;
 	}
@@ -324,10 +344,12 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 	   * If it is less than the minimum possible number of control
 	   * lines, then there is something wrong...
 	   */
-	  free(contents);
-	  free((char *)d.offset);
+	  free((void*)d.contents);
+	  d.contents = NULL;
+	  free((void*)d.offset);
+	  d.offset = NULL;
 	  close(d.ctlfd);
-#ifdef HAVE_MMAP
+#if defined(HAVE_MMAP) && defined(TA_USE_MMAP)
 	  if (d.ctlmap != NULL)
 	    munmap((void*)d.ctlmap, d.contentsize);
 #endif	  
@@ -352,6 +374,17 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 	d.dsnretmode = NULL;
 	d.verbose    = NULL;
 
+	headers_cnt = 0;
+	headers_spc = 1;
+	for (i = 0; i < n; ++i)
+	  if (contents[ d.offset[i] ] == _CF_MSGHEADERS)
+	    ++headers_spc;
+
+	msgheaders = (char***)malloc(sizeof(char***) *
+				     (headers_spc+1));
+	msgheaderscvt = (char***)malloc(sizeof(char***) *
+					(headers_spc+1));
+
 	/* run through the file and set up the information we need */
 	for (i = 0; i < n; ++i) {
 	  if (*exitflagp && d.recipients == NULL)
@@ -359,6 +392,17 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 	  /* Shudder... we trash the memory block here.. */
 	  s = contents + d.offset[i];
 	  switch (*s) {
+	  case _CF_FORMAT:
+	    ++s;
+	    format = 0;
+	    sscanf(s,"%li",&format);
+	    if (format & (~_CF_FORMAT_KNOWN_SET)) {
+	      warning("Unsupported SCHEDULER file format flags seen: 0x%x at file '%s'",
+		      format, file);
+	      *exitflagp = 1;
+	      break;
+	    }
+	    break;
 	  case _CF_SENDER:
 	    ap = ctladdr(s+2);
 	    ap->link  = d.senders;
@@ -383,19 +427,23 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 
 	    if (*s != _CFTAG_NORMAL || d.senders == NULL)
 	      break;
+
 	    ++s;
-	    if (*s >= 'a' && *s <= 'z') {
-	      /* The old style file.. w/o PID-lock */
-	      if ((ap = ctladdr(s)) == NULL) {
-		warning("Out of virtual memory!", (char *)NULL);
-		*exitflagp = 1;
-		break;
-	      }
-	    } else if ((ap = ctladdr(s+_CFTAG_RCPTPIDSIZE)) == NULL) {
+	    /* Unconditionally expecting _CF_FORMAT_TA_PID !! */
+	    s += _CFTAG_RCPTPIDSIZE;
+	    delayslot = NULL;
+	    if ((format & _CF_FORMAT_DELAY1) || *s == ' ' ||
+		(*s >= '0' && *s <= '9')) {
+	      /* Newer DELAY data slot - _CFTAG_RCPTDELAYSIZE bytes */
+	      delayslot = s;
+	      s += _CFTAG_RCPTDELAYSIZE;
+	    }
+	    if ((ap = ctladdr(s)) == NULL) {
 	      warning("Out of virtual memory!", (char *)NULL);
 	      *exitflagp = 1;
 	      break;
 	    }
+
 	    /* [mea] understand  'host' of type:
 	       "((mxer)(mxer mxer))"   */
 	    if ((channel != NULL
@@ -417,7 +465,7 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 	      break;
 	    }
 	    ap->link = d.senders; /* point at sender address */
-	    rp = (struct rcpt *)emalloc(sizeof (struct rcpt));
+	    rp = (struct rcpt *)malloc(sizeof (struct rcpt));
 	    if (rp == NULL) {
 	      lockaddr(d.ctlfd, d.ctlmap, d.offset[i]+1,
 		       _CFTAG_LOCK, _CFTAG_DEFER, file, host, mypid);
@@ -426,12 +474,14 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 	      break;
 	    }
 	    rp->addr = ap;
+	    rp->delayslot = delayslot;
 	    rp->id = d.offset[i];
 	    /* XX: XOR locks are different */
 	    rp->lockoffset = rp->id + 1;
 	    rp->next = d.recipients;
 	    rp->desc = &d;
 	    rp->orcpt = NULL;
+	    rp->inrcpt = NULL;
 	    rp->notify = NULL;
 	    rp->notifyflgs = _DSN_NOTIFY_FAILURE; /* Default behaviour */
 	    d.recipients = rp;
@@ -481,6 +531,13 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 		  if (*s) *s++ = 0;
 		  continue;
 		}
+		if (cistrncmp("INRCPT=",s,7)==0) {
+		  s += 7;
+		  prevrp->inrcpt = s;
+		  while (*s && *s != ' ' && *s != '\t') ++s;
+		  if (*s) *s++ = 0;
+		  continue;
+		}
 		/* XX: BOO! Unknown value! */
 		while (*s && *s != ' ' && *s != '\t') ++s;
 	      }
@@ -507,38 +564,23 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 		 individual lines.. [mea] */
 	      while (*s) {
 		if (headerlines == 0) {
-		  msgheader = (char **)malloc(sizeof(char**)*8);
+		  msgheader = (char **)malloc(sizeof(char**) * 8);
 		  headerspace = 7;
 		}
-		if (headerlines+1 >= headerspace) {
+		if (headerlines >= headerspace) {
 		  headerspace += 8;
-		  msgheader =
-		    (char**)realloc(msgheader,
-				    sizeof(char**)*headerspace);
+		  msgheader = (char**)realloc((void*) msgheader,
+					      sizeof(void*) * (headerspace+1));
 		}
 		ss = s;
 		while (*ss && *ss != '\n') ++ss;
 		if (*ss == '\n') *ss++ = '\0';
 		msgheader[headerlines++] = s;
-		msgheader[headerlines] = NULL;
+		msgheader[headerlines  ] = NULL;
 		s = ss;
 	      }
 
 	      /* And the global connection.. */
-	      if (headers_cnt == 0) {
-		msgheaders = (char***)malloc(sizeof(char***)*8);
-		msgheaderscvt = (char***)malloc(sizeof(char***)*8);
-		headers_spc = 7;
-	      }
-	      if (headers_cnt+1 >= headers_spc) {
-		headers_spc += 8;
-		msgheaders =
-		  (char***)realloc(msgheaders,
-				   sizeof(char***)*headers_spc);
-		msgheaderscvt =
-		  (char***)realloc(msgheaderscvt,
-				   sizeof(char***)*headers_spc);
-	      }
 	      msgheaders   [headers_cnt] = msgheader;
 	      msgheaderscvt[headers_cnt] = NULL;
 
@@ -590,8 +632,8 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 	mfpath = alloca((u_int)5 + sizeof(QUEUEDIR)
 			+ strlen(dirprefix) + strlen(d.msgfile));
 #else
-	mfpath = emalloc((u_int)5 + sizeof(QUEUEDIR)
-			 + strlen(dirprefix) + strlen(d.msgfile));
+	mfpath = malloc((u_int)5 + sizeof(QUEUEDIR)
+			+ strlen(dirprefix) + strlen(d.msgfile));
 #endif
 	sprintf(mfpath, "../%s/%s%s", QUEUEDIR, dirprefix, d.msgfile);
 	if ((d.msgfd = open(mfpath, O_RDONLY, 0)) < 0) {
@@ -626,9 +668,11 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 	  return NULL;
 	}
 
+	d.msginonumber = (long)stbuf.st_ino;
+
 	fcntl(d.msgfd, F_SETFD, 1); /* Close-on-exec */
 
-#ifdef	HAVE_MMAP
+#if defined(HAVE_MMAP) && defined(TA_USE_MMAP)
 	d.let_buffer = (char *)mmap(NULL, stbuf.st_size, PROT_READ,
 				    MAP_FILE|MAP_SHARED|MAP_VARIABLE,
 				    d.msgfd, 0);
