@@ -23,13 +23,13 @@
 #include <fcntl.h>
 #include <sys/stat.h>
 
+#include "ta.h"
+
 #include "mail.h"
-#include "malloc.h"
+#include "zmalloc.h"
 #include "libz.h"
 #include "libc.h"
 
-#include "ta.h"
-
 /*
 
    If WE log the diagnostic status information into the transport
@@ -48,6 +48,14 @@
 
 int ta_logs_diagnostics = 1;
 
+/*
+
+  If we encounter malloc failure, we diagnose everything grimly as
+  EX_TEMPFAIL - always.. and hope the TA caller understands to exit..
+  
+*/
+int zmalloc_failure = 0;
+
 char *notarybuf = NULL;
 time_t retryat_time = 0; /* Used at SMTP to avoid trying same host too soon.. */
 
@@ -110,18 +118,31 @@ int xdly;
 	xdelay = xdly;
 }
 
+static char *A1 = NULL, *A2 = NULL, *A3 = NULL, *A4 = NULL;
+static const char *notaryfmt = "%s\001%s\001%s\001%s\001%s";
+
 void
 notaryreport(arg1,arg2,arg3,arg4)
      const char *arg1, *arg2, *arg3, *arg4;
 {
-	const char *fmt = "%s\001%s\001%s\001%s\001%s";
 	int len;
-	static char *A1 = NULL, *A2 = NULL, *A3 = NULL, *A4 = NULL;
 
-	if (arg1) { if (A1) free(A1); A1 = strdup(arg1); }
-	if (arg2) { if (A2) free(A2); A2 = strdup(arg2); }
-	if (arg3) { if (A3) free(A3); A3 = strdup(arg3); }
-	if (arg4) { if (A4) free(A4); A4 = strdup(arg4); }
+	if (arg1) {
+	  if (A1) free(A1);
+	  A1 = strdup(arg1);
+	}
+	if (arg2) {
+	  if (A2) free(A2);
+	  A2 = strdup(arg2);
+	}
+	if (arg3) {
+	  if (A3) free(A3);
+	  A3 = strdup(arg3);
+	}
+	if (arg4) {
+	  if (A4) free(A4);
+	  A4 = strdup(arg4);
+	}
 
 	len = 5; /* "\001\001\001\001" */
 	if (A1) len += strlen(A1);
@@ -129,14 +150,24 @@ notaryreport(arg1,arg2,arg3,arg4)
 	if (A3) len += strlen(A3);
 	if (A4) len += strlen(A4);
 	if (wtthost) len += strlen(wtthost);
+	if (wttip) len += strlen(wttip)+5;
 
 	if (!notarybuf)
-	  notarybuf = (char*) emalloc(len);
+	  notarybuf = (char*) malloc(len);
 	else
-	  notarybuf = (char*) erealloc(notarybuf,len);
-	sprintf(notarybuf,fmt,
-		A1?A1:"",A2?A2:"",A3?A3:"",A4?A4:"",
+	  notarybuf = (char*) realloc(notarybuf,len);
+
+	if (! notarybuf) {
+	  zmalloc_failure = 1;
+	  return;
+	}
+
+	sprintf(notarybuf, notaryfmt,
+		A1?A1:"", A2?A2:"", A3?A3:"", A4?A4:"",
 		wtthost?wtthost:"");
+	if (wttip) {
+	  sprintf(notarybuf + strlen(notarybuf), " (%s)", wttip);
+	}
 }
 
 
@@ -170,6 +201,8 @@ diagnostic(rp, rc, timeout, fmt, va_alist) /* (rp, rc, timeout, "fmtstr", remote
 	int report_notary = 1;
 	int logreport = 0;
 
+	if (!rp->lockoffset) return; /* Don't re-report... */
+
 	rp->status = rc;
 
 #ifdef HAVE_STDARG_H
@@ -212,6 +245,29 @@ diagnostic(rp, rc, timeout, fmt, va_alist) /* (rp, rc, timeout, "fmtstr", remote
 	*s = '\0';
 	va_end(ap);
 
+	/* If we had ZMALLOC_FAILURE -> ABORT!  */
+	if (zmalloc_failure && !(rp->notifyflgs & _DSN__DIAGDELAYMODE)) {
+
+	  fprintf(stdout,"%d/%d\t%s\t%s %s\n",
+		  rp->desc->ctlid, rp->id,
+		  (notarybuf && report_notary) ? notarybuf : "",
+		  
+		  "deferred", "MALLOC FAILURE!");
+	  fflush(stdout);
+
+	  if (!lockaddr(rp->desc->ctlfd, rp->desc->ctlmap,
+			rp->lockoffset, _CFTAG_LOCK, _CFTAG_DEFER,
+			(char*)rp->desc->msgfile, rp->addr->host, getpid())) {
+	    /* something went wrong in unlocking it, concurrency problem? */
+	  }
+	  rp->lockoffset = 0;	/* mark this recipient unlocked */
+
+	  tasyslog(rp, xdelay, wtthost, wttip, "deferred", "MALLOC FAILURE!");
+
+	  return;
+	}
+
+
 	retryat_time = 0;
 
 	/* Optimize the common case -- less stuff into back-report
@@ -280,71 +336,92 @@ diagnostic(rp, rc, timeout, fmt, va_alist) /* (rp, rc, timeout, "fmtstr", remote
 	  /* Right, we have a honour to append our diagnostics to the
 	     transport specification file ourselves */
 	  int oldfl = fcntl(rp->desc->ctlfd, F_GETFL);
-	  int len = 80 + strlen(notarybuf) + strlen(message);
+	  int newfl;
+	  int len = 80 + strlen(notarybuf ? notarybuf : "") + strlen(message);
 	  int rc2;
+	  off_t ctlsize;
 	  long oldalarm;
 	  char *sbuf;
 
 	  /* Set the APPEND-mode on.. We need it now !
 	     (and make sure the non-blocking mode is NOT on!) */
-	  if ((oldfl & O_APPEND) == 0 || (oldfl & O_NONBLOCK) != 0)
-	    fcntl(rp->desc->ctlfd, F_SETFL, (oldfl & ~O_NONBLOCK) | O_APPEND);
+	  newfl = (oldfl & ~O_NONBLOCK) | O_APPEND;
+	  if (oldfl != newfl)
+	    fcntl(rp->desc->ctlfd, F_SETFL, newfl);
 
 #ifdef HAVE_ALLOCA
 	  sbuf = alloca(len);
 #else
-	  sbuf = emalloc(len);
+	  sbuf = malloc(len);
+	  if (!sbuf)
+	    zmalloc_failure = 1;
 #endif
 
-	  /* Log the diagnostic string to the file */
-	  len = -1;
+	  if (sbuf) {
+	    /* Log the diagnostic string to the file */
 #ifndef SPRINTF_CHAR
-	  len = 
+	    len = 
 #endif
-	    sprintf(sbuf, "%c%c%d:%d:%d::%ld\t%s\t%s\n",
-		    _CF_DIAGNOSTIC, _CFTAG_NORMAL,
-		    rp->id, rp->headeroffset, rp->drptoffset,
-		    (long)time(NULL), notarybuf, message);
+	      sprintf(sbuf, "%c%c%d:%d:%d::%ld\t%s\t%s\n",
+		      _CF_DIAGNOSTIC, _CFTAG_NORMAL,
+		      rp->id, rp->headeroffset, rp->drptoffset,
+		      (long)time(NULL), notarybuf ? notarybuf : "", message);
 #ifdef SPRINTF_CHAR
-	  len = strlen(sbuf);
+	    len = strlen(sbuf);
 #endif
-	  oldalarm = alarm(0);	/* We do NOT want to be alarmed while
-				   writing to the log! */
 
-	  rc2 = write(rp->desc->ctlfd, sbuf, len);
-	  if (rc2 != len || rc2 < 0 || len < 0) {
-	    /* UAARGH! -- write failed, must have disk full! */
-	    printf("#HELP! diagnostic writeout with bad results!: len=%d, rc=%d\n", len, rc2);
-	    exit(EX_DATAERR);
-	  }
-	  fsync(rp->desc->ctlfd);
-	  if (oldalarm)		/* Restore it, if it was ticking. */
-	    alarm(oldalarm);
+	    oldalarm = alarm(0);	/* We do NOT want to be alarmed while
+					   writing to the log! */
+
+	    ctlsize = lseek(rp->desc->ctlfd, 0, SEEK_END);
+
+	    rc2 = write(rp->desc->ctlfd, sbuf, len);
+	    if (rc2 != len || rc2 < 0 || len < 0) {
+	      /* UAARGH! -- write failed, must have disk full! */
+#ifdef HAVE_FTRUNCATE
+	      ftruncate(rp->desc->ctlfd, ctlsize); /* Sigh.. */
+#endif
+	      fprintf(stdout,"#HELP! diagnostic writeout with bad results!: len=%d, rc=%d\n", len, rc2);
+	      exit(EX_DATAERR);
+	    }
+#ifdef HAVE_FSYNC
+	    fsync(rp->desc->ctlfd);
+#endif
+	    if (oldalarm)		/* Restore it, if it was ticking. */
+	      alarm(oldalarm);
 
 #ifndef HAVE_ALLOCA
-	  free(sbuf);
+	    free(sbuf);
 #endif
+	  }
 
 	  /* If we had to set the APPEND mode previously, clear it now! */
-	  if ((oldfl & O_APPEND) == 0 || (oldfl & O_NONBLOCK) != 0)
+	  if (oldfl != newfl)
 	    fcntl(rp->desc->ctlfd, F_SETFL, oldfl);
 
 	  /* Now we have no reason to send also the NOTARY report up.. */
-	  *notarybuf = 0;
+	  if (notarybuf != NULL)
+	    *notarybuf = 0;
 	}
 
-	printf("%d/%d\t%s\t%s %s\n",
-	       rp->desc->ctlid, rp->id,
-	       (notarybuf && report_notary) ? notarybuf : "",
-	       statmsg, message);
+	/* "Delay" the diagnostics from mailbox sieve subprocessing.
+	   Actually DON'T do then at all! */
+	if (!(rp->notifyflgs & _DSN__DIAGDELAYMODE)) {
 
-	if (!lockaddr(rp->desc->ctlfd, rp->desc->ctlmap,
-		      rp->lockoffset, _CFTAG_LOCK, mark,
-		      (char*)rp->desc->msgfile, "?host?", getpid())) {
-	  /* something went wrong in unlocking it, concurrency problem? */
+	  fprintf(stdout,"%d/%d\t%s\t%s %s\n",
+		  rp->desc->ctlid, rp->id,
+		  (notarybuf && report_notary) ? notarybuf : "",
+		  statmsg, message);
+	  fflush(stdout);
+
+	  if (!lockaddr(rp->desc->ctlfd, rp->desc->ctlmap,
+			rp->lockoffset, _CFTAG_LOCK, mark,
+			(char*)rp->desc->msgfile, rp->addr->host, getpid())) {
+	    /* something went wrong in unlocking it, concurrency problem? */
+	  }
+	  rp->lockoffset = 0;	/* mark this recipient unlocked */
+
+	  tasyslog(rp, xdelay, wtthost, wttip, statmsg, message);
 	}
-	rp->lockoffset = 0;	/* mark this recipient unlocked */
 	fflush(stdout);
-
-	tasyslog(rp, xdelay, wtthost, wttip, statmsg, message);
 }
