@@ -14,8 +14,8 @@
  * There are multiple-kind headers depending upon how they have been
  * rewritten, and those do tack together for each recipients (rp->)
  * There
- *    rp->newmsgheader    is a pointer to an element on  rp->desc->headers[]
- *    rp->newmsgheadercvt is respectively an elt on  rp->desc->headerscvt[]
+ *  rp->newmsgheader    is a pointer to an element on rp->desc->msgheaders[]
+ *  rp->newmsgheadercvt is respectively an element on rp->desc->msgheaderscvt[]
  *
  * The routine-collection   mimeheaders.c  creates converted headers,
  * if the receiving system needs them. Converted header data is created
@@ -26,7 +26,6 @@
  */
 
 #include "hostenv.h"
-#include <stdio.h>
 #ifdef HAVE_STDARG_H
 # include <stdarg.h>
 #else
@@ -42,9 +41,10 @@
 #endif
 #endif
 
+#include "ta.h"
+
 #include "zmalloc.h"
 #include "libz.h"
-#include "ta.h"
 
 #ifndef strdup
 extern char *strdup();
@@ -126,8 +126,19 @@ strqcpy(buf,buflen,str)
 }
 
 
+void cvtspace_free(rp)
+     struct rcpt *rp;
+{
+	char ***ss;
+
+	if (!rp->newmsgheadercvt) return;
+
+	ss = rp->newmsgheadercvt;
+}
+
+
 int cvtspace_copy(rp)
-struct rcpt *rp;
+     struct rcpt *rp;
 {
 	int hdrcnt = 0;
 	char **probe = *(rp->newmsgheader);
@@ -141,7 +152,7 @@ struct rcpt *rp;
 
 	/* Allocate, and copy ! */
 
-	newcvt = (char **)emalloc(sizeof(char *)*(hdrcnt+1));
+	newcvt = (char **)malloc(sizeof(char *)*(hdrcnt+1));
 	if (newcvt != NULL) {
 
 	  char **ss = newcvt;
@@ -149,8 +160,8 @@ struct rcpt *rp;
 
 	  while (*probe) { /* Copy over */
 	    int len = strlen(*probe)+1;
-	    *ss = (char *)emalloc(len);
-	    /* emalloc() exits at the 'out of memory' condition */
+	    *ss = (char *)malloc(len);
+	    if (! *ss) return 0;
 	    memcpy(*ss,*probe,len);
 	    ++hdrcnt;
 	    ++probe;
@@ -224,21 +235,24 @@ append_header(va_alist)
 
 	while ((*hdrpp)[linecnt] != NULL) ++linecnt;
 
-	hdrp2 = (char**)erealloc(*hdrpp,sizeof(char **) * (linecnt+3));
+	hdrp2 = (char**)realloc(*hdrpp,sizeof(char **) * (linecnt+3));
 
 	if (!hdrp2) return -1;
-	hdrp2[linecnt] = (char*) emalloc(linelen+3);
-	memcpy(hdrp2[linecnt],lbuf,linelen+2);
-	hdrp2[++linecnt] = NULL;
-	*hdrpp = hdrp2;
+	hdrp2[linecnt] = (char*) malloc(linelen+3);
+	if (! hdrp2[linecnt]) zmalloc_failure = 1;
+	else {
+	  memcpy(hdrp2[linecnt],lbuf,linelen+2);
+	  hdrp2[++linecnt] = NULL;
+	  *hdrpp = hdrp2;
+	}
 	return linecnt;
 }
 
 void
 output_content_type(rp,ct,old)
-struct rcpt *rp;
-struct ct_data *ct;
-char **old;
+     struct rcpt *rp;
+     struct ct_data *ct;
+     char **old;
 {
 	/* Output the  Content-Type: with  append_header()
 	   Sample output:
@@ -349,7 +363,11 @@ char **old;
 	    ++o2, ++oldcnt;
 	}
 	newlines = hdrlines + linecnt - oldcnt + 1;
-	newmsgheaders = (char**)emalloc(sizeof(char*)*newlines);
+	newmsgheaders = (char**)malloc(sizeof(char*)*newlines);
+	if (! newmsgheaders) {
+	  zmalloc_failure = 1;
+	  return;
+	}
 	o2 = newmsgheaders;
 	h1 = *basep;
 	while (*h1 && h1 != old) {
@@ -374,10 +392,10 @@ char **old;
 
 struct ct_data *
 parse_content_type(ct_linep)
-char **ct_linep;	/* Could be multiline! */
+     char **ct_linep;	/* Could be multiline! */
 {
 	char *s, *p, *pv, *ss;
-	struct ct_data *ct = (struct ct_data*)emalloc(sizeof(struct ct_data));
+	struct ct_data *ct = (struct ct_data*)malloc(sizeof(struct ct_data));
 	int unknowncount = 0;
 
 	if (!ct) return NULL; /* Failed to parse it! */
@@ -396,7 +414,8 @@ char **ct_linep;	/* Could be multiline! */
 	p = s;
 	while (*s && *s != ' ' && *s != '/' && *s != '\t' && *s != ';')
 	  ++s;
-	ct->basetype = emalloc((s - p)+2);
+	ct->basetype = malloc((s - p)+2);
+	/* FIXME: malloc problem check ?? */
 	ss = ct->basetype;
 	/* Copy over the basetype */
 	while (p < s) *ss++ = *p++;
@@ -408,7 +427,8 @@ char **ct_linep;	/* Could be multiline! */
 	  p = s;
 	  while (*s && *s != ' ' && *s != '/' && *s != '\t' && *s != ';')
 	    ++s;
-	  ct->subtype = emalloc((s - p)+2);
+	  ct->subtype = malloc((s - p)+2);
+	  /* FIXME: malloc problem check ?? */
 	  ss = ct->subtype;
 	  /* Copy over the subtype */
 	  while (p < s) *ss++ = *p++;
@@ -499,7 +519,8 @@ char **ct_linep;	/* Could be multiline! */
 	    }
 	  }
 
-	  parval = emalloc((s - pv) + 2);
+	  parval = malloc((s - pv) + 2);
+	  /* FIXME: malloc problem check ?? */
 	  ss = parval;
 	  if (*pv == '"') { /* Copy the quoted string to parval */
 	    int quoted = 0;
@@ -532,13 +553,16 @@ char **ct_linep;	/* Could be multiline! */
 	    int unklen = strlen(parval)+5+strlen(paramname);
 	    int unkpos;
 	    if (!ct->unknown) {
-	      ct->unknown = (char**)emalloc(sizeof(char*)*2);
+	      ct->unknown = (char**)malloc(sizeof(char*)*2);
+	/* FIXME: malloc problem check ?? */
 	      ct->unknown[1] = NULL;
 	    } else {
-	      ct->unknown = (char**)erealloc(ct->unknown,
-					     sizeof(char*)*(unknowncount+2));
+	      ct->unknown = (char**)realloc(ct->unknown,
+					    sizeof(char*)*(unknowncount+2));
+	/* FIXME: malloc problem check ?? */
 	    }
-	    ct->unknown[unknowncount] = emalloc(unklen);
+	    ct->unknown[unknowncount] = malloc(unklen);
+	/* FIXME: malloc problem check ?? */
 	    sprintf(ct->unknown[unknowncount],"%s=",paramname);
 	    unkpos = strlen(ct->unknown[unknowncount]);
 	    strqcpy(ct->unknown[unknowncount]+unkpos,unklen-unkpos,parval);
@@ -550,10 +574,12 @@ char **ct_linep;	/* Could be multiline! */
 
 struct cte_data *
 parse_content_encoding(cte_linep)
-char **cte_linep;	/* Propably is not a multiline entry.. */
+     char **cte_linep;	/* Propably is not a multiline entry.. */
 {
 	char *line, *s;
-	struct cte_data *cte = emalloc(sizeof(struct cte_data));
+	struct cte_data *cte = malloc(sizeof(struct cte_data));
+
+	if (!cte) return NULL;
 
 	s = (*cte_linep) + 26;
 	/* Skip over the 'Content-Transfer-Encoding:' */
@@ -561,13 +587,15 @@ char **cte_linep;	/* Propably is not a multiline entry.. */
 	line = s;
 	if (*s == '"') {
 	  char *p;
-	  cte->encoder = p = emalloc(strlen(s));
+	  cte->encoder = p = malloc(strlen(s));
+	/* FIXME: malloc problem check ?? */
 	  while (*s && *s != '"')
 	    *p++ = *s++;
 	  *p = 0;
 	} else {
 	  char *p;
-	  cte->encoder = p = emalloc(strlen(s));
+	  cte->encoder = p = malloc(strlen(s));
+	/* FIXME: malloc problem check ?? */
 	  while (*s && *s != ' ' && *s != '\t')
 	    *p++ = *s++;
 	  *p = 0;
@@ -586,7 +614,7 @@ char **cte_linep;	/* Propably is not a multiline entry.. */
  */
 int
 check_conv_prohibit(rp)
-struct rcpt *rp;
+     struct rcpt *rp;
 {
 	char **hdrs = *(rp->newmsgheader);
 	if (!hdrs) return 0;
@@ -609,7 +637,7 @@ static const char *cCT  = "Content-Type:";
 
 int
 cte_check(rp)
-struct rcpt *rp;
+     struct rcpt *rp;
 {	/* "Content-Transfer-Encoding: 8BIT" */
 
 	char **hdrs = *(rp->newmsgheader);
@@ -617,7 +645,8 @@ struct rcpt *rp;
 	int mime = 0;
 
 	/* if (*(rp->newmsgheadercvt) != NULL)
-	   hdrs = *(rp->newmsgheadercvt); */ /* here we check the ORIGINAL headers.. */
+	   hdrs = *(rp->newmsgheadercvt); */
+	/* here we check the ORIGINAL headers.. */
 
 	if (!hdrs) return 0;
 
@@ -642,8 +671,8 @@ struct rcpt *rp;
 
 char **  /* Return a pointer to header line pointer */
 has_header(rp,keystr)
-struct rcpt *rp;
-const char *keystr;
+     struct rcpt *rp;
+     const char *keystr;
 {
 	char **hdrs = *(rp->newmsgheader);
 	int keylen = strlen(keystr);
@@ -662,8 +691,8 @@ const char *keystr;
 void
 delete_header(rp,hdrp)	/* Delete the header, and its possible
 			   continuation lines */
-struct rcpt *rp;
-char **hdrp;
+     struct rcpt *rp;
+     char **hdrp;
 {
 	char **h1 = hdrp;
 	char **h2 = hdrp+1;
@@ -680,8 +709,8 @@ char **hdrp;
 
 int
 downgrade_charset(rp, verboselog)
-struct rcpt *rp;
-FILE *verboselog;
+     struct rcpt *rp;
+     FILE *verboselog;
 {
 	char **CT   = NULL;
 	char **CTE  = NULL;
@@ -718,9 +747,9 @@ FILE *verboselog;
 
 int
 downgrade_headers(rp, convertmode, verboselog)
-struct rcpt *rp;
-int convertmode;
-FILE *verboselog;
+     struct rcpt *rp;
+     CONVERTMODE convertmode;
+     FILE *verboselog;
 {
 	char ***oldmsgheader;
 	char **CT   = NULL;
@@ -754,28 +783,32 @@ FILE *verboselog;
 	  /* We downgrade by changing it to Q-P as per RFC 1428/Appendix A */
 	  static const char *warning_lines[] = {
 "X-Warning: Original message contained 8-bit characters, however during",
-"           the SMTP transport session the receiving system was unable to",
-"           announce capability of receiving 8-bit SMTP (RFC 1651-1653),",
-"           and as this message does not have MIME headers (RFC 2045-2049)",
-"           to enable encoding change, we had very little choices.",
+"\t   the SMTP transport session the receiving system did not announce",
+"\t   capability of receiving 8-bit SMTP (RFC 1651-1653), and as this",
+"\t   message does not have MIME headers (RFC 2045-2049) to enable",
+"\t   encoding change, we had very little choice.",
 "X-Warning: We ASSUME it is less harmful to add the MIME headers, and",
-"           convert the text to Quoted-Printable, than not to do so,",
-"           and to strip the message to 7-bits.. (RFC 1428 Appendix A)",
+"\t   convert the text to Quoted-Printable, than not to do so,",
+"\t   and to strip the message to 7-bits.. (RFC 1428 Appendix A)",
 "X-Warning: We don't know what character set the user used, thus we had to",
-"           write these MIME-headers with our local system default value.",
+"\t   write these MIME-headers with our local system default value.",
 "MIME-Version: 1.0",
 "Content-Transfer-Encoding: QUOTED-PRINTABLE",
 NULL };
 
-	  char **newmsgheaders = (char**)emalloc(sizeof(char**)*(lines+15));
+	  char **newmsgheaders = (char**)malloc(sizeof(char**)*(lines+15));
 	  char *defcharset = getzenv("DEFCHARSET");
 	  char *newct;
+
+	/* FIXME: malloc problem check ?? */
+
 	  if (!defcharset)
 	    defcharset = "ISO-8859-1";
 #ifdef HAVE_ALLOCA
 	  newct = alloca(strlen(defcharset)+2+sizeof("Content-Type: TEXT/PLAIN; charset="));
 #else
-	  newct = emalloc(strlen(defcharset)+2+sizeof("Content-Type: TEXT/PLAIN; charset="));
+	  newct = malloc(strlen(defcharset)+2+sizeof("Content-Type: TEXT/PLAIN; charset="));
+	/* FIXME: malloc problem check ?? */
 #endif
 	  sprintf(newct,"Content-Type: TEXT/PLAIN; charset=%s",defcharset);
 
@@ -792,14 +825,10 @@ NULL };
 	    delete_header(rp,CT);
 	  /* These most propably won't happen, but the delete_header()
 	     does scram the pointers anyway.. */
-	  if (MIME) {
-	    MIME = has_header(rp,"MIME-Version:");
+	  if (MIME)
 	    delete_header(rp,MIME);
-	  }
-	  if (CTE) {
-	    CTE  = has_header(rp,cCTE);
+	  if (CTE)
 	    delete_header(rp,CTE);
-	  }
 
 	  for (i = 0; (*oldmsgheader)[i] != NULL; ++i)
 	    newmsgheaders[lines+i] = (*oldmsgheader)[i];
@@ -827,11 +856,10 @@ NULL };
 
 	/* Ok, this was C-T-E: 7BIT, turn charset to US-ASCII if it
 	   was  ISO-*  */
-	if (CT == NULL && verboselog) {
-	  fprintf(verboselog,"Had Content-Transfer-Encoding -header, but no Content-Type header ???  Adding C-T..\n");
-	}
 
 	if (CT == NULL) { /* ???? Had C-T-E, but no C-T ?? */
+	  if (verboselog)
+	    fprintf(verboselog,"Had Content-Transfer-Encoding -header, but no Content-Type header ???  Adding C-T..\n");
 	  append_header(rp,"Content-Type: TEXT/PLAIN; charset=US-ASCII");
 	  return 0;
 	}
@@ -865,6 +893,10 @@ NULL };
 
 	if (convertmode == _CONVERT_QP) {
 
+	  /* Preceding 'output_content_type()' call scrambled lists,
+	     we have to reget the CTE pointer */
+	  CTE  = has_header(rp,cCTE);
+
 	  strcpy(*CTE, "Content-Transfer-Encoding: QUOTED-PRINTABLE");
 	  mime_received_convert(rp," convert rfc822-to-quoted-printable");
 
@@ -880,39 +912,65 @@ mime_received_convert(rp, convertstr)
 	char *convertstr;
 {
 	int convertlen = strlen(convertstr);
+	char *semic = NULL;
+	char **schdr = NULL;
+	int semicindex = -1;
 
 	char **inhdr = *(rp->newmsgheadercvt);
+	char **hdro = NULL;
 
-	/* We have one advantage: The "Received:" header we want to
-	   fiddle with is the first of them at all. */
+	/* We have one advantage: The "Received:" header we
+	   want to fiddle with is the first one of them. */
 
-	while (inhdr && cistrncmp(*inhdr,"Received:",9) == 0) {
-	  int receivedlen = strlen(*inhdr);
-	  char *newreceived = NULL;
-	  char *semicpos;
-	  int  semicindex = receivedlen;
+	if (!inhdr || cistrncmp(*inhdr,"Received:",9) != 0) {
+	  return 0; /* D'uh ??  Not 'Received:' ??? */
+	}
+
+	/* Look for the LAST semicolon in this Received: header.. */
+
+	do {
+	  char *sc = strrchr(*inhdr, ';');
+	  if (sc) {
+	    semic = sc;
+	    schdr = inhdr;
+	    semicindex = sc - *inhdr;
+	  }
+	  hdro = inhdr;
+	  ++inhdr;
+	} while (*inhdr && (**inhdr == ' ' || **inhdr == '\t'));
+
+	/* Now 'semic' is either set, or not; if set, 'schdr' points
+	   to the begin of the line where it is.
+	   If 'semic' is not set, 'hdro' will point to the last line
+	   of the 'Received:' header in question */
+
+	if (!semic) {
+	  schdr = hdro;
+	  semicindex = strlen(*schdr);
+	  semic = *schdr + semicindex;
+	}
+
+	{
+	  int receivedlen = strlen(*schdr);
+	  char *newreceived = malloc(receivedlen + convertlen + 1);
+	  int  semicindex;
 
-	  newreceived = emalloc(receivedlen + convertlen + 1);
 	  if (!newreceived) return 0; /* Failed malloc.. */
-	  strcpy(newreceived,*inhdr);
-	  semicpos = strchr(newreceived,';');
-	  if (semicpos != NULL) {
-	    semicindex = semicpos - newreceived;
-	    strcpy(newreceived+semicindex+convertlen,(*inhdr)+semicindex);
-	  } else
-	    semicpos = newreceived+semicindex;
-	  memcpy(semicpos,convertstr,convertlen);
 
-	  ctlfree(rp->desc,*inhdr);
-	  *inhdr = newreceived;
+	  semicindex = semic - *schdr;
+	  memcpy(newreceived, *schdr, semicindex);
+	  memcpy(newreceived+semicindex, convertstr, convertlen);
+	  strcpy(newreceived+semicindex+convertlen, (*schdr) + semicindex);
+	  newreceived[receivedlen + convertlen] = '\0';
+
+	  ctlfree(rp->desc,*schdr);
+	  *schdr = newreceived;
 
 	  /* if (verboselog) {
 	     fprintf(verboselog,"Rewriting 'Received:' headers.\n");
-	     fprintf(verboselog,"The new line is: '%s'\n",*inhdr);
+	     fprintf(verboselog,"The new line is: '%s'\n",*hdro);
 	     }
-	   */
-
-	  inhdr = NULL; /* quit this header munging.. */
+	  */
 	}
 	return 1;
 }
