@@ -1,6 +1,6 @@
 /*
  * tasyslog() -- support routine for ZMailer transport agents.
- * Copyright 1997, Matti Aarnio <mea@nic.funet.fi>
+ * Copyright 1997-1999, Matti Aarnio <mea@nic.funet.fi>
  *
  * The purpose of this routine is to produce similar syslog entries
  * to those that sendmail(8) does for its message processing.
@@ -39,6 +39,17 @@ const char *msg;
   static char *syslogflg = NULL;
   char *t;
 
+  /* Syslogflag 'T' for classical format, and 't' for TAB-separated format */
+
+  static const char *fmt1c = "%s: to=<%.200s>, delay=%s, xdelay=%s, mailer=%.80s, stat=%.80s %.200s";
+  static const char *fmt1t = "%s:\tto=<%.200s>\tdelay=%s\txdelay=%s\tmailer=%.80s\tstat=%.80s\t%.200s";
+
+  static const char *fmt2c = "%s: to=<%.200s>, delay=%s, xdelay=%s, mailer=%.80s, relay=%.200s ([%.80s]), stat=%.80s %.400s";
+  static const char *fmt2t = "%s:\tto=<%.200s>\tdelay=%s\txdelay=%s\tmailer=%.80s\trelay=%.200s ([%.80s])\tstat=%.80s\t%.400s";
+
+  static const char *fmt3c = "%s: to=<%.200s>, delay=%s, xdelay=%s, mailer=%.80s, relay=%.200s, stat=%.80s %.400s";
+  static const char *fmt3t = "%s:\tto=<%.200s>\tdelay=%s\txdelay=%s\tmailer=%.80s\trelay=%.200s\tstat=%.80s\t%.400s";
+
   if (syslogflg == NULL) {
     syslogflg = getzenv("SYSLOGFLG");
     if (syslogflg == NULL)
@@ -79,15 +90,15 @@ const char *msg;
   }
 #endif
   if (wtthost == NULL)
-    sprintf(linebuf, "%s: to=<%.200s>, delay=%s, xdelay=%s, mailer=%.80s, stat=%.80s %.200s",
+    sprintf(linebuf, ((*t == 't') ? fmt1t : fmt1c),
 	    spoolid, rp->addr->user, delays, xdelays, rp->addr->channel, statstr, msg);
   else {
     if (wttip != NULL)
-      sprintf(linebuf, "%s: to=<%.200s>, delay=%s, xdelay=%s, mailer=%.80s, relay=%.200s ([%.80s]), stat=%.80s %.400s",
+      sprintf(linebuf, ((*t == 't') ? fmt2t : fmt2c),
 	    spoolid, rp->addr->user, delays, xdelays, rp->addr->channel,
 	      wtthost, wttip, statstr, msg);
     else
-      sprintf(linebuf, "%s: to=<%.200s>, delay=%s, xdelay=%s, mailer=%.80s, relay=%.200s, stat=%.80s %.400s",
+      sprintf(linebuf, ((*t == 't') ? fmt3t : fmt3c),
 	      spoolid, rp->addr->user, delays, xdelays, rp->addr->channel,
 	      wtthost, statstr, msg);
   }
