@@ -7,7 +7,6 @@
 #define DefCharset "ISO-8859-1"
 
 #include "hostenv.h"
-#include <stdio.h>
 #include <ctype.h>
 #include <pwd.h>
 #include "zmsignal.h"
@@ -15,14 +14,16 @@
 #include <sys/file.h>
 #include <sys/param.h>
 #include <sys/stat.h>
-#include "mail.h"
-#include "malloc.h"
-#include "zsyslog.h"
 #include "ta.h"
+#include "mail.h"
+#include "zmalloc.h"
+#include "zsyslog.h"
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 #include <errno.h>
+#include "libz.h"
+
 
 #ifdef  HAVE_WAITPID
 # include <sys/wait.h>
@@ -74,6 +75,18 @@ int	keep_header8 = 0;	/* Don't do "MIME-2" to the headers */
 int	D_alloc = 0;		/* Memory debugging */
 char    *defcharset;
 
+
+extern RETSIGTYPE sigpipe();
+int gotsigpipe = 0;
+
+RETSIGTYPE
+sigpipe(sig)
+int sig;
+{
+	gotsigpipe = 1;
+	SIGNAL_HANDLE(SIGPIPE, sigpipe);
+}
+
 extern RETSIGTYPE wantout();
 #ifndef MALLOC_TRACE
 extern univptr_t emalloc();
@@ -99,7 +112,8 @@ struct maildesc {
 	char	*name;
 	short	flags;
 	char	*command;
-	char	*argv[20];
+#define MD_ARGVMAX 20
+	char	*argv[MD_ARGVMAX];
 };
 
 extern int writemimeline __(( struct maildesc *mp, FILE *fp, const char *buf, int len, int convertmode));
@@ -121,6 +135,7 @@ extern int writemimeline __(( struct maildesc *mp, FILE *fp, const char *buf, in
 #define MO_BEDSMTP		0x04000 /* EBSMTP + DSN */
 #define MO_WANTSDATE		0x08000 /* Wants "Date:" -header */
 #define MO_WANTSFROM		0x10000 /* Wants "From:" -header */
+#define MO_BSMTPHELO		0x20000 /* Add HELO/EHLO to the BSMTP */
 
 struct exmapinfo {
 	int	origstatus;
@@ -183,7 +198,7 @@ main(argc, argv)
 	if (oldsig != SIG_IGN)
 	  SIGNAL_HANDLE(SIGHUP, wantout);
 
-	SIGNAL_IGNORE(SIGPIPE);
+	SIGNAL_HANDLE(SIGPIPE, sigpipe);
 
 	if ((progname = strrchr(argv[0], '/')) == NULL)
 	  progname = argv[0];
@@ -344,17 +359,18 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	struct rcpt *startrp, *endrp;
 	FILE *verboselog;
 {
-	struct rcpt *rp;
+	struct rcpt *rp = NULL;
 	struct exmapinfo *exp;
 	const char *exs, *exd;
-	int i, j, pid, in[2], out[2];
+	int i, j, pid = 0, in[2], out[2], ii = 0;
 	unsigned int avsize;
-	FILE *tafp, *errfp;
-	char *cp, buf[BUFSIZ], buf2[BUFSIZ];
-	char *ws;
+	FILE *tafp = NULL, *errfp = NULL;
+	char *cp = NULL, buf[BUFSIZ], buf2[BUFSIZ];
+	char *ws = NULL, *we = NULL;
 	const char *ds, **av, *s;
 	int status;
-	int content_kind, conversion_prohibited, convertmode, ascii_clean = 0;
+	int content_kind, conversion_prohibited, ascii_clean = 0;
+	CONVERTMODE convertmode = _CONVERT_NONE;
 
 	if (lseek(dp->msgfd, (off_t)(dp->msgbodyoffset), SEEK_SET) < 0L)
 		warning("Cannot seek to message body! (%m)", (char *)NULL);
@@ -376,24 +392,31 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	  else
 	    av[i++] = startrp->addr->link->user;
 	}
-	for (j = 1; mp->argv[j] != NULL; ++j) {
+	for (j = 0; mp->argv[j] != NULL; ++j) {
 	  while (i+2 >= avsize) {
 	    avsize *= 2;
 	    av = (const char **)erealloc((char *)av,
 					 sizeof av[0] * avsize);
 	  }
-	  if (strchr(mp->argv[j], '$') == 0) {
-	    av[i++] = mp->argv[j];
+	  if (strchr(mp->argv[j], '$') == NULL) {
+	    if (j > 0)
+	      av[i++] = mp->argv[j];
 	    continue;
 	  }
 	  rp = startrp;
 	  do {
+
+	    /* Even argv[0] MAY have $-expansions.. */
+	    ii = i; if (j == 0) ii = 0;
+
 	    while (i+2 >= avsize) {
 	      avsize <<= 1;
 	      av = (const char **)erealloc((char *)av,
 					   sizeof av[0] * avsize);
 	    }
-	    for (cp = mp->argv[j], ws = buf; *cp != '\0'; ++cp) {
+	    ws = buf;
+	    we = buf + sizeof(buf);
+	    for (cp = mp->argv[j]; *cp != '\0'; ++cp) {
 	      if (*cp == '$') {
 		switch (*++cp) {
 		case 'g':
@@ -407,11 +430,23 @@ deliver(dp, mp, startrp, endrp, verboselog)
 		  rp = rp->next;
 		  break;
 		case 'U':
-		  strcpy(buf2, rp->addr->user);
+		  strncpy(buf2, rp->addr->user, sizeof(buf2));
+		  buf2[sizeof(buf2)-1] = 0;
 		  strlower(buf2);
 		  ds = buf2;
 		  rp = rp->next;
 		  break;
+		case '{':
+		  s = ++cp;
+		  while (*cp != 0 && *cp != '}') ++cp;
+		  if (*cp != 0) {
+		    *cp = 0;
+		    ds = getzenv(s);
+		    *cp = '}';
+		  } else {
+		    ds = getzenv(s);
+		  }
+		  break;
 		default:
 		  ds = NULL;
 		  break;
@@ -424,20 +459,32 @@ deliver(dp, mp, startrp, endrp, verboselog)
 			  *cp, dp->msgfile);
 		  warning(msg, mp->name);
 		} else {
-		  strcpy(ws, ds);
-		  ws += strlen(ds);
+		  int len = strlen(ds);
+		  if (ws + len >= we)
+		    break; /* D'uh :-( */
+		  memcpy(ws, ds, len+1);
+		  ws += len;
 		}
 	      } else
-		*ws++ = *cp;
+		if (ws < we)
+		  *ws++ = *cp;
 	    }
-	    *ws = '\0';
-	    av[i] = emalloc((u_int)(strlen(buf)+1));
+	    if (ws < we)
+	      *ws = '\0';
+	    else
+	      we[-1] = '\0'; /* Trunk in all cases */
 	    /* not worth freeing this stuff */
-	    strcpy((char*)av[i], buf);
-	    ++i;
+	    av[ii] = strdup(buf);
+	    if (j > 0)
+	      ++i;
 	  } while (rp != startrp && rp != endrp);
+	  /* End of: "do {" */
 	}
+	/* End of: "for (j = ...) {" */
 	av[i] = NULL;
+
+	gotsigpipe = 0;
+
 	/* now we can fork off and run the command... */
 	if (pipe(out) < 0) {
 	  for (rp = startrp; rp != endrp; rp = rp->next) {
@@ -464,7 +511,7 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	if (verboselog) {
 	  const char **p = av;
 	  fprintf(verboselog,"To run UID=%d GID=%d ARGV[] =",
-		  getuid(), getgid());
+		  (int)getuid(), (int)getgid());
 	  for ( ;*p != NULL; ++p) {
 	    fprintf(verboselog," '%s'", *p);
 	  }
@@ -509,9 +556,9 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	  }
 	  return;
 	}
-	close(out[0]);
+	close(out[0]); /* child ends.. */
 	close(in[1]);
-	tafp = fdopen(out[1], "w");
+	tafp = fdopen(out[1], "w"); /* parent ends .. */
 	errfp = fdopen(in[0], "r");
 	/* read any messages from its stdout/err on in[0] */
 
@@ -529,12 +576,14 @@ deliver(dp, mp, startrp, endrp, verboselog)
 
 	if (mp->flags & MO_BSMTP) {
 
-	  if (mp->flags & MO_BESMTP)
-	    fprintf(tafp,"EHLO %s",myhostname);
-	  else
-	    fprintf(tafp,"HELO %s",myhostname);
-	  if (mp->flags & MO_CRLF) putc('\r',tafp);
-	  putc('\n',tafp);
+	  if (mp->flags & MO_BSMTPHELO) {
+	    if (mp->flags & MO_BESMTP)
+	      fprintf(tafp,"EHLO %s",myhostname);
+	    else
+	      fprintf(tafp,"HELO %s",myhostname);
+	    if (mp->flags & MO_CRLF) putc('\r',tafp);
+	    putc('\n',tafp);
+	  }
 
 	  if (strcmp(startrp->addr->link->channel,"error")==0)
 	    fprintf(tafp,"MAIL From:<>");
@@ -625,7 +674,8 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	  case 8:		/* 8BIT */
 	    if (!can_8bit && !ascii_clean) {
 	      convertmode = _CONVERT_QP;
-	      downgrade_headers(startrp, convertmode, verboselog);
+	      if (!downgrade_headers(startrp, convertmode, verboselog))
+		convertmode = _CONVERT_NONE;
 	    }
 	    break;
 	  case 9:		/* QUOTED-PRINTABLE */
@@ -633,7 +683,8 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	      /* Force(d) to decode Q-P while transfer.. */
 	      convertmode = _CONVERT_8BIT;
 	      /*  UPGRADE TO 8BIT !  */
-	      qp_to_8bit(startrp);
+	      if (!qp_to_8bit(startrp))
+		convertmode = _CONVERT_NONE;
 	      content_kind = 10;
 	      ascii_clean = 0;
 	    }
@@ -650,28 +701,32 @@ deliver(dp, mp, startrp, endrp, verboselog)
  	/* Add the "Return-Path:" is it is desired, but does not yet
 	   exist.. */
 	if (mp->flags & MO_RETURNPATH) {
-	  char **hdrs = has_header(startrp,"Return-Path:");
 	  const char *uu = startrp->addr->link->user;
-	  if (hdrs) delete_header(startrp, hdrs);
+	  char **hdrs;
+	  do {
+	    hdrs = has_header(startrp,"Return-Path:");
+	    if (hdrs) delete_header(startrp, hdrs);
+	  } while (hdrs);
 	  if (strcmp(startrp->addr->link->channel,"error")==0)
 	    uu = "";
-	  append_header(startrp,"Return-Path: <%s>", uu);
+	  append_header(startrp,"Return-Path: <%.999s>", uu);
 	}
 	if (mp->flags & MO_CRLF) {
-	  writeheaders(startrp, tafp, "\r\n", convertmode, maxwidth);
+	  fwriteheaders(startrp, tafp, "\r\n", convertmode, maxwidth, NULL);
 	  fprintf(tafp, "\r\n");
 	} else {
-	  writeheaders(startrp, tafp, "\n",   convertmode, maxwidth);
+	  fwriteheaders(startrp, tafp, "\n",   convertmode, maxwidth, NULL);
 	  fprintf(tafp, "\n");
 	}
 
 	if (verboselog) {
-	  writeheaders(startrp, verboselog, "\n", convertmode, maxwidth);
+	  fwriteheaders(startrp, verboselog, "\n", convertmode, maxwidth, NULL);
 	  fprintf(verboselog, "\n");
 	}
 
 	/* append message body itself */
-	if ((i = appendlet(dp, mp, tafp, verboselog, convertmode)) != EX_OK) {
+	i = appendlet(dp, mp, tafp, verboselog, convertmode);
+	if (i != EX_OK && !gotsigpipe) {
 	  for (rp = startrp; rp != endrp; rp = rp->next) {
 	    notaryreport(rp->addr->user,"failed",
 			 /* Could indicate: 4.3.1 - mail system full ?? */
@@ -684,6 +739,8 @@ deliver(dp, mp, startrp, endrp, verboselog)
 	  sleep(1);
 	  kill(pid, SIGKILL);
 	  wait(NULL);
+	  fclose(tafp); /* FP/pipe cleanups */
+	  fclose(errfp);
 	  return;
 	}
 
@@ -827,14 +884,10 @@ appendlet(dp, mp, fp, verboselog, convertmode)
 	     We are better to feed writemimeline() with LINES
 	     instead of blocks of data.. */
 #if !(defined(HAVE_MMAP) && defined(TA_USE_MMAP))
-	  char iobuf[BUFSIZ];
-	  FILE *mfp = fdopen(mfd,"r");
+	  Sfio_t *mfp = sfnew(NULL, NULL, 64*1024, mfd, SF_READ|SF_WHOLE);
+	  sfseek(mfp, dp->msgbodyoffset, SEEK_SET);
 
-	  fseek(mfp,(off_t)(dp->msgbodyoffset),SEEK_SET);
-
-	  setvbuf(mfp, iobuf, _IOFBF, sizeof(iobuf));
-
-#define MFPCLOSE i = dup(mfd); fclose(mfp); dup2(i,mfd); close(i);
+#define MFPCLOSE sfsetfd(mfp, -1); sfclose(mfp);
 
 	  readalready = 0;
 #else
@@ -846,7 +899,7 @@ appendlet(dp, mp, fp, verboselog, convertmode)
 	  /*
 	     if(verboselog) fprintf(verboselog,
 	     "sm: Convert mode: %d, fd=%d, fdoffset=%d, bodyoffset=%d\n",
-	     convertmode, mfd, (int)lseek(mfd,(off_t)0,1),
+	     convertmode, mfd, (int)lseek(mfd, (off_t)0, SEEK_CUR),
 	     dp->msgbodyoffset);
 	   */
 
@@ -877,7 +930,7 @@ appendlet(dp, mp, fp, verboselog, convertmode)
 	    }
 	  }
 #if !(defined(HAVE_MMAP) && defined(TA_USE_MMAP))
-	  if (i == EOF && !feof(mfp)) {
+	  if (i == EOF && !sfeof(mfp) && !sferror(mfp)) {
 	    MFPCLOSE
 	    return EX_IOERR;
 	  }
@@ -1182,6 +1235,7 @@ readsmcf(file, mailer)
 	  case 'n':	m.flags &= ~MO_UNIXFROM;	break;
 	  case 'r':	m.flags |= MO_RFROMFLAG;	break;
 	  case 's':	m.flags |= MO_STRIPQUOTES;	break;
+	  case 'H':     m.flags |= MO_BSMTPHELO;	break;
 	  case 'b':	m.flags |= (MO_BSMTP|MO_HIDDENDOT); break;
 	  case 'B':	if (m.flags & MO_BESMTP) /* -BB */
 			    m.flags |= MO_BEDSMTP;
@@ -1232,7 +1286,7 @@ readsmcf(file, mailer)
 	}
 	SKIPWHILE(isspace, cp);
 	i = 0;
-	while (isascii(*cp) && !isspace(*cp)) {
+	while (isascii(*cp) && !isspace(*cp) && i < MD_ARGVMAX) {
 	  if (*cp == '\0')
 	    break;
 	  m.argv[i++] = cp;
@@ -1275,7 +1329,6 @@ struct ctldesc *dp;
 	register int i;
 	register int bufferfull;
 	int lastwasnl;
-	off_t mfd_pos;
 	int mfd = dp->msgfd;
 
 /* can we use cache of message body data */
@@ -1294,7 +1347,7 @@ struct ctldesc *dp;
 	    if (errno == EINTR)
 	      continue;
 	    readalready = 0;
-	    lseek(mfd,dp->msgbodyoffset,0);
+	    lseek(mfd, dp->msgbodyoffset, SEEK_SET);
 	    return 0;
 	  }
 	  lastwasnl = (let_buffer[i-1] == '\n');
@@ -1302,14 +1355,14 @@ struct ctldesc *dp;
 	  bufferfull++;
 	  for (i=0; i < readalready; ++i)
 	    if (128 & (let_buffer[i])) {
-	      lseek(mfd,dp->msgbodyoffset,0);
+	      lseek(mfd, dp->msgbodyoffset, SEEK_SET);
 	      /* We propably have not read everything of the file! */
 	      readalready = 0;
 	      return 0;		/* Not clean ! */
 	    }
 	}
 	/* Got to EOF, and still it is clean 7-BIT! */
-	lseek(mfd,dp->msgbodyoffset,0);
+	lseek(mfd, dp->msgbodyoffset, SEEK_SET);
 
 	if (bufferfull > 1)	/* not all in memory, need to reread */
 	  readalready = 0;
