@@ -33,6 +33,9 @@
 #endif
 #endif
 
+#define _POLICYTEST_INTERNAL_
+#include "policy.h"
+
 #define PROG "makedb"
 
 #include <errno.h>
@@ -135,6 +138,10 @@ static int store_db(dbf, typ, overwritemode, linenum, t, tlen, s, slen)
 		Ndat.dsize = slen;
 		rc = dbm_store(ndbmfile, Nkey, Ndat,
 			       overwritemode ? DBM_REPLACE : DBM_INSERT);
+
+		if (rc < 0 && errno == ENOSPC) {
+		  fprintf(stderr,"NDBM does not allow KEY.LEN + DATA.LEN to exceed 1024 bytes!  linenro=%d\n",linenum);
+		}
 	}
 #endif
 #ifdef HAVE_GDBM_H
@@ -164,6 +171,10 @@ static int store_db(dbf, typ, overwritemode, linenum, t, tlen, s, slen)
 #ifdef HAVE_DB_OPEN2
 		rc = (dbfile->put) (dbfile, NULL, &Bkey, &Bdat,
 				    overwritemode ? 0: DB_NOOVERWRITE);
+
+		/* emulate BSD DB 1.85 return values */
+		rc = -rc;
+
 #else
 		rc = (dbfile->put) (dbfile, &Bkey, &Bdat,
 				    overwritemode ? 0: R_NOOVERWRITE);
@@ -261,6 +272,35 @@ static int store_db(dbf, typ, overwritemode, linenum, t, tlen, s, slen)
 extern int parsepolicykey __((void *, char *));
 extern int parseattributepair __((void *, char *, char *));
 
+/* KK() and KA() macroes are at "policy.h" */
+
+static char *showkey __((const char *key));
+static char *showkey(key)
+const char *key;
+{
+    static char buf[256];
+
+    if (key[1] != P_K_IPv4 && key[1] != P_K_IPv6) {
+	if (strlen(key+2) > (sizeof(buf) - 200))
+	    sprintf(buf,"%d/%s/'%s'", key[0], KK(key[1]), "<too long name>");
+	else
+	    sprintf(buf,"%d/%s/'%s'", key[0], KK(key[1]), key+2);
+    } else
+      if (key[1] == P_K_IPv4)
+	sprintf(buf,"%d/%s/%u.%u.%u.%u/%d",
+		key[0], KK(key[1]),
+		key[2] & 0xff, key[3] & 0xff, key[4] & 0xff, key[5] & 0xff,
+		key[6] & 0xff);
+      else
+	sprintf(buf,"%d/%s/%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x/%d",
+		key[0], KK(key[1]),
+		key[2] & 0xff, key[3] & 0xff, key[4] & 0xff, key[5] & 0xff,
+		key[6] & 0xff, key[7] & 0xff, key[8] & 0xff, key[9] & 0xff,
+		key[10] & 0xff, key[11] & 0xff, key[12] & 0xff, key[13] & 0xff,
+		key[14] & 0xff, key[15] & 0xff, key[16] & 0xff, key[17] & 0xff,
+		key[18] & 0xff);
+    return buf;
+}
 
 /* Scan over quoted string with embedded white spaces, or
    in case the object does not start with a double quote,
@@ -430,17 +470,13 @@ const int typ;
 		rc = store_db(dbf, typ, 0, linenum,
 			      t, tlen, s, slen);
 		if (rc > 0) {
-			int tl = tlen;
-			fprintf(stderr, "WARNING: Duplicate key at line %d: \"",
+			char *s;
+			fprintf(stderr, "WARNING: Duplicate key at line %d: ",
 				linenum);
-			for (;tl > 0; --tl,++t) {
-				unsigned char c = *t;
-				if (c < ' ' || c > 126 || c == '\\')
-					fprintf(stderr, "\\%03o", c);
-				else
-					fprintf(stderr, "%c", c);
-			}
-			fprintf(stderr, "\"\n");
+
+			s = showkey(policykeybuf);
+
+			fprintf(stderr, "%s\n", s);
 		}
 	}
 	return errflag;
@@ -892,5 +928,10 @@ char *argv[];
     }
 #endif
 
+    if (store_errors) {
+      fprintf(stderr,"STORE ERRORS DURING DATABASE WRITE!\n");
+      rc = 1;
+    }
+
     return (rc ? 1 : 0);
 }
